/* eslint  no-console:0 */
const AWS = require('aws-sdk');
const axios = require('axios');
const { compact } = require('lodash');

const INSTANCE_TYPE = 'm5n.xlarge';
const FALLBACK_INSTANCE_TYPE = 'm5.xlarge';

async function getRegions() {
    const EC2 = new AWS.EC2({ region: 'us-east-1' });
    const response = await EC2.describeRegions({ AllRegions: true }).promise();

    return response.Regions.map(region => region.RegionName);
}

async function generateData(amis) {
    const promises = amis.map(async ami => {
        console.log('Handling region:', ami.region);

        const response = await axios.get(`https://pricing.us-east-1.amazonaws.com/offers/v1.0/aws/AmazonEC2/current/${ami.region}/index.json`);

        const filtered = Object.values(response.data.products).filter(product => product.productFamily === 'Compute Instance'
            && product.attributes.operatingSystem === 'Linux'
            && product.attributes.instanceFamily === 'General purpose'
            && product.attributes.vcpu === '4'
            && product.attributes.memory === '16 GiB'
            && product.attributes.tenancy === 'Shared'
            && product.attributes.storage === 'EBS only'
            && product.attributes.instanceType === INSTANCE_TYPE);

        if (filtered.length === 0) {
            console.error(`Instance type ${INSTANCE_TYPE} not found in region ${ami.region}, using ${FALLBACK_INSTANCE_TYPE}`);
        }

        return {
            region: ami.region,
            ami: ami.imageId,
            instanceType: filtered.length === 0 ? FALLBACK_INSTANCE_TYPE : INSTANCE_TYPE
        };
    });

    return Promise.all(promises);
}

async function getLatestAmazonLinuxAmi(region) {
    console.log('Getting images for region:', region);

    const ec2 = new AWS.EC2({ region });

    const response = await ec2.describeImages({
        Filters:
            [
                {
                    Name: 'virtualization-type',
                    Values: ['hvm']
                },
                {
                    Name: 'architecture',
                    Values: ['x86_64']
                },
                {
                    Name: 'state',
                    Values: ['available']
                },
                {
                    Name: 'root-device-type',
                    Values: ['ebs']
                },
                {
                    Name: 'hypervisor',
                    Values: ['xen']
                },
                {
                    Name: 'image-type',
                    Values: ['machine']
                },
                {
                    Name: 'block-device-mapping.volume-type',
                    Values: ['gp2']
                },
                {
                    Name: 'name',
                    Values: ['amzn2-ami-hvm-*']
                },
                {
                    Name: 'owner-id',
                    Values: ['137112412989', '656109587541', '910595266909', '071630900071', '210953353124'] // 656109587541 - for Bahrain, 910595266909 - for Hong Kong, 071630900071 - Milan, 210953353124 - Cape Town
                }
            ],
        Owners: ['amazon']
    })
        .promise();

    console.log(`Found ${response.Images.length} images in region:`, region);

    // 2016-02-10T23:44:07.000Z <- ISO_8601
    const sorted = response.Images.sort((a, b) => Date.parse(b.CreationDate) - Date.parse(a.CreationDate));

    return {
        region,
        imageId: sorted[0].ImageId
    };
}

async function main() {
    console.log('Starting...');

    console.log('Getting all AWS regions...');
    const regions = await getRegions();
    console.log('Got regions:', regions);

    const promises = regions.map(async region => {
        try {
            return await getLatestAmazonLinuxAmi(region);
        } catch (err) {
            console.error(`Failed to get latest AMIs for region ${region}. Make sure that the region is enabled.`, err);
        }
    });
    const latestAmazonLinuxAMIs = compact(await Promise.all(promises));
    console.log('Got latest AMIs:', latestAmazonLinuxAMIs);

    const data = await generateData(latestAmazonLinuxAMIs);
    const regionMap = {};

    data.forEach(x => {
        regionMap[x.region] = {
            ami: x.ami,
            InstanceType: x.instanceType
        };
    });

    console.log(JSON.stringify(regionMap));
}

main();
