const {
    intersectionWith, pullAllWith, last
} = require('lodash');
const AWS = require('aws-sdk');
const moment = require('moment');
const { readFileSync } = require('fs-extra');

const CHANGE_LOG_BUCKET_NAME = 'eyaltest5';

function fillDependencyObjectsSet(dependencies, dependenciesSet) {
    Object.entries(dependencies).forEach(entry => {
        const [key, value] = entry;
        if (!value.dev) {
            const { dependencies: entryDependencies, version } = value;
            dependenciesSet.add(`${key}@${version}`);
            if (entryDependencies) {
                fillDependencyObjectsSet(entryDependencies, dependenciesSet);
            }
        }
    });
}

function createDependencyObjectsArray(packageLockJSON) {
    const { dependencies } = packageLockJSON;
    const dependenciesSet = new Set();
    fillDependencyObjectsSet(dependencies, dependenciesSet);
    return [...dependenciesSet];
}

async function getS3ObjectAsJSON(s3, bucketName, key) {
    const s3Object = await s3.getObject({ Bucket: bucketName, Key: key }).promise();
    const s3ObjectStringified = s3Object.Body.toString();
    return JSON.parse(s3ObjectStringified);
}

async function upload(s3, data, bucket, key) {
    return s3.upload({
        Bucket: bucket,
        Key: key,
        Body: data,
        ACL: 'public-read' // TODO Needed
    }).promise();
}

function comparePackageLockFiles(oldPackageLockJSON, newPackageLockJSON) {
    const oldDependencyObjectsArray = createDependencyObjectsArray(oldPackageLockJSON);
    const newDependencyObjectsArray = createDependencyObjectsArray(newPackageLockJSON);
    const equalDependencies = intersectionWith(oldDependencyObjectsArray, newDependencyObjectsArray);
    pullAllWith(oldDependencyObjectsArray, equalDependencies);
    pullAllWith(newDependencyObjectsArray, equalDependencies);
    return { Added: newDependencyObjectsArray, Deleted: oldDependencyObjectsArray };
}

async function uploadPackageLockChangelog(env, accessKeyId, secretAccessKey) {
    const s3 = new AWS.S3({ accessKeyId, secretAccessKey });
    const PACKAGE_LOCK_JSON = JSON.parse(readFileSync('../package-lock.json'));
    if (env !== 'dev') { return; } // TODO delete after testing
    const currentVersion = JSON.parse(readFileSync('../about.json')).version;
    const versionsObject = await getS3ObjectAsJSON(s3, CHANGE_LOG_BUCKET_NAME, `changelog/${env}/versions.json`);
    const lastVersion = last(versionsObject.VERSIONS).version;
    const lastPackageLockObject = await getS3ObjectAsJSON(s3, CHANGE_LOG_BUCKET_NAME, `changelog/${env}/${lastVersion}/package-lock.json`);
    const differences = comparePackageLockFiles(lastPackageLockObject, PACKAGE_LOCK_JSON);
    versionsObject.VERSIONS.push({ version: currentVersion, 'release-date': moment().format().slice(0, -6) });

    await upload(s3, JSON.stringify(differences), CHANGE_LOG_BUCKET_NAME, `changelog/${env}/${currentVersion}/difference.json`);
    await upload(s3, JSON.stringify(PACKAGE_LOCK_JSON), CHANGE_LOG_BUCKET_NAME, `changelog/${env}/${currentVersion}/package-lock.json`);
    await upload(s3, JSON.stringify(versionsObject), CHANGE_LOG_BUCKET_NAME, `changelog/${env}/versions.json`);
}

module.exports = uploadPackageLockChangelog;
