const config = require('config');
const Promise = require('bluebird');
const {
    AnonymousCredential, StorageSharedKeyCredential, BlobServiceClient
} = require('@azure/storage-blob');
const mime = require('mime-types');
const {
    last, compact, map, sortBy, zipWith, get
} = require('lodash');
const prettyBytes = require('bytes');
const {
    TYPE_FILE, TYPE_DIRECTORY, TYPE_SYMBOLIC, MOUNT_TARGET_SUFFIX, MOUNT_SOURCE_SUFFIX
} = require('../consts');
const { getMetadata } = require('../vault/metadata-store.js');

const getLogger = require('../logger/logger');

const logger = getLogger();

// SAS:
// connection string -> BlobEndpoint=https://ortestazurev10.blob.core.windows.net/;TableEndpoint=https://ortestazurev10.table.core.windows.net/;SharedAccessSignature=sv=2018-03-28&ss=b&srt=sco&sp=rwdlac&se=2019-08-28T19:56:10Z&st=2019-08-28T11:56:10Z&spr=https&sig=KgdHYR4B262eu3gzcShGTCQdPGoyzSEUKRnfQH83h98%3D
// blob service SAS URL -> https://ortestazurev10.blob.core.windows.net/?sv=2018-03-28&ss=b&srt=sco&sp=rwdlac&se=2019-08-28T19:56:10Z&st=2019-08-28T11:56:10Z&spr=https&sig=KgdHYR4B262eu3gzcShGTCQdPGoyzSEUKRnfQH83h98%3D
//
// Access key:
// connection string -> DefaultEndpointsProtocol=https;AccountName=ortestazurev10;AccountKey=NJlcynokuLetWI7+YjnihFsOwqpbI20yDMvSRG2DfRLJlCByb/DXbM9nUl7inN96X19bfUTu1I6ZzpdcrTBRJg==;EndpointSuffix=core.windows.net
// account name + key
//
// Managed identity:
//
// AD (service principal):
// tenant id + storage account name + client id + client secret

function isURL(string) {
    try {
        return new URL(string);
    } catch (err) {
        return false;
    }
}

function parseUrlString(string) {
    let credential;
    let url = string;

    if (isURL(string)) {
        // SAS URL
        credential = new AnonymousCredential();
    } else {
        const accountNamePattern = /AccountName=(.*?)(;|$)/;
        const accountKeyPattern = /AccountKey=(.*?)(;|$)/;

        const accountName = string.match(accountNamePattern);
        const accountKey = string.match(accountKeyPattern);

        if (accountName && accountKey) {
            // Shared key constructor
            credential = new StorageSharedKeyCredential(accountName[1], accountKey[1]);
            url = `https://${accountName[1]}.blob.core.windows.net`;
        }

        const blobEndPointPattern = /BlobEndpoint=(.*?)(;|$)/;
        const sharedAccessSignaturePattern = /SharedAccessSignature=(.*?)(;|$)/;

        const blobEndPoint = string.match(blobEndPointPattern);
        const sasEndPoint = string.match(sharedAccessSignaturePattern);

        if (blobEndPoint && sasEndPoint) {
            // build SAS URL
            credential = new AnonymousCredential();
            url = `${blobEndPoint[1]}?${sasEndPoint[1]}`;
        }
    }

    if (!credential) {
        throw new Error('Invalid credentials. Please check your credentials and try again');
    }
    return { url, credential };
}

function getBlobServiceClientByConnectionString(string) {
    logger.debug('Getting blob service client by connection string');

    const { url, credential } = parseUrlString(string);
    let proxyOptions;
    const proxy = process.env.https_proxy || process.env.http_proxy;

    if (proxy) {
        const regexPattern = /\/\/(?:(\w+):(\w+)@?)?(.*):(\d{1,5})/;
        const [, username, password, host, port] = regexPattern.exec(proxy);
        proxyOptions = {
            host, port, username, password
        };
    }

    return new BlobServiceClient(url, credential, { keepAliveOptions: { enable: false }, proxyOptions });
}

async function getBlobServiceClientByData(relationshipId, enumDirection) {
    logger.debug('Getting blob service client by data: ', { relationshipId, enumDirection });

    const { storageAccountConnectionString: metadata } = await getMetadata(`${relationshipId}.${enumDirection}.azure`);

    if (metadata) {
        return getBlobServiceClientByConnectionString(metadata);
    }

    throw new Error('Failed to find connection-string');
}

async function getBlobClient(relationshipId, enumDirection, containerName, blobName) {
    logger.debug('Getting blob client by data: ', { relationshipId, enumDirection });

    const blobServiceClient = await getBlobServiceClientByData(relationshipId, enumDirection);
    const containerClient = blobServiceClient.getContainerClient(containerName);

    return containerClient.getBlobClient(blobName);
}

async function getBlockBlobClient(relationshipId, enumDirection, containerName, blobName) {
    logger.debug('Getting block blob client by data: ', { relationshipId, enumDirection });

    const blobServiceClient = await getBlobServiceClientByData(relationshipId, enumDirection);
    const containerClient = blobServiceClient.getContainerClient(containerName);

    return containerClient.getBlockBlobClient(blobName);
}

async function listContainersSegmented(blobServiceClient) {
    logger.debug('List containers segmented');

    const containers = [];
    const iter = blobServiceClient.listContainers();

    for await (const container of iter) {
        containers.push(container.name);
    }

    return containers;
}

async function getContainerProperties(connectionString, containerName) {
    logger.debug('Get container properties: ', containerName);

    const blobServiceClient = getBlobServiceClientByConnectionString(connectionString);
    const containerClient = blobServiceClient.getContainerClient(containerName);

    return containerClient.getProperties();
}

async function listContainers(connectionString) {
    logger.debug('List containers');

    const blobServiceClient = getBlobServiceClientByConnectionString(connectionString);

    const containers = await listContainersSegmented(blobServiceClient);
    const properties = await Promise.all(containers.map(container => getContainerProperties(connectionString, container)));

    const result = zipWith(containers, properties, (container, property) => ({
        name: container,
        tags: map(property.metadata, (value, key) => ({ Key: key, Value: value }))
    }));

    return sortBy(result, 'name');
}

async function listBlobsByHierarchy(connectionString, containerName, prefix, options, dataKind) {
    logger.debug('List blobs by hierarchy: ', { containerName, prefix });

    const data = [];
    const blobServiceClient = getBlobServiceClientByConnectionString(connectionString);
    const containerClient = blobServiceClient.getContainerClient(containerName);
    const iter = containerClient.listBlobsByHierarchy('/',
        options);
    for await (const item of iter) {
        const { kind } = item;
        if (kind === dataKind) {
            data.push(dataKind === 'prefix' ? { name: item.name } : item);
        }
    }

    return data;
}

async function listPrefixByHierarchyByConnectionString(containerName, prefix, connectionString) {
    logger.debug('List prefix by hierarchy by connection string: ', { containerName, prefix });

    return listBlobsByHierarchy(connectionString, containerName, prefix, { prefix: prefix ? `${prefix}/` : prefix }, 'prefix');
}

async function listPrefixByHierarchyByMetadata(containerName, prefix, relationshipId, srcOrtrg) {
    logger.debug('List prefix by hierarchy by metadata: ', {
        containerName, prefix, relationshipId, srcOrtrg
    });
    const { storageAccountConnectionString: connectionString } = await getMetadata(`${relationshipId}.${srcOrtrg}.azure`);
    return listBlobsByHierarchy(connectionString, containerName, prefix, { prefix: prefix ? `${prefix}/` : prefix }, 'prefix');
}

async function listBlobsByHierarchyByMetaData(containerName, prefix, relationshipId, srcOrtrg) {
    logger.debug('List blob by hierarchy by metadata: ', {
        containerName, prefix, relationshipId, srcOrtrg
    });

    const { storageAccountConnectionString: connectionString } = await getMetadata(`${relationshipId}.${srcOrtrg}.azure`);

    return listBlobsByHierarchy(connectionString, containerName, prefix, {
        prefix: prefix ? `${prefix}/` : undefined,
        includeMetadata: true
    }, 'blob');
}

async function getBlobProperties(relationshipId, srcOrTrg, containerName, blobName) {
    logger.debug('Getting blob properties: ', {
        relationshipId, srcOrTrg, containerName, blobName
    });

    const blobClient = await getBlobClient(relationshipId, srcOrTrg, containerName, blobName);

    return blobClient.getProperties();
}

async function listBlobs(containerName, prefix, relationshipId, srcOrtrg) {
    logger.debug('List blobs: ', {
        containerName, prefix, relationshipId, srcOrtrg
    });

    const [directories, files] = await Promise.all([
        listPrefixByHierarchyByMetadata(containerName, prefix, relationshipId, srcOrtrg),
        listBlobsByHierarchyByMetaData(containerName, prefix, relationshipId, srcOrtrg)
    ]);
    logger.debug(`found ${directories.length} directories and ${files.length} files `);

    const filesMetadata = compact(files.map(file => {
        const path = last(file.name.split('/'));
        // blob might be without name - displayed as "<no name>" in azure console
        if (path) {
            return {
                path,
                type: get(file, 'metadata.netappcsslink') === 'True' ? TYPE_SYMBOLIC : TYPE_FILE,
                size: parseInt(file.properties.contentLength),
                mtimeSec: parseInt(get(file, 'metadata.netappcsmtimesec', new Date(file.properties.lastModified) / 1000)),
                mode: parseInt(get(file, 'metadata.netappcsmode', config.get('protocols.azure.default-file-mode'))),
                uid: parseInt(get(file, 'metadata.netappcsuid', 0)),
                gid: parseInt(get(file, 'metadata.netappcsgid', 0))
            };
        }
        return undefined;
    }));

    const directoriesMetadata = await Promise.map(directories, async directory => {
        let blobProperties;
        try {
            blobProperties = await getBlobProperties(relationshipId, srcOrtrg, containerName, directory.name);
        } catch (err) {
            if (err.statusCode !== '404') {
                logger.debug('Failed to get blob metadata', err);
            }
        }

        return {
            path: last(compact(directory.name.split('/'))),
            type: TYPE_DIRECTORY,
            size: 0,
            mtimeSec: blobProperties ? parseInt(blobProperties.metadata.netappcsmtimesec) || parseInt(new Date(blobProperties.lastModified) / 1000) : 0,
            mode: parseInt(get(blobProperties, 'metadata.netappcsmode', config.get('protocols.azure.default-directory-mode'))),
            uid: parseInt(get(blobProperties, 'metadata.netappcsuid', 0)),
            gid: parseInt(get(blobProperties, 'metadata.netappcsgid', 0))
        };
    }, { concurrency: 100 });

    return filesMetadata.concat(directoriesMetadata);
}

async function deleteBlob(containerName, blobName, relationshipId, mountSuffix) {
    logger.debug('Deleting blob: ', { containerName, blobName, relationshipId });

    const blobClient = await getBlobClient(relationshipId, mountSuffix, containerName, blobName);

    await blobClient.delete();
}

async function streamToString(readableStream) {
    return new Promise((resolve, reject) => {
        const chunks = [];
        readableStream.on('data', data => chunks.push(data.toString()));
        readableStream.on('end', () => resolve(chunks.join('')));
        readableStream.on('error', err => reject(err));
    });
}

async function createBlobReadStream(containerName, blobName, relationshipId, size, abortController) {
    logger.debug('Creating blob read stream: ', { containerName, blobName, relationshipId });

    const blobClient = await getBlobClient(relationshipId, MOUNT_SOURCE_SUFFIX, containerName, blobName);

    const { readableStreamBody } = await blobClient.download(0, size, { abortSignal: abortController });

    return readableStreamBody;
}

async function getBlobToText(containerName, blobName, relationshipId, abortController) {
    logger.debug('Getting blob to text: ', { containerName, blobName, relationshipId });
    const readableStreamBody = await createBlobReadStream(containerName, blobName, relationshipId, abortController);
    return streamToString(readableStreamBody);
}

async function createBlockBlobFromText(containerName, blobName, metadata, relationshipId, abortController, text = undefined) {
    logger.debug('Creating block blob from text: ', {
        containerName, blobName, text, metadata, relationshipId
    });

    const blockBlobClient = await getBlockBlobClient(relationshipId, MOUNT_TARGET_SUFFIX, containerName, blobName);

    // when blob name parameter ends with / <no name> blob is created inside of the folder
    return blockBlobClient.upload(
        text || '',
        text ? text.length : 0,
        { abortSignal: abortController, metadata }
    );
}

async function createWriteStreamToBlockBlobByFile(containerName, blobName, metadata, relationshipId, sourcePath, abortController) {
    logger.debug('Creating write stream to block blob by file: ', {
        containerName, blobName, metadata, relationshipId, sourcePath
    });

    const contentType = mime.lookup(blobName) || 'application/octet-stream';
    const blockBlobClient = await getBlockBlobClient(relationshipId, MOUNT_TARGET_SUFFIX, containerName, blobName);

    return blockBlobClient.uploadFile(
        sourcePath, { abortSignal: abortController, metadata, blobHTTPHeaders: { blobContentType: contentType } }
    );
}

async function createWriteStreamToBlockBlobByReadStream(containerName, blobName, metadata, relationshipId, readStream, abortController) {
    logger.debug('Creating write stream to block blob by read stream: ', {
        containerName, blobName, metadata, relationshipId
    });

    const contentType = mime.lookup(blobName) || 'application/octet-stream';
    const blockBlobClient = await getBlockBlobClient(relationshipId, MOUNT_TARGET_SUFFIX, containerName, blobName);

    return blockBlobClient.uploadStream(
        readStream, prettyBytes(config.get('protocols.azure.buffer-size')), config.get('protocols.azure.upload-stream-concurrency'), {
            abortSignal: abortController,
            metadata,
            blobHTTPHeaders: { blobContentType: contentType }
        }
    );
}

async function setBlobAccessTier(containerName, blobName, relationshipId, blobTier, abortController) {
    logger.debug('Setting blob tier: ', {
        containerName, blobName, relationshipId, blobTier
    });

    if (blobTier !== 'HOT') {
        const blobClient = await getBlobClient(relationshipId, MOUNT_TARGET_SUFFIX, containerName, blobName);

        return blobClient.setAccessTier(blobTier, { abortSignal: abortController });
    }
}

function createMetadata(requestMetadata, entry) {
    logger.debug('Creating metadata: ', { requestMetadata, entry });
    const metadata = {
        netappcsuid: entry.uid.toString(),
        netappcsgid: entry.gid.toString(),
        netappcsmode: entry.mode.toString(),
        netappcsmtimesec: entry.mtimeSec ? entry.mtimeSec.toString() : 'undefined'
    };

    if (entry.type === TYPE_SYMBOLIC) {
        metadata.netappcsslink = 'True';
    }

    if (requestMetadata) {
        requestMetadata.forEach(rm => {
            metadata[rm.Key] = rm.Value;
        });
    }

    return metadata;
}

function createSasUrl(storageConnectionString, targetContainerName, targetBlobName) {
    logger.debug('Creating SAS URL: ', { targetContainerName, targetBlobName });

    const sharedAccessSignaturePattern = /SharedAccessSignature=(.*?)(;|$)/;
    const blobEndPointPattern = /BlobEndpoint=(.*?)(;|$)/;
    const sas = sharedAccessSignaturePattern.exec(storageConnectionString) ? sharedAccessSignaturePattern.exec(storageConnectionString)[1] : undefined;

    if (!sas) {
        throw new Error('source connection-string should be SAS Connection-string and not regular connection-string');
    }
    const blobEndPoint = blobEndPointPattern.exec(storageConnectionString) ? blobEndPointPattern.exec(storageConnectionString)[1] : undefined;

    if (!blobEndPoint) {
        throw new Error('An error occurred- received invalid connection string, Expected a connection-string with BlobEndPoint');
    }

    return `${blobEndPoint}${targetContainerName}/${targetBlobName}?${sas}`;
}

async function copyBlobFromUrl(relationshipId, sourceContainerName, targetContainerName, sourceBlobName, targetBlobName) {
    logger.debug('Starting blob copy: ', {
        relationshipId,
        sourceContainerName,
        targetContainerName,
        sourceBlobName,
        targetBlobName
    });

    const blobClient = await getBlobClient(relationshipId, MOUNT_TARGET_SUFFIX, targetContainerName, targetBlobName);

    const { storageAccountConnectionString: sourceStorageConnectionString } = await getMetadata(`${relationshipId}.${MOUNT_SOURCE_SUFFIX}.azure`);

    const sourceSasUrl = createSasUrl(sourceStorageConnectionString, sourceContainerName, sourceBlobName);
    const copyPoller = await blobClient.beginCopyFromURL(sourceSasUrl, {
        onProgress(state) {
            logger.debug(`CopyId: ${state.copyId} progress: ${state.copyProgress}`);
        }
    });

    return copyPoller.pollUntilDone();
}

async function setBlobMetadata(relationshipId, srgOrTrg, containerName, blobName, metadata, abortController) {
    logger.debug('Setting blob meta data: ', {
        relationshipId, srgOrTrg, containerName, blobName, metadata
    });

    const blobClient = await getBlobClient(relationshipId, srgOrTrg, containerName, blobName);

    return blobClient.setMetadata(metadata, { abortSignal: abortController });
}

module.exports = {
    listContainers,
    listPrefixByHierarchyByConnectionString,
    listBlobs,
    deleteBlob,
    createBlobReadStream,
    getBlobToText,
    createWriteStreamToBlockBlobByFile,
    createWriteStreamToBlockBlobByReadStream,
    createMetadata,
    createBlockBlobFromText,
    copyBlobFromUrl,
    getBlobProperties,
    setBlobAccessTier,
    setBlobMetadata,
    getBlobServiceClientByData
};
