const config = require('config');
const { mkdirs } = require('fs-extra');
const { compact, trim } = require('lodash');
const ms = require('ms');
const os = require('os');
const getLogger = require('../logger/logger');
const execute = require('../executor');

const SHOWMOUNT_REGEX = /(^\S+)(\s+)(\S+)/;
const LIST_REGEX = /(^Disk\|)(.+)(\|)/gm;

const logger = getLogger();

// return kernel version and major version (<kernel version>.<major version> for example 4.5)
function getKernelVersion() {
    const release = os.release();
    const versionArray = release.split('.');
    if (isNaN(versionArray[0]) || isNaN(versionArray[1])) { // eslint-disable-line no-restricted-globals
        return undefined;
    }
    return parseFloat(`${versionArray[0]}.${versionArray[1]}`);
}

async function mountNFS(mountPoint, host, nfsExport, version) {
    let { 'mount-parameters': mountParameters, timeout } = config.get('protocols.nfs');
    await mkdirs(mountPoint);

    if (getKernelVersion() >= 5.3) { // nconnect header is featured only in kernel version of 5.3 or more
        mountParameters = 'nconnect=16,'.concat(mountParameters);
    }
    let protocolVersion = version;
    if (!version) {
        ({ 'default-version': protocolVersion } = config.get('protocols.nfs'));
    }

    const command = `mount ${host}:${nfsExport} ${mountPoint} -o ${mountParameters},vers=${protocolVersion}`;
    await execute(command, ms(timeout));
}

async function mountCIFS(mountPoint, host, share, version, username, password, workgroup = 'WORKGROUP') {
    await mkdirs(mountPoint);

    let protocolVersion = version;
    if (!version) {
        ({ 'default-version': protocolVersion } = config.get('protocols.cifs'));
    }

    const { timeout, 'default-file-mode': defaultFileMode, 'mount-parameters': defaultMountParameters } = config.get('protocols.cifs');
    const command = `mount -t cifs //${host}/'${trim(share, '/')}' ${mountPoint} -o domain=${workgroup},vers=${protocolVersion},file_mode=${defaultFileMode},${defaultMountParameters}`;
    await execute(command, ms(timeout), { PASSWD: password, USER: username });
}

async function umount(mountPoint) {
    const { timeout } = config.get('protocols.nfs');
    const command = `umount -f ${mountPoint}`;
    try {
        await execute(command, ms(timeout));
    } catch (e) {
        logger.warn(`Failed to umount ${mountPoint}`, e);
    }
}

async function showmount(host) {
    const { timeout } = config.get('protocols.nfs');
    const command = `showmount -e ${host} --no-header`;
    const exports = await execute(command, ms(timeout));
    const split = compact(exports.split('\n'));

    return split.map(entry => SHOWMOUNT_REGEX.exec(entry)[1]);
}

async function listShares(host, version, user, password, workgroup = 'WORKGROUP') {
    const { timeout, 'list-parameters': parameters } = config.get('protocols.cifs');

    // smbclient supports the following versions - NT1 (default), SMB2, SMB3
    const protocolVersionsMapping = {
        '1.0': 'NT1',
        '2.0': 'SMB2',
        '2.1': 'SMB2',
        '3.0': 'SMB3',
        '3.11': 'SMB3'
    };

    let protocolVersion;
    if (version && Object.keys(protocolVersionsMapping).includes(version)) {
        protocolVersion = protocolVersionsMapping[version];
    } else {
        const defaultVersion = config.get('protocols.cifs.default-version');
        protocolVersion = protocolVersionsMapping[defaultVersion];
    }

    const command = `smbclient --list=${host} ${parameters} --timeout=${ms(timeout) / 1000} --workgroup=${workgroup} --max-protocol=${protocolVersion}`;
    const output = await execute(command, ms(timeout), { PASSWD: password, USER: user });

    let matches;
    const disks = [];

    // eslint-disable-next-line no-cond-assign
    while (matches = LIST_REGEX.exec(output)) {
        LIST_REGEX.lastIndex += 1;
        disks.push(matches[2]);
    }

    return disks.sort();
}

module.exports = {
    mountNFS, mountCIFS, umount, showmount, listShares
};
