const { remove, rmdir } = require('fs-extra');
const { gracefulJoin } = require('./fs');
const { MOUNT_FOLDER } = require('../consts');
const { umount, mountNFS, mountCIFS } = require('./mount');
const { mountSecureNFS, umountSecureNFS } = require('../snfs/snfs');
const getLogger = require('../logger/logger');
const { getMetadata } = require('../vault/metadata-store.js');

const logger = getLogger();

function getMountPoint(relationshipId, requestId, partialSuffix) {
    const fullSuffix = partialSuffix ? `${requestId}_${partialSuffix}` : requestId;
    return gracefulJoin(MOUNT_FOLDER, relationshipId, fullSuffix);
}

function getNewEndPointIfNeeded(relationshipId, requestId, suffix, endPoint) {
    let path;

    if (endPoint && endPoint.nfs) {
        ({ nfs: { path } } = endPoint);
        const mountPoint = getMountPoint(relationshipId, requestId, suffix);
        path = gracefulJoin(mountPoint, path);
    } else if (endPoint && endPoint.cifs) {
        ({ cifs: { path } } = endPoint);
        const mountPoint = getMountPoint(relationshipId, requestId, suffix);
        path = gracefulJoin(mountPoint, path);
    }

    if (path) {
        return { local: { path } };
    }

    return endPoint;
}

async function mountIfNeeded(relationshipId, mountFolderName, mountSuffix, endPoint, endPointPassword) {
    let mountPoint;

    if (endPoint.nfs) {
        const {
            nfs: {
                host, export: nfsExport, version
            }
        } = endPoint;
        mountPoint = getMountPoint(relationshipId, mountFolderName, mountSuffix);
        await mountNFS(mountPoint, host, nfsExport, version);
    } else if (endPoint.snfs) {
        const {
            snfs: {
                serverIP, port, certificate, serverName, host, export: nfsExport, version
            }
        } = endPoint;

        await mountSecureNFS(relationshipId, serverIP, port, certificate, serverName, mountFolderName, host, nfsExport, version);
    } else if (endPoint.cifs) {
        let password = endPointPassword;

        let {
            cifs: {
                host, share, version, username, workgroup
            }
        } = endPoint;
        mountPoint = getMountPoint(relationshipId, mountFolderName, mountSuffix);

        if (!username || !password) {
            const credentials = await getMetadata(`${relationshipId}.${mountSuffix}.cifs`);
            ({ username, password, workgroup } = credentials);
        }

        await mountCIFS(mountPoint, host, share, version, username, password, workgroup);
    }

    return mountPoint;
}

async function umountAndDelete(mountPoint) {
    await umount(mountPoint);
    try {
        await rmdir(mountPoint);
    } catch (err) {
        logger.warn(`Failed to rmdir ${mountPoint}`, err);
    }
}

async function umountAndDeleteRelationshipFolder(relationshipId, sourceEndpoint, targetEndpoint) {
    if (sourceEndpoint.snfs) {
        await umountSecureNFS(relationshipId, sourceEndpoint.snfs.serverIP,
            sourceEndpoint.snfs.port, sourceEndpoint.snfs.certificate, sourceEndpoint.snfs.serverName);
    }

    if (targetEndpoint.snfs) {
        await umountSecureNFS(relationshipId, targetEndpoint.snfs.serverIP,
            targetEndpoint.snfs.port, targetEndpoint.snfs.certificate, targetEndpoint.snfs.serverName);
    }

    await umount(gracefulJoin(MOUNT_FOLDER, relationshipId, '*'));
    await remove(gracefulJoin(MOUNT_FOLDER, relationshipId));
}

module.exports = {
    getNewEndPointIfNeeded, mountIfNeeded, umountAndDeleteRelationshipFolder, umountAndDelete, getMountPoint
};
