const config = require('config');
const { Storage } = require('@google-cloud/storage');
const mimeType = require('mime-types');
const {
    get, keys, last, compact, isNil, sortBy
} = require('lodash');
const {
    TYPE_FILE, TYPE_DIRECTORY, TYPE_SYMBOLIC
} = require('../consts');
const getLogger = require('../logger/logger');

const logger = getLogger();

let storageInstance;

function getStorage() {
    if (!storageInstance) {
        storageInstance = new Storage();
        storageInstance.interceptors.push({
            request(reqOpts) {
                reqOpts.forever = config.get('protocols.gcp.forever');
                return reqOpts;
            }
        });
    }
    return storageInstance;
}

async function getBucketLocationAndTags(bucketName) {
    const storage = getStorage();
    const bucket = storage.bucket(bucketName);
    const metadata = await bucket.getMetadata();
    const { location, labels: tags } = metadata[0];
    const tagsArr = tags ? keys(tags).map(key => ({ Key: key, Value: tags[key] })) : [];

    return { location, tags: tagsArr };
}

async function listGcpBuckets() {
    const storage = getStorage();
    const buckets = await storage.getBuckets();
    const promises = buckets[0].map(async bucket => {
        try {
            const { location, tags } = await getBucketLocationAndTags(bucket.name);
            return { name: bucket.name, region: location.toLowerCase(), tags };
        } catch (err) {
            logger.warn(`Failed to get bucket location/tagging, bucket name ${bucket.name}`, err);
            return { name: bucket.name, region: '', tags: [] };
        }
    });

    const bucketsResponse = await Promise.all(promises);
    return sortBy(bucketsResponse, 'name');
}

async function deleteGcpObject(bucket, file) {
    const storage = getStorage();
    return storage.bucket(bucket).file(file).delete();
}

function getGcpReadStream(bucket, file) {
    const storage = getStorage();
    return storage.bucket(bucket).file(file).createReadStream({ validation: false });
}

function getGcpWriteStream(bucket, file, tags) {
    const contentType = mimeType.lookup(file) || 'application/octet-stream';
    const storage = getStorage();
    const options = {
        resumable: false,
        metadata: {
            contentType,
            metadata: tags
        }
    };
    return storage.bucket(bucket).file(file).createWriteStream(options);
}

async function isObjectExists(bucketName, fileName) {
    const storage = getStorage();
    const objectExist = await storage.bucket(bucketName).file(fileName).exists();
    return objectExist[0];
}

async function getGcpObjectTagging(bucketName, fileName, objectType) {
    const storage = getStorage();
    let noNameFileExists = false;
    const noNameFile = '<no name>';

    // the purpose of the if is to avoid redundant exists() calls in case of TYPE_FILE/SYMBOLIC
    if (objectType === TYPE_DIRECTORY) { // if the object is a directory - we need to check if there's a <no name> file that contains the directory's permissions
        noNameFileExists = await isObjectExists(bucketName, `${fileName}/${noNameFile}`);
    }
    const file = noNameFileExists ? storage.bucket(bucketName).file(`${fileName}/${noNameFile}`) : storage.bucket(bucketName).file(fileName);

    try {
        const metadata = await file.getMetadata();
        const tags = metadata[0].metadata;
        return tags ? keys(tags).map(key => ({ Key: key, Value: tags[key] })) : [];
    } catch (err) {
        if (err.code === 404) {
            return [];
        }
        throw err;
    }
}

async function createGcpFolder(bucketName, folderName, tags) {
    const contentType = mimeType.lookup(folderName) || 'application/octet-stream';
    const storage = getStorage();
    const metadata = {
        contentType,
        metadata: tags
    };
    await storage.bucket(bucketName).file(`${folderName}/`).save('');
    await storage.bucket(bucketName).file(`${folderName}/<no name>`).save('', { metadata });
}

async function createGcpSymlink(bucketName, fileName, fileContent, tags) {
    const contentType = mimeType.lookup(fileName) || 'application/octet-stream';
    const storage = getStorage();
    const options = {
        metadata: {
            contentType,
            metadata: tags
        }
    };
    return storage.bucket(bucketName).file(fileName).save('', options);
}

async function getGcpObjectMetadata(bucketName, fileName) {
    const storage = getStorage();
    const noNameFileExists = await isObjectExists(bucketName, `${fileName}/<no name>`);
    const file = noNameFileExists ? storage.bucket(bucketName).file(`${fileName}/<no name>`) : storage.bucket(bucketName).file(fileName);
    const result = await file.getMetadata();
    return result[0];
}

function createGcpMetadata(requestMetadata, entry, symlinkContent) {
    const metadata = {
        'netapp-cs-uid': entry.uid,
        'netapp-cs-gid': entry.gid,
        'netapp-cs-mode': entry.mode,
        'netapp-cs-mtimeSec': entry.mtimeSec
    };

    if (entry.type === TYPE_SYMBOLIC) {
        metadata['netapp-cs-slink'] = 'True';
        metadata.symlink = symlinkContent;
    }

    if (requestMetadata) {
        requestMetadata.forEach(rm => {
            metadata[rm.Key] = rm.Value;
        });
    }

    return metadata;
}

async function getGcpObjectsAndTags(bucket, key, size, objectType, mode, lastModified) {
    const tagsArray = await getGcpObjectTagging(bucket, key, objectType);
    const tagsObject = tagsArray.reduce((accumulator, currentValue) => (Object.assign(accumulator, { [currentValue.Key]: currentValue.Value })), {});
    const modifiedTime = lastModified === undefined ? Date.now() : lastModified.getTime();

    if (!key.endsWith('/') || objectType !== TYPE_FILE) {
        return {
            path: key === '/' ? '/' : last(compact(key.split('/'))),
            type: tagsObject['netapp-cs-slink'] === 'True' ? TYPE_SYMBOLIC : objectType,
            size: parseInt(size),
            mtimeSec: parseInt(tagsObject['netapp-cs-mtimeSec']) || parseInt(modifiedTime / 1000),
            mode: parseInt(tagsObject['netapp-cs-mode']) || mode,
            uid: parseInt(tagsObject['netapp-cs-uid']) || 0,
            gid: parseInt(tagsObject['netapp-cs-gid']) || 0
        };
    }

    return null;
}

async function listGcpBucketFolders(bucketName, prefix, nextQuery, dirs = []) {
    const storage = getStorage();
    const bucket = storage.bucket(bucketName);
    let fixedPrefix = isNil(prefix) || prefix === '' ? undefined : prefix.replace(/^\//, '');
    if (fixedPrefix) {
        fixedPrefix = fixedPrefix.endsWith('/') ? fixedPrefix : `${fixedPrefix}/`;
    }

    const response = await bucket.getFiles({
        prefix: fixedPrefix, autoPaginate: false, pageToken: nextQuery, delimiter: '/'
    });
    const responseDirs = response[2].prefixes ? response[2].prefixes : [];
    const updatedDirs = compact(dirs.concat(compact(responseDirs.map(dir => dir.replace(fixedPrefix, '').replace(/\/$/, '')))));
    if (response[2].nextPageToken) {
        return listGcpBucketFolders(bucketName, fixedPrefix, response[2].nextPageToken, updatedDirs);
    }

    return updatedDirs;
}

async function listGcpObjects(bucketName, prefix, nextQuery, dirsAndFiles = []) {
    const storage = getStorage();
    const bucket = storage.bucket(bucketName);
    let fixedPrefix = isNil(prefix) || prefix === '' ? undefined : prefix;
    if (fixedPrefix) {
        fixedPrefix = fixedPrefix.endsWith('/') ? fixedPrefix : `${fixedPrefix}/`;
    }

    const response = await bucket.getFiles({
        prefix: fixedPrefix, autoPaginate: false, pageToken: nextQuery, delimiter: '/'
    });
    const responseFiles = compact(response[0].map(file => file.name.replace(fixedPrefix, '').replace(/\/$/, '')).filter(file => file !== '<no name>'));
    let responseDirs = response[2].prefixes ? response[2].prefixes : [];
    responseDirs = compact(responseDirs.map(dir => dir.replace(fixedPrefix, '').replace(/\/$/, '')));

    const filesPromises = responseFiles.map(async file => {
        const filePath = fixedPrefix ? fixedPrefix + file : file;
        const metadata = await getGcpObjectMetadata(bucketName, filePath);
        const mtimeSec = new Date(get(metadata, 'metadata.netapp-cs-mtimeSec', get(metadata, 'updated', '')));
        return getGcpObjectsAndTags(bucketName, filePath, metadata.size, TYPE_FILE,
            config.get('protocols.gcp.default-file-mode'), mtimeSec);
    });

    const directoriesPromises = responseDirs.map(async dir => {
        const dirPath = fixedPrefix ? fixedPrefix + dir : dir;
        return getGcpObjectsAndTags(bucketName, dirPath, 0, TYPE_DIRECTORY, config.get('protocols.gcp.default-directory-mode'));
    });

    const result = await Promise.all(filesPromises.concat(compact(directoriesPromises)));
    const updatedDirsAndFiles = compact(dirsAndFiles.concat(compact(result)));
    if (response[1] !== null && 'pageToken' in response[1]) {
        return listGcpObjects(bucketName, prefix, response[1].pageToken, updatedDirsAndFiles);
    }

    return updatedDirsAndFiles;
}

async function isGcpBucketExists(bucketName) {
    const storage = getStorage();
    const bucket = storage.bucket(bucketName);
    const result = await bucket.exists();
    if (result[0] === false) {
        throw new Error(`Bucket ${bucketName} doesn't exist`);
    }
    return true;
}

async function copyGcpObject(sourceBucketName, sourcePath, targetBucketName, targetPath) {
    const storage = getStorage();
    const destination = storage.bucket(targetBucketName).file(targetPath);
    await storage
        .bucket(sourceBucketName)
        .file(sourcePath)
        .copy(destination);
}

async function setMetadata(bucket, path, metadata) {
    const storage = getStorage();
    await storage.bucket(bucket).file(path).setMetadata({ metadata });
}

module.exports = {
    getStorage,
    listGcpBuckets,
    deleteGcpObject,
    getGcpReadStream,
    getGcpWriteStream,
    createGcpSymlink,
    createGcpFolder,
    createGcpMetadata,
    listGcpBucketFolders,
    listGcpObjects,
    isGcpBucketExists,
    getBucketLocationAndTags,
    getGcpObjectMetadata,
    isObjectExists,
    copyGcpObject,
    setMetadata
};
