const moment = require('moment');
const { isValidSSLCert, validateSSLCert } = require('ssl-validator');
const ms = require('ms');
const {
    mkdirs, remove, readFile, pathExists
} = require('fs-extra');
const { CERT_AND_KEY_FOLDER, DATA_BROKER_RUNTIME_CONFIG } = require('../consts');
const sleep = require('../../utils/utils');
const getLogger = require('../logger/logger');
const execute = require('../executor');

const logger = getLogger();

async function createCertAndRestartVault(databrokerId, keyPath, certPath) {
    const cmd = `openssl req -newkey rsa:2048 -subj '/CN=${databrokerId}' -nodes -keyout ${keyPath} -x509 -days 365 -out ${certPath}`;
    await execute(cmd);
    // after generating certificate, we need to restart the vault
    await execute('pm2 restart vault');
    await sleep(ms('5s')); // give some time for vault to reboot
}

async function generateCertificate() {
    try {
        const { 'data-broker-id': databrokerId } = DATA_BROKER_RUNTIME_CONFIG;
        const certPath = `${CERT_AND_KEY_FOLDER}/dataBrokerCert.crt`;

        const isCertExists = await pathExists(certPath);
        if (isCertExists) {
            const currentCert = await readFile(certPath);
            const [isCertValid, certInfo] = await Promise.all([isValidSSLCert(currentCert), validateSSLCert(currentCert, { skipDateValidation: true })]);
            const threeMonthsInMs = moment.duration(3, 'months').asMilliseconds();
            const now = Date.now();
            // check if certificate expiration date is less then 2 months from now.
            if (!isCertValid || (certInfo.validity.end - now < threeMonthsInMs)) {
                logger.info('*** Certificate is invalid or about to expire, generating new one.');
                await remove(certPath);
                const keyPath = `${CERT_AND_KEY_FOLDER}/dataBrokerKey.key`;
                await remove(keyPath);
                await createCertAndRestartVault(databrokerId, keyPath, certPath);
            }
        } else {
            // at this point, certificate does not exists in initiator/regular data broker, and also in listener data broker (because we changed the name)
            // after the update (August 2020), all data brokers must have certificate and key.
            // making sure we create certificate directory if not exists
            await mkdirs(CERT_AND_KEY_FOLDER);
            const keyPath = `${CERT_AND_KEY_FOLDER}/dataBrokerKey.key`;
            await createCertAndRestartVault(databrokerId, keyPath, certPath);
        }
    } catch (error) {
        logger.error('Could not generate certificate', error);
        throw error; // data broker does not deployed if certificate fails.
    }
}

module.exports = generateCertificate;
