const config = require('config');
const AWS = require('aws-sdk');
const proxy = require('proxy-agent');
const https = require('https');
const http = require('http');
const {
    ONTAP_S3_PROVIDER,
    SQS_DEFAULT_REGION, OSS_DEFAULT_REGION, IBM_PROVIDER, SGWS_PROVIDER, ALIBABA_PROVIDER, S3_PROVIDER
} = require('./consts');
const { getMetadata } = require('./vault/metadata-store.js');

/*
const getLogger = require('../lib/logger/logger');

const logger = getLogger();
*/

function createHttpOptions(provider, host) {
    const useProxy = config.get(`protocols.${provider}.use-proxy`);
    const rejectInvalidCertificates = config.get(`protocols.${provider}.reject-invalid-certificates`);
    const maxVersion = config.get(`protocols.${provider}.max-tls-version`) ? config.get(`protocols.${provider}.max-tls-version`) : undefined;
    let agent;
    if (process.env.http_proxy && useProxy) {
        agent = proxy(process.env.http_proxy, { rejectUnauthorized: rejectInvalidCertificates, maxVersion });
    } else if ((host && host.startsWith('http://') || provider === ONTAP_S3_PROVIDER)) {
        agent = new http.Agent();
    } else {
        agent = new https.Agent({ rejectUnauthorized: rejectInvalidCertificates, maxVersion });
    }
    return { agent };
}

function getS3(accessKeyId, secretAccessKey, region = SQS_DEFAULT_REGION, sessionToken) {
    const endpoint = config.get('protocols.s3.endpoint');
    const httpOptions = createHttpOptions('s3');

    return new AWS.S3({
        accessKeyId,
        secretAccessKey,
        sessionToken,
        region,
        correctClockSkew: true,
        useDualstack: config.get('protocols.s3.use-dual-stack'),
        httpOptions,
        endpoint
    });
}

async function getSgws(host, port, accessKey, secretKey, isS3EndPoint, relationshipId, srcOrTrg) {
    let accessKeyId = accessKey;
    let secretAccessKey = secretKey;
    const httpOptions = createHttpOptions(SGWS_PROVIDER, host);

    if (!accessKeyId || !secretAccessKey) {
        const credentials = await getMetadata(`${relationshipId}.${srcOrTrg}.sgws`);
        ({ accessKey: accessKeyId, secretKey: secretAccessKey } = credentials);
    }

    return new AWS.S3({
        endpoint: (!port || port === 443 || port === '443') ? `${host}` : `${host}:${port}`,
        accessKeyId,
        secretAccessKey,
        httpOptions,
        s3BucketEndpoint: isS3EndPoint,
        s3ForcePathStyle: true,
        correctClockSkew: true,
        retryDelayOptions: {
            base: 1000
        }
    });
}

async function getIbm(host, port, accessKey, secretKey, isS3EndPoint, relationshipId, srcOrtrg) {
    let accessKeyId = accessKey;
    let secretAccessKey = secretKey;
    const httpOptions = createHttpOptions(IBM_PROVIDER, host);

    if (!accessKeyId || !secretAccessKey) {
        let credentials = await getMetadata(`${relationshipId}.${srcOrtrg}.ibm`);

        if (!credentials) {
            // for backward compatibility - ibm was once the same as sgws
            credentials = await getMetadata(`${relationshipId}.${srcOrtrg}.sgws`);
        }

        ({ accessKey: accessKeyId, secretKey: secretAccessKey } = credentials);
    }

    return new AWS.S3({
        endpoint: `${host}:${port}`,
        accessKeyId,
        secretAccessKey,
        httpOptions,
        s3BucketEndpoint: isS3EndPoint,
        s3ForcePathStyle: true,
        correctClockSkew: true
    });
}

async function getOntapS3(host, accessKey, secretKey, relationshipId, srcOrtrg, protocol) {
    let accessKeyId = accessKey;
    let secretAccessKey = secretKey;
    let endpoint = host;
    const httpOptions = createHttpOptions(ONTAP_S3_PROVIDER, host);

    if (!accessKeyId || !secretAccessKey) {
        const credentials = await getMetadata(`${relationshipId}.${srcOrtrg}.ontap`);
        ({ accessKey: accessKeyId, secretKey: secretAccessKey } = credentials);
    }
    if (protocol) {
        if (host.startsWith('http://')) {
            endpoint = host.replace('http://', 'https://');
        } else if (!host.includes('https://')) {
            endpoint = `https://${host}`;
        }
    }

    return new AWS.S3({
        endpoint,
        accessKeyId,
        secretAccessKey,
        httpOptions,
        s3ForcePathStyle: true,
        s3BucketEndpoint: false,
        sslEnabled: false,
        isS3EndPoint: false,
        correctClockSkew: true
    });
}

async function getAlibaba(accessKey, secretKey, region = OSS_DEFAULT_REGION, relationshipId, srcOrTrg) {
    let accessKeyId = accessKey;
    let secretAccessKey = secretKey;
    const httpOptions = createHttpOptions(ALIBABA_PROVIDER);

    if (!accessKeyId || !secretAccessKey) {
        const credentials = await getMetadata(`${relationshipId}.${srcOrTrg}.alibaba`);
        ({ accessKey: accessKeyId, secretKey: secretAccessKey } = credentials);
    }

    return new AWS.S3({
        endpoint: `${region || 'oss-me-east-1'}.aliyuncs.com`,
        accessKeyId,
        secretAccessKey,
        httpOptions,
        s3ForcePathStyle: false,
        correctClockSkew: true
    });
}

async function getClient(provider, options = {}) {
    switch (provider) {
        case S3_PROVIDER: {
            const {
                region, secretAccessKey, accessKeyId, sessionToken
            } = options;
            return getS3(accessKeyId, secretAccessKey, region, sessionToken);
        }
        case ALIBABA_PROVIDER: {
            const {
                accessKey, secretKey, region, relationshipId, srcOrTrg
            } = options;
            return getAlibaba(accessKey, secretKey, region, relationshipId, srcOrTrg);
        }
        case SGWS_PROVIDER: {
            const {
                host, port, accessKey, secretKey, isS3EndPoint, relationshipId, srcOrTrg
            } = options;
            return getSgws(host, port, accessKey, secretKey, isS3EndPoint, relationshipId, srcOrTrg);
        }
        case IBM_PROVIDER: {
            const {
                host, port, accessKey, secretKey, isS3EndPoint, relationshipId, srcOrTrg
            } = options;
            return getIbm(host, port, accessKey, secretKey, isS3EndPoint, relationshipId, srcOrTrg);
        }
        case ONTAP_S3_PROVIDER: {
            const {
                host, accessKey, secretKey, relationshipId, srcOrTrg, protocol
            } = options;
            return getOntapS3(host, accessKey, secretKey, relationshipId, srcOrTrg, protocol);
        }
        default:
            throw new Error(`unknown provider ${JSON.stringify(provider)}`);
    }
}

module.exports = {
    getSgws, getS3, getClient, getAlibaba, getIbm
};
