const http2 = require('http2');
const Promise = require('bluebird');
const config = require('config');

const {
    HTTP2_HEADER_METHOD,
    HTTP2_HEADER_PATH,
    HTTP2_METHOD_POST,
    HTTP2_HEADER_STATUS
} = http2.constants;

const {
    OPER_LIST_WITH_STATS, OPER_GET_FILE, OPER_GET_SYMLINK, OPER_UPLOAD_FILE, OPER_UPLOAD_SYMLINK, OPER_UPLOAD_DIR,
    OPER_MOUNT, OPER_UMOUNT, OPER_DELETE_FILE, OPER_SECURE_MKDIRS
} = require('../consts');
const getLogger = require('../logger/logger');

const logger = getLogger();

function buildRequest(client, operation, relId, body) {
    logger.debug(`Building request: operation: ${operation}, relationshipId: ${relId}, body: `, body);
    let request;
    switch (operation) {
        case OPER_GET_SYMLINK:
        case OPER_DELETE_FILE:
        case OPER_GET_FILE:
        case OPER_SECURE_MKDIRS:
        case OPER_LIST_WITH_STATS: {
            const { requestId, path } = body;
            request = client.request({
                [HTTP2_HEADER_PATH]: `/${relId}`,
                [HTTP2_HEADER_METHOD]: HTTP2_METHOD_POST,
                operation,
                'request-id': requestId,
                path
            });
            break;
        }
        case OPER_UPLOAD_SYMLINK:
        case OPER_UPLOAD_DIR:
        case OPER_UPLOAD_FILE: {
            const { requestId, path, entry } = body;
            const entryValue = JSON.stringify(entry);
            request = client.request({
                [HTTP2_HEADER_PATH]: `/${relId}`,
                [HTTP2_HEADER_METHOD]: HTTP2_METHOD_POST,
                operation,
                'request-id': requestId,
                path,
                entry: entryValue
            });
            break;
        }
        case OPER_MOUNT: {
            const {
                requestId, host, nfsExport, version
            } = body;
            request = client.request({
                [HTTP2_HEADER_PATH]: `/${relId}`,
                [HTTP2_HEADER_METHOD]: HTTP2_METHOD_POST,
                operation,
                'request-id': requestId,
                'source-ip': host,
                'nfs-export': nfsExport,
                version
            });
            break;
        }
        case OPER_UMOUNT: {
            request = client.request({
                [HTTP2_HEADER_PATH]: `/${relId}`,
                [HTTP2_HEADER_METHOD]: HTTP2_METHOD_POST,
                operation
            });
            break;
        }
        default:
            throw new Error(`Operation ${operation} is unknown. Can't build request.`);
    }
    return request;
}

function connectToServer(serverIP, port, certificate, serverName) {
    // on client side - http2.connect method creates Http2Session instance.

    let listenerIp;
    try {
        listenerIp = config.get('snfs.listener-ip');
    } catch (e) {
        // do nothing
    }

    const finalIp = listenerIp || serverIP;
    return http2.connect(`https://${finalIp}:${port}`, {
        ca: certificate,
        servername: serverName
    });
}

async function getData(stream) {
    return new Promise((resolve, reject) => {
        let data = '';
        stream.on('data', chunk => {
            data += chunk;
        });
        stream.on('end', () => {
            resolve(data);
        });
        stream.on('error', err => {
            logger.error(`Error while receiving data from server. error: ${err}`);
            reject(new Error(err));
        });
    });
}

async function checkResponse(request, client) {
    return new Promise((resolve, reject) => {
        client.on('error', err => {
            client.close();
            reject(new Error(`Client session error: ${err}`));
        });

        request.on('response', async headers => {
            logger.debug(`Got response from server.\n request: ${request} STATUS CODE: ${headers[HTTP2_HEADER_STATUS]}.`);
            if (headers[HTTP2_HEADER_STATUS] !== 200 && headers[HTTP2_HEADER_STATUS] !== 202) {
                const errorFromServer = await getData(request);
                switch (headers[HTTP2_HEADER_STATUS]) {
                    case 500:
                    case 400:
                        client.close();
                        reject(JSON.parse(errorFromServer));
                        break;
                    default:
                        reject(new Error('Received unknown status code from server.'));
                }
            } else {
                resolve();
            }
        });
    });
}

async function listSecureNFS(relationshipId, serverIP, port, certificate, serverName, requestId, path) {
    const client = connectToServer(serverIP, port, certificate, serverName);
    const request = buildRequest(client, OPER_LIST_WITH_STATS, relationshipId,
        { requestId, path });
    request.on('error', err => {
        logger.error('Request had an error:', err);
        client.close();
    });
    await checkResponse(request, client);
    const data = await getData(request);
    client.close();
    return data ? JSON.parse(data) : [];
}

async function getSecureReadStream(relationshipId, client, requestId, path) {
    const serverStream = buildRequest(client, OPER_GET_FILE, relationshipId, { requestId, path });
    serverStream.on('error', err => {
        logger.error('Request had an error:', err);
        client.close();
    });
    await checkResponse(serverStream, client);
    return serverStream;
}

async function getSecureUploadStream(relationshipId, client, requestId, path, entry) {
    const uploadStream = buildRequest(client, OPER_UPLOAD_FILE, relationshipId, { requestId, path, entry });
    uploadStream.on('error', err => {
        logger.error('Request had an error:', err);
        client.close();
    });
    await checkResponse(uploadStream, client);
    return uploadStream;
}

async function uploadDirectory(relationshipId, serverIP, port, certificate, serverName, requestId, path, entry) {
    const client = connectToServer(serverIP, port, certificate, serverName);
    const request = buildRequest(client, OPER_UPLOAD_DIR, relationshipId, { requestId, path, entry });
    request.end();
    request.on('error', err => {
        logger.error('Request had an error:', err);
        client.close();
    });
    await checkResponse(request, client);
    client.close();
}

async function secureMkdirs(relationshipId, serverIP, port, certificate, serverName, requestId, path) {
    const client = connectToServer(serverIP, port, certificate, serverName);
    const request = buildRequest(client, OPER_SECURE_MKDIRS, relationshipId, { requestId, path });
    request.end();
    request.on('error', err => {
        logger.error('Request had an error:', err);
        client.close();
    });
    await checkResponse(request, client);
    client.close();
}

async function deleteFile(relationshipId, serverIP, port, certificate, serverName, requestId, path) {
    const client = connectToServer(serverIP, port, certificate, serverName);
    const request = buildRequest(client, OPER_DELETE_FILE, relationshipId, { requestId, path });
    request.end();
    await checkResponse(request, client);
    client.close();
}

async function getSecureSymlinkContent(relationshipId, serverIP, port, certificate, serverName, requestId, path) {
    const client = connectToServer(serverIP, port, certificate, serverName);
    const request = buildRequest(client, OPER_GET_SYMLINK, relationshipId, { requestId, path });
    request.on('error', err => {
        logger.error('Request had an error:', err);
        client.close();
    });
    await checkResponse(request, client);
    const data = await getData(request);
    client.close();
    return data;
}

async function uploadSymlink(relationshipId, serverIP, port, certificate, serverName, requestId, path, entry, symlinkContent) {
    const client = connectToServer(serverIP, port, certificate, serverName);
    const request = buildRequest(client, OPER_UPLOAD_SYMLINK, relationshipId, {
        requestId, path, entry
    });
    request.end(symlinkContent);
    request.on('error', err => {
        logger.error('Request had an error:', err);
        client.close();
    });
    await checkResponse(request, client);
    client.close();
}

async function mountSecureNFS(relationshipId, serverIP, port, certificate, serverName, requestId, host, nfsExport, version) {
    const client = connectToServer(serverIP, port, certificate, serverName);
    const request = buildRequest(client, OPER_MOUNT, relationshipId, {
        requestId, host, nfsExport, version
    });
    request.on('error', err => {
        logger.error('Request had an error:', err);
        client.close();
    });
    await checkResponse(request, client);
    client.close();
}

async function umountSecureNFS(relationshipId, serverIP, port, certificate, serverName) {
    const client = connectToServer(serverIP, port, certificate, serverName);
    const request = buildRequest(client, OPER_UMOUNT, relationshipId);
    request.on('error', err => {
        logger.error('Request had an error:', err);
        client.close();
    });
    await checkResponse(request, client);
    client.close();
}

module.exports = {
    secureMkdirs,
    deleteFile,
    uploadSymlink,
    uploadDirectory,
    getSecureUploadStream,
    connectToServer,
    mountSecureNFS,
    listSecureNFS,
    getSecureReadStream,
    umountSecureNFS,
    getSecureSymlinkContent
};
