const axios = require('axios');
const config = require('config');
const {
    head, keys, get
} = require('lodash');
const { readFile } = require('fs-extra');
const retry = require('async-retry');
const { writeFile } = require('fs-extra');
const https = require('https');
const AWS = require('aws-sdk');
const aws4 = require('aws4');
const getLogger = require('../logger/logger');

const logger = getLogger();

const LOCAL_VAULT_BASE_URL = 'https://127.0.0.1:8200';
const VAULT_BASE_URL = config.get('external-integrations.hashicorp.url');
const VAULT_ROLE_ID = config.get('external-integrations.hashicorp.app-role.root_id');
const VAULT_SECRET_ID = config.get('external-integrations.hashicorp.app-role.secret_id');
const VAULT_PATH = config.get('external-integrations.hashicorp.path');
const VAULT_NAMESPACE = config.get('external-integrations.hashicorp.namespace');
const VAULT_ROLE_NAME = config.get('external-integrations.hashicorp.aws-iam.role-name');
const VAULT_AUTH_METHOD = config.get('external-integrations.hashicorp.auth-method');
const DATA_BROKER_SECRET_ENGINE_PATH = 'secret';
const AWS_REQUEST_BODY = 'Action=GetCallerIdentity&Version=2011-06-15';
const VAULT_KEYS_FILE = './vault/keys.json';
const VALID_HEALTH_STATUSES = [200, 429, 472, 473, 501, 503]; // https://www.vaultproject.io/api-docs/system/health

// https agent for local vault
const localAgent = new https.Agent({
    rejectUnauthorized: false
});

const externalAgent = new https.Agent({
    rejectUnauthorized: config.get('external-integrations.hashicorp.reject-unauthorized')
});

// region Vault External Integration
async function getVaultHealthStatus() {
    logger.debug('Getting Vault health status');

    const { data } = await axios.get(`${VAULT_BASE_URL}/v1/sys/health`, {
        httpsAgent: externalAgent,
        validateStatus: status => (status >= 200 && status < 300) || VALID_HEALTH_STATUSES.includes(status)
    });

    return data;
}

async function appRoleLogin() {
    logger.debug('Login with role, secret and namespace:', { VAULT_ROLE_ID, VAULT_NAMESPACE });
    const path = `${VAULT_BASE_URL}/v1/auth/approle/login`;

    const { data: { auth } } = await axios.post(path,
        {
            role_id: VAULT_ROLE_ID,
            secret_id: VAULT_SECRET_ID
        }, {
            httpsAgent: externalAgent,
            headers: {
                'X-Vault-Namespace': VAULT_NAMESPACE
            }
        });

    return auth;
}

async function getProviderCredentials(clientToken, requestId, endPoint) {
    logger.debug('Getting Vault provider credentials');
    const protocol = head(keys(endPoint)).toUpperCase();
    const path = `${VAULT_PATH}/${requestId}/${protocol}Creds`;

    try {
        const { data } = await axios.get(`${VAULT_BASE_URL}/${path}`, {
            httpsAgent: externalAgent,
            headers: {
                'X-Vault-Token': clientToken,
                'X-Vault-Namespace': VAULT_NAMESPACE
            }
        });

        // if user store secrets at the "default" secret/* path of vault, then we need to access data.data.data,
        // else, secret is stored in a different path that the user enabled himself, then we need to get them from data.data
        return data.data.data || data.data;
    } catch (err) {
        logger.warn(`Unable to find data for path: ${path}`, err);
    }
}

function getSignedRequest(creds) {
    let awsCreds = {
        accessKeyId: creds.accessKeyId,
        secretAccessKey: creds.secretAccessKey
    };
    if (creds.sessionToken) {
        awsCreds.sessionToken = creds.sessionToken;
    }
    return aws4.sign({
            service: 'sts',
            headers: {'X-Vault-AWS-IAM-Server-ID': VAULT_BASE_URL},
            body: AWS_REQUEST_BODY
    }, awsCreds);
}

function getSignedHeaders (creds) {
    const signedRequest = getSignedRequest(creds);
    logger.info('Got signed request from aws4');
    const headers = signedRequest.headers;
    Object.keys(headers).forEach(header => {
        if (typeof headers[header] === 'number') {
            headers[header] = headers[header].toString();
        }
        headers[header] = [headers[header]];
    });
    return headers;
}

async function awsLogin() {
    try {
        logger.info('Logging into vault with AWS authentication');
        const providerChain = new AWS.CredentialProviderChain();
        const creds = await providerChain.resolvePromise();
        const awsRequestUrl = 'https://sts.amazonaws.com/';

        const requestHeaders = getSignedHeaders(creds);
        const body = {
            role: VAULT_ROLE_NAME,
            iam_http_request_method: 'POST',
            iam_request_url: new Buffer.from(awsRequestUrl, 'utf-8').toString('base64'),
            iam_request_body: new Buffer.from(AWS_REQUEST_BODY, 'utf-8').toString('base64'),
            iam_request_headers: new Buffer.from(JSON.stringify(requestHeaders), 'utf-8').toString('base64')
        };

        const { data: { auth } } = await axios.post(`${VAULT_BASE_URL}/v1/auth/aws/login`, body,{
            httpsAgent: externalAgent,
            headers: {
                'X-Vault-Namespace': VAULT_NAMESPACE
            }
        });
        return auth;
    } catch (err) {
        logger.error('Failed to authenticate to external vault using AWS authentication', JSON.stringify(err));
        throw err;
    }
}

function updateProviderCredentials(endPoint, credentials) {
    logger.debug('Updating Vault provider credentials', endPoint);
    const protocol = head(keys(endPoint));

    switch (protocol) {
        case 's3':
            endPoint.s3.accessKey = credentials.accessKey;
            endPoint.s3.secretKey = credentials.secretKey;
            break;
        case 'azure':
            endPoint.azure.storageAccountName = credentials.storageAccountName;
            endPoint.azure.storageAccountConnectionString = credentials.storageAccountConnectionString;
            break;
        case 'gcp':
            endPoint.gcp.projectId = credentials.projectId;
            endPoint.gcp.clientEmail = credentials.clientEmail;
            endPoint.gcp.privateKey = credentials.privateKey;
            break;
        default:
            logger.error(`Couldn't find matching protocol for endPoint ${endPoint}`);
            break;
    }

    return endPoint;
}

async function loginToVault() {
    switch (VAULT_AUTH_METHOD) {
        case('app-role'):
            return appRoleLogin();
        case ('aws-iam'):
            return awsLogin();
        default:
            throw new Error('Authentication method for vault is not defined:', VAULT_AUTH_METHOD);
    }
}

async function bootVaultMode(source, target, copyMessageAttributes) {
    logger.debug('Booting Vault mode');


    const { client_token: clientToken } = await loginToVault();
    const { initialized, sealed } = await getVaultHealthStatus();

    if (initialized && !sealed) {
        logger.debug('Vault is initialized and unsealed');

        const externalRequestIdSrc = copyMessageAttributes['x-netapp-external-request-id-src'];
        const externalRequestIdTrg = copyMessageAttributes['x-netapp-external-request-id-trg'];

        let updatedSrc;
        let updatedTrg;

        if (externalRequestIdSrc) {
            const srcProviderCredentials = await getProviderCredentials(clientToken, externalRequestIdSrc, source);

            if (srcProviderCredentials) {
                updatedSrc = updateProviderCredentials(source, srcProviderCredentials);
            }
        }

        if (externalRequestIdTrg) {
            const trgProviderCredentials = await getProviderCredentials(clientToken, externalRequestIdTrg, target);

            if (trgProviderCredentials) {
                updatedTrg = updateProviderCredentials(target, trgProviderCredentials);
            }
        }

        return { updatedSrc, updatedTrg };
    }
}

// endregion

// region Vault Internal Integration As metadata store

async function getLocalVaultHealthStatus() {
    logger.debug('Get local vault health status');
    /*
       200 - if initialized, unsealed, and active
       501 - if not initialized
       503 - if sealed
       */
    const { data } = await axios.get(`${LOCAL_VAULT_BASE_URL}/v1/sys/health`, {
        httpsAgent: localAgent,
        validateStatus: status => (status >= 200 && status < 300) || VALID_HEALTH_STATUSES.includes(status)
    });

    return data;
}

async function initSecretsEngine() {
    logger.debug('Initialize Vault secrets engine');

    const { data } = await axios.put(`${LOCAL_VAULT_BASE_URL}/v1/sys/init`, {
        secret_shares: 5,
        secret_threshold: 5
    }, { httpsAgent: localAgent });

    await writeFile(VAULT_KEYS_FILE, JSON.stringify(data, null, 4));
}

async function getUnsealKeys() {
    logger.debug('Getting unseal keys');

    const keysFile = await readFile('./vault/keys.json');
    const data = JSON.parse(keysFile);
    const { keys: unsealKeys } = data;

    return unsealKeys;
}

async function unsealVault(unsealKeys) {
    await retry(async () => {
        logger.debug('Unseal Vault keys');

        await Promise.all(unsealKeys.map(async key => {
            try {
                await axios.put(`${LOCAL_VAULT_BASE_URL}/v1/sys/unseal`,
                    {
                        key
                    }, { httpsAgent: localAgent });
            } catch (err) {
                logger.error('Failed to unseal vault key. Error: ', err);
                throw err;
            }
        }));
    }, {
        retries: 3,
        onRetry: err => logger.warn('Retrying unseal Vault keys. Errors: ', err)
    });
}

async function getRootToken() {
    logger.debug('Getting root token');

    const keysFile = await readFile('./vault/keys.json');
    const data = JSON.parse(keysFile);
    const { root_token: rootToken } = data;

    return rootToken;
}

async function getSecretsEngineMounts() {
    logger.debug('Getting Vault secrets engine mounts');

    const rootToken = await getRootToken();

    const { data } = await axios.get(`${LOCAL_VAULT_BASE_URL}/v1/sys/mounts`, {
        httpsAgent: localAgent,
        headers: {
            'X-Vault-Token': rootToken
        }
    });

    return data;
}

async function isSecretsEngineMounted() {
    logger.debug('Verify if Vault secrets engine enabled');

    const { 'secret/': secretMount } = await getSecretsEngineMounts();

    return secretMount;
}

async function enableSecretsEngine() {
    logger.debug('Enabling Vault secrets engine');

    const rootToken = await getRootToken();

    const { data } = await axios.post(`${LOCAL_VAULT_BASE_URL}/v1/sys/mounts/${DATA_BROKER_SECRET_ENGINE_PATH}`, {
        type: 'kv'
    }, {
        httpsAgent: localAgent,
        headers: {
            'X-Vault-Token': rootToken
        }
    });

    return data;
}

async function generateNewTokenIfNeeded() {
    const keysFile = await readFile(VAULT_KEYS_FILE);
    const file = JSON.parse(keysFile);
    const { root_token: token } = file;

    const { data: { data: { orphan } } } = await axios.post(`${LOCAL_VAULT_BASE_URL}/v1/auth/token/lookup`, { token }, {
        httpsAgent: localAgent,
        headers: {
            'X-Vault-Token': token
        }
    });

    // if root token, then true, else false
    if (orphan) { // if orphan, which means root, create new one and save it in the keys file
        logger.info('Vault token is root, generating new one');

        const { data: { auth: { client_token: clientToken } } } = await axios.post(`${LOCAL_VAULT_BASE_URL}/v1/auth/token/create`, {}, {
            httpsAgent: localAgent,
            headers: {
                'X-Vault-Token': token
            }
        });
        // update token
        file.root_token = clientToken;
        await writeFile(VAULT_KEYS_FILE, JSON.stringify(file, null, 4));
    }
}

async function prepareVault() {
    logger.info('Prepare vault for data broker');

    try {
        const { initialized: isVaultInitialized, sealed: isVaultSealed } = await getLocalVaultHealthStatus();

        logger.info('Vault health: ', { isVaultInitialized, isVaultSealed });

        if (!isVaultInitialized) {
            logger.info('Vault secret engine is not initialized');

            await initSecretsEngine();
        }

        if (isVaultSealed) {
            logger.info('Vault is sealed');

            const unsealKeys = await getUnsealKeys();
            await unsealVault(unsealKeys);
        }

        await generateNewTokenIfNeeded();

        const secretsEngineMount = await isSecretsEngineMounted();

        if (!secretsEngineMount) {
            logger.info('Vault secret engine is unmount, going to mount secret path');

            await enableSecretsEngine();
        }
    } catch (err) {
        logger.error('Failed to prepare vault for data broker', err);
    }
}

async function getSecret(secretPath) {
    logger.debug('Getting secret from Vault by path: ', secretPath);

    const rootToken = await getRootToken();
    const path = `${LOCAL_VAULT_BASE_URL}/v1/${DATA_BROKER_SECRET_ENGINE_PATH}/${secretPath}`;

    try {
        const { data } = await axios.get(path, {
            httpsAgent: localAgent,
            headers: {
                'X-Vault-Token': rootToken
            }
        });

        return data.data;
    } catch (err) {
        const responseStatus = get(err, 'response.status');

        if (responseStatus === 404) {
            return undefined;
        }
    }
}

async function createSecret(secretPath, secretData) {
    logger.debug('Creating Vault secret to path: ', { secretPath, secretData });

    const rootToken = await getRootToken();

    const { data } = await axios.post(`${LOCAL_VAULT_BASE_URL}/v1/${DATA_BROKER_SECRET_ENGINE_PATH}/${secretPath}`, secretData, {
        httpsAgent: localAgent,
        headers: {
            'X-Vault-Token': rootToken
        }
    });

    return data;
}

async function listSecretKeys() {
    logger.debug('Listing Vault secret keys');

    const rootToken = await getRootToken();

    const { data } = await axios({
        method: 'list',
        url: `${LOCAL_VAULT_BASE_URL}/v1/${DATA_BROKER_SECRET_ENGINE_PATH}`,
        headers: {
            'X-Vault-Token': rootToken
        },
        httpsAgent: localAgent
    });

    return data;
}

async function exportVaultSecretData() {
    logger.debug('Exporting Vault secret data');

    const { data: { keys: secretKeys } } = await listSecretKeys();

    if (secretKeys) {
        return Promise.all(secretKeys.map(async key => {
            const data = await getSecret(key);

            return { k: key, v: data };
        }));
    }

    return [];
}

// endregion

module.exports = {
    bootVaultMode,
    prepareVault,
    getSecret,
    createSecret,
    exportVaultSecretData,
    getLocalVaultHealthStatus
};
