const Joi = require('@hapi/joi');
const endPointSchema = require('../schemas/endPoint-schema');
const metadataSchema = require('../schemas/metadata-schema');
const settingsSchema = require('../schemas/settings-schema');
const {
    ONTAP_S3_PROVIDER, NFS_VERSIONS, CIFS_VERSIONS, ALIBABA_PROVIDER, S3_PROVIDER, SGWS_PROVIDER, IBM_PROVIDER
} = require('../../lib/consts');

const masterCopySchema = Joi.object().keys({
    relationshipId: Joi.string().required(),
    source: endPointSchema.required(),
    target: endPointSchema.required(),
    scannerQueue: Joi.string().required(),
    transferrerQueue: Joi.string().required(),
    settings: settingsSchema.required()
});

const masterSaveMetadataSchema = Joi.object().keys({
    relationshipId: Joi.string().required(),
    source: endPointSchema.required(),
    target: endPointSchema.required()
});

const masterDownloadLogsSchema = Joi.object().keys({
    name: Joi.string().required()
});

const masterListKmsKeysSchema = Joi.object().keys({
    bucket: Joi.string().required()
});

const masterChangeStatusesQueueSchema = Joi.object().keys({
    statusesQueueUrl: Joi.string().required()
});

const masterListS3Buckets = Joi.object().keys({
    provider: Joi.string().valid(ONTAP_S3_PROVIDER, ALIBABA_PROVIDER, S3_PROVIDER, SGWS_PROVIDER, IBM_PROVIDER),
    host: Joi.when('provider', { is: [SGWS_PROVIDER, IBM_PROVIDER], then: Joi.string().required() }),
    port: Joi.when('provider', { is: [IBM_PROVIDER], then: Joi.string().required() }),
    accessKey: Joi.when('provider', { is: [ALIBABA_PROVIDER, SGWS_PROVIDER, IBM_PROVIDER], then: Joi.string().required() }),
    secretKey: Joi.when('provider', { is: [ALIBABA_PROVIDER, SGWS_PROVIDER, IBM_PROVIDER], then: Joi.string().required() })
});

const masterListS3BucketFolders = Joi.object().keys({
    bucket: Joi.string().required(),
    prefix: Joi.string().allow(''),
    provider: Joi.string().valid(ONTAP_S3_PROVIDER, ALIBABA_PROVIDER, S3_PROVIDER, SGWS_PROVIDER, IBM_PROVIDER),
    host: Joi.when('provider', { is: [SGWS_PROVIDER, IBM_PROVIDER], then: Joi.string().required() }),
    port: Joi.when('provider', { is: [IBM_PROVIDER], then: Joi.string().required() }),
    accessKey: Joi.when('provider', { is: [ALIBABA_PROVIDER, SGWS_PROVIDER, IBM_PROVIDER], then: Joi.string().required() }),
    secretKey: Joi.when('provider', { is: [ALIBABA_PROVIDER, SGWS_PROVIDER, IBM_PROVIDER], then: Joi.string().required() })
});

const masterRelationshipsListSchema = Joi.object().keys({
    relationships: Joi.array().items(Joi.object({
        relationshipId: Joi.string(),
        scannerQueue: Joi.string().required(),
        transferrerQueue: Joi.string().required(),
        source: endPointSchema.required(),
        target: endPointSchema.required(),
        settings: settingsSchema.required()
    })),
    metadata: metadataSchema
});

const masterListAzureBlobStorageContainersSchema = Joi.object().keys({
    storageAccountName: Joi.string().required(),
    storageAccountConnectionString: Joi.string()
});

const masterListAzureBlobStorageContainerFoldersSchema = Joi.object().keys({
    container: Joi.string().required(),
    prefix: Joi.string(),
    storageAccountName: Joi.string().required(),
    storageAccountConnectionString: Joi.string()
});

const masterListExportsSchema = Joi.object().keys({
    host: Joi.string().required()
});

const masterListExportFoldersSchema = Joi.object().keys({
    host: Joi.string().required(),
    path: Joi.string().allow(''),
    export: Joi.string().required(),
    version: Joi.number().valid(...NFS_VERSIONS)
});

const doneRemoveAbortSchema = Joi.object().keys({
    relationshipId: Joi.string().required(),
    source: endPointSchema.required(),
    target: endPointSchema.required()
});

const updateSchema = Joi.object().keys({
    bundleUrl: Joi.string().uri({ scheme: ['https'] }).required(),
    bundleVersion: Joi.string().required(),
    relationships: Joi.array().items(Joi.object({
        relationshipId: Joi.string(),
        scannerQueue: Joi.string().required(),
        transferrerQueue: Joi.string().required(),
        source: endPointSchema.required(),
        target: endPointSchema.required()
    }))
});

const masterSharesSchema = Joi.object().keys({
    host: Joi.string().required(),
    username: Joi.string().required(),
    password: Joi.string().required(),
    workgroup: Joi.string(),
    version: Joi.number().valid(...CIFS_VERSIONS)
});

const masterListShareFoldersSchema = Joi.object().keys({
    host: Joi.string().required(),
    share: Joi.string().required(),
    path: Joi.string(),
    username: Joi.string().required(),
    password: Joi.string().required(),
    workgroup: Joi.string(),
    version: Joi.number().valid(...CIFS_VERSIONS)
});

const masterListGcpBucketsFoldersSchema = Joi.object().keys({
    bucket: Joi.string().required(),
    prefix: Joi.string().allow('')
});

const masterShutdownSchema = Joi.object().keys({});

const masterMetadataRequestSchema = Joi.object().keys({
    targetDataBrokerId: Joi.string().required()
});

const masterSendAsupRequestSchema = Joi.object().keys({
    serialNumber: Joi.string().required(),
    asupUrl: Joi.string().required()
});

const sqsMasterIncomingMessagesSchemes = {
    startup: masterRelationshipsListSchema,
    copy: masterCopySchema,
    's3-buckets': masterListS3Buckets,
    's3-bucket-folders': masterListS3BucketFolders,
    'gcp-buckets': Joi.object(),
    'gcp-bucket-folders': masterListGcpBucketsFoldersSchema,
    exports: masterListExportsSchema,
    'export-folders': masterListExportFoldersSchema,
    'azure-blob-storage-containers': masterListAzureBlobStorageContainersSchema,
    'azure-blob-storage-container-folders': masterListAzureBlobStorageContainerFoldersSchema,
    add: masterRelationshipsListSchema,
    done: doneRemoveAbortSchema,
    remove: doneRemoveAbortSchema,
    abort: doneRemoveAbortSchema,
    update: updateSchema,
    shares: masterSharesSchema,
    'share-folders': masterListShareFoldersSchema,
    shutdown: masterShutdownSchema,
    restart: masterRelationshipsListSchema,
    metadata: masterMetadataRequestSchema,
    'send-asup': masterSendAsupRequestSchema,
    'save-metadata': masterSaveMetadataSchema,
    'download-logs': masterDownloadLogsSchema,
    'kms-keys': masterListKmsKeysSchema,
    'change-statuses-queue': masterChangeStatusesQueueSchema
};

module.exports = sqsMasterIncomingMessagesSchemes;
