const { URL } = require('url');
const { cpus } = require('os');
const cluster = require('cluster');
const https = require('https');
const http = require('http');
const { join } = require('path');
const config = require('config');
const Promise = require('bluebird');
const { v4 } = require('uuid');
const {
    mkdirs, remove, emptyDir, readFile, pathExists, stat, writeFile
} = require('fs-extra');
const prettyBytes = require('bytes');
const {
    update, head, keys, compact, sortBy,
    merge, trim, get, isEmpty, last, pick
} = require('lodash');
const si = require('systeminformation');
const generateCertificate = require('../lib/phoenix/generate-certificate');
const getLogger = require('../lib/logger/logger');
const { receiveMessages, sendMessage, deleteMessage } = require('../lib/aws/sqs');
const { getAwsAccountId, assumeRole } = require('../lib/aws/sts');
const { listContainers, listPrefixByHierarchyByConnectionString } = require('../lib/azure/blob');
const { registerMetadata, exportMetadata, importMetadata } = require('../lib/vault/metadata-store.js');
const { isGcpBucketExists } = require('../lib/gcp/storage');
const getAwsMetadata = require('../lib/aws/metadata');
const getGcpMetadata = require('../lib/gcp/metadata');
const getAzureMetadata = require('../lib/azure/metadata');
const getOsMetadata = require('../lib/os/metadata');
const getProjectId = require('../lib/gcp/auth');
const sqsMasterIncomingMessagesSchemes = require('../messages/master/incoming-messages');
const sqsMasterOutgoingMessagesSchemes = require('../messages/master/outgoing-messages');
const {
    WORKER_TYPE_SCANNER, WORKER_TYPE_TRANSFERRER, WORKER_TYPE_SERVER, WORKER_TYPE_ACL_SERVER, DATA_BROKER_RUNTIME_CONFIG, CERT_AND_KEY_FOLDER,
    DATA_BROKER_RUNTIME_ABOUT, STATUS_FAILED, DEPLOYMENT_TYPES, INITIALIZE_STEP, TYPE_DIRECTORY, TYPE_SYMBOLIC, STATUS_SUCCEEDED,
    LIST_STEP, TRANSFER_STEP, ACTION_DELETE_TARGET, ACTION_DELETE_FAILED, ACTION_COPY_FAILED, CLEAN_RELATIONSHIP_RESOURCES_TIMEOUT,
    MOUNT_SOURCE_SUFFIX, MOUNT_TARGET_SUFFIX, MOUNT_EXPORTS_FOLDER, LOGS_DIRECTORY_RELATIVE_PATH,
    MOUNT_SHARES_FOLDER, ONTAP_S3_PROVIDER, S3_PROVIDER, ALIBABA_PROVIDER, SGWS_PROVIDER, IBM_PROVIDER, LOGS_BUCKET_NAME,
    AWS_SQS_CREDENTIALS, LOCAL_CONFIG_FILE
} = require('../lib/consts');
const execute = require('../lib/executor');
const { failedSchema } = require('../messages/schemas/statuses-outgoing-messages');
const sendAccessListRequest = require('../lib/access-list');
const { bootVaultMode, prepareVault } = require('../lib/vault/vault-operations');

const { statusesQueue } = DATA_BROKER_RUNTIME_CONFIG;
const { showmount, listShares } = require('../lib/fs/mount');
const { mountIfNeeded, umountAndDeleteRelationshipFolder, umountAndDelete } = require('../lib/fs/mount_auxiliary');
const {
    list, gracefulJoin, removeCoreFiles, getCifsAcl, setCifsAcl
} = require('../lib/fs/fs');
const validateMessage = require('../messages/schemas/message-validator');
const performUpdate = require('../lib/phoenix/update');
const { startHealthLogInterval } = require('../lib/health');
const { listGcpBuckets, listGcpBucketFolders, getBucketLocationAndTags } = require('../lib/gcp/storage');
const { secureMkdirs } = require('../lib/snfs/snfs');
const {
    upload,
    headBucket, getBucketLocation, getBucketTagging, listBuckets, listObjectsV2, listObjects
} = require('../lib/aws/s3_base_operations');
const {
    getClient
} = require('../lib/s3Clients');
const handleAsupRequest = require('../lib/phoenix/asup');
const createZipDirectory = require('../lib/phoenix/zip-directory');
const { listAllKmsAliases, describeKmsKey } = require('../lib/aws/kms');

const logger = getLogger();
startHealthLogInterval();
const relationshipsLogger = getLogger('relationships');

const META_DATA = {};
const RELATIONSHIPS_STATUS = {};
const TOTALS = {};
const ACL_SERVERS = {};
const WORKERS_IDS = {};

http.globalAgent.maxSockets = 500;
https.globalAgent.maxSockets = 500;

function forkProcess(file, type) {
    logger.info('Forking file:', file);

    let execArgv = [];
    if (type === WORKER_TYPE_SCANNER) {
        const heapLimitInMb = prettyBytes(config.get('workers.scanner.heap-limit')) / 1024 / 1024;
        execArgv = [`--max_old_space_size=${heapLimitInMb}`];
    }

    cluster.setupMaster({
        exec: file,
        silent: true,
        execArgv
    });

    const childProcess = cluster.fork({ UV_THREADPOOL_SIZE: config.get('uv-thread-pool-size') });
    childProcess.workerType = type;
    return childProcess;
}

function startTransferrers() {
    logger.info('Starting transferrers...');
    const processesCpuRatio = config.get('workers.transferrer.processes-cpu-ratio');
    const processesLimitNumber = config.get('workers.transferrer.processes-limit');
    const numberOfTransferrers = Math.min(cpus().length * processesCpuRatio, processesLimitNumber);
    logger.info('Process to CPU ratio', processesCpuRatio);
    logger.info('The max number of processes allowed', processesLimitNumber);
    logger.info('Number of transferrers', numberOfTransferrers);
    for (let i = 0; i < numberOfTransferrers; i++) {
        forkProcess(join(__dirname, 'transferrer.js'), WORKER_TYPE_TRANSFERRER);
    }
}

function startServer() {
    logger.info('Starting server...');
    const processesCpuRatio = config.get('workers.server.processes-cpu-ratio');
    const processesLimitNumber = config.get('workers.server.processes-limit');
    logger.info('Process to CPU ratio', processesCpuRatio);
    logger.info('The max number of processes allowed', processesLimitNumber);
    logger.info('Number of servers', processesLimitNumber);
    forkProcess(join(__dirname, 'server.js'), WORKER_TYPE_SERVER);
}

function startScanners() {
    logger.info('Starting scanners...');
    const processesCpuRatio = config.get('workers.scanner.processes-cpu-ratio');
    const processesLimitNumber = config.get('workers.scanner.processes-limit');
    const numberOfScanners = Math.min(cpus().length * processesCpuRatio, processesLimitNumber);
    logger.info('Process to CPU ratio', processesCpuRatio);
    logger.info('The max number of processes allowed', processesLimitNumber);
    logger.info('Number of scanners', numberOfScanners);
    for (let i = 0; i < numberOfScanners; i++) {
        const scannerProcess = forkProcess(join(__dirname, 'scanner.js'), WORKER_TYPE_SCANNER);
        WORKERS_IDS[scannerProcess.id] = scannerProcess.process.pid;
        scannerProcess.send({
            workerId: scannerProcess.id,
            type: 'create-scanner'
        });
    }
}

function startAclServers() {
    try {
        require('cifs-acl'); /* eslint-disable-line global-require */
    } catch (e) {
        logger.info('Could not require cifs-acl package');
        return;
    }
    logger.info('Starting acl-servers...');
    const processesCpuRatio = config.get('workers.acl-server.processes-cpu-ratio');
    const processesLimitNumber = config.get('workers.acl-server.processes-limit');
    const numberOfAclServers = Math.min(cpus().length * processesCpuRatio, processesLimitNumber);
    for (let i = 0; i < numberOfAclServers; i++) {
        const aclServerProcess = forkProcess(join(__dirname, 'acl-server.js'), WORKER_TYPE_ACL_SERVER);
        aclServerProcess.send({
            port: config.get('workers.acl-server.port-start-range') + i,
            type: 'create-acl-server'
        });
        ACL_SERVERS[aclServerProcess.process.pid] = config.get('workers.acl-server.port-start-range') + i;
    }
}

function getEndPointString(endPoint) {
    if (endPoint.nfs || endPoint.cifs) {
        return 'local';
    }
    return head(keys(endPoint));
}

async function mkdirsIfNeeded(relationshipId, requestId, endPoint, mountPoint = undefined) {
    if (endPoint.local) {
        const { local: { path } } = endPoint;
        await mkdirs(path);
    } else if (endPoint.nfs) {
        const { nfs: { path } } = endPoint;
        await mkdirs(gracefulJoin(mountPoint, path));
    } else if (endPoint.cifs) {
        const { cifs: { path } } = endPoint;
        await mkdirs(gracefulJoin(mountPoint, path));
    } else if (endPoint.snfs) {
        const {
            snfs: {
                serverIP, port, certificate, serverName, path
            }
        } = endPoint;
        await secureMkdirs(relationshipId, serverIP, port, certificate, serverName, requestId, path);
    }
}

async function findRegionS3Provider(endPoint, relationshipId, srcOrTrg) {
    const { provider, bucket } = endPoint;
    const s3 = await getClient(provider, { ...endPoint, relationshipId, srcOrTrg });
    return headBucket(s3, bucket);
}

async function prepareBucketIfNeeded(endPoint, relationshipId, srcOrTrg) {
    // done to make sure the bucket exists, otherwise a fatally failed error will be sent from the master
    if (endPoint) {
        if (endPoint.s3) {
            if (endPoint.s3.provider.includes([ALIBABA_PROVIDER, S3_PROVIDER])) {
                await findRegionS3Provider(endPoint.s3, relationshipId, srcOrTrg);
            }
        }
        if (endPoint.gcp) {
            const { gcp: { bucket } } = endPoint;
            return isGcpBucketExists(bucket);
        }
    }
}

async function sendStatistics(relationshipId) {
    const statistics = RELATIONSHIPS_STATUS[relationshipId];
    if (statistics) {
        // sent statistics should be deleted synchronously since new statistics message are comming through IPC and we don't want to send the same counters more then once
        delete RELATIONSHIPS_STATUS[relationshipId];
        statistics.relationshipId = relationshipId;
        const { attributes } = statistics;
        const { 'data-broker-id': dataBrokerId } = DATA_BROKER_RUNTIME_CONFIG;
        Object.assign(attributes, { dataBrokerId });
        delete statistics.attributes;
        if (validateMessage(statistics, 'Outgoing-SQS', sqsMasterOutgoingMessagesSchemes.statistics)) {
            await sendMessage(statusesQueue, statistics, attributes);
            logger.debug(`Sent statistics for relationship ${relationshipId} statistics: ${JSON.stringify(statistics)} and attributes ${JSON.stringify(attributes)}`);
        } else {
            logger.error(`Failed to send statistics ${JSON.stringify(statistics)} for relationship ${relationshipId} due to validation error`);
        }
    }
}

async function cleanRelationshipResources(relationshipId, sourceEndpoint, targetEndpoint) {
    logger.info(`Cleaning resources for rid: ${relationshipId}`);
    await sendStatistics(relationshipId);
    const totals = TOTALS[relationshipId];
    if (totals) {
        logger.info(`before cleaning relationship ${relationshipId}, current totals ${JSON.stringify(totals)}`);
    }
    // delete relationship's queues
    delete META_DATA[relationshipId];
    delete RELATIONSHIPS_STATUS[relationshipId];
    delete TOTALS[relationshipId];

    // unmount if mount point exists
    const [sourceProtocol, targetProtocol] = [head(keys(sourceEndpoint)), head(keys(targetEndpoint))];
    if (['nfs', 'cifs', 'snfs'].includes(sourceProtocol) || ['nfs', 'cifs', 'snfs'].includes(targetProtocol)) {
        await umountAndDeleteRelationshipFolder(relationshipId, sourceEndpoint, targetEndpoint);
    }
}

function getWorkersIdsByType(type) {
    return Object.keys(cluster.workers).filter(id => cluster.workers[id].workerType === type);
}

function sendWorkersMessage(message) {
    const workersToIgnore = [WORKER_TYPE_ACL_SERVER];
    Object.keys(cluster.workers).filter(id => !workersToIgnore.includes(cluster.workers[id].workerType)).forEach(id => cluster.workers[id].send(message));
}

function initializeScannerAndTrnasferrer(relationshipId, attributes, scannerQueue, transferrerQueue, settings, serverName, certificate) {
    const { requestId, receiptHandle } = attributes;
    const transferrerSettings = pick(settings, ['retries', 'copyAcl', 'deleteOnSource', 'objectTagging']);
    getWorkersIdsByType(WORKER_TYPE_SCANNER).forEach(id => cluster.workers[id].send({
        relationshipId, requestId, receiptHandle, scannerQueue, transferrerQueue, serverName, certificate, settings, type: 'copy'
    }));

    getWorkersIdsByType(WORKER_TYPE_TRANSFERRER).forEach(id => cluster.workers[id].send({
        relationshipId,
        requestId,
        transferrerQueue,
        serverName,
        certificate,
        settings: transferrerSettings,
        type: 'copy'
    }));
}

async function handleStartUpAdd(message) {
    const { body: { relationships, metadata }, attributes, attributes: { requestId } } = message;
    if (!isEmpty(metadata)) {
        await importMetadata(metadata);
    }
    const promises = relationships.map(async relationship => {
        const {
            relationshipId, scannerQueue, transferrerQueue, source, target, settings
        } = relationship;
        try {
            relationshipsLogger.info(`[${relationshipId}] unmount source and target if needed`);
            await umountAndDeleteRelationshipFolder(relationshipId, source, target);
            relationshipsLogger.info(`[${relationshipId}] mount source and target if needed`);
            await mountIfNeeded(relationshipId, requestId, MOUNT_SOURCE_SUFFIX, source);
            await mountIfNeeded(relationshipId, requestId, MOUNT_TARGET_SUFFIX, target);
            META_DATA[relationshipId] = {
                scannerQueue, transferrerQueue, requestId, settings
            };
            if (source.snfs || target.snfs) {
                META_DATA[relationshipId].certificate = get(source, 'snfs.certificate') || get(target, 'snfs.certificate');
                META_DATA[relationshipId].serverName = get(source, 'snfs.serverName') || get(target, 'snfs.serverName');
            }
            if (source.snfs) {
                initializeScannerAndTrnasferrer(relationshipId, attributes, scannerQueue, transferrerQueue, settings,
                    source.snfs.serverName, source.snfs.certificate);
            } else if (target.snfs) {
                initializeScannerAndTrnasferrer(relationshipId, attributes, scannerQueue, transferrerQueue, settings,
                    target.snfs.serverName, target.snfs.certificate);
            } else {
                initializeScannerAndTrnasferrer(relationshipId, attributes, scannerQueue, transferrerQueue, settings);
            }
        } catch (err) {
            const errorMessage = merge({
                step: INITIALIZE_STEP,
                relationshipId,
                status: STATUS_FAILED,
                reason: err.message || err
            }, message.body);
            if (validateMessage(errorMessage, 'SQS', failedSchema)) {
                sendMessage(statusesQueue, errorMessage, message.attributes);
            }
        }
    });
    await Promise.all(promises);
}

async function handleCopyMessage(copyMessage) {
    logger.info('Handling copy message');
    const sourcePassword = get(copyMessage, `body.source.${Object.keys(copyMessage.body.source)[0]}.password`);
    const targetPassword = get(copyMessage, `body.target.${Object.keys(copyMessage.body.target)[0]}.password`);
    const {
        body: {
            relationshipId, scannerQueue, transferrerQueue, source, target, settings
        },
        attributes,
        attributes: { requestId }
    } = copyMessage;

    if (source.snfs || target.snfs) {
        try {
            const listenerIp = config.get('snfs.listener-ip');
            logger.info('********* Got data broker listener ip from config file *********');
            logger.info(`********* data broker listener ip inserted by user: ${listenerIp}, connection to the listener data broker will be established using this ip *********`);
        } catch (err) {
            logger.info('********* Couldnt locate listener ip from config file:', err, '*********');
            logger.info('********* relationship will use private ip located in the relationship message *********');
        }
    }
    if (source.cifs && target.cifs && settings.copyAcl) {
        settings.copyAcl = 'cifs';
    }
    if (source.nfs && target.nfs && settings.copyAcl) {
        settings.copyAcl = 'nfs';
    }

    META_DATA[relationshipId] = {
        scannerQueue, transferrerQueue, requestId, settings
    };

    if (source.snfs || target.snfs) {
        META_DATA[relationshipId].certificate = get(source, 'snfs.certificate') || get(target, 'snfs.certificate');
        META_DATA[relationshipId].serverName = get(source, 'snfs.serverName') || get(target, 'snfs.serverName');
    }

    const vaultModeEnabled = config.get('external-integrations.hashicorp.enabled');

    let updatedSrc;
    let updatedTrg;

    if (vaultModeEnabled) {
        logger.info('Vault mode is enabled, getting credentials from vault');

        const vaultUpdatedValues = await bootVaultMode(source, target, copyMessage.attributes);

        if (vaultUpdatedValues) {
            ({ updatedSrc, updatedTrg } = vaultUpdatedValues);
        }
    }

    await registerMetadata(updatedSrc || source, updatedTrg || target, sourcePassword, targetPassword, relationshipId);

    // prepare source
    const sourceMountPoint = await mountIfNeeded(relationshipId, requestId, MOUNT_SOURCE_SUFFIX, source, sourcePassword);
    await prepareBucketIfNeeded(source, relationshipId, MOUNT_SOURCE_SUFFIX);

    // prepare target
    const targetMountPoint = await mountIfNeeded(relationshipId, requestId, MOUNT_TARGET_SUFFIX, target, targetPassword);
    await mkdirsIfNeeded(relationshipId, requestId, target, targetMountPoint);
    await prepareBucketIfNeeded(target, relationshipId, MOUNT_TARGET_SUFFIX);

    // both root directories should have the same acl in order to avoid mismatches in their files' acl
    if (source.cifs && target.cifs && settings.copyAcl === 'cifs') {
        try {
            await setCifsAcl(gracefulJoin(targetMountPoint, target.cifs.path), await getCifsAcl(gracefulJoin(sourceMountPoint, source.cifs.path)));
        } catch (err) {
            logger.error(err);
        }
    }

    if (source.snfs) {
        initializeScannerAndTrnasferrer(relationshipId, attributes, scannerQueue, transferrerQueue, settings,
            source.snfs.serverName, source.snfs.certificate);
    } else if (target.snfs) {
        initializeScannerAndTrnasferrer(relationshipId, attributes, scannerQueue, transferrerQueue, settings,
            target.snfs.serverName, target.snfs.certificate);
    } else {
        initializeScannerAndTrnasferrer(relationshipId, attributes, scannerQueue, transferrerQueue, settings);
    }
    // for debug - to make sure we finished the initialization
    logger.info('Done handling copy message on master');
}

async function handleSaveMetadataMessage(message) {
    const sourcePassword = get(message, `body.source.${Object.keys(message.body.source)[0]}.password`);
    const targetPassword = get(message, `body.target.${Object.keys(message.body.target)[0]}.password`);
    const {
        body: {
            relationshipId, source, target
        }
    } = message;

    if (source.snfs || target.snfs) {
        try {
            const listenerIp = config.get('snfs.listener-ip');
            logger.info('********* Got data broker listener ip from config file *********');
            logger.info(`********* data broker listener ip inserted by user: ${listenerIp}, connection to the listener data broker will be established using this ip *********`);
        } catch (err) {
            logger.info('********* Couldnt locate listener ip from config file:', err, '*********');
            logger.info('********* relationship will use private ip located in the relationship message *********');
        }
    }

    await registerMetadata(source, target, sourcePassword, targetPassword, relationshipId);
}

async function handleExportFoldersMessage(message) {
    const {
        body: {
            host, path, export: exportName, version
        }, attributes: { requestId }
    } = message;
    const mountPoint = await mountIfNeeded(MOUNT_EXPORTS_FOLDER, requestId, '', {
        nfs: {
            host,
            export: exportName,
            version
        }
    });

    const content = await list(gracefulJoin(mountPoint, path));
    const folders = content.filter(entry => entry.type === TYPE_DIRECTORY).filter(directory => !config.get('workers.scanner.ignore.directories').includes(directory.path)).map(directory => `/${directory.path}`);
    const response = {
        host, path, export: exportName, folders: folders.sort(), status: STATUS_SUCCEEDED
    };

    if (validateMessage(response, 'Outgoing-SQS', sqsMasterOutgoingMessagesSchemes['export-folders'])) {
        await sendMessage(statusesQueue, response, message.attributes);
    }
    await umountAndDelete(mountPoint);
}

async function handleDoneRemoveAbortMessage(message) {
    const {
        body: {
            relationshipId, source, target, scannerQueue, transferrerQueue
        }, attributes: { type }
    } = message;
    const { scannerQueue: relationshipScannerQueue, transferrerQueue: relationshipTransferrerQueue } = META_DATA[relationshipId] || {};

    if (relationshipScannerQueue === scannerQueue && relationshipTransferrerQueue === transferrerQueue) {
        logger.info(`Handling ${type} message for relationshipId: ${relationshipId} with scanner queue ${scannerQueue} and transferrer queue ${transferrerQueue}`);
        sendWorkersMessage({
            source: getEndPointString(source), target: getEndPointString(target), relationshipId, type
        });
        // TODO - clean relationship from realtime when all items Done aborting (For now "X" sec - need fixing).
        setTimeout(async () => {
            try {
                await cleanRelationshipResources(relationshipId, source, target);
            } catch (err) {
                logger.error(`failed cleaning relationship: ${relationshipId} resources with scanner queue ${scannerQueue} and transferrer queue ${transferrerQueue}`, err);
            }
        }, CLEAN_RELATIONSHIP_RESOURCES_TIMEOUT);

        if (type === 'done' && head(keys(source)) === 'cifs' && head(keys(target)) === 'cifs') {
            const { cifs: { aclServer } } = source;

            if (aclServer) {
                await sendAccessListRequest(source, target, aclServer, relationshipId);
            } else {
                logger.info('SMB ACL server is not defined');
            }
        }
    } else if (scannerQueue && transferrerQueue) { // on remove there are no queues on the message no need to log this case
        logger.warn(`received ${type} message with scannerQueue ${scannerQueue} instead of ${relationshipScannerQueue} 
        and transferrerQueue ${transferrerQueue} instead of ${relationshipTransferrerQueue} for rid ${relationshipId}`);
    }
}

async function handleUpdateMessage(message, queueUrl) {
    if (config.get('update-new-version')) {
        const { body: { bundleUrl, bundleVersion, relationships } } = message;
        // verify we dont try to update the same version more than once
        const aboutFile = await readFile(config.get('about-file'));
        const about = JSON.parse(aboutFile);
        logger.info('ABOUT file is', about);
        const { version } = about;
        if (version !== bundleVersion) {
            logger.info('Cleaning all resources');
            relationships.forEach(relationship => {
                const { source, target, relationshipId } = relationship;
                sendWorkersMessage({
                    source: getEndPointString(source), target: getEndPointString(target), relationshipId, type: 'update'
                });
            });
            // TODO - clean relationship from realtime when all items Done aborting (For now "X" sec - need fixing).
            setTimeout(async () => {
                try {
                    const promises = relationships.map(async relationship => {
                        const { relationshipId, source, target } = relationship;
                        try {
                            await cleanRelationshipResources(relationshipId, source, target);
                        } catch (err) {
                            logger.error(`failed cleaning relationship: ${relationshipId} resources`, err);
                        }
                    });
                    await Promise.all(promises);
                } finally {
                    // deleting message before restart
                    logger.info('done cleaning resources..., deleting update message & updating & restart');
                    await deleteMessage(queueUrl, message.receiptHandle);
                    await performUpdate(bundleVersion, bundleUrl);
                }
            }, CLEAN_RELATIONSHIP_RESOURCES_TIMEOUT);
        } else {
            logger.info('Version update is canceled - data broker already has the same version', version);
            await deleteMessage(queueUrl, message.receiptHandle);
        }
    } else {
        logger.info('Version update is disabled, ignoring and deleting update message');
        await deleteMessage(queueUrl, message.receiptHandle);
    }
}

async function handleExportsMessage(message) {
    const { body: { host } } = message;
    const exports = await showmount(host);
    const response = { host, exports: exports.sort(), status: STATUS_SUCCEEDED };
    if (validateMessage(response, 'Outgoing-SQS', sqsMasterOutgoingMessagesSchemes.exports)) {
        await sendMessage(statusesQueue, response, message.attributes);
    }
}

async function lists3ProviderBucket(provider, options) {
    const result = { provider, status: STATUS_SUCCEEDED };
    let buckets;
    switch (provider) {
        case S3_PROVIDER: {
            const s3 = await getClient(provider);
            const response = await listBuckets(s3);
            buckets = await compact(await Promise.all(response.map(async bucket => {
                const { name } = bucket;
                try {
                    const region = await getBucketLocation(s3, name);
                    const s3ObjectWithRegion = await getClient(provider, { region });
                    const tags = await getBucketTagging(s3ObjectWithRegion, name);
                    return { name, region, tags };
                } catch (err) {
                    logger.warn(`Failed to get bucket location/tagging, bucket name ${name}`, err);
                }
            })));
            break;
        }
        case ALIBABA_PROVIDER: {
            const { accessKey, secretKey } = options;
            const s3 = await getClient(provider, { accessKey, secretKey });

            const response = await listBuckets(s3);
            buckets = await compact(await Promise.all(response.map(async bucket => {
                const { name } = bucket;
                try {
                    const region = await getBucketLocation(s3, name);
                    return { name, region };
                } catch (err) {
                    logger.warn(`Failed to get bucket location/tagging, bucket name ${name}`, err);
                }
            })));
            break;
        }
        case IBM_PROVIDER:
        case SGWS_PROVIDER: {
            const { host, port } = options;
            const s3 = await getClient(provider, options);
            buckets = await listBuckets(s3);
            result.host = host;
            result.port = port;
            break;
        }
        case ONTAP_S3_PROVIDER: {
            const { host } = options;
            const s3 = await getClient(provider, options);
            buckets = await listBuckets(s3);
            result.host = host;
            break;
        }
        default: {
            throw new Error(`Unknown provider ${provider} received`);
        }
    }
    result.buckets = sortBy(buckets.filter(bucket => !bucket.name.startsWith('fabric-pool-')), 'name');
    return result;
}

async function listS3ProviderBucketFolder(provider, bucket, prefix, options) {
    let result;
    const s3 = await getClient(provider, options);
    switch (provider) {
        case ALIBABA_PROVIDER:
        case S3_PROVIDER: {
            result = await listObjectsV2(s3, bucket, prefix, true);
            break;
        }
        case ONTAP_S3_PROVIDER:
        case IBM_PROVIDER:
        case SGWS_PROVIDER: {
            result = listObjects(s3, bucket, prefix, true);
            break;
        }
        default: {
            throw new Error(`Unknown provider ${provider} received`);
        }
    }
    return result;
}

async function handleS3BucketsMessage(message) {
    const { body: { provider }, attributes: { type } } = message;
    const response = await lists3ProviderBucket(provider, message.body);
    if (validateMessage(response, 'Outgoing-SQS', sqsMasterOutgoingMessagesSchemes[type])) {
        await sendMessage(statusesQueue, response, message.attributes);
    }
}

async function handleGcpBucketsMessage(message) {
    const buckets = await listGcpBuckets();
    const response = { buckets, status: STATUS_SUCCEEDED };
    if (validateMessage(response, 'Outgoing-SQS', sqsMasterOutgoingMessagesSchemes['gcp-buckets'])) {
        await sendMessage(statusesQueue, response, message.attributes);
    }
}

async function handleGcpBucketFoldersMessage(message) {
    const {
        body: {
            bucket, prefix
        }
    } = message;
    const { location } = await getBucketLocationAndTags(bucket);
    const folders = await listGcpBucketFolders(bucket, prefix);
    const response = {
        bucket: { name: bucket, region: location }, prefix, folders: folders.sort(), status: STATUS_SUCCEEDED
    };
    if (validateMessage(response, 'Outgoing-SQS', sqsMasterOutgoingMessagesSchemes['gcp-bucket-folders'])) {
        await sendMessage(statusesQueue, response, message.attributes);
    }
}

async function handleS3BucketFoldersMessage(message) {
    const {
        body: {
            bucket, prefix, host, port, provider
        }, attributes: { type }
    } = message;
    let region;
    if ([S3_PROVIDER, ALIBABA_PROVIDER].includes(provider)) {
        region = await findRegionS3Provider({ provider: S3_PROVIDER, bucket });
    }
    const folders = await listS3ProviderBucketFolder(provider, bucket, prefix, {
        ...(message.body),
        isS3EndPoint: true,
        region
    });
    const response = {
        bucket: { name: bucket, region }, prefix, folders: folders.filter(folder => folder !== '.'), status: STATUS_SUCCEEDED, host, port, provider
    };
    if (validateMessage(response, 'Outgoing-SQS', sqsMasterOutgoingMessagesSchemes[type])) {
        await sendMessage(statusesQueue, response, message.attributes);
    }
}

async function handleAzureBlobStorageContainersMessage(message) {
    const { body: { storageAccountName, storageAccountConnectionString } } = message;
    const containers = await listContainers(storageAccountConnectionString);
    const response = { storageAccountName, containers, status: STATUS_SUCCEEDED };
    if (validateMessage(response, 'Outgoing-SQS', sqsMasterOutgoingMessagesSchemes['azure-blob-storage-containers'])) {
        await sendMessage(statusesQueue, response, message.attributes);
    }
}

async function handleAzureBlobStorageContainerFoldersMessage(message) {
    const {
        body: {
            container, prefix, storageAccountName, storageAccountConnectionString
        }
    } = message;

    const directories = await listPrefixByHierarchyByConnectionString(container, prefix, storageAccountConnectionString);
    const folders = directories.map(directory => trim(directory.name.slice(prefix ? prefix.length : 0), '/'));

    const response = {
        storageAccountName, container, prefix, folders, status: STATUS_SUCCEEDED
    };

    if (validateMessage(response, 'Outgoing-SQS', sqsMasterOutgoingMessagesSchemes['azure-blob-storage-container-folders'])) {
        await sendMessage(statusesQueue, response, message.attributes);
    }
}

async function handleSharesMessage(message) {
    const password = get(message, 'body.password');
    const {
        body: {
            host, version, username, workgroup
        }
    } = message;

    const shares = await listShares(host, version, username, password, workgroup);
    const response = { shares, host, status: STATUS_SUCCEEDED };

    if (validateMessage(response, 'Outgoing-SQS', sqsMasterOutgoingMessagesSchemes.shares)) {
        await sendMessage(statusesQueue, response, message.attributes);
    }
}

async function handleKmsKeysMessage(message) {
    const {
        body: {
            bucket
        }, attributes: { type }
    } = message;
    const s3 = await getClient(S3_PROVIDER);
    const region = await headBucket(s3, bucket);
    const kmsKeysList = await listAllKmsAliases(region);
    const kmsKeys = await Promise.map(kmsKeysList.filter(key => !key.AliasName.includes('/aws/')), async key => {
        const { TargetKeyId: keyId, AliasName: name, AliasArn: arn } = key;
        const metadata = await describeKmsKey(region, keyId);
        return { name, arn, isEnabled: metadata.Enabled };
    }, { concurrency: 10 });
    const response = { bucket, status: STATUS_SUCCEEDED, keys: kmsKeys.map(key => ({ isEnabled: key.isEnabled, name: key.name.substr(6), arn: key.arn })) };

    if (validateMessage(response, 'Outgoing-SQS', sqsMasterOutgoingMessagesSchemes[type])) {
        await sendMessage(statusesQueue, response, message.attributes);
    }
}

async function handleShareFoldersMessage(message) {
    const password = get(message, 'body.password');
    const {
        body: {
            host, path, share, username, workgroup, version
        }, attributes: { requestId }
    } = message;

    const mountPoint = await mountIfNeeded(MOUNT_SHARES_FOLDER, requestId, '', {
        cifs: {
            host, share, username, workgroup, version
        }
    },
    password);

    const content = await list(gracefulJoin(mountPoint, path));

    // In SMB we support symbolic links to directories as target
    // there is a special parameter in SMB mount - "mfsymlinks"
    // which allows us to "see" symbolic links over mount
    const contentSymlinksResolved = await Promise.map(content, async entry => {
        if (entry.type === TYPE_SYMBOLIC) {
            const targetStat = await stat(gracefulJoin(mountPoint, path, entry.path));
            if (targetStat.isDirectory()) { entry.type = TYPE_DIRECTORY; }
        }
        return entry;
    }, { concurrency: 100 });
    const folders = contentSymlinksResolved.filter(entry => entry.type === TYPE_DIRECTORY).filter(directory => !config.get('workers.scanner.ignore.directories').includes(directory.path)).map(directory => `/${directory.path}`);
    const response = {
        host, path, share, folders: folders.sort(), status: STATUS_SUCCEEDED
    };

    if (validateMessage(response, 'Outgoing-SQS', sqsMasterOutgoingMessagesSchemes['share-folders'])) {
        await sendMessage(statusesQueue, response, message.attributes);
    }
    await umountAndDelete(mountPoint);
}

async function handleShutdownMessage() {
    logger.info('Shutting down...');
    await execute('pm2 delete all');
}

async function handleRestartMessage(message, queueUrl) {
    logger.info('Restarting  ...');
    const { body: { relationships } } = message;

    logger.info('Cleaning all resources');
    relationships.forEach(relationship => {
        const { source, target, relationshipId } = relationship;
        sendWorkersMessage({
            source: getEndPointString(source), target: getEndPointString(target), relationshipId, type: 'update'
        });
    });
    // TODO - clean relationship from realtime when all items Done aborting (For now "X" sec - need fixing).
    setTimeout(async () => {
        try {
            const promises = relationships.map(async relationship => {
                const { relationshipId, source, target } = relationship;
                try {
                    await cleanRelationshipResources(relationshipId, source, target);
                } catch (e) {
                    logger.error(`failed cleaning relationship: ${relationshipId} resources`, e);
                }
            });
            await Promise.all(promises);
        } finally {
            // deleting message before restart
            logger.info('done cleaning resources..., deleting update message & updating & restart');
            await deleteMessage(queueUrl, message.receiptHandle);
            await execute('pm2 restart all');
        }
    }, CLEAN_RELATIONSHIP_RESOURCES_TIMEOUT);
}

async function handleMetadataMessage(message) {
    logger.info('handle metadataRequestMessage', message);
    const { body: { targetDataBrokerId } } = message;
    // create metadata response
    const metadata = await exportMetadata();
    const response = { targetDataBrokerId, metadata, status: STATUS_SUCCEEDED };

    if (validateMessage(response, 'Outgoing-SQS', sqsMasterOutgoingMessagesSchemes.metadata)) {
        await sendMessage(statusesQueue, response, message.attributes);
    }
}

async function handleSendAsupMessage(message) {
    logger.info('Handle send-asup message: ', message);

    const {
        body: {
            serialNumber, asupUrl
        }, attributes: { requestId, dataBrokerId }
    } = message;

    // do not wait until the ASUP is sent because it might take more then 5 minutes - delete sqs
    handleAsupRequest(requestId, serialNumber, dataBrokerId, asupUrl);
}

async function uploadZipDirectoryToS3(directoryName, fileName, uploadedObjectName) {
    const data = await readFile(`${directoryName}/${fileName}`);
    const assumeRoleResult = await assumeRole(AWS_SQS_CREDENTIALS.ACCESS_KEY_ID, AWS_SQS_CREDENTIALS.SECRET_ACCESS_KEY, v4());
    const s3 = await getClient('s3', {
        accessKeyId: assumeRoleResult.AccessKeyId,
        secretAccessKey: assumeRoleResult.SecretAccessKey,
        sessionToken: assumeRoleResult.SessionToken
    });
    const managedUpload = upload(s3, LOGS_BUCKET_NAME, data, uploadedObjectName, {});
    await managedUpload.promise();
}

async function handleChangeStatusesQueue(message, queueUrl) {
    logger.info('Handle change statuses queue message: ', message);
    const {
        body: { statusesQueueUrl }
    } = message;
    const configFile = await readFile(config.get('config-file'));
    const configuration = JSON.parse(configFile);
    const oldStatusesQueue = configuration.statusesQueue;
    logger.info('changing statuses queue from', oldStatusesQueue, 'to', statusesQueueUrl);
    configuration.statusesQueue = statusesQueueUrl;
    configuration.oldStatusesQueue = oldStatusesQueue;
    await writeFile(config.get('config-file'), JSON.stringify(configuration));
    await deleteMessage(queueUrl, message.receiptHandle);
    logger.info('Restarting...');
    await execute('pm2 restart data-broker');
}

async function handleDownloadLogsMessage(message) {
    logger.info('Handle download-logs message: ', message);
    const {
        body: { name },
        attributes: { dataBrokerId }
    } = message;
    const downloadLogsTempDirectory = `./tmp/downloadLogs/${last(name.split('/'))}`;
    try {
        await emptyDir(downloadLogsTempDirectory);
        const zipDirectoryName = `${dataBrokerId}-logs`;
        const isLocalConfigFileExist = await pathExists(LOCAL_CONFIG_FILE);

        const content = (isLocalConfigFileExist) ? [`${LOGS_DIRECTORY_RELATIVE_PATH}`, `${LOCAL_CONFIG_FILE}`] : [`${LOGS_DIRECTORY_RELATIVE_PATH}`];

        await createZipDirectory(downloadLogsTempDirectory, zipDirectoryName, content);
        logger.info(`uploading zip directory  ${zipDirectoryName}.7z to S3 ${name}`);
        await uploadZipDirectoryToS3(downloadLogsTempDirectory, `${zipDirectoryName}.7z`,
            `data-broker-logs/${name}`);
    } catch (err) {
        logger.error('Failed to upload data broker logs');
        throw err;
    } finally {
        await remove(downloadLogsTempDirectory);
        logger.info('Done handling download-logs message');
    }
}

async function handleMessage(message, queueUrl) {
    try {
        if (validateMessage(message.body, 'SQS', sqsMasterIncomingMessagesSchemes[message.attributes.type])) {
            switch (message.attributes.type) {
                case 'copy':
                    await handleCopyMessage(message);
                    break;
                case 'done':
                case 'remove':
                case 'abort':
                    await handleDoneRemoveAbortMessage(message);
                    break;
                case 'exports':
                    await handleExportsMessage(message);
                    break;
                case 'export-folders':
                    await handleExportFoldersMessage(message);
                    break;
                case 's3-buckets':
                    await handleS3BucketsMessage(message);
                    break;
                case 's3-bucket-folders':
                    await handleS3BucketFoldersMessage(message);
                    break;
                case 'gcp-buckets':
                    await handleGcpBucketsMessage(message);
                    break;
                case 'gcp-bucket-folders':
                    await handleGcpBucketFoldersMessage(message);
                    break;
                case 'startup':
                case 'add':
                    await handleStartUpAdd(message);
                    break;
                case 'update':
                    await handleUpdateMessage(message, queueUrl);
                    break;
                case 'azure-blob-storage-containers':
                    await handleAzureBlobStorageContainersMessage(message);
                    break;
                case 'azure-blob-storage-container-folders':
                    await handleAzureBlobStorageContainerFoldersMessage(message);
                    break;
                case 'shares':
                    await handleSharesMessage(message);
                    break;
                case 'share-folders':
                    await handleShareFoldersMessage(message);
                    break;
                case 'shutdown':
                    await handleShutdownMessage();
                    break;
                case 'kms-keys':
                    await handleKmsKeysMessage(message);
                    break;
                case 'restart':
                    await handleRestartMessage(message, queueUrl);
                    break;
                case 'metadata':
                    await handleMetadataMessage(message);
                    break;
                case 'send-asup':
                    await handleSendAsupMessage(message);
                    break;
                case 'save-metadata':
                    await handleSaveMetadataMessage(message);
                    break;
                case 'download-logs':
                    await handleDownloadLogsMessage(message);
                    break;
                case 'change-statuses-queue':
                    await handleChangeStatusesQueue(message, queueUrl);
                    break;
                default:
                    logger.error(`Got unknown command type ${message.attributes.type}:`, message);
            }
        }
    } catch (e) {
        const { body: { relationshipId } } = message;
        const errorMessage = merge({
            step: INITIALIZE_STEP,
            relationshipId,
            status: STATUS_FAILED,
            reason: e.message || e
        }, message.body);
        if (validateMessage(errorMessage, 'SQS', failedSchema)) {
            if (relationshipId) {
                relationshipsLogger.error(`[${relationshipId}] fatally failed due to error:`, e);
            } else {
                logger.error(`[${message.attributes.type}]  fatally failed due to error:`, e);
            }
            sendMessage(statusesQueue, errorMessage, message.attributes);
        }
    } finally {
        await deleteMessage(queueUrl, message.receiptHandle);
    }
}

async function poll(queueUrl) {
    logger.debug('Polling queue', queueUrl);
    try {
        const messages = await receiveMessages(queueUrl);
        messages.forEach(msg => handleMessage(msg, queueUrl));
        setImmediate(() => poll(queueUrl));
    } catch (err) {
        if (err.code === 'AWS.SimpleQueueService.NonExistentQueue') {
            logger.warn(`'${queueUrl}' queue does not exist -> stop polling`);
        } else if (err.code === 'InvalidClientTokenId') {
            // data broker user was deleted
            logger.warn(`'${queueUrl}' queue invalid client token -> stop polling`);
        } else {
            logger.warn(`Delaying polling for SQS queue '${queueUrl}' due to error`, err);
            setTimeout(() => poll(queueUrl), 5000);
        }
    }
}

async function sendPlacement() {
    const { type, 'data-broker-id': dataBrokerId } = DATA_BROKER_RUNTIME_CONFIG;
    const awsAccountId = await getAwsAccountId();
    const gcpProjectId = await getProjectId();
    const osMetadata = await getOsMetadata();
    const { version } = DATA_BROKER_RUNTIME_ABOUT;
    const proxy = process.env.http_proxy;
    let proxyString;
    if (proxy) {
        const proxyUrl = new URL(proxy);
        if (proxyUrl.username) {
            proxyUrl.username = '*****';
            proxyUrl.password = '*****';
        }
        proxyString = proxyUrl.toString();
    }
    const updateNewVersion = config.get('update-new-version');

    const message = Object.assign(osMetadata, {
        version, awsAccountId, proxy: proxyString, updateNewVersion, gcpProjectId
    });

    const certPath = `${CERT_AND_KEY_FOLDER}/dataBrokerCert.crt`;
    // read certificate and send it.
    message.certificate = await readFile(certPath, 'utf8');

    try {
        let messageType;
        switch (type) {
            case DEPLOYMENT_TYPES.AWS: {
                const awsMetadata = await getAwsMetadata();
                Object.assign(message, awsMetadata);
                messageType = sqsMasterOutgoingMessagesSchemes['placement-aws'];
                break;
            }
            case DEPLOYMENT_TYPES.ONPREM: {
                messageType = sqsMasterOutgoingMessagesSchemes['placement-on-prem'];
                break;
            }
            case DEPLOYMENT_TYPES.AZURE: {
                const azureMetadata = await getAzureMetadata();
                Object.assign(message, azureMetadata);
                messageType = sqsMasterOutgoingMessagesSchemes['placement-azure'];
                break;
            }
            case DEPLOYMENT_TYPES.GCP: {
                let gcpMetadata;
                try {
                    gcpMetadata = await getGcpMetadata();
                } catch (err) {
                    logger.warn('Couldn\'t get GCP metadata. ', err);
                }
                if (gcpMetadata) {
                    Object.assign(message, gcpMetadata);
                    messageType = sqsMasterOutgoingMessagesSchemes['placement-gcp'];
                } else {
                    messageType = sqsMasterOutgoingMessagesSchemes['placement-on-prem'];
                }
                break;
            }
            default:
                logger.error(`Failed to send placement - unknown deployement type: ${type}`);
        }

        if (validateMessage(message, 'Outgoing-SQS', messageType)) {
            await sendMessage(statusesQueue, message, { type: 'placement', dataBrokerId });
        }
    } catch (err) {
        logger.error('Failed to send placement', err);
    }
}

async function getTransferRate() {
    try {
        const networkStats = await si.networkStats('*');

        return networkStats
            .filter(networkStat => networkStat.iface !== 'lo')
            .filter(networkStat => networkStat.tx_sec !== -1)
            .map(networkStat => {
                if (Object.is(Infinity, networkStat.tx_sec)) {
                    return 0;
                }
                return networkStat.tx_sec;
            })
            .reduce((accumulator, value) => accumulator + value, 0);
    } catch (err) {
        logger.warn('Unable to get transfer rate', err);
        return 0;
    }
}

async function sendHealth() {
    const { 'data-broker-id': dataBrokerId } = DATA_BROKER_RUNTIME_CONFIG;
    const message = {
        transferrers: getWorkersIdsByType(WORKER_TYPE_TRANSFERRER).length,
        scanners: getWorkersIdsByType(WORKER_TYPE_SCANNER).length,
        memory: process.memoryUsage(),
        cpuUsage: process.cpuUsage(),
        transferRate: await getTransferRate()
    };

    try {
        if (validateMessage(message, 'Outgoing-SQS', sqsMasterOutgoingMessagesSchemes.health)) {
            sendMessage(statusesQueue, message, { type: 'health', dataBrokerId });
        }
    } catch (err) {
        logger.error('Failed to send health message', err);
    }
}

function findTargetRequiredAction(status, step, targetRequiredAction) {
    if (status === STATUS_FAILED && step === LIST_STEP) {
        return STATUS_FAILED;
    }
    if (status === STATUS_FAILED && step === TRANSFER_STEP) {
        if (targetRequiredAction === ACTION_DELETE_TARGET) {
            return ACTION_DELETE_FAILED;
        }
        return ACTION_COPY_FAILED;
    }

    return targetRequiredAction;
}

async function boot() {
    logger.info('Booting, number of CPUs', cpus().length);

    cluster
        .on('exit', async (worker, code, signal) => {
            const { workerType } = worker;
            logger.warn(`Worker #${worker.id}, PID ${worker.process.pid} died ${(signal || code)}.`);
            await remove(`core.${worker.process.pid}`);

            if (workerType === WORKER_TYPE_SCANNER) {
                logger.info(`Restarting worker of type ${workerType}.`);
                const scannerProcess = forkProcess(join(__dirname, 'scanner.js'), workerType);
                WORKERS_IDS[worker.id] = scannerProcess.process.pid;
                scannerProcess.send({
                    workerId: scannerProcess.id,
                    type: 'create-scanner'
                });
                Object.keys(META_DATA).forEach(async relationshipId => {
                    scannerProcess.send({
                        relationshipId,
                        scannerQueue: META_DATA[relationshipId].scannerQueue,
                        transferrerQueue: META_DATA[relationshipId].transferrerQueue,
                        requestId: META_DATA[relationshipId].requestId,
                        settings: META_DATA[relationshipId].settings,
                        certificate: META_DATA[relationshipId].certificate,
                        serverName: META_DATA[relationshipId].serverName,
                        type: 'copy'
                    });
                });
            } else if (workerType === WORKER_TYPE_TRANSFERRER) {
                logger.info(`Restarting worker of type ${workerType}.`);
                const transferrerProcess = forkProcess(join(__dirname, 'transferrer.js'), workerType);
                Object.keys(META_DATA).forEach(async relationshipId => {
                    transferrerProcess.send({
                        relationshipId,
                        transferrerQueue: META_DATA[relationshipId].transferrerQueue,
                        requestId: META_DATA[relationshipId].requestId,
                        certificate: META_DATA[relationshipId].certificate,
                        serverName: META_DATA[relationshipId].serverName,
                        settings: {
                            retries: META_DATA[relationshipId].settings.retries
                        },
                        type: 'copy'
                    });
                });
            } else if (workerType === WORKER_TYPE_SERVER && code !== 8) {
                logger.info(`Restarting worker of type ${workerType}.`);
                forkProcess(join(__dirname, 'server.js'), workerType);
            } else if (workerType === WORKER_TYPE_ACL_SERVER) {
                logger.info(`Restarting worker of type ${workerType}.`);
                const aclServerProcess = forkProcess(join(__dirname, 'acl-server.js'), workerType);
                aclServerProcess.send({
                    port: ACL_SERVERS[worker.process.pid],
                    type: 'create-acl-server'
                });
                ACL_SERVERS[aclServerProcess.process.pid] = ACL_SERVERS[worker.process.pid];
                delete ACL_SERVERS[worker.process.pid];
            }
        })
        .on('message', (worker, message) => {
            if (message.topic !== 'log4js:message') {
                const {
                    status, relationshipId, objectType, step, bytes, items, reason
                } = message;

                const targetRequiredAction = findTargetRequiredAction(status, step, message.targetRequiredAction);

                update(RELATIONSHIPS_STATUS, `${relationshipId}.${step}.${targetRequiredAction}.${objectType}`, obj => (obj ? {
                    bytes: obj.bytes + bytes,
                    items: obj.items + items
                } : { bytes, items }));
                RELATIONSHIPS_STATUS[relationshipId].attributes = message.attributes;

                update(TOTALS, `${relationshipId}.${step}.${targetRequiredAction}.${objectType}`, obj => (obj ? {
                    bytes: obj.bytes + bytes,
                    items: obj.items + items
                } : { bytes, items }));

                if (status === STATUS_FAILED) {
                    update(RELATIONSHIPS_STATUS, `${relationshipId}.${step}.relationshipErrors.${reason.code}`, obj => {
                        if (obj) {
                            obj.counter += 1;
                            return obj;
                        }
                        const res = { counter: 1 };
                        if (reason.description) {
                            res.description = reason.description;
                        }
                        return res;
                    });
                }
            }
        });

    if (prettyBytes(config.get('protocols.s3.max-object-size')) < prettyBytes('1TB')) {
        logger.warn('consider increasing protocols.s3.max-object-size in order to avoid batching problems');
    }

    await generateCertificate();

    try {
        await prepareVault();
    } catch (err) {
        logger.error('Failed to install vault and register it as pm2 process', err);
    }

    startTransferrers();
    startScanners();
    startAclServers();

    if (DATA_BROKER_RUNTIME_CONFIG.port) {
        startServer();
    }

    removeCoreFiles();
    setInterval(() => sendHealth(), config.get('statuses.health.send-interval'));
    setInterval(() => Object.keys(META_DATA).forEach(relationshipId => sendStatistics(relationshipId)), config.get('statuses.statistics.interval'));

    await poll(DATA_BROKER_RUNTIME_CONFIG.commandsQueue);
    await sendPlacement();
}

module.exports = boot;
