const cluster = require('cluster');
const https = require('https');
const http = require('http');
const config = require('config');
const bytes = require('bytes');
const randomize = require('randomatic');
const moment = require('moment');
const { readFileSync } = require('fs-extra');
const { queue } = require('async');
const {
    _, shuffle, head, partition, keyBy, last, mapValues, keys, compact, isEqualWith, has, set, get, includes, sumBy, groupBy, isEmpty, chunk, update, flatten, find, pullAll
} = require('lodash');
const ms = require('ms');
const Promise = require('bluebird');
const retry = require('async-retry');
const { join } = require('path');
const getLogger = require('../lib/logger/logger');
const {
    listObjectsOnce, isEndpointEmpty, filterEqualObjects, areCloudObjectsEqual
} = require('../lib/cloud-to-cloud/cloud-scan');
const {
    closeDB, openDB, putIntoDB, removeFromDB, readFromDB
} = require('../lib/cloud-to-cloud/db');
const { list, gracefulJoin, generateErrorStatisticsObject } = require('../lib/fs/fs');
const {
    receiveMessages, sendMessage, deleteMessage, changeMessageVisibilityBatch
} = require('../lib/aws/sqs');
const {
    ONTAP_S3_PROVIDER, TYPE_FILE, TYPE_DIRECTORY, TYPE_OTHER, TYPE_SYMBOLIC, SCAN_MODE_SOURCE_ONLY, SCAN_MODE_TARGET_ONLY, SCAN_MODE_SOURCE_AND_TARGET, SCAN_MODE_CLOUD_TO_CLOUD,
    ACTION_DELETE_SOURCE, ACTION_DELETE_TARGET, ACTION_COPY, ACTION_IGNORE, ACTION_LIST, STATUS_SUCCEEDED, STATUS_FAILED,
    STATUS_IGNORED, SCAN_STEP, LIST_STEP, IPC_TYPE_DONE, IPC_TYPE_UPDATE, IPC_TYPE_ABORT, IPC_TYPE_REMOVE, IPC_TYPE_COPY,
    NETAPP_ENTRY_SUFFIX, NETAPP_ENTRY_RANDOM_SIZE, NETAPP_ENTRY_RANDOM_PATTERN,
    MOUNT_SOURCE_SUFFIX, MOUNT_TARGET_SUFFIX, SQS_DEFAULT_WAIT_TIMEOUT, SQS_MESSAGE_UNCHANGED_VISIBILITY_TIMEOUT_LIMIT, SQS_CHANGE_MESSAGE_VISIBILITY_TIMEOUT_INTERVAL,
    ALIBABA_PROVIDER, IBM_PROVIDER, SGWS_PROVIDER, S3_PROVIDER
} = require('../lib/consts');
const { listBlobs, getBlobServiceClientByData } = require('../lib/azure/blob');
const { listGcpObjects, getStorage } = require('../lib/gcp/storage');
const scannerIncomingSQSSchema = require('../messages/scanner/sqs-incoming-messages');
const scannerIncomingIPCMessagesSchemes = require('../messages/scanner/ipc-incoming-messages');
const { copyStatisticsSchema } = require('../messages/schemas/statuses-outgoing-messages');
const validateMessage = require('../messages/schemas/message-validator');
const { getNewEndPointIfNeeded } = require('../lib/fs/mount_auxiliary');
const { startHealthLogInterval } = require('../lib/health');
const { listSecureNFS } = require('../lib/snfs/snfs');
const { calculateObjectStoragePath, getObjectTags } = require('../lib/aws/s3_base_operations');
const {
    getClient
} = require('../lib/s3Clients');
const {
    listObjectsV2, headObject, listObjects
} = require('../lib/aws/s3_base_operations');

const logger = getLogger();
const relationshipsLogger = getLogger('relationships');
const failedRelationshipLogger = getLogger('failed-relationships');

http.globalAgent.maxSockets = 500;
https.globalAgent.maxSockets = 500;

startHealthLogInterval();

process.on('uncaughtException', err => logger.error('Got uncaught exception', err));
process.on('unhandledRejection', event => logger.error('Got unhandled rejection', event));

const META_DATA = {}; // relationship id to map metadata
let INTERNAL_DISPATCHER_QUEUE;
let WORKER_ID;

function createAclSet(acl) { // Acl example: 'A::OWNER@:rwa\nA::GROUP@:rtc\nA::EVERYONE@:rty'
    const acesArray = acl.split('\n'); // breaks the acl to entities (aces) [A::OWNER@:rwa, A::GROUP@:rtc, A::EVERYONE@:rty]
    acesArray.pop(); // Pops last ace which is empty string
    const aceRegex = /(.*):([^:]*)$/;
    return acesArray.reduce((accumulator, ace) => {
        const splittedAce = aceRegex.exec(ace); // splits the ace to 2 parts and creates the following array - [A::OWNER@:rwa, A::OWNER@, rwa]
        if (splittedAce[1].indexOf('GROUP') > splittedAce[1].indexOf('g') && splittedAce[1].indexOf('g') > -1) { // delete header g if exists before GROUP
            splittedAce[1] = splittedAce[1].substring(0, splittedAce[1].indexOf('g')) + splittedAce[1].substring(splittedAce[1].indexOf('g') + 1, splittedAce[1].length);
        }
        Array.from(splittedAce[2]).forEach(permission => { // converts the permissions string to array and add them one by one to the accumulator set
            accumulator.add(`${splittedAce[1]}:${permission}`);
        });
        return accumulator;
    }, new Set());
}

async function getObjectDetails(s3, provider, bucket, key, useTags, size, objectType, mode, lastModified) {
    let metadata = {};
    if (provider === S3_PROVIDER && useTags) {
        metadata = await getObjectTags(s3, bucket, key);
    } else if (provider !== ONTAP_S3_PROVIDER) {
        metadata = await headObject(s3, bucket, key);
    }
    const modifiedTime = lastModified ? lastModified.getTime() : Date.now();
    if (!key.endsWith('/') || objectType !== TYPE_FILE) {
        return {
            path: key === '/' ? '/' : last(compact(key.split('/'))),
            type: metadata['netapp-cs-slink'] === 'True' ? TYPE_SYMBOLIC : objectType,
            size,
            mtimeSec: parseInt(metadata['netapp-cs-mtimesec']) || parseInt(modifiedTime / 1000),
            mode: parseInt(metadata['netapp-cs-mode']) || mode,
            uid: parseInt(metadata['netapp-cs-uid']) || 0,
            gid: parseInt(metadata['netapp-cs-gid']) || 0
        };
    }
}

async function listObjectsOperation(endPoint, relationshipId, srcOrTrg) {
    const { provider } = endPoint;
    const s3 = await getClient(provider, {
        ...endPoint, relationshipId, srcOrTrg, isS3EndPoint: true
    });
    const useTags = get(META_DATA, `${relationshipId}.settings.objectTagging`, true);
    switch (provider) {
        case ALIBABA_PROVIDER:
        case S3_PROVIDER: {
            const { bucket, path } = endPoint;
            const allObjects = await listObjectsV2(s3, bucket, path);
            const objectsWithTags = await Promise.map(allObjects, object => getObjectDetails(s3, provider, bucket, object.key, useTags, object.size, object.type, object.type === TYPE_DIRECTORY ? config.get('protocols.s3.default-directory-mode') : config.get('protocols.s3.default-file-mode'), object.lastModified), { concurrency: 100 });
            return compact(objectsWithTags);
        }
        case ONTAP_S3_PROVIDER:
        case IBM_PROVIDER:
        case SGWS_PROVIDER: {
            const {
                bucket, path
            } = endPoint;
            const allObjects = await listObjects(s3, bucket, path);
            const objectsWithTags = await Promise.map(allObjects, object => getObjectDetails(s3, provider, bucket, object.key, useTags, object.size, object.type, object.type === TYPE_DIRECTORY ? config.get('protocols.sgws.default-directory-mode') : config.get('protocols.sgws.default-file-mode'), object.lastModified), { concurrency: 100 });
            return compact(objectsWithTags);
        }
        default: {
            throw new Error(`Unknown provider ${provider}`);
        }
    }
}

function isEqualSets(set1, set2) {
    if (set1.size !== set2.size) {
        return false;
    }
    return [...set1].reduce((accumulator, elem) => {
        if (!set2.has(elem)) {
            accumulator = false; // eslint-disable-line no-param-reassign
        }
        return accumulator;
    }, true);
}

function addMinimalPermissions(acl) {
    return config.get('protocols.nfs.acl.minimal-acl').reduce((accumulator, ace) => accumulator.concat(`${ace}\n`), acl);
}

function compareAcls(sourceAcl, targetAcl) {
    return isEqualSets(createAclSet(targetAcl), createAclSet(addMinimalPermissions(sourceAcl)));
}

function compareFiles(sourceFiles, targetFiles, sourceProtocol, targetProtocol, targetProvider) {
    const sourceFilesObj = keyBy(sourceFiles, 'path');
    const targetFilesObj = keyBy(targetFiles, 'path');

    const result = keys(sourceFilesObj).reduce((accumulator, path) => {
        const src = sourceFilesObj[path];
        const trg = targetFilesObj[path];
        if (config.has(`protocols.${targetProtocol}.${targetProvider}.max-object-size`) && src.size > bytes(config.get(`protocols.${targetProtocol}.${targetProvider}.max-object-size`))) {
            // object stores usually have object size limit -> we have to ignore such files
            src.reason = 'exceeding size limit';
            accumulator.TASK_TYPE_IGNORE.push(src);
        } else if (src.type === TYPE_OTHER) {
            // CS does not support special files, such as socket, pipe, etc.
            src.reason = 'special file';
            accumulator.TASK_TYPE_IGNORE.push(src);
        } else if (!trg || !isEqualWith(src, trg, (val1, val2, key) => {
            switch (key) {
                case 'nfs4Acl':
                    return compareAcls(val1, val2);
                case 'uid':
                case 'gid':
                case 'mode':
                    return targetProtocol === 'cifs' ? true : undefined;
                default:
                    return undefined;
            }
        })) {
            // same file appears in source but not in target -> copy
            // same file appears in source and target, but has different properties -> copy
            accumulator.TASK_TYPE_COPY.push(src);
        } else {
            accumulator.TASK_TYPE_DELETE_SOURCE.push(src); // the files are equal
        }

        return accumulator;
    }, {
        TASK_TYPE_COPY: [],
        TASK_TYPE_DELETE_SOURCE: [],
        TASK_TYPE_DELETE_TARGET: [],
        TASK_TYPE_IGNORE: []
    });

    return keys(targetFilesObj).reduce((accumulator, path) => {
        const src = sourceFilesObj[path];
        const trg = targetFilesObj[path];
        if (trg.type === TYPE_OTHER) {
            trg.reason = 'special file';
            accumulator.TASK_TYPE_IGNORE.push(trg);
        } else if (!src) {
            // file in target but not in source -> delete
            accumulator.TASK_TYPE_DELETE_TARGET.push(trg);
        }

        return accumulator;
    }, result);
}

function compareDirs(sourceDirs, targetDirs, sourceProtocol, targetProtocol) {
    let sourceDirsObj = keyBy(sourceDirs, 'path');
    let targetDirsObj = keyBy(targetDirs, 'path');

    const cifsInvolved = sourceProtocol === 'cifs' || targetProtocol === 'cifs';
    if (targetProtocol === 'cifs') { // Cifs is case-insensitive. This solves slb use case where they have "ep" folder on GCP and "EP" folder on SMB which is marked for copy again and again.
        sourceDirsObj = Object.fromEntries(Object.entries(sourceDirsObj).map(([k, v]) => [k.toLowerCase(), v]));
        targetDirsObj = Object.fromEntries(Object.entries(targetDirsObj).map(([k, v]) => [k.toLowerCase(), v]));
    }

    const result = keys(sourceDirsObj).reduce((accumulator, path) => {
        const src = sourceDirsObj[path];
        const trg = targetDirsObj[path];
        if (includes(config.get('workers.scanner.ignore.directories'), path)) {
            src.reason = 'ignored directory';
            accumulator.TASK_TYPE_IGNORE.push(src);
        } else if (!trg) {
            // in source, not in target -> copy all
            accumulator.TASK_TYPE_SOURCE_ONLY.push(src);
        } else if (!isEqualWith(src, trg, (val1, val2, key) => {
            switch (key) {
                case 'path':
                    return targetProtocol === 'cifs' ? val1.toLowerCase() === val2.toLowerCase() : undefined;
                case 'mtimeSec':
                    return true;
                case 'nfs4Acl':
                    return compareAcls(val1, val2);
                case 'uid':
                case 'gid':
                case 'mode':
                    return cifsInvolved ? true : undefined;
                default:
                    return undefined;
            }
        })) {
            // we have to ignore modification time of directories since there is not way (currently, without creating additional SQS for directories only) to make sure it is correct
            // creation/modification of file in target changes modification time of the parent directory
            // moreover - we want to compare acls with our customized function because acls behave differently on different OS distributions

            // source and target directories exist but have not equal attributes -> dive in and copy the attributes of the directory
            accumulator.TASK_TYPE_EQUAL.push(src);
            accumulator.TASK_TYPE_EQUAL_AND_COPY.push(src);
        } else {
            // source and target directories are equal -> dive in
            accumulator.TASK_TYPE_EQUAL.push(src);
        }

        return accumulator;
    }, {
        TASK_TYPE_EQUAL: [],
        TASK_TYPE_SOURCE_ONLY: [],
        TASK_TYPE_TARGET_ONLY: [],
        TASK_TYPE_IGNORE: [],
        TASK_TYPE_EQUAL_AND_COPY: []
    });

    return keys(targetDirsObj).reduce((accumulator, path) => {
        const src = sourceDirsObj[path];
        const trg = targetDirsObj[path];
        if (includes(config.get('workers.scanner.ignore.directories'), path)) {
            trg.reason = 'ignored directory';
            accumulator.TASK_TYPE_IGNORE.push(trg);
        } else if (!src) {
            // directory exists in target but not in source -> scan target in order to remove
            accumulator.TASK_TYPE_TARGET_ONLY.push(trg);
        }

        return accumulator;
    }, result);
}

function transformPath(endPoint, directory) {
    const protocol = head(keys(endPoint));
    switch (protocol) {
        case 'cifs':
        case 'nfs':
        case 'local':
        case 'snfs':
            return mapValues(endPoint, value => ({ ...value, path: gracefulJoin(value.path, directory.path) }));
        case 's3':
        case 'sgws':
        case 'azure':
        case 'gcp':
            return mapValues(endPoint, value => ({
                ...value,
                path: `${calculateObjectStoragePath(value.path, directory.path)}`
            }));
        default:
            throw new Error(`transformPath, got wrong protocol for endPoint ${endPoint}`);
    }
}

async function sendMessagesInChunks(messages, workerQueue, attributes) {
    const entriesPerChunk = config.get('workers.scanner.messages-zip-ratio'); // Tests on long paths entries showed that the average is ~2000 entries per chunk (taking into account they will be zipped)
    await Promise.map(chunk(messages, entriesPerChunk), chunky => {
        attributes.MessagesCount = chunky.length.toString();
        return sendMessage(workerQueue, chunky, attributes);
    }, { concurrency: 100 });
}

async function sendScannerMessages(message, type, directories) {
    const {
        body: {
            source, target, relationshipId
        }
    } = message;
    const queues = META_DATA[relationshipId] ? META_DATA[relationshipId].queues : undefined;
    if (queues) {
        await sendMessagesInChunks(directories.map(dir => {
            const scannerMessage = {
                relationshipId,
                source: transformPath(source, dir),
                target: transformPath(target, dir),
                type
            };
            logger.debug(`Send message to scanner got type: ${type}, message: ${scannerMessage}`);
            return scannerMessage;
        }), queues.scannerQueue, message.attributes);
    }
}

function updateStatisticsMessages(statisticsMessages, key, entry) {
    if (has(statisticsMessages, key)) {
        update(statisticsMessages, key, msg => ({ items: msg.items + 1, bytes: msg.bytes + entry.size }));
    } else {
        set(statisticsMessages, key, { items: 1, bytes: entry.size });
    }
}

function sendStatisticsMessage(statisticsMessage) {
    const isValid = validateMessage(statisticsMessage, 'SQS', copyStatisticsSchema);
    if (isValid) {
        logger.debug('Sending statistics message:', JSON.stringify(statisticsMessage, null, 4));
        process.send(statisticsMessage);
    }
}

const createListStatsMessage = (itemsBytes, items, objectType, relationshipId, attributes) => ({
    step: LIST_STEP,
    relationshipId,
    targetRequiredAction: ACTION_LIST,
    bytes: itemsBytes,
    items,
    objectType,
    status: STATUS_SUCCEEDED,
    attributes
});

function getEndPointProtocol(endPoint) {
    return _(endPoint).keys().head();
}

async function sendTransferrerMessages(message, type, entries, source, target) {
    if (isEmpty(entries)) {
        return;
    }
    const protocolsForMultipart = ['local', 'nfs', 'cifs'];
    const { body: { relationshipId } } = message;
    const queues = META_DATA[relationshipId] ? META_DATA[relationshipId].queues : undefined;

    if (queues) {
        const statisticsMessages = {};
        const transferrerMessages = await Promise.map(entries, entry => {
            updateStatisticsMessages(statisticsMessages, JSON.stringify([type, entry.type]), entry);
            if (protocolsForMultipart.includes(getEndPointProtocol(source)) && protocolsForMultipart.includes(getEndPointProtocol(target))) {
                const partSize = bytes(config.get('protocols.nfs.max-entry-part-size'));
                if (entry.size > partSize) {
                    logger.info(`entry that can be copied as multipart source ${JSON.stringify(source)} target ${JSON.stringify(target)}, ${JSON.stringify(entry)}`);
                    entry.partsDetails = {
                        partNumber: 1,
                        start: 0,
                        end: Math.min(partSize, entry.size),
                        partSize
                    };
                }
            }
            if (protocolsForMultipart.includes(getEndPointProtocol(target))) {
                entry.tempEntryName = `.${entry.path}_${randomize(NETAPP_ENTRY_RANDOM_PATTERN, NETAPP_ENTRY_RANDOM_SIZE)}${NETAPP_ENTRY_SUFFIX}`;
                if (entry.tempEntryName.length > 255) {
                    entry.tempEntryName = entry.path;
                }
            }

            return {
                type,
                source: type === ACTION_COPY || ACTION_DELETE_SOURCE ? source : undefined,
                target,
                entry,
                relationshipId
            };
        }, { concurrency: 100 });
        Object.entries(statisticsMessages).forEach(([key, value]) => {
            const parsedKey = JSON.parse(`[${key}]`);
            sendStatisticsMessage({
                step: SCAN_STEP,
                relationshipId,
                targetRequiredAction: parsedKey[0],
                bytes: value.bytes,
                items: value.items,
                objectType: parsedKey[1],
                status: STATUS_SUCCEEDED,
                attributes: message.attributes
            });
        });
        logger.debug('length of transferrerMessages to send', transferrerMessages.length);
        await sendMessagesInChunks(transferrerMessages, queues.transferrerQueue, message.attributes);
    }
}

async function filterEqualObjectsFromDB(objects, DB, isTargetDB, message, deleteOnSource, minimumValue = '') {
    const DBobjects = await readFromDB(DB, minimumValue);
    if (DBobjects.length === 0) { return objects; }
    const equalObjects = [];
    await Promise.map(objects, async object => {
        try {
            const stringifiedDBObject = await DB.get(object.path); // If the key doesn't exist in DB, we fall to catch
            try {
                if (areCloudObjectsEqual(object, JSON.parse(stringifiedDBObject))) {
                    equalObjects.push(object);
                    await removeFromDB(DB, [object]);
                } else if (isTargetDB) {
                    await removeFromDB(DB, [object]); // In the end targetDB should contain only keys eligible for deleteOnTarget
                }
            } catch (e) {
                relationshipsLogger.error(`[${message.body.relationshipId}] Failed to remove entry from DB: `, object.path); // TODO What is the expected behavior?
            }
        } catch (e) {
            if (e.notFound) { logger.debug(`Could not find entry ${object.path} in target DB`); }
        }
    }, { concurrency: 100 });
    if (deleteOnSource) {
        await sendTransferrerMessages(message, ACTION_DELETE_SOURCE, equalObjects, message.body.source, message.body.target);
    }
    return filterEqualObjectsFromDB(pullAll(objects, equalObjects), DB, isTargetDB, message, deleteOnSource, DBobjects[DBobjects.length - 1].path);
}
function getDatabrokerId() {
    return JSON.parse(readFileSync('config/data-broker.json'))['data-broker-id'];
}

async function hasMessagePassedInFlightLimit(message) {
    const { body: { relationshipId } } = message;
    const hoursDiff = moment().diff(META_DATA[relationshipId]['cloud-to-cloud'].receiveTimeStamp, 'hours');
    if (hoursDiff >= config.get('workers.scanner.msg-in-flight-hours-limit')) {
        logger.info(`Sending refreshed cloud-to-cloud message for relationship ${relationshipId} after ${hoursDiff} hours`);
        META_DATA[relationshipId]['cloud-to-cloud'].receiveTimeStamp = moment();
        await sendMessage(META_DATA[relationshipId].queues.scannerQueue, [message.body], Object.assign(message.attributes, { addresseeScanner: `${getDatabrokerId()}-${WORKER_ID}` }));
        await deleteMessage(META_DATA[relationshipId].queues.scannerQueue, META_DATA[relationshipId]['cloud-to-cloud'].receiptHandle);
    }
}

async function fillDBs(sourceProvider, targetProvider, sourceProtocol, targetProtocol, sourceDB, targetDB, sourceClient, targetClient, sourceBucket, targetBucket, sourcePrefix, targetPrefix, message, deleteOnSource, sourceToken, targetToken) {
    if (META_DATA[message.body.relationshipId]) {
        await hasMessagePassedInFlightLimit(message);
        const useEtags = sourceProtocol === targetProtocol;
        const [[sourceObjects, sourceNewToken], [targetObjects, targetNewToken]] = await Promise.all([listObjectsOnce(sourceProvider, sourceProtocol, sourceClient, sourceBucket, sourcePrefix, sourceToken, useEtags), listObjectsOnce(targetProvider, targetProtocol, targetClient, targetBucket, targetPrefix, targetToken, useEtags)]);
        sendStatisticsMessage(createListStatsMessage(sumBy(sourceObjects, object => object.size), sourceObjects.length, TYPE_FILE, message.body.relationshipId, message.attributes));
        let [filteredSourceObjects, filteredTargetObjects, equalObjects] = filterEqualObjects(sourceObjects, targetObjects);
        if (deleteOnSource) {
            await sendTransferrerMessages(message, ACTION_DELETE_SOURCE, equalObjects, message.body.source, message.body.target);
        }
        [filteredSourceObjects, filteredTargetObjects] = await Promise.all([filterEqualObjectsFromDB(filteredSourceObjects, targetDB, true, message, deleteOnSource), filterEqualObjectsFromDB(filteredTargetObjects, sourceDB, false, message, deleteOnSource)]);
        await Promise.all([putIntoDB(sourceDB, filteredSourceObjects), putIntoDB(targetDB, filteredTargetObjects)]);
        if (sourceNewToken || targetNewToken) {
            return fillDBs(sourceProvider, targetProvider, sourceProtocol, targetProtocol, sourceDB, targetDB, sourceClient, targetClient, sourceBucket, targetBucket, sourcePrefix, targetPrefix, message, deleteOnSource, sourceNewToken || null, targetNewToken || null);
        }
    }
}

async function readDBAndSendToTransferrer(DB, action, message, minimumValue = '') {
    if (META_DATA[message.body.relationshipId]) {
        const objects = await readFromDB(DB, minimumValue);
        if (objects.length === 0) {
            return;
        }
        sendTransferrerMessages(message, action, objects.map(object => Object.assign(object, {
            type: TYPE_FILE, uid: 0, gid: 0, mode: 33206
        })), message.body.source, message.body.target); // we don't want to save TYPE_FILE in the DB for each object
        return readDBAndSendToTransferrer(DB, action, message, objects[objects.length - 1].path);
    }
}

async function useAndCloseDBs(sourceDB, targetDB, relationshipId, message, settings) {
    await Promise.all([readDBAndSendToTransferrer(sourceDB, ACTION_COPY, message), settings.deleteOnTarget ? readDBAndSendToTransferrer(targetDB, ACTION_DELETE_TARGET, message) : undefined]);
    await Promise.all([closeDB(sourceDB, join('db', `${relationshipId}_src`)), closeDB(targetDB, join('db', `${relationshipId}_trg`))]);
}

async function readBucketAndSendToTransferrer(provider, protocol, client, bucket, prefix, message, continuationToken) {
    if (META_DATA[message.body.relationshipId]) {
        await hasMessagePassedInFlightLimit(message);
        const result = await listObjectsOnce(provider, protocol, client, bucket, prefix, continuationToken);
        sendStatisticsMessage(createListStatsMessage(sumBy(result[0], object => object.size), result[0].length, TYPE_FILE, message.body.relationshipId, message.attributes));
        sendTransferrerMessages(message, ACTION_COPY, result[0].map(object => Object.assign(object, {
            path: object.path, type: TYPE_FILE, uid: 0, gid: 0, mode: 33206
        })), message.body.source, message.body.target);
        if (result[1]) {
            return readBucketAndSendToTransferrer(provider, protocol, client, bucket, prefix, message, result[1]);
        }
    }
}

const CLOUD_TO_CLOUD_TABLE = [
    // s3 -> s3
    {
        source: 's3',
        target: 's3',
        prepareScan: async (source, target, relationshipId) => {
            const { bucket: sourceBucket, path: sourcePrefix, provider: sourceProvider } = source.s3;
            const { bucket: targetBucket, path: targetPrefix, provider: targetProvider } = target.s3;
            const sourceClient = await getClient(sourceProvider, {
                ...source.s3, relationshipId, srcOrTrg: MOUNT_SOURCE_SUFFIX, isS3EndPoint: true
            });
            const targetClient = await getClient(targetProvider, {
                ...target.s3, relationshipId, srcOrTrg: MOUNT_TARGET_SUFFIX, isS3EndPoint: true
            });
            return {
                sourceProvider, targetProvider, sourceClient, targetClient, sourceBucket, targetBucket, sourcePrefix, targetPrefix
            };
        }
    },
    // s3->azure
    {
        source: 's3',
        target: 'azure',
        prepareScan: async (source, target, relationshipId) => {
            const { bucket: sourceBucket, path: sourcePrefix, provider: sourceProvider } = source.s3;
            const { container: targetContainer, path: targetPrefix } = target.azure;
            const sourceClient = await getClient(sourceProvider, {
                ...source.s3, relationshipId, srcOrTrg: MOUNT_SOURCE_SUFFIX, isS3EndPoint: true
            });
            const blobServiceClient = await getBlobServiceClientByData(relationshipId, MOUNT_TARGET_SUFFIX);
            const targetClient = blobServiceClient.getContainerClient(targetContainer);
            return {
                sourceProvider, sourceClient, targetClient, sourceBucket, targetBucket: targetContainer, sourcePrefix, targetPrefix
            };
        }
    },
    {
        source: 's3',
        target: 'gcp',
        prepareScan: async (source, target, relationshipId) => {
            const { bucket: sourceBucket, path: sourcePrefix, provider: sourceProvider } = source.s3;
            const { bucket: targetBucket, path: targetPrefix } = target.gcp;
            const sourceClient = await getClient(sourceProvider, {
                ...source.s3, relationshipId, srcOrTrg: MOUNT_SOURCE_SUFFIX, isS3EndPoint: true
            });
            const targetClient = getStorage();
            return {
                sourceProvider, sourceClient, targetClient, sourceBucket, targetBucket, sourcePrefix, targetPrefix
            };
        }
    },
    {
        source: 'azure',
        target: 's3',
        prepareScan: async (source, target, relationshipId) => {
            const { container: sourceContainer, path: sourcePrefix } = source.azure;
            const { bucket: targetBucket, path: targetPrefix, provider: targetProvider } = target.s3;
            const blobServiceClient = await getBlobServiceClientByData(relationshipId, MOUNT_SOURCE_SUFFIX);
            const sourceClient = blobServiceClient.getContainerClient(sourceContainer);
            const targetClient = await getClient(targetProvider, {
                ...target.s3, relationshipId, srcOrTrg: MOUNT_TARGET_SUFFIX, isS3EndPoint: true
            });
            return {
                targetProvider, sourceClient, targetClient, sourceBucket: sourceContainer, targetBucket, sourcePrefix, targetPrefix
            };
        }
    },
    {
        source: 'azure',
        target: 'azure',
        prepareScan: async (source, target, relationshipId) => {
            const { container: sourceContainer, path: sourcePrefix } = source.azure;
            const { container: targetContainer, path: targetPrefix } = target.azure;
            const sourceBlobServiceClient = await getBlobServiceClientByData(relationshipId, MOUNT_SOURCE_SUFFIX);
            const sourceClient = sourceBlobServiceClient.getContainerClient(sourceContainer);
            const targetBlobServiceClient = await getBlobServiceClientByData(relationshipId, MOUNT_TARGET_SUFFIX);
            const targetClient = targetBlobServiceClient.getContainerClient(targetContainer);
            return {
                sourceClient, targetClient, sourceBucket: sourceContainer, targetBucket: targetContainer, sourcePrefix, targetPrefix
            };
        }
    },
    {
        source: 'azure',
        target: 'gcp',
        prepareScan: async (source, target, relationshipId) => {
            const { container: sourceContainer, path: sourcePrefix } = source.azure;
            const { bucket: targetBucket, path: targetPrefix } = target.gcp;
            const blobServiceClient = await getBlobServiceClientByData(relationshipId, MOUNT_SOURCE_SUFFIX);
            const sourceClient = blobServiceClient.getContainerClient(sourceContainer);
            const targetClient = getStorage();
            return {
                sourceClient, targetClient, sourceBucket: sourceContainer, targetBucket, sourcePrefix, targetPrefix
            };
        }
    },
    {
        source: 'gcp',
        target: 's3',
        prepareScan: async (source, target, relationshipId) => {
            const { bucket: sourceBucket, path: sourcePrefix } = source.gcp;
            const { bucket: targetBucket, path: targetPrefix, provider: targetProvider } = target.s3;
            const sourceClient = await getStorage();
            const targetClient = await getClient(targetProvider, {
                ...target.s3, relationshipId, srcOrTrg: MOUNT_TARGET_SUFFIX, isS3EndPoint: true
            });
            return {
                targetProvider, sourceClient, targetClient, sourceBucket, targetBucket, sourcePrefix, targetPrefix
            };
        }
    },
    {
        source: 'gcp',
        target: 'azure',
        prepareScan: async (source, target, relationshipId) => {
            const { bucket: sourceBucket, path: sourcePrefix } = source.gcp;
            const { container: targetContainer, path: targetPrefix } = target.azure;
            const sourceClient = await getStorage();
            const blobServiceClient = await getBlobServiceClientByData(relationshipId, MOUNT_TARGET_SUFFIX);
            const targetClient = blobServiceClient.getContainerClient(targetContainer);
            return {
                sourceClient, targetClient, sourceBucket, targetBucket: targetContainer, sourcePrefix, targetPrefix
            };
        }
    },
    {
        source: 'gcp',
        target: 'gcp',
        prepareScan: async (source, target) => {
            const { bucket: sourceBucket, path: sourcePrefix } = source.gcp;
            const { bucket: targetBucket, path: targetPrefix } = target.gcp;
            const sourceClient = await getStorage();
            const targetClient = await getStorage();
            return {
                sourceClient, targetClient, sourceBucket, targetBucket, sourcePrefix, targetPrefix
            };
        }
    }
];

async function cloudToCloudScan(source, target, relationshipId, message, settings) {
    const [sourceProtocol, targetProtocol] = [head(keys(source)), head(keys(target))];
    const match = find(CLOUD_TO_CLOUD_TABLE, {
        source: sourceProtocol,
        target: targetProtocol
    });
    const {
        sourceProvider, targetProvider, sourceClient, targetClient, sourceBucket, targetBucket, sourcePrefix, targetPrefix
    } = await match.prepareScan(source, target, relationshipId);
    const isTargetEmpty = await isEndpointEmpty(targetProvider, targetProtocol, targetClient, targetBucket, targetPrefix);
    if (isTargetEmpty) {
        await readBucketAndSendToTransferrer(sourceProvider, sourceProtocol, sourceClient, sourceBucket, sourcePrefix, message);
    } else {
        const [sourceDB, targetDB] = await Promise.all([openDB(join('db', `${relationshipId}_src`)), openDB(join('db', `${relationshipId}_trg`))]);
        await fillDBs(sourceProvider, targetProvider, sourceProtocol, targetProtocol, sourceDB, targetDB, sourceClient, targetClient, sourceBucket, targetBucket, sourcePrefix, targetPrefix, message, settings.deleteOnSource && config.get('workers.transferrer.delete-on-source'));
        await useAndCloseDBs(sourceDB, targetDB, relationshipId, message, settings);
    }
}

function ignoreEntries(relationshipId, attributes, ...entries) {
    const statisticsMessages = {};
    flatten(entries).forEach(entry => {
        updateStatisticsMessages(statisticsMessages, entry.type, entry);
        relationshipsLogger.info(`[${relationshipId}] IGNORING`, entry);
    });
    Object.entries(statisticsMessages).forEach(([key, value]) => {
        sendStatisticsMessage({
            step: SCAN_STEP,
            relationshipId,
            targetRequiredAction: ACTION_IGNORE,
            bytes: value.bytes,
            items: value.items,
            objectType: key,
            status: STATUS_IGNORED,
            attributes
        });
    });
}

function calculateListStatistics(sourceFiles, relationshipId, attributes) {
    // directories
    const dirStatisticsMessage = createListStatsMessage(0, 1, TYPE_DIRECTORY, relationshipId, attributes);
    sendStatisticsMessage(dirStatisticsMessage);
    // partition between files and symbolic links
    // files
    const [symbolic, files] = partition(sourceFiles, entry => entry.type === TYPE_SYMBOLIC);
    const sizeOfFiles = sumBy(files, file => file.size);

    const filesStatisticsMessage = createListStatsMessage(sizeOfFiles, files.length, TYPE_FILE, relationshipId, attributes);
    sendStatisticsMessage(filesStatisticsMessage);

    // symbolic link
    const symbolicStatisticsMessage = createListStatsMessage(sumBy(symbolic, file => file.size), symbolic.length, TYPE_SYMBOLIC, relationshipId, attributes);
    sendStatisticsMessage(symbolicStatisticsMessage);
}

function filterFunction(endpointEntries, settings, targetProtocol, targetPath) {
    const {
        fileTypes, files: {
            excludeExtensions, minSize, maxSize, minDate, maxDate
        }
    } = settings;
    return endpointEntries.reduce((result, entry) => {
        // filter relevant types:
        if (!((fileTypes.files && entry.type === TYPE_FILE)
            || (fileTypes.directories && entry.type === TYPE_DIRECTORY)
            || (fileTypes.symlinks && entry.type === TYPE_SYMBOLIC))) {
            entry.reason = `filtered out - type ${entry.type}`;
            result.TASK_TYPE_IGNORE.push(entry);
            return result;
        }

        // filter out entries that has object name greater than 1024 - relevant for objectStore
        if ([ALIBABA_PROVIDER, S3_PROVIDER].includes(targetProtocol)) {
            const objectStoragePath = calculateObjectStoragePath(targetPath, entry.path);
            logger.debug('objectStoragePath is', objectStoragePath.length);
            if (objectStoragePath.length > 1024) {
                logger.debug(`Ignore entry ${entry.path}`);
                entry.reason = 'filtered out - exceeding object name';
                result.TASK_TYPE_IGNORE.push(entry);
                return result;
            }
        }

        if (entry.type !== TYPE_DIRECTORY) {
            // filter files with specific extensions
            if (excludeExtensions.some(ext => entry.path.endsWith(ext))) {
                entry.reason = 'filtered out - excluded extensions';
                result.TASK_TYPE_IGNORE.push(entry);
                return result;
            }

            // filter files with specific size
            if (!(entry.size >= minSize && entry.size <= maxSize)) {
                entry.reason = 'filtered out - size limits';
                result.TASK_TYPE_IGNORE.push(entry);
                return result;
            }

            // filter files with specific date
            const fileMoment = moment.unix(entry.mtimeSec);
            if (fileMoment.isSameOrAfter(minDate, 'day')) {
                if (maxDate) {
                    if (!(fileMoment.isSameOrBefore(maxDate, 'day'))) {
                        entry.reason = 'filtered out - date limits';
                        result.TASK_TYPE_IGNORE.push(entry);
                        return result;
                    }
                }
            } else {
                entry.reason = 'filtered out - date limits';
                result.TASK_TYPE_IGNORE.push(entry);
                return result;
            }
            result.TASK_TYPE_FILES_COPY.push(entry);
        } else {
            result.TASK_TYPE_DIRS_COPY.push(entry);
        }
        return result;
    }, { TASK_TYPE_IGNORE: [], TASK_TYPE_FILES_COPY: [], TASK_TYPE_DIRS_COPY: [] });
}

async function getEntriesList(endPoint, relationshipId, srcOrTrg, retries) {
    const endPointProtocol = head(keys(endPoint));
    return retry(async () => {
        switch (endPointProtocol) {
            case 'local':
                return list(endPoint.local.path, META_DATA[relationshipId].settings.copyAcl);
            case 'snfs':
                return listSecureNFS(relationshipId, endPoint.snfs.serverIP, endPoint.snfs.port,
                    META_DATA[relationshipId].snfs.certificate, META_DATA[relationshipId].snfs.serverName, META_DATA[relationshipId].requestId, endPoint.snfs.path);
            case 's3':
                return listObjectsOperation(endPoint.s3, relationshipId, srcOrTrg);
            case 'azure':
                return listBlobs(endPoint.azure.container, endPoint.azure.path, relationshipId, srcOrTrg);
            case 'gcp': {
                return listGcpObjects(endPoint.gcp.bucket, endPoint.gcp.path);
            }
            default:
                throw new Error(`Got unsupported protocol ${endPoint}`);
        }
    }, {
        retries,
        onRetry: err => logger.info('Retrying listing of', endPoint, err)
    });
}

function filterGracePeriod(message, entries, settings) {
    const currentTimeMilliSec = new Date().getTime();
    const { body: { source, relationshipId } } = message;
    const grace = settings.gracePeriod;
    return partition(entries, entry => {
        const notGrace = ((entry.mtimeSec + grace) * 1000) < currentTimeMilliSec;
        if (!notGrace) {
            relationshipsLogger.info(`[${relationshipId}] IGNORING - GRACE: ${grace} sec, time: ${currentTimeMilliSec}, source:`, source, 'entry:', entry);
        }
        return notGrace;
    });
}

async function handleCloudToCloudScannerMessage(relationshipId, attributes, receiptHandle) {
    if (attributes.addresseeScanner) { // if that attribute doesn't exist - it's the initial message
        logger.info(`WorkerId that got the cloud to cloud message is ${WORKER_ID} and the message is addressed to ${attributes.addresseeScanner}`);
        if (attributes.addresseeScanner === `${getDatabrokerId()}-${WORKER_ID}`) { // checks if that is the desired process
            if (META_DATA[relationshipId]['cloud-to-cloud'] && META_DATA[relationshipId]['cloud-to-cloud'].receiveTimeStamp) { // that means it hasn't died and the timestamp is still in memory (we want to return and not to start the scan again)
                await Promise.map(INTERNAL_DISPATCHER_QUEUE.workersList(), task => { if (task.data.receiptHandle === META_DATA[relationshipId]['cloud-to-cloud'].receiptHandle) { task.data.receiptHandle = receiptHandle; } });
                return Object.assign(META_DATA[relationshipId], { 'cloud-to-cloud': { receiptHandle, receiveTimeStamp: moment() } });
            }
        } else { // retrieve the message to the queue
            await changeMessageVisibilityBatch(META_DATA[relationshipId].queues.scannerQueue, [receiptHandle], 0);
            return true;
        }
    }
    Object.assign(META_DATA[relationshipId], { 'cloud-to-cloud': { receiptHandle, receiveTimeStamp: moment() } });
    return false;
}

async function handleScannerMessage(message) {
    const {
        body: {
            source, target, relationshipId, type
        },
        receiptHandle, attributes
    } = message;

    if (!META_DATA[relationshipId]) {
        return;
    }
    // mechanism to handle messages that stay longer than 12 hours in scanner queue  (cloud to cloud)
    if (config.get('workers.scanner.cloud-to-cloud')) {
        const shouldExit = await handleCloudToCloudScannerMessage(relationshipId, attributes, receiptHandle);
        if (shouldExit) { return; }
    }

    try {
        const newSource = getNewEndPointIfNeeded(relationshipId, META_DATA[relationshipId].requestId, MOUNT_SOURCE_SUFFIX, source);
        const newTarget = getNewEndPointIfNeeded(relationshipId, META_DATA[relationshipId].requestId, MOUNT_TARGET_SUFFIX, target);
        const sourceProtocol = head(keys(source));
        const targetProtocol = head(keys(target));
        const targetPath = target[targetProtocol].path;
        const promises = [];
        const CLOUDS = ['s3', 'azure', 'gcp']; // TODO delete after QA checks
        const { settings } = META_DATA[relationshipId];
        switch (type) {
            case SCAN_MODE_SOURCE_AND_TARGET: {
                if (!(config.get('workers.scanner.cloud-to-cloud') && CLOUDS.includes(sourceProtocol) && CLOUDS.includes(targetProtocol))) { // TODO delete after QA checks
                    // list both sides
                    const [sourceEntries, targetEntries] = await Promise.all([
                        getEntriesList(newSource, relationshipId, MOUNT_SOURCE_SUFFIX, settings.retries),
                        getEntriesList(newTarget, relationshipId, MOUNT_TARGET_SUFFIX, settings.retries)
                    ]);

                    // get all source entries that are not directories for statistics calculations and send them.
                    const allSourceFiles = sourceEntries.filter(entry => entry.type !== TYPE_DIRECTORY);
                    calculateListStatistics(allSourceFiles, relationshipId, attributes);

                    if (sourceEntries.length < 20) {
                        logger.debug(`Source Entries ${JSON.stringify(sourceEntries, null, 4)}`);
                    }
                    if (targetEntries.length < 20) {
                        logger.debug(`Target Entries ${JSON.stringify(targetEntries, null, 4)}`);
                    }

                    const sourceFilteredEntries = filterFunction(sourceEntries, settings, targetProtocol, targetPath);
                    const targetFilteredEntries = filterFunction(targetEntries, settings, targetProtocol, targetPath);

                    // compare files and send messages to transferrer queue
                    const filesDifferences = compareFiles(sourceFilteredEntries.TASK_TYPE_FILES_COPY, targetFilteredEntries.TASK_TYPE_FILES_COPY,
                        sourceProtocol, targetProtocol, target[targetProtocol].provider);
                    const dirDifferences = compareDirs(sourceFilteredEntries.TASK_TYPE_DIRS_COPY, targetFilteredEntries.TASK_TYPE_DIRS_COPY, sourceProtocol, targetProtocol);
                    const [filesFiltered, graceFiles] = filterGracePeriod(message, filesDifferences.TASK_TYPE_COPY, settings);
                    promises.push(sendTransferrerMessages(message, ACTION_COPY, filesFiltered, source, target));
                    if (settings.deleteOnTarget) {
                        promises.push(sendTransferrerMessages(message, ACTION_DELETE_TARGET, filesDifferences.TASK_TYPE_DELETE_TARGET, source, target));
                    }
                    if (settings.deleteOnSource && config.get('workers.transferrer.delete-on-source')) {
                        promises.push(sendTransferrerMessages(message, ACTION_DELETE_SOURCE, filesDifferences.TASK_TYPE_DELETE_SOURCE, source, target));
                    }

                    // compare directories - equal directories - dive in
                    promises.push(sendScannerMessages(message, SCAN_MODE_SOURCE_AND_TARGET, dirDifferences.TASK_TYPE_EQUAL));

                    // compare directories - directories existing in the source, but not in the target - copy from source to target
                    if (dirDifferences.TASK_TYPE_SOURCE_ONLY.length < 10) {
                        logger.debug('2 side - directories to copy', dirDifferences.TASK_TYPE_SOURCE_ONLY, ' sending message', JSON.stringify(message, null, 4), ' to scanner with type ', SCAN_MODE_SOURCE_ONLY);
                    }
                    promises.push(sendScannerMessages(message, SCAN_MODE_SOURCE_ONLY, dirDifferences.TASK_TYPE_SOURCE_ONLY));
                    promises.push(sendTransferrerMessages(message, ACTION_COPY, dirDifferences.TASK_TYPE_SOURCE_ONLY, source, target));
                    promises.push(sendTransferrerMessages(message, ACTION_COPY, dirDifferences.TASK_TYPE_EQUAL_AND_COPY, source, target));

                    // compare directories - directories existing in the target, but not in the source - delete from target
                    if (settings.deleteOnTarget) {
                        if (dirDifferences.TASK_TYPE_TARGET_ONLY.length < 10) {
                            logger.debug('2 side - directories to delete', dirDifferences.TASK_TYPE_TARGET_ONLY, ' sending message', JSON.stringify(message, null, 4), ' to scanner with type ', SCAN_MODE_TARGET_ONLY);
                        }
                        promises.push(sendScannerMessages(message, SCAN_MODE_TARGET_ONLY, dirDifferences.TASK_TYPE_TARGET_ONLY));
                        promises.push(sendTransferrerMessages(message, ACTION_DELETE_TARGET, dirDifferences.TASK_TYPE_TARGET_ONLY, source, target));
                    }
                    ignoreEntries(relationshipId, attributes, filesDifferences.TASK_TYPE_IGNORE, dirDifferences.TASK_TYPE_IGNORE, sourceFilteredEntries.TASK_TYPE_IGNORE, targetFilteredEntries.TASK_TYPE_IGNORE, graceFiles);
                    break;
                }
            } // fallthrough until we remove the cloud to cloud from config file
            case SCAN_MODE_CLOUD_TO_CLOUD: { // eslint-disable-line no-fallthrough
                await cloudToCloudScan(source, target, relationshipId, message, settings);
                break;
            }
            case SCAN_MODE_SOURCE_ONLY: {
                const sourceEntries = await getEntriesList(newSource, relationshipId, MOUNT_SOURCE_SUFFIX, settings.retries);

                // get all source entries that are not directories for statistics calculations and send them.
                const allSourceFiles = sourceEntries.filter(entry => entry.type !== TYPE_DIRECTORY);
                calculateListStatistics(allSourceFiles, relationshipId, attributes);

                if (sourceEntries.length < 20) {
                    logger.debug(`Source Entries ${JSON.stringify(sourceEntries, null, 4)}`);
                }

                const sourceFilteredEntries = filterFunction(sourceEntries, settings, targetProtocol, targetPath);

                const filesDifferences = compareFiles(sourceFilteredEntries.TASK_TYPE_FILES_COPY, [], sourceProtocol, targetProtocol, target[targetProtocol].provider);
                const dirDifferences = compareDirs(sourceFilteredEntries.TASK_TYPE_DIRS_COPY, [], sourceProtocol, targetProtocol);

                const [filesFiltered, graceFiles] = filterGracePeriod(message, filesDifferences.TASK_TYPE_COPY, settings);
                promises.push(sendTransferrerMessages(message, ACTION_COPY, filesFiltered, source, target));
                promises.push(sendTransferrerMessages(message, ACTION_COPY, dirDifferences.TASK_TYPE_SOURCE_ONLY, source, target));
                promises.push(sendScannerMessages(message, SCAN_MODE_SOURCE_ONLY, dirDifferences.TASK_TYPE_SOURCE_ONLY));
                ignoreEntries(relationshipId, attributes, filesDifferences.TASK_TYPE_IGNORE, dirDifferences.TASK_TYPE_IGNORE, sourceFilteredEntries.TASK_TYPE_IGNORE, graceFiles);
                break;
            }
            case SCAN_MODE_TARGET_ONLY: {
                // if delete on target is false, should never reach here
                const targetEntries = await getEntriesList(newTarget, relationshipId, MOUNT_TARGET_SUFFIX, settings.retries);

                if (targetEntries.length < 20) {
                    logger.debug(`Target Entries ${JSON.stringify(targetEntries, null, 4)}`);
                }

                const targetFilteredEntries = filterFunction(targetEntries, settings, targetProtocol, targetPath);

                const filesDifferences = compareFiles([], targetFilteredEntries.TASK_TYPE_FILES_COPY, sourceProtocol, targetProtocol, target[targetProtocol].provider);
                const dirDifferences = compareDirs([], targetFilteredEntries.TASK_TYPE_DIRS_COPY, sourceProtocol, targetProtocol);

                promises.push(sendTransferrerMessages(message, ACTION_DELETE_TARGET, filesDifferences.TASK_TYPE_DELETE_TARGET, source, target));
                promises.push(sendTransferrerMessages(message, ACTION_DELETE_TARGET, dirDifferences.TASK_TYPE_TARGET_ONLY, source, target));
                promises.push(sendScannerMessages(message, SCAN_MODE_TARGET_ONLY, dirDifferences.TASK_TYPE_TARGET_ONLY));
                ignoreEntries(relationshipId, attributes, filesDifferences.TASK_TYPE_IGNORE, dirDifferences.TASK_TYPE_IGNORE, targetFilteredEntries.TASK_TYPE_IGNORE);
                break;
            }
            default:
                logger.error(`[${relationshipId}] unknown scan mode ${type}`);
        }
        await Promise.all(promises);
    } catch (err) {
        const errorObject = generateErrorStatisticsObject(err);
        const errorMessage = {
            step: LIST_STEP,
            relationshipId,
            status: STATUS_FAILED,
            objectType: TYPE_DIRECTORY,
            bytes: 0,
            items: 1,
            reason: errorObject,
            targetRequiredAction: STATUS_FAILED,
            attributes: message.attributes
        };

        sendStatisticsMessage(errorMessage);
        logger.error(`scan operation failed for rid: ${relationshipId} with error ${err} sending errorMessage`, errorMessage);
        failedRelationshipLogger.addContext('relationshipId', relationshipId);
        failedRelationshipLogger.info(`[${relationshipId}] [${LIST_STEP}] FAILED source`, source, ' and target: ', target, err);
    } finally {
        if (META_DATA[relationshipId]) {
            logger.debug('Deleting message from scanner queue', receiptHandle);
            const newReceiptHandle = config.get('workers.scanner.cloud-to-cloud') && META_DATA[relationshipId]['cloud-to-cloud'].receiptHandle !== receiptHandle;
            await deleteMessage(META_DATA[relationshipId].queues.scannerQueue, newReceiptHandle ? META_DATA[relationshipId]['cloud-to-cloud'].receiptHandle : receiptHandle);
        }
    }
}

async function handleMessageVisibility() {
    const runningTasks = INTERNAL_DISPATCHER_QUEUE.workersList().map(task => task.data);
    const waitingTasks = INTERNAL_DISPATCHER_QUEUE._tasks.toArray(); // eslint-disable-line no-underscore-dangle
    const queues = compact(Object.values(META_DATA).map(value => get(value, 'queues.scannerQueue')));
    logger.info('Internal dispatcher queue:', {
        runningTasks: runningTasks.length,
        waitingTasks: waitingTasks.length,
        queues: queues.length
    });
    const groupedMessagesByRelationshipId = groupBy(runningTasks.concat(waitingTasks), message => message.body.relationshipId);
    Object.keys(groupedMessagesByRelationshipId).forEach(relationshipId => {
        if (META_DATA[relationshipId]) {
            const receiptHandles = groupedMessagesByRelationshipId[relationshipId]
                .filter(task => Date.now() - task.lastVisibilityTimeoutChange > ms(SQS_MESSAGE_UNCHANGED_VISIBILITY_TIMEOUT_LIMIT))
                .map(task => task.receiptHandle);

            INTERNAL_DISPATCHER_QUEUE._tasks.toArray().forEach(message => { if (receiptHandles.includes(message.receiptHandle)) { message.lastVisibilityTimeoutChange = Date.now(); } }); // eslint-disable-line no-param-reassign, no-underscore-dangle
            INTERNAL_DISPATCHER_QUEUE.workersList().forEach(task => { if (receiptHandles.includes(task.data.receiptHandle)) { task.data.lastVisibilityTimeoutChange = Date.now(); } }); // eslint-disable-line no-param-reassign
            changeMessageVisibilityBatch(META_DATA[relationshipId].queues.scannerQueue, receiptHandles);
        }
    });
}

async function getWork(concurrency) {
    const queues = compact(Object.values(META_DATA).map(value => get(value, 'queues.scannerQueue')));

    if (queues.length > 0) {
        const waitingWorkersThreshold = config.get('workers.scanner.concurrency-eagerness-ratio') * concurrency;
        if (waitingWorkersThreshold > INTERNAL_DISPATCHER_QUEUE.length()) {
            const scannerQueue = head(shuffle(queues));
            try {
                const messages = await receiveMessages(scannerQueue);

                logger.debug(`Got ${messages.length} messages`, {
                    running: INTERNAL_DISPATCHER_QUEUE.running(),
                    length: INTERNAL_DISPATCHER_QUEUE.length()
                });
                messages.forEach(async message => {
                    try {
                        if (validateMessage(message.body, 'SQS', scannerIncomingSQSSchema)) {
                            INTERNAL_DISPATCHER_QUEUE.push(message);
                        }
                    } catch (err) {
                        logger.warn('Failed to validate scanner message', err);
                        deleteMessage(scannerQueue, message.receiptHandle);
                    }
                });
            } catch (err) {
                logger.error('Failed to receive messages from queue', scannerQueue, err);
            }
        }
        setImmediate(() => getWork(concurrency));
    } else {
        setTimeout(() => getWork(concurrency), SQS_DEFAULT_WAIT_TIMEOUT * 1000);
    }
}

function handleCopyMessage(message) {
    const {
        relationshipId, requestId, scannerQueue, transferrerQueue, settings, serverName, certificate
    } = message;
    logger.info(`handling IPC copy message for rid: ${relationshipId}, requestId: ${requestId}`);
    //TODO - Delete when CS-4986 is resolved
    if (config.get('protocols.nfs.acl.active')) {
        settings.copyAcl = 'nfs';
    }
    META_DATA[relationshipId] = {
        queues: {
            scannerQueue,
            transferrerQueue
        },
        status: IPC_TYPE_COPY,
        requestId,
        settings
    };
    if (serverName && certificate) {
        META_DATA[relationshipId].snfs = { serverName, certificate };
    }
}

function handleDoneRemoveAbortUpdateMessage(message) {
    const { relationshipId, type } = message;
    delete META_DATA[relationshipId];
    INTERNAL_DISPATCHER_QUEUE.remove(task => task.data.body.relationshipId === relationshipId);
    logger.info(`Done handling command: ${type} for rid: ${relationshipId}`);
}

function boot() {
    const concurrency = config.get('workers.scanner.concurrency');
    logger.info(`Starting scanner #${cluster.worker.id}, PID ${process.pid}, concurrency ${concurrency} ...`);

    // queue should have 'async' keyword otherwise it does not work
    INTERNAL_DISPATCHER_QUEUE = queue(async scannerMessage => handleScannerMessage(scannerMessage), concurrency);
    INTERNAL_DISPATCHER_QUEUE.drain = () => logger.info('Drain -> no more work');
    INTERNAL_DISPATCHER_QUEUE.error = (err, scannerMessage) => logger.error(`Got error for scanner message ${scannerMessage}`, err);
    getWork(concurrency);
    setInterval(() => handleMessageVisibility(), ms(SQS_CHANGE_MESSAGE_VISIBILITY_TIMEOUT_INTERVAL));
}

process.on('message', message => {
    logger.info('Got IPC message', message);
    if (validateMessage(message, 'IPC', scannerIncomingIPCMessagesSchemes[message.type])) {
        switch (message.type) {
            case IPC_TYPE_COPY:
                handleCopyMessage(message);
                break;
            case IPC_TYPE_DONE:
            case IPC_TYPE_UPDATE:
            case IPC_TYPE_REMOVE:
            case IPC_TYPE_ABORT:
                handleDoneRemoveAbortUpdateMessage(message);
                break;
            case 'create-scanner':
                WORKER_ID = message.workerId;
                break;
            default:
                logger.error('Got unexpected message type', message.type);
        }
    }
});

if (process.env.NODE_ENV !== 'test') {
    boot();
}

module.exports = {
    compareFiles, compareDirs, listObjectsOperation, sendTransferrerMessages
};
