const http2 = require('http2');
const { readFileSync, createWriteStream } = require('fs');
const {
    open, readlink
} = require('fs').promises;
const { pipeline } = require('stream');
const { exec } = require('child_process');
const mimeType = require('mime-types');
const ms = require('ms');
const Promise = require('bluebird');
const config = require('config');
const {
    mkdirs, remove, symlink, chmod, utimes, rename, constants
} = require('fs-extra');
const randomize = require('randomatic');
const { umount } = require('../lib/fs/mount');

const {
    HTTP2_HEADER_STATUS,
    HTTP2_HEADER_CONTENT_TYPE,
    HTTP2_HEADER_PATH,
    HTTP2_HEADER_METHOD,
    HTTP2_METHOD_POST
} = http2.constants;

const validateMessage = require('../messages/schemas/message-validator');
const schemas = require('../messages/server/messagesSchema');
const getLogger = require('../lib/logger/logger');

const logger = getLogger();

const {
    OPER_LIST_WITH_STATS, OPER_GET_FILE,
    OPER_MOUNT, OPER_UMOUNT, OPER_GET_SYMLINK, OPER_SECURE_MKDIRS,
    MOUNT_FOLDER, OPER_UPLOAD_FILE, OPER_UPLOAD_SYMLINK, OPER_UPLOAD_DIR,
    MOUNT_SOURCE_SUFFIX, OPER_DELETE_FILE, DATA_BROKER_RUNTIME_CONFIG, CERT_AND_KEY_FOLDER, NETAPP_ENTRY_SUFFIX, NETAPP_ENTRY_RANDOM_SIZE, NETAPP_ENTRY_RANDOM_PATTERN
} = require('../lib/consts');

const { getMountPoint } = require('../lib/fs/mount_auxiliary');
const {
    gracefulJoin, list, chmodAndChown, chownIfNeeded
} = require('../lib/fs/fs');

process.on('uncaughtException', err => logger.error('Got uncaught exception', err));
process.on('unhandledRejection', event => logger.error('Got unhandled rejection', event));

function messageApproved(data, operation) {
    return validateMessage(data, undefined, schemas[operation]);
}

async function pipe(readStream, writeStream) {
    return new Promise((resolve, reject) => {
        pipeline(
            readStream,
            writeStream,
            err => {
                if (err) {
                    reject(err);
                } else {
                    resolve();
                }
            }
        );
    });
}

function initServer() {
    const { port } = DATA_BROKER_RUNTIME_CONFIG;
    const keyPath = `${CERT_AND_KEY_FOLDER}/dataBrokerKey.key`;
    const certPath = `${CERT_AND_KEY_FOLDER}/dataBrokerCert.crt`;

    const server = http2.createSecureServer({
        key: readFileSync(keyPath),
        cert: readFileSync(certPath)
    });

    server.listen(port);
    logger.info(`Server deployed. Listening on port number: ${port}`);

    return server;
}

async function getData(stream) {
    return new Promise((resolve, reject) => {
        let data = '';
        stream.on('data', chunk => {
            data += chunk;
        });
        stream.on('end', () => {
            resolve(data);
        });
        stream.on('error', err => {
            logger.error(`Error while receiving data from server. error: ${err}`);
            reject(err);
        });
    });
}

async function execute(command, timeout, env) {
    logger.info('Executing command', { command, timeout, env });
    return new Promise((resolve, reject) => {
        exec(command, { timeout, env }, (error, stdout, stderr) => {
            if (error) {
                reject(stderr || stdout || `Command execution: ${command} - timed out`);
            } else {
                resolve(stdout);
            }
        });
    });
}

async function mountNFS(mountPoint, host, nfsExport, version) {
    const { 'mount-parameters': mountParameters, timeout } = config.get('protocols.nfs');
    await mkdirs(mountPoint);

    let protocolVersion = version;
    if (!version) {
        ({ 'default-version': protocolVersion } = config.get('protocols.nfs'));
    }

    const command = `mount ${host}:${nfsExport} ${mountPoint} -o ${mountParameters},vers=${protocolVersion}`;
    await execute(command, ms(timeout));
}

async function closeFDonStreamClose(stream, fileHandle) {
    return new Promise(resolve => {
        stream.on('close', async () => {
            logger.debug('got close from client, closing fd');
            await fileHandle.close();
            resolve();
        });
    });
}

async function dispatch(stream, operation, headers) {
    const method = headers[HTTP2_HEADER_METHOD];

    if (method === HTTP2_METHOD_POST) {
        switch (operation) {
            case OPER_LIST_WITH_STATS: {
                const { 'request-id': requestId, path, [HTTP2_HEADER_PATH]: relId } = headers;
                const mountPoint = getMountPoint(relId, requestId, MOUNT_SOURCE_SUFFIX);
                const joinedPath = gracefulJoin(mountPoint, path);
                const result = await list(joinedPath);
                logger.debug(`Files list and stats: ${JSON.stringify(result)}`);
                stream.respond({
                    [HTTP2_HEADER_STATUS]: 200
                });
                stream.end(JSON.stringify(result));
                break;
            }

            case OPER_GET_FILE: {
                const { 'request-id': requestId, path, [HTTP2_HEADER_PATH]: relId } = headers;
                const mountPoint = getMountPoint(relId, requestId, MOUNT_SOURCE_SUFFIX);
                const joinedPath = gracefulJoin(mountPoint, path);
                const contentType = mimeType.lookup(joinedPath) || 'application/octet-stream';
                const fileHandle = await open(joinedPath, 'r');
                logger.debug(`FD: ${fileHandle.fd}`);
                stream.respondWithFD(fileHandle.fd, {
                    [HTTP2_HEADER_CONTENT_TYPE]: contentType,
                    [HTTP2_HEADER_STATUS]: 200
                });
                await closeFDonStreamClose(stream, fileHandle);
                break;
            }

            case OPER_DELETE_FILE: {
                const { 'request-id': requestId, path, [HTTP2_HEADER_PATH]: relId } = headers;
                const mountPoint = getMountPoint(relId, requestId, MOUNT_SOURCE_SUFFIX);
                const joinedPath = gracefulJoin(mountPoint, path);
                await remove(joinedPath);
                stream.respond({
                    [HTTP2_HEADER_STATUS]: 200
                });
                stream.end();
                break;
            }

            case OPER_MOUNT: {
                const {
                    'request-id': requestId, [HTTP2_HEADER_PATH]: relId, 'source-ip': host, 'nfs-export': nfsExport, version
                } = headers;
                const mountPoint = getMountPoint(relId, requestId, MOUNT_SOURCE_SUFFIX);
                await mountNFS(mountPoint, host, nfsExport, version);
                stream.respond({
                    [HTTP2_HEADER_STATUS]: 200
                });
                stream.end();
                logger.info('Host mounted successfully.');
                break;
            }

            case OPER_UMOUNT: {
                const { [HTTP2_HEADER_PATH]: relId } = headers;
                const mountPoint = gracefulJoin(MOUNT_FOLDER, relId, '*');
                await umount(mountPoint);
                await remove(gracefulJoin(MOUNT_FOLDER, relId));
                stream.respond({
                    [HTTP2_HEADER_STATUS]: 200
                });
                stream.end();
                logger.info('Host umounted successfully.');
                break;
            }

            case OPER_GET_SYMLINK: {
                const { 'request-id': requestId, path, [HTTP2_HEADER_PATH]: relId } = headers;
                logger.debug(`request ID: ${requestId}, path: ${path}.`);
                const mountPoint = getMountPoint(relId, requestId, MOUNT_SOURCE_SUFFIX);
                const pathToFile = gracefulJoin(mountPoint, path);
                const symlinkContent = await readlink(pathToFile);
                stream.respond({
                    [HTTP2_HEADER_STATUS]: 200
                });
                stream.end(symlinkContent);
                break;
            }

            case OPER_UPLOAD_FILE: {
                const {
                    'request-id': requestId, path, [HTTP2_HEADER_PATH]: relId, entry: entryHeader
                } = headers;
                const entry = JSON.parse(entryHeader);

                const mountPoint = getMountPoint(relId, requestId, MOUNT_SOURCE_SUFFIX);
                const targetLocalPath = gracefulJoin(mountPoint, path);

                await mkdirs(targetLocalPath);
                const targetPath = gracefulJoin(targetLocalPath, entry.path);

                if (entry.size === 0) {
                    const fileHandle = await open(targetPath, constants.O_CREAT);
                    await Promise.all([
                        utimes(targetPath, entry.mtimeSec, entry.mtimeSec),
                        chmod(targetPath, entry.mode),
                        chownIfNeeded(targetPath, entry.uid, entry.gid)
                    ]);
                    await fileHandle.close();
                    stream.respond({
                        [HTTP2_HEADER_STATUS]: 200
                    });
                } else {
                    const tempEntryName = `.${entry.path}_${randomize(NETAPP_ENTRY_RANDOM_PATTERN, NETAPP_ENTRY_RANDOM_SIZE)}${NETAPP_ENTRY_SUFFIX}`;
                    const tempTargetPath = gracefulJoin(targetLocalPath, tempEntryName);
                    const writeStream = createWriteStream(tempTargetPath, { mode: entry.mode });

                    stream.respond({
                        [HTTP2_HEADER_STATUS]: 202
                    });

                    await pipe(stream, writeStream);
                    stream.close();

                    await Promise.all([
                        utimes(tempTargetPath, entry.mtimeSec, entry.mtimeSec),
                        chmod(tempTargetPath, entry.mode),
                        chownIfNeeded(tempTargetPath, entry.uid, entry.gid)
                    ]);
                    await rename(tempTargetPath, targetPath);
                }

                break;
            }

            case OPER_UPLOAD_SYMLINK: {
                const {
                    'request-id': requestId, path, [HTTP2_HEADER_PATH]: relId, entry: entryHeader
                } = headers;
                const entry = JSON.parse(entryHeader);
                const mountPoint = getMountPoint(relId, requestId, MOUNT_SOURCE_SUFFIX);
                const targetLocalPath = gracefulJoin(mountPoint, path);

                await mkdirs(targetLocalPath);

                const targetPath = gracefulJoin(targetLocalPath, entry.path);
                await remove(targetPath);

                const symlinkContent = await getData(stream);
                await symlink(symlinkContent, targetPath);
                stream.respond({
                    [HTTP2_HEADER_STATUS]: 200
                });
                stream.end();
                break;
            }

            case OPER_UPLOAD_DIR: {
                const {
                    'request-id': requestId, path, [HTTP2_HEADER_PATH]: relId, entry: entryHeader
                } = headers;
                const entry = JSON.parse(entryHeader);
                const mountPoint = getMountPoint(relId, requestId, MOUNT_SOURCE_SUFFIX);
                const targetLocalPath = gracefulJoin(mountPoint, path);
                await mkdirs(gracefulJoin(targetLocalPath, entry.path));
                await chmodAndChown(targetLocalPath, entry);
                stream.respond({
                    [HTTP2_HEADER_STATUS]: 200
                });
                stream.end();
                break;
            }

            case OPER_SECURE_MKDIRS: {
                const {
                    'request-id': requestId, path, [HTTP2_HEADER_PATH]: relId
                } = headers;
                const mountPoint = getMountPoint(relId, requestId, MOUNT_SOURCE_SUFFIX);
                const targetLocalPath = gracefulJoin(mountPoint, path);
                await mkdirs(targetLocalPath);
                stream.respond({
                    [HTTP2_HEADER_STATUS]: 200
                });
                stream.end();
                break;
            }

            default:
                throw new Error(`unexpected operation: ${operation}.`);
        }
    }
}

function boot() {
    let server;
    // deploy server
    try {
        server = initServer();
    } catch (err) {
        logger.error('Server deployment failed: ', err);
        process.exit(8);
    }

    server.on('error', err => {
        logger.error(`Server has encountered an error: ${err}.\nClosing server...`);
        server.close();
        process.exit(1);
    });

    server.on('sessionError', err => {
        logger.error('Server has session error: ', err);
    });

    // event emitted when server gets data from client through http2clientstream.
    server.on('stream', async (stream, headers) => {
        const { operation } = headers;
        logger.debug(`operation received from client: ${operation}`);

        if (!operation) {
            logger.error(`Server didn't receive valid operation... operation received: ${operation}`);
            stream.respond({
                [HTTP2_HEADER_STATUS]: 400
            });
            stream.end(`Server didn't receive valid operation... operation received: ${operation}`);
        }

        let isApproved = false;

        // Validate data:
        try {
            isApproved = messageApproved(headers, operation);
        } catch (err) {
            logger.error('Message validation failed. Error: ', err);
            stream.respond({
                [HTTP2_HEADER_STATUS]: 400
            });
            stream.end(JSON.stringify(err));
        }

        if (isApproved) {
            try {
                await dispatch(stream, operation, headers);
            } catch (err) {
                logger.error('dispatch failed: ', err);
                stream.respond({
                    [HTTP2_HEADER_STATUS]: 500
                });
                stream.end(JSON.stringify(err));
            }
        }
    });
}

if (process.env.NODE_ENV !== 'test') {
    boot();
}
