const cluster = require('cluster');
const config = require('config');
const bytes = require('bytes');
const Promise = require('bluebird');
const { queue } = require('async');
const { pipeline, PassThrough } = require('stream');
const retry = require('async-retry');
const {
    readlink, symlink, mkdirs, createReadStream, createWriteStream, remove, rename, chown, chmod, utimes,
    pathExists, open, close, constants
} = require('fs-extra');
const { setTimeout } = require('timers');
const ms = require('ms');
const {
    shuffle, head, find, keys, _, compact, get, groupBy, uniq, isNil, cloneDeep, range,
    zip, dropRight, tail, first, last
} = require('lodash');
const https = require('https');
const http = require('http');
const prettyBytes = require('bytes');
const { escape } = require('querystring');
const {
    AbortController
} = require('@azure/abort-controller');
const {
    gracefulJoin, chmodAndChown, chownIfNeeded, setNfsAcl, setCifsAcl, generateErrorStatisticsObject
} = require('../lib/fs/fs');
const { getNewEndPointIfNeeded } = require('../lib/fs/mount_auxiliary');
const getLogger = require('../lib/logger/logger');
const {
    receiveMessages, changeMessageVisibilityBatch, deleteMessage, sendMessage
} = require('../lib/aws/sqs');
const {
    deleteBlob, createBlobReadStream, createWriteStreamToBlockBlobByFile, createWriteStreamToBlockBlobByReadStream, getBlobToText, createMetadata, createBlockBlobFromText,
    copyBlobFromUrl, setBlobMetadata, getBlobProperties, setBlobAccessTier
} = require('../lib/azure/blob');
const {
    deleteGcpObject, getGcpReadStream, getGcpObjectMetadata, getGcpWriteStream, createGcpFolder, createGcpSymlink,
    createGcpMetadata, isObjectExists, copyGcpObject, setMetadata
} = require('../lib/gcp/storage');

const {
    getSecureReadStream, getSecureSymlinkContent, connectToServer, uploadSymlink,
    uploadDirectory, getSecureUploadStream, deleteFile
} = require('../lib/snfs/snfs');
const {
    ONTAP_S3_PROVIDER,
    TYPE_FILE,
    TYPE_DIRECTORY,
    TYPE_SYMBOLIC,
    IPC_TYPE_COPY,
    IPC_TYPE_REMOVE,
    IPC_TYPE_ABORT,
    IPC_TYPE_DONE,
    IPC_TYPE_UPDATE,
    ACTION_DELETE_SOURCE,
    ACTION_DELETE_TARGET,
    ACTION_COPY,
    STATUS_SUCCEEDED,
    STATUS_FAILED,
    TRANSFER_STEP,
    CLEAN_RELATIONSHIP_RESOURCES_TIMEOUT,
    MOUNT_SOURCE_SUFFIX,
    MOUNT_TARGET_SUFFIX,
    SQS_MESSAGE_UNCHANGED_VISIBILITY_TIMEOUT_LIMIT,
    SQS_CHANGE_MESSAGE_VISIBILITY_TIMEOUT_INTERVAL,
    SQS_DEFAULT_WAIT_TIMEOUT, S3_PROVIDER, ALIBABA_PROVIDER, SGWS_PROVIDER, IBM_PROVIDER
} = require('../lib/consts');
const sqsTransferrerIncomingMessagesSchemes = require('../messages/transferrer/sqs-incoming-messages');
const transferrerIncomingIPCMessagesSchemes = require('../messages/transferrer/ipc-incoming-messages');
const { copyStatisticsSchema } = require('../messages/schemas/statuses-outgoing-messages');
const validateMessage = require('../messages/schemas/message-validator');
const { startHealthLogInterval } = require('../lib/health');
const {
    getClient
} = require('../lib/s3Clients');
const {
    upload, getReadStream, deleteObject, getObject, createTagsArray,
    createS3Metadata, copyRegularObject, copyPart,
    calculateObjectStoragePath, convertTagsArrayToString, completeMultipartUpload, createMultipartUpload,
    restoreS3Object, headObject, putS3Object
} = require('../lib/aws/s3_base_operations');

const logger = getLogger();
const relationshipsLogger = getLogger('relationships');
const failedRelationshipLogger = getLogger('failed-relationships');

http.globalAgent.maxSockets = 500;
https.globalAgent.maxSockets = 500;

startHealthLogInterval();

process.on('uncaughtException', err => logger.error('Got uncaught exception', err));
process.on('unhandledRejection', event => logger.error('Got unhandled rejection', event));

const NETAPP_TEMP_FILE_REGEX = /^[.].+_[a-zA-Z\d]{7}_netapp$/;

const META_DATA = {};
let INTERNAL_DISPATCHER_QUEUE;

function isNetappEntry(path) {
    return NETAPP_TEMP_FILE_REGEX.test(path);
}

function isRelationshipAborted(relationshipId) {
    const metaData = META_DATA[relationshipId];
    return !metaData || metaData.status === IPC_TYPE_ABORT;
}

function addInTransitItem(relationshipId, path, item) {
    META_DATA[relationshipId].items[path] = item;
}

function removeInTransitItem(relationshipId, targetPath) {
    delete META_DATA[relationshipId].items[targetPath];
}

async function triggerAbort(item) {
    const {
        type, resource
    } = item;

    logger.debug('Trigger abort for item type: ', type);

    try {
        switch (type) {
            case 'stream':
                logger.debug('Destroying temporary file due to copy abort');
                resource.destroy(new Error('Aborted'));
                break;
            case 'path':
                logger.debug('Removing temporary file due to copy abort: ', resource);
                await remove(resource);
                break;
            case 'client':
                logger.debug('Closing client connection due to copy abort');
                resource.close();
                break;
            case 'aborter':
                logger.debug('Aborting AZURE action due to copy abort');
                resource.abort();
                break;
            case 'managedUpload':
                logger.debug('Aborting AWS action due to copy abort');
                resource.abort();
                break;
            default:
                logger.warn('Got unexpected operation type: ', type);
                break;
        }
    } catch (err) {
        logger.warn('Failed to abort item type:', type, err);
    }
}

async function pipe(readStream, writeStream) {
    return new Promise((resolve, reject) => {
        pipeline(
            readStream,
            writeStream,
            err => {
                if (err) {
                    reject(err);
                } else {
                    resolve();
                }
            }
        );
    });
}

async function copyMultipartObject(s3, targetBucket, targetKey, copySource, sourceObjectSize, uploadId) {
    // calculate size
    const partSize = bytes(config.get('protocols.s3.part-size-s3-to-s3'));
    const rangeArr = range(0, sourceObjectSize, partSize);
    rangeArr.push(sourceObjectSize);
    // create array of objects - contains the start and end of each range and the partNumber
    const copyParts = zip(dropRight(rangeArr, 1), tail(rangeArr)).map((bytesPair, index) => ({
        bytesPair,
        partNum: index + 1
    }));

    return Promise.map(copyParts, (async partDetails => {
        const nextStartBytesRange = first(partDetails.bytesPair);
        const partNumber = partDetails.partNum;
        const nextEndBytesRange = last(partDetails.bytesPair) - 1; // because the last in the range in inclusive
        try {
            const result = await copyPart(s3, targetBucket, targetKey, copySource, partNumber, uploadId, nextStartBytesRange, nextEndBytesRange);
            return {
                ETag: result.CopyPartResult.ETag,
                PartNumber: partNumber
            };
        } catch (err) {
            logger.error(`Failed upload part number ${partNumber} for target from ${copySource} to ${targetKey} with error ${err}`);
            throw err;
        }
    }), { concurrency: 10 }); // when there are a lot of parts - receive reduce your rate
}

async function copyObject(s3, destinationBucket, copySource, destinationKey, extraParams, entrySize) {
    if (entrySize < bytes(config.get('protocols.s3.part-size-s3-to-s3'))) {
        await copyRegularObject(s3, destinationBucket, copySource, destinationKey, (extraParams.Tagging) ? Object.assign(extraParams, { TaggingDirective: 'REPLACE' }) : Object.assign(extraParams, { MetadataDirective: 'REPLACE' }));
    } else {
        logger.debug('Create multipart upload');
        // create multipart upload
        const uploadId = await createMultipartUpload(s3, destinationBucket, destinationKey, extraParams);
        const copyPartResults = await copyMultipartObject(s3, destinationBucket, destinationKey, copySource, entrySize, uploadId);
        // send complete multipart upload request
        await completeMultipartUpload(s3, destinationBucket, destinationKey, copyPartResults, uploadId);
    }
}

// function for s3 -> s3 and alibaba -> alibaba
async function copyObjectBetweenIdenticalS3Providers(source, target, entry, relationshipId) {
    const { provider: sourceProvider, bucket: sourceBucket } = source;
    const s3 = await getClient(sourceProvider, {
        ...target, relationshipId, srcOrTrg: MOUNT_SOURCE_SUFFIX, isS3EndPoint: true
    });

    const { bucket: targetBucket } = target;
    const sourcePath = calculateObjectStoragePath(source.path, entry.path);
    const targetPath = calculateObjectStoragePath(target.path, entry.path);
    const copySource = entry.type === TYPE_DIRECTORY ? encodeURI(`${sourceBucket}/${sourcePath}/`) : escape(`${sourceBucket}/${sourcePath}`);
    const destinationKey = entry.type === TYPE_DIRECTORY ? `${targetPath}/` : targetPath;
    if (sourceProvider === S3_PROVIDER) {
        const storageClass = get(target, 'storageClass');

        let SSEprotocol;
        let kmsKeyId;
        if (target.encryption) {
            if (target.encryption.kms) {
                kmsKeyId = get(target, 'encryption.kms.keyId');
                SSEprotocol = 'aws:kms';
            } else { // then protocol is AES256
                SSEprotocol = target.encryption.protocol;
            }
        }

        const extraParams = {
            StorageClass: storageClass || 'STANDARD',
            SSEKMSKeyId: kmsKeyId,
            ServerSideEncryption: SSEprotocol,
            ACL: config.get(`protocols.${sourceProvider}.object-acl`)
        };

        if (META_DATA[relationshipId].settings.objectTagging) {
            extraParams.Tagging = convertTagsArrayToString(createTagsArray(target.tags, entry));
        } else {
            extraParams.Metadata = createS3Metadata(entry, target.tags || []);
        }
        return copyObject(s3, targetBucket, copySource, destinationKey, extraParams, entry.size, entry.type);
    }
    if (sourceProvider === ALIBABA_PROVIDER) {
        // alibaba supports only Standard storageClass with S3 sdk
        const extraParams = {
            Metadata: createS3Metadata(entry, target.tags || []),
            StorageClass: 'STANDARD',
            ACL: config.get(`protocols.${sourceProvider}.object-acl`)
        };
        return copyObject(s3, targetBucket, copySource, destinationKey, extraParams, entry.size, entry.type);
    }
}

async function sendTransferrerMessages(source, target, relationshipId, entry, attributes) {
    const queues = META_DATA[relationshipId] ? META_DATA[relationshipId].queues : undefined;

    if (queues) {
        if (entry.partsDetails) {
            const messageToSend = {
                type: ACTION_COPY,
                source,
                target,
                entry,
                relationshipId
            };
            return sendMessage(queues.transferrerQueue, messageToSend, attributes);
        }
    }
}

async function deleteFromS3Provider(endPoint, key, relationshipId, mountSuffix) {
    const { provider, bucket } = endPoint;
    const s3 = await getClient(provider, {
        ...endPoint, relationshipId, srcOrTrg: mountSuffix, isS3EndPoint: true
    });
    return deleteObject(s3, bucket, key);
}

function buildExtraParams(provider, entry, endPoint, metadata, useTags) {
    switch (provider) {
        case S3_PROVIDER: {
            const { tags, storageClass } = endPoint;
            let SSEprotocol;
            let kmsKeyId;
            if (endPoint.encryption) {
                if (endPoint.encryption.kms) {
                    kmsKeyId = get(endPoint, 'encryption.kms.keyId');
                    SSEprotocol = 'aws:kms';
                } else { // then protocol is AES256
                    SSEprotocol = endPoint.encryption.protocol;
                }
            }
            const s3Metadata = createS3Metadata(entry, tags);
            // in order to put in the metadata only the values that are not on tags
            let metadataToCopy;
            if (useTags) {
                if (metadata) {
                    metadataToCopy = Object.keys(metadata).reduce((accumulator, key) => {
                        if (!s3Metadata[key]) {
                            accumulator[key] = metadata[key];
                        }
                        return accumulator;
                    }, {});
                }
            } else {
                metadataToCopy = { ...metadata, ...s3Metadata };
            }
            return {
                ACL: config.get(`protocols.${provider}.object-acl`),
                StorageClass: storageClass || 'STANDARD',
                SSEKMSKeyId: kmsKeyId,
                ServerSideEncryption: SSEprotocol,
                tags: useTags ? createTagsArray(tags, entry) : undefined,
                Metadata: metadataToCopy
            };
        }
        case ALIBABA_PROVIDER: {
            const { tags } = endPoint;

            return {
                ACL: config.get(`protocols.${provider}.object-acl`),
                Metadata: createS3Metadata(entry, tags || [])
            };
        }
        case SGWS_PROVIDER:
        case IBM_PROVIDER: {
            const { tags } = endPoint;
            const mergedMetadata = { ...metadata, ...createS3Metadata(entry, tags || []) };
            return { Metadata: mergedMetadata };
        }
        default: {
            throw new Error(`Received unknown provider ${provider}`);
        }
    }
}

async function uploadToS3Providers(endPoint, entry, bodyStream, targetPath, relationshipId, sourceMetadata) {
    const { provider, bucket } = endPoint;
    const s3 = await getClient(provider, {
        ...endPoint, relationshipId, srcOrTrg: MOUNT_TARGET_SUFFIX, isS3EndPoint: true, protocol: Boolean(provider === ONTAP_S3_PROVIDER)
    });
    if (provider === ONTAP_S3_PROVIDER) {
        // ontap S3 doesnt support user defined metadata
        logger.info('Uploading to S3 ONTAP', entry.path, entry.size);
        return putS3Object(s3, bucket, bodyStream, targetPath);
    }
    const extraParams = buildExtraParams(provider, entry, endPoint, sourceMetadata, META_DATA[relationshipId].settings.objectTagging);
    return upload(s3, bucket, bodyStream, targetPath, extraParams);
}

async function createReadStreamFromS3Provider(endPoint, entry, relationshipId) {
    const { provider, bucket, path } = endPoint;
    const s3 = await getClient(provider, {
        ...endPoint, relationshipId, srcOrTrg: MOUNT_SOURCE_SUFFIX, isS3EndPoint: true
    });
    const newPath = calculateObjectStoragePath(path, entry.path);
    return getReadStream(s3, bucket, newPath, entry.size);
}

async function getMetadataFromS3Provider(endPoint, entry, relationshipId) {
    const { provider, bucket, path } = endPoint;
    const s3 = await getClient(provider, {
        ...endPoint, relationshipId, srcOrTrg: MOUNT_SOURCE_SUFFIX, isS3EndPoint: true
    });
    const newPath = calculateObjectStoragePath(path, entry.path);
    return headObject(s3, bucket, newPath);
}

async function restoreS3ObjectIfNeeded(relationshipId, source, entry) {
    const {
        s3: {
            bucket, path, provider, glacier, deepArchive
        }
    } = source;
    const s3 = await getClient(provider, {
        ...source, relationshipId, srcOrTrg: MOUNT_SOURCE_SUFFIX, isS3EndPoint: true
    });
    const newPath = calculateObjectStoragePath(path, entry.path);
    const objectDetails = await s3.headObject({ Bucket: bucket, Key: newPath }).promise();
    if (objectDetails.StorageClass === 'GLACIER' && glacier && !objectDetails.Restore) {
        await restoreS3Object(s3, bucket, newPath, glacier);
        return true;
    }
    if (objectDetails.StorageClass === 'DEEP_ARCHIVE' && deepArchive && !objectDetails.Restore) {
        await restoreS3Object(s3, bucket, newPath, deepArchive);
        return true;
    }
    return false;
}

async function getObjectFromS3Provider(endPoint, sourcePath, relationshipId) {
    const { provider, bucket } = endPoint;
    const s3 = await getClient(provider, {
        ...endPoint, relationshipId, srcOrTrg: MOUNT_SOURCE_SUFFIX, isS3EndPoint: true
    });
    return getObject(s3, bucket, sourcePath);
}

const DELETE_TABLE = [
    {
        endpoint: 's3',
        type: TYPE_FILE,
        deleteHandler: async (endpoint, entry, relationshipId, mountSuffix) => {
            const newPath = calculateObjectStoragePath(endpoint.s3.path, entry.path);
            await deleteFromS3Provider(endpoint.s3, `${newPath}`, relationshipId, mountSuffix);
        }
    },
    {
        endpoint: 's3',
        type: TYPE_DIRECTORY,
        deleteHandler: async (endpoint, entry, relationshipId, mountSuffix) => {
            const newPath = calculateObjectStoragePath(endpoint.s3.path, entry.path);
            await deleteFromS3Provider(endpoint.s3, `${newPath}/`, relationshipId, mountSuffix);
        }
    },
    {
        endpoint: 's3',
        type: TYPE_SYMBOLIC,
        deleteHandler: async (endpoint, entry, relationshipId, mountSuffix) => {
            const newPath = calculateObjectStoragePath(endpoint.s3.path, entry.path);
            await deleteFromS3Provider(endpoint.s3, `${newPath}`, relationshipId, mountSuffix);
        }
    },
    {
        endpoint: 'snfs',
        type: TYPE_FILE,
        deleteHandler: async (endpoint, entry, relationshipId) => {
            await deleteFile(relationshipId, endpoint.snfs.serverIP, endpoint.snfs.port, META_DATA[relationshipId].snfs.certificate,
                META_DATA[relationshipId].snfs.serverName, META_DATA[relationshipId].requestId, gracefulJoin(endpoint.snfs.path, entry.path));
        }
    },
    {
        endpoint: 'snfs',
        type: TYPE_DIRECTORY,
        deleteHandler: async (endpoint, entry, relationshipId) => {
            await deleteFile(relationshipId, endpoint.snfs.serverIP, endpoint.snfs.port, META_DATA[relationshipId].snfs.certificate,
                META_DATA[relationshipId].snfs.serverName, META_DATA[relationshipId].requestId, gracefulJoin(endpoint.snfs.path, entry.path));
        }
    },
    {
        endpoint: 'snfs',
        type: TYPE_SYMBOLIC,
        deleteHandler: async (endpoint, entry, relationshipId) => {
            await deleteFile(relationshipId, endpoint.snfs.serverIP, META_DATA[relationshipId].requestId, gracefulJoin(endpoint.snfs.path, entry.path));
        }
    },

    {
        endpoint: 'local',
        type: TYPE_FILE,
        deleteHandler: async (endpoint, entry) => {
            await remove(gracefulJoin(endpoint.local.path, entry.path));
        }
    },
    {
        endpoint: 'local',
        type: TYPE_DIRECTORY,
        deleteHandler: async (endpoint, entry) => {
            await remove(gracefulJoin(endpoint.local.path, entry.path));
        }
    },
    {
        endpoint: 'local',
        type: TYPE_SYMBOLIC,
        deleteHandler: async (endpoint, entry) => {
            await remove(gracefulJoin(endpoint.local.path, entry.path));
        }
    },
    {
        endpoint: 'azure',
        type: TYPE_FILE,
        deleteHandler: async (endpoint, entry, relationshipId, mountSuffix) => {
            const newPath = calculateObjectStoragePath(endpoint.azure.path, entry.path);
            await deleteBlob(endpoint.azure.container, newPath, relationshipId, mountSuffix);
        }
    },
    {
        endpoint: 'azure',
        type: TYPE_DIRECTORY,
        deleteHandler: async (endpoint, entry, relationshipId, mountSuffix) => {
            const newPath = calculateObjectStoragePath(endpoint.azure.path, entry.path);
            await deleteBlob(endpoint.azure.container, `${newPath}/`, relationshipId, mountSuffix);
        }
    },
    {
        endpoint: 'azure',
        type: TYPE_SYMBOLIC,
        deleteHandler: async (endpoint, entry, relationshipId, mountSuffix) => {
            const newPath = calculateObjectStoragePath(endpoint.azure.path, entry.path);
            await deleteBlob(endpoint.azure.container, newPath, relationshipId, mountSuffix);
        }
    },
    {
        endpoint: 'gcp',
        type: TYPE_FILE,
        deleteHandler: async (endpoint, entry) => {
            const newPath = calculateObjectStoragePath(endpoint.gcp.path, entry.path);
            await deleteGcpObject(endpoint.gcp.bucket, `${newPath}`);
        }
    },
    {
        endpoint: 'gcp',
        type: TYPE_DIRECTORY,
        deleteHandler: async (endpoint, entry) => {
            const newPath = calculateObjectStoragePath(endpoint.gcp.path, entry.path);
            const noNameFileExists = await isObjectExists(endpoint.gcp.bucket, `${newPath}/<no name>`);
            if (noNameFileExists) {
                await deleteGcpObject(endpoint.gcp.bucket, `${newPath}/<no name>`);
            }
            await deleteGcpObject(endpoint.gcp.bucket, `${newPath}/`);
        }
    },
    {
        endpoint: 'gcp',
        type: TYPE_SYMBOLIC,
        deleteHandler: async (endpoint, entry) => {
            const newPath = calculateObjectStoragePath(endpoint.gcp.path, entry.path);
            await deleteGcpObject(endpoint.gcp.bucket, `${newPath}`);
        }
    }
];

const COPY_TABLE = [

    // local -> s3
    {
        source: 'local',
        target: 's3',
        type: TYPE_FILE,
        copyHandler: async (source, target, entry, relationshipId) => {
            const targetPath = calculateObjectStoragePath(target.s3.path, entry.path);
            const readStream = createReadStream(gracefulJoin(source.local.path, entry.path));
            const managedUpload = await uploadToS3Providers(target.s3, entry, readStream, targetPath, relationshipId);

            addInTransitItem(relationshipId, targetPath, [{
                type: 'stream', resource: readStream
            }, {
                type: 'managedUpload', resource: managedUpload
            }]);

            await managedUpload.promise();
            removeInTransitItem(relationshipId, targetPath);
        }
    },
    {
        source: 'local',
        target: 's3',
        type: TYPE_DIRECTORY,
        copyHandler: async (source, target, entry, relationshipId) => {
            const newPath = calculateObjectStoragePath(target.s3.path, entry.path);
            const managedUpload = await uploadToS3Providers(target.s3, entry, '', `${newPath}/`, relationshipId);
            await managedUpload.promise();
        }
    },
    {
        source: 'local',
        target: 's3',
        type: TYPE_SYMBOLIC,
        copyHandler: async (source, target, entry, relationshipId) => {
            const linkFileContent = await readlink(gracefulJoin(source.local.path, entry.path));
            const newPath = calculateObjectStoragePath(target.s3.path, entry.path);
            const managedUpload = await uploadToS3Providers(target.s3, entry, linkFileContent, `${newPath}`, relationshipId);
            await managedUpload.promise();
        }
    },

    // snfs -> local
    {
        source: 'snfs',
        target: 'local',
        type: TYPE_FILE,
        copyHandler: async (source, target, entry, relationshipId) => {
            await mkdirs(target.local.path);
            const targetPath = gracefulJoin(target.local.path, entry.path);

            if (entry.size === 0) {
                const fd = await open(targetPath, constants.O_CREAT);
                await Promise.all([
                    utimes(targetPath, entry.mtimeSec, entry.mtimeSec),
                    chmod(targetPath, entry.mode),
                    chownIfNeeded(targetPath, entry.uid, entry.gid)
                ]);
                await close(fd);
            } else {
                const tempTargetPath = gracefulJoin(target.local.path, entry.tempEntryName);
                const client = connectToServer(source.snfs.serverIP, source.snfs.port,
                    META_DATA[relationshipId].snfs.certificate, META_DATA[relationshipId].snfs.serverName);
                const secureReadStream = await getSecureReadStream(relationshipId, client,
                    META_DATA[relationshipId].requestId, gracefulJoin(source.snfs.path, entry.path));
                const writeStream = createWriteStream(tempTargetPath, { mode: entry.mode });

                addInTransitItem(relationshipId, targetPath, [{
                    type: 'stream', resource: secureReadStream
                }, {
                    type: 'stream', resource: writeStream
                }, {
                    type: 'path', resource: tempTargetPath
                }, {
                    type: 'client', resource: client
                }]);

                await pipe(secureReadStream, writeStream);
                secureReadStream.close();
                client.close();
                removeInTransitItem(relationshipId, targetPath);

                await Promise.all([
                    utimes(tempTargetPath, entry.mtimeSec, entry.mtimeSec),
                    chmod(tempTargetPath, entry.mode),
                    chownIfNeeded(tempTargetPath, entry.uid, entry.gid)
                ]);
                await rename(tempTargetPath, targetPath);
            }
        }
    },
    {
        source: 'snfs',
        target: 'local',
        type: TYPE_DIRECTORY,
        copyHandler: async (source, target, entry) => {
            await mkdirs(gracefulJoin(target.local.path, entry.path));
            await chmodAndChown(target.local.path, entry);
        }
    },
    {
        source: 'snfs',
        target: 'local',
        type: TYPE_SYMBOLIC,
        copyHandler: async (source, target, entry, relationshipId) => {
            await mkdirs(target.local.path);

            const sourcePath = gracefulJoin(source.snfs.path, entry.path);
            const targetPath = gracefulJoin(target.local.path, entry.path);

            const linkFileContent = await getSecureSymlinkContent(relationshipId, source.snfs.serverIP, source.snfs.port,
                META_DATA[relationshipId].snfs.certificate, META_DATA[relationshipId].snfs.serverName, META_DATA[relationshipId].requestId, sourcePath);

            await remove(targetPath);
            await symlink(linkFileContent, targetPath);
        }
    },

    // local -> snfs
    {
        source: 'local',
        target: 'snfs',
        type: TYPE_FILE,
        copyHandler: async (source, target, entry, relationshipId) => {
            const client = connectToServer(target.snfs.serverIP, target.snfs.port, META_DATA[relationshipId].snfs.certificate,
                META_DATA[relationshipId].snfs.serverName);
            const writeStream = await getSecureUploadStream(relationshipId, client,
                META_DATA[relationshipId].requestId, target.snfs.path, entry);
            const readStream = createReadStream(gracefulJoin(source.local.path, entry.path));
            const targetPath = gracefulJoin(target.snfs.path, entry.path);

            addInTransitItem(relationshipId, targetPath, [{
                type: 'stream', resource: readStream
            }, {
                type: 'stream', resource: writeStream
            }]);

            await pipe(readStream, writeStream);
            removeInTransitItem(relationshipId, targetPath);
            client.close();
        }
    },
    {
        source: 'local',
        target: 'snfs',
        type: TYPE_DIRECTORY,
        copyHandler: async (source, target, entry, relationshipId) => {
            await uploadDirectory(relationshipId, target.snfs.serverIP, target.snfs.port, META_DATA[relationshipId].snfs.certificate,
                META_DATA[relationshipId].snfs.serverName, META_DATA[relationshipId].requestId, target.snfs.path, entry);
        }
    },
    {
        source: 'local',
        target: 'snfs',
        type: TYPE_SYMBOLIC,
        copyHandler: async (source, target, entry, relationshipId) => {
            const symlinkContent = await readlink(gracefulJoin(source.local.path, entry.path));
            await uploadSymlink(relationshipId, target.snfs.serverIP, target.snfs.port, META_DATA[relationshipId].snfs.certificate,
                META_DATA[relationshipId].snfs.serverName, META_DATA[relationshipId].requestId,
                target.snfs.path, entry, symlinkContent);
        }
    },

    // local -> local
    {
        source: 'local',
        target: 'local',
        type: TYPE_FILE,
        copyHandler: async (source, target, entry, relationshipId, attributes, originalCopy, originalTarget) => {
            await mkdirs(target.local.path);
            const sourcePath = gracefulJoin(source.local.path, entry.path);
            const targetPath = gracefulJoin(target.local.path, entry.path);
            const tempTargetPath = gracefulJoin(target.local.path, entry.tempEntryName);
            if (entry.partsDetails) {
                const {
                    partsDetails: {
                        start, end, partNumber, partSize
                    }, size
                } = entry;
                logger.info('Received message parts', JSON.stringify({
                    partNumber, start, end, partSize
                }));
                const readStream = createReadStream(sourcePath, {
                    start,
                    end
                });
                const writeStream = createWriteStream(tempTargetPath, {
                    mode: entry.mode,
                    start: entry.start,
                    flags: 'a'
                });

                addInTransitItem(relationshipId, targetPath, [{
                    type: 'stream', resource: readStream
                }, {
                    type: 'stream', resource: writeStream
                }, {
                    type: 'path', resource: tempTargetPath
                }]);

                await pipe(readStream, writeStream);
                removeInTransitItem(relationshipId, targetPath);
                // calcualte new partDetails
                const newPartDetails = {
                    partNumber: partNumber + 1,
                    start: Math.min(end + 1, size), // because both start and end are inclusive
                    end: Math.min(end + partSize, size), // on node.js start and end are inclusive + we will always have start
                    partSize
                };
                logger.info('calculating new part details', JSON.stringify(newPartDetails));
                if (newPartDetails.start === size) {
                    logger.info(`Done coping parts of file ${entry.path}`);
                    if (META_DATA[relationshipId].settings.copyAcl === 'nfs') {
                        setNfsAcl(tempTargetPath, entry.nfs4Acl, 0);
                    }
                    await Promise.all([
                        utimes(tempTargetPath, entry.mtimeSec, entry.mtimeSec),
                        chmod(tempTargetPath, entry.mode),
                        chownIfNeeded(tempTargetPath, entry.uid, entry.gid)
                    ]);
                    await rename(tempTargetPath, targetPath);
                    if (META_DATA[relationshipId].settings.copyAcl === 'cifs') {
                        await setCifsAcl(targetPath, entry.cifsAcl);
                    }
                } else {
                    const entryToSend = cloneDeep(entry);
                    entryToSend.partsDetails = newPartDetails;
                    await sendTransferrerMessages(originalCopy, originalTarget, relationshipId, entryToSend, attributes);
                }
            } else {
                const readStream = createReadStream(gracefulJoin(source.local.path, entry.path));
                const writeStream = createWriteStream(tempTargetPath, { mode: entry.mode });

                // TODO - race condition - abort command between removeItem and copyToFS will delete file and create chmod/n rename errors..
                addInTransitItem(relationshipId, targetPath, [{
                    type: 'stream', resource: readStream
                }, {
                    type: 'stream', resource: writeStream
                }, {
                    type: 'path', resource: tempTargetPath
                }]);
                await pipe(readStream, writeStream);
                removeInTransitItem(relationshipId, targetPath);

                if (META_DATA[relationshipId].settings.copyAcl === 'nfs') {
                    setNfsAcl(tempTargetPath, entry.nfs4Acl, 0);
                }
                await Promise.all([
                    utimes(tempTargetPath, entry.mtimeSec, entry.mtimeSec),
                    chmod(tempTargetPath, entry.mode),
                    chownIfNeeded(tempTargetPath, entry.uid, entry.gid)
                ]);
                await rename(tempTargetPath, targetPath);
                if (META_DATA[relationshipId].settings.copyAcl === 'cifs') {
                    await setCifsAcl(targetPath, entry.cifsAcl);
                }
            }
        }
    },
    {
        source: 'local',
        target: 'local',
        type: TYPE_DIRECTORY,
        copyHandler: async (source, target, entry, relationshipId) => {
            const targetPath = gracefulJoin(target.local.path, entry.path);
            await mkdirs(targetPath);
            if (META_DATA[relationshipId].settings.copyAcl === 'nfs') {
                setNfsAcl(targetPath, entry.nfs4Acl, 1);
            }
            if (META_DATA[relationshipId].settings.copyAcl === 'cifs') {
                await setCifsAcl(targetPath, entry.cifsAcl);
            }
            await chmodAndChown(target.local.path, entry);
        }
    },
    // We can't change mtime to symlink and as a result, it will be copied again and again
    {
        source: 'local',
        target: 'local',
        type: TYPE_SYMBOLIC,
        copyHandler: async (source, target, entry) => {
            await mkdirs(target.local.path);
            const sourcePath = gracefulJoin(source.local.path, entry.path);
            const targetPath = gracefulJoin(target.local.path, entry.path);
            const [linkFileContent] = await Promise.all([readlink(sourcePath), remove(targetPath)]);
            await symlink(linkFileContent, targetPath);
        }
    },

    // s3 -> local
    {
        source: 's3',
        target: 'local',
        type: TYPE_FILE,
        copyHandler: async (source, target, entry, relationshipId) => {
            const { s3: { glacier, deepArchive } } = source;
            if (glacier || deepArchive) {
                const isRestored = await restoreS3ObjectIfNeeded(relationshipId, source, entry);
                if (isRestored) {
                    return;
                }
            }
            await mkdirs(target.local.path);
            const tempTargetPath = gracefulJoin(target.local.path, entry.tempEntryName);
            const targetPath = gracefulJoin(target.local.path, entry.path);
            const readStream = await createReadStreamFromS3Provider(source.s3, entry, relationshipId);
            const writeStream = createWriteStream(tempTargetPath, { mode: entry.mode });

            addInTransitItem(relationshipId, targetPath, [{
                type: 'stream', resource: readStream
            }, {
                type: 'stream', resource: writeStream
            }, {
                type: 'path', resource: tempTargetPath
            }]);

            await pipe(readStream, writeStream);
            removeInTransitItem(relationshipId, targetPath);
            await Promise.all([utimes(tempTargetPath, entry.mtimeSec, entry.mtimeSec), chmod(tempTargetPath, entry.mode), chown(tempTargetPath, entry.uid, entry.gid)]);
            await rename(tempTargetPath, targetPath);
        }
    },
    {
        source: 's3',
        target: 'local',
        type: TYPE_DIRECTORY,
        copyHandler: async (source, target, entry) => {
            await mkdirs(gracefulJoin(target.local.path, entry.path));
            await chmodAndChown(target.local.path, entry);
        }
    },
    {
        source: 's3',
        target: 'local',
        type: TYPE_SYMBOLIC,
        copyHandler: async (source, target, entry, relationshipId) => {
            const { s3: { glacier, deepArchive } } = source;
            if (glacier || deepArchive) {
                const isRestored = await restoreS3ObjectIfNeeded(relationshipId, source, entry);
                if (isRestored) {
                    return;
                }
            }
            await mkdirs(target.local.path);
            const symlinkSrcPath = calculateObjectStoragePath(source.s3.path, entry.path);
            const FilePointedFromSymlinkObject = await getObjectFromS3Provider(source.s3, symlinkSrcPath, relationshipId);
            const FilePointedFromSymlinkRelativePath = FilePointedFromSymlinkObject.toString();
            const symlinkTrgPath = gracefulJoin(target.local.path, entry.path);
            await remove(symlinkTrgPath);
            await symlink(FilePointedFromSymlinkRelativePath, symlinkTrgPath);
        }
    },

    // azure -> local
    {
        source: 'azure',
        target: 'local',
        type: TYPE_FILE,
        copyHandler: async (source, target, entry, relationshipId) => {
            await mkdirs(target.local.path);
            const tempTargetPath = gracefulJoin(target.local.path, entry.tempEntryName);
            const targetPath = gracefulJoin(target.local.path, entry.path);
            const abortController = new AbortController();

            const newPath = calculateObjectStoragePath(source.azure.path, entry.path);
            const readStream = await createBlobReadStream(source.azure.container, newPath, relationshipId, entry.size, abortController.signal);
            const writeStream = createWriteStream(tempTargetPath, { mode: entry.mode });

            addInTransitItem(relationshipId, targetPath, [{
                type: 'aborter', resource: abortController
            }, {
                type: 'stream', resource: writeStream
            }, {
                type: 'path', resource: tempTargetPath
            }]);

            await pipe(readStream, writeStream);
            await Promise.all([utimes(tempTargetPath, entry.mtimeSec, entry.mtimeSec), chmod(tempTargetPath, entry.mode), chown(tempTargetPath, entry.uid, entry.gid)]);
            await rename(tempTargetPath, targetPath);
            removeInTransitItem(relationshipId, targetPath);
        }
    },
    {
        source: 'azure',
        target: 'local',
        type: TYPE_DIRECTORY,
        copyHandler: async (source, target, entry) => {
            await mkdirs(gracefulJoin(target.local.path, entry.path));
            await chmodAndChown(target.local.path, entry);
        }
    },
    {
        source: 'azure',
        target: 'local',
        type: TYPE_SYMBOLIC,
        copyHandler: async (source, target, entry, relationshipId) => {
            await mkdirs(target.local.path);
            const newPath = calculateObjectStoragePath(source.azure.path, entry.path);
            const abortController = new AbortController();
            addInTransitItem(relationshipId, newPath, [{
                type: 'aborter', resource: abortController
            }]);

            const linkFileContent = await getBlobToText(source.azure.container, newPath, relationshipId, abortController.signal);
            const targetPath = gracefulJoin(target.local.path, entry.path);
            await remove(targetPath);
            await symlink(linkFileContent, targetPath);
            removeInTransitItem(relationshipId, targetPath);
        }
    },

    // local -> azure
    {
        source: 'local',
        target: 'azure',
        type: TYPE_FILE,
        copyHandler: async (source, target, entry, relationshipId) => {
            const targetPath = calculateObjectStoragePath(target.azure.path, entry.path);
            const metadata = createMetadata(target.azure.tags, entry);
            const abortController = new AbortController();

            addInTransitItem(relationshipId, targetPath, [{
                type: 'aborter', resource: abortController
            }]);

            await createWriteStreamToBlockBlobByFile(target.azure.container, targetPath, metadata, relationshipId, gracefulJoin(source.local.path, entry.path), abortController.signal);
            await setBlobAccessTier(target.azure.container, targetPath, relationshipId, get(target, 'azure.blobTier'), abortController.signal);
            removeInTransitItem(relationshipId, targetPath);
        }
    },
    {
        source: 'local',
        target: 'azure',
        type: TYPE_DIRECTORY,
        copyHandler: async (source, target, entry, relationshipId) => {
            const newPath = calculateObjectStoragePath(target.azure.path, entry.path);
            const metadata = createMetadata(target.azure.tags, entry);
            const abortController = new AbortController();

            addInTransitItem(relationshipId, newPath, [{
                type: 'aborter', resource: abortController
            }]);

            await createBlockBlobFromText(target.azure.container, `${newPath}/`, metadata, relationshipId, abortController.signal);
            await setBlobAccessTier(target.azure.container, `${newPath}/`, relationshipId, get(target, 'azure.blobTier'), abortController.signal);
            removeInTransitItem(relationshipId, newPath);
        }
    },
    {
        source: 'local',
        target: 'azure',
        type: TYPE_SYMBOLIC,
        copyHandler: async (source, target, entry, relationshipId) => {
            const linkFileContent = await readlink(gracefulJoin(source.local.path, entry.path));
            const newPath = calculateObjectStoragePath(target.azure.path, entry.path);
            const metadata = createMetadata(target.azure.tags, entry);
            const abortController = new AbortController();

            addInTransitItem(relationshipId, newPath, [{
                type: 'aborter', resource: abortController
            }]);

            await createBlockBlobFromText(target.azure.container, newPath, metadata, relationshipId, abortController.signal, linkFileContent);
            await setBlobAccessTier(target.azure.container, newPath, relationshipId, get(target, 'azure.blobTier'), abortController.signal);
            removeInTransitItem(relationshipId, newPath);
        }
    },

    // gcp -> local
    {
        source: 'gcp',
        target: 'local',
        type: TYPE_FILE,
        copyHandler: async (source, target, entry, relationshipId) => {
            await mkdirs(target.local.path);
            const tempTargetPath = gracefulJoin(target.local.path, entry.tempEntryName);
            const targetPath = gracefulJoin(target.local.path, entry.path);

            const newPath = calculateObjectStoragePath(source.gcp.path, entry.path);
            const readStream = getGcpReadStream(source.gcp.bucket, newPath, relationshipId);
            const writeStream = createWriteStream(tempTargetPath, { mode: entry.mode });

            addInTransitItem(relationshipId, targetPath, [{
                type: 'stream', resource: readStream
            }, {
                type: 'stream', resource: writeStream
            }, {
                type: 'path', resource: tempTargetPath
            }]);

            await pipe(readStream, writeStream);
            removeInTransitItem(relationshipId, targetPath);
            await Promise.all([utimes(tempTargetPath, entry.mtimeSec, entry.mtimeSec), chmod(tempTargetPath, entry.mode), chown(tempTargetPath, entry.uid, entry.gid)]);
            await rename(tempTargetPath, targetPath);
        }
    },
    {
        source: 'gcp',
        target: 'local',
        type: TYPE_DIRECTORY,
        copyHandler: async (source, target, entry) => {
            await mkdirs(gracefulJoin(target.local.path, entry.path));
            await chmodAndChown(target.local.path, entry);
        }
    },
    {
        source: 'gcp',
        target: 'local',
        type: TYPE_SYMBOLIC,
        copyHandler: async (source, target, entry) => {
            // TODO check if we could use getFile or how we could get the file content ( not stream )
            await mkdirs(target.local.path);
            const newPath = calculateObjectStoragePath(source.gcp.path, entry.path);
            const linkFileMetaData = await getGcpObjectMetadata(source.gcp.bucket, newPath);
            const linkFileContent = linkFileMetaData.metadata.symlink;
            const targetPath = gracefulJoin(target.local.path, entry.path);
            await remove(targetPath);
            await symlink(linkFileContent, targetPath);
        }
    },

    // local -> gcp
    {
        source: 'local',
        target: 'gcp',
        type: TYPE_FILE,
        copyHandler: async (source, target, entry, relationshipId) => {
            const readStream = createReadStream(gracefulJoin(source.local.path, entry.path));
            const targetPath = calculateObjectStoragePath(target.gcp.path, entry.path);
            const writeStream = getGcpWriteStream(target.gcp.bucket, targetPath, createGcpMetadata(target.gcp.tags, entry));

            addInTransitItem(relationshipId, targetPath, [{
                type: 'stream', resource: readStream
            }, {
                type: 'stream', resource: writeStream
            }]);

            await pipe(readStream, writeStream);
            removeInTransitItem(relationshipId, targetPath);
        }
    },
    {
        source: 'local',
        target: 'gcp',
        type: TYPE_DIRECTORY,
        copyHandler: async (source, target, entry) => {
            const targetPath = calculateObjectStoragePath(target.gcp.path, entry.path);
            await createGcpFolder(target.gcp.bucket, targetPath, createGcpMetadata(target.gcp.tags, entry));
        }
    },
    {
        source: 'local',
        target: 'gcp',
        type: TYPE_SYMBOLIC,
        copyHandler: async (source, target, entry, relationshipId) => {
            const linkFileContent = await readlink(gracefulJoin(source.local.path, entry.path));
            const targetPath = calculateObjectStoragePath(target.gcp.path, entry.path);
            await createGcpSymlink(target.gcp.bucket, targetPath, linkFileContent, createGcpMetadata(target.gcp.tags, entry, linkFileContent), relationshipId);
        }
    },

    // azure -> azure
    {
        source: 'azure',
        target: 'azure',
        type: TYPE_FILE,
        copyHandler: async (source, target, entry, relationshipId) => {
            const targetPath = calculateObjectStoragePath(target.azure.path, entry.path);
            const sourcePath = calculateObjectStoragePath(source.azure.path, entry.path);
            const metadata = createMetadata(target.azure.tags, entry);
            // finding the current metadata of the copied blob
            const abortController = new AbortController();
            const { metadata: sourceBlobMetadata } = await getBlobProperties(relationshipId, MOUNT_SOURCE_SUFFIX, source.azure.container, sourcePath);
            const targetBlobMetadata = { ...metadata, ...sourceBlobMetadata };
            await copyBlobFromUrl(relationshipId, source.azure.container, target.azure.container, sourcePath, targetPath);

            addInTransitItem(relationshipId, targetPath, [{
                type: 'aborter', resource: abortController
            }]);

            await setBlobMetadata(relationshipId, MOUNT_TARGET_SUFFIX,
                target.azure.container, targetPath, targetBlobMetadata, abortController.signal);
            await setBlobAccessTier(target.azure.container, targetPath, relationshipId, get(target, 'azure.blobTier'), abortController.signal);
            removeInTransitItem(relationshipId, targetPath);
        }
    },
    {
        source: 'azure',
        target: 'azure',
        type: TYPE_DIRECTORY,
        copyHandler: async (source, target, entry, relationshipId) => {
            const newPath = calculateObjectStoragePath(target.azure.path, entry.path);
            const metadata = createMetadata(target.azure.tags, entry);
            const abortController = new AbortController();

            addInTransitItem(relationshipId, newPath, [{
                type: 'aborter', resource: abortController
            }]);
            await createBlockBlobFromText(target.azure.container, `${newPath}/`, metadata, relationshipId, abortController.signal);
            await setBlobAccessTier(target.azure.container, `${newPath}/`, relationshipId, get(target, 'azure.blobTier'), abortController.signal);
            removeInTransitItem(relationshipId, newPath);
        }
    },
    {
        source: 'azure',
        target: 'azure',
        type: TYPE_SYMBOLIC,
        copyHandler: async (source, target, entry, relationshipId) => {
            const targetPath = calculateObjectStoragePath(target.azure.path, entry.path);
            const sourcePath = calculateObjectStoragePath(source.azure.path, entry.path);
            const metadata = createMetadata(target.azure.tags, entry);
            // finding the current metadata of the copied blob
            const abortController = new AbortController();
            const { metadata: sourceBlobMetadata } = await getBlobProperties(relationshipId, MOUNT_SOURCE_SUFFIX, source.azure.container, sourcePath);
            const targetBlobMetadata = { ...metadata, ...sourceBlobMetadata };
            addInTransitItem(relationshipId, targetPath, [{
                type: 'aborter', resource: abortController
            }]);

            await copyBlobFromUrl(relationshipId,
                source.azure.container, target.azure.container, sourcePath, targetPath);

            await setBlobMetadata(relationshipId, MOUNT_TARGET_SUFFIX,
                target.azure.container, targetPath, targetBlobMetadata, abortController.signal);
            await setBlobAccessTier(target.azure.container, targetPath, relationshipId, get(target, 'azure.blobTier'), abortController.signal);
            removeInTransitItem(relationshipId, targetPath);
        }
    },

    // s3 -> s3
    {
        source: 's3',
        target: 's3',
        type: TYPE_FILE,
        copyHandler: async (source, target, entry, relationshipId) => {
            const { s3: { glacier, deepArchive } } = source;
            if (glacier || deepArchive) {
                const isRestored = await restoreS3ObjectIfNeeded(relationshipId, source, entry);
                if (isRestored) {
                    return;
                }
            }
            const { s3: { provider: sourceProvider } } = source;
            const { s3: { provider: targetProvider } } = target;
            if (sourceProvider === targetProvider && [S3_PROVIDER, ALIBABA_PROVIDER].includes(sourceProvider)) {
                await copyObjectBetweenIdenticalS3Providers(source.s3, target.s3, entry, relationshipId);
            } else {
                const targetPath = calculateObjectStoragePath(target.s3.path, entry.path);
                const readStream = await createReadStreamFromS3Provider(source.s3, entry, relationshipId);
                const sourceMetadata = await getMetadataFromS3Provider(source.s3, entry, relationshipId);
                const managedUpload = await uploadToS3Providers(target.s3, entry, readStream, targetPath, relationshipId, sourceMetadata);

                addInTransitItem(relationshipId, targetPath, [{
                    type: 'stream', resource: readStream
                }, {
                    type: 'managedUpload', resource: managedUpload
                }]);

                await managedUpload.promise();
                removeInTransitItem(relationshipId, targetPath);
            }
        }
    },
    {
        source: 's3',
        target: 's3',
        type: TYPE_DIRECTORY,
        copyHandler: async (source, target, entry, relationshipId) => {
            const { s3: { provider: sourceProvider } } = source;
            // If the "directory" is not a directory we uploaded the copyObject operation will fail because the directory is not an actual key but a prefix
            const newPath = calculateObjectStoragePath(target.s3.path, entry.path);
            let sourceMetadata;
            if ([SGWS_PROVIDER, IBM_PROVIDER].includes(sourceProvider)) { // copy metadata only relevant for SGWS and IBM
                sourceMetadata = await getMetadataFromS3Provider(source.s3, entry, relationshipId);
            }
            const managedUpload = await uploadToS3Providers(target.s3, entry, '', `${newPath}/`, relationshipId, sourceMetadata);
            await managedUpload.promise();
        }
    },
    {
        source: 's3',
        target: 's3',
        type: TYPE_SYMBOLIC,
        copyHandler: async (source, target, entry, relationshipId) => {
            const { s3: { glacier, deepArchive } } = source;
            if (glacier || deepArchive) {
                const isRestored = await restoreS3ObjectIfNeeded(relationshipId, source, entry);
                if (isRestored) {
                    return;
                }
            }
            const { provider: sourceProvider } = source;
            const { provider: targetProvider } = target;
            if (sourceProvider === targetProvider && [S3_PROVIDER, ALIBABA_PROVIDER].includes(sourceProvider)) {
                await copyObjectBetweenIdenticalS3Providers(source.s3, target.s3, entry, relationshipId);
            } else {
                const symlinkSrcPath = calculateObjectStoragePath(source.s3.path, entry.path);
                const FilePointedFromSymlinkObject = await getObjectFromS3Provider(source.s3, symlinkSrcPath, relationshipId);
                const linkFileContent = FilePointedFromSymlinkObject.toString();
                const newPath = calculateObjectStoragePath(target.s3.path, entry.path);
                const sourceMetadata = await getMetadataFromS3Provider(source.s3, entry, relationshipId);
                const managedUpload = await uploadToS3Providers(target.s3, entry, linkFileContent, newPath, relationshipId, sourceMetadata);
                await managedUpload.promise();
            }
        }
    },

    // s3 -> azure
    {
        source: 's3',
        target: 'azure',
        type: TYPE_FILE,
        copyHandler: async (source, target, entry, relationshipId) => {
            const { s3: { glacier, deepArchive } } = source;
            if (glacier || deepArchive) {
                const isRestored = await restoreS3ObjectIfNeeded(relationshipId, source, entry);
                if (isRestored) {
                    return;
                }
            }
            const targetPath = calculateObjectStoragePath(target.azure.path, entry.path);
            const metadata = createMetadata(target.azure.tags, entry);
            const readStream = await createReadStreamFromS3Provider(source.s3, entry, relationshipId);
            const abortController = new AbortController();

            addInTransitItem(relationshipId, targetPath, [{
                type: 'stream', resource: readStream
            }, {
                type: 'aborter', resource: abortController
            }]);

            await createWriteStreamToBlockBlobByReadStream(target.azure.container, targetPath, metadata, relationshipId, readStream, abortController.signal);
            await setBlobAccessTier(target.azure.container, targetPath, relationshipId, get(target, 'azure.blobTier'), abortController.signal);
            removeInTransitItem(relationshipId, targetPath);
        }
    },
    {
        source: 's3',
        target: 'azure',
        type: TYPE_DIRECTORY,
        copyHandler: async (source, target, entry, relationshipId) => {
            const newPath = calculateObjectStoragePath(target.azure.path, entry.path);
            const metadata = createMetadata(target.azure.tags, entry);
            const abortController = new AbortController();

            addInTransitItem(relationshipId, newPath, [{
                type: 'aborter', resource: abortController
            }]);

            await createBlockBlobFromText(target.azure.container, `${newPath}/`, metadata, relationshipId, abortController.signal);
            await setBlobAccessTier(target.azure.container, `${newPath}/`, relationshipId, get(target, 'azure.blobTier'), abortController.signal);
            removeInTransitItem(relationshipId, newPath);
        }
    },
    {
        source: 's3',
        target: 'azure',
        type: TYPE_SYMBOLIC,
        copyHandler: async (source, target, entry, relationshipId) => {
            const { s3: { glacier, deepArchive } } = source;
            if (glacier || deepArchive) {
                const isRestored = await restoreS3ObjectIfNeeded(relationshipId, source, entry);
                if (isRestored) {
                    return;
                }
            }
            const symlinkSrcPath = calculateObjectStoragePath(source.s3.path, entry.path);
            const FilePointedFromSymlinkObject = await getObjectFromS3Provider(source.s3, symlinkSrcPath, relationshipId);
            const linkFileContent = FilePointedFromSymlinkObject.toString();
            const newPath = calculateObjectStoragePath(target.azure.path, entry.path);
            const metadata = createMetadata(target.azure.tags, entry);
            const abortController = new AbortController();

            addInTransitItem(relationshipId, newPath, [{
                type: 'aborter', resource: abortController
            }]);

            await createBlockBlobFromText(target.azure.container, newPath, metadata, relationshipId, abortController.signal, linkFileContent);
            await setBlobAccessTier(target.azure.container, newPath, relationshipId, get(target, 'azure.blobTier'), abortController.signal);
            removeInTransitItem(relationshipId, newPath);
        }
    },

    // azure -> s3
    {
        source: 'azure',
        target: 's3',
        type: TYPE_FILE,
        copyHandler: async (source, target, entry, relationshipId) => {
            const targetPath = calculateObjectStoragePath(target.s3.path, entry.path);
            const newPath = calculateObjectStoragePath(source.azure.path, entry.path);
            const abortController = new AbortController();
            const readStream = await createBlobReadStream(source.azure.container, newPath, relationshipId, entry.size, abortController.signal);
            const pass = new PassThrough();
            const managedUpload = await uploadToS3Providers(target.s3, entry, readStream.pipe(pass), targetPath, relationshipId);

            addInTransitItem(relationshipId, targetPath, [{
                type: 'aborter', resource: abortController
            }, {
                type: 'managedUpload', resource: managedUpload
            }]);

            await managedUpload.promise();
            removeInTransitItem(relationshipId, targetPath);
        }
    },
    {
        source: 'azure',
        target: 's3',
        type: TYPE_DIRECTORY,
        copyHandler: async (source, target, entry, relationshipId) => {
            const newPath = calculateObjectStoragePath(target.s3.path, entry.path);
            const managedUpload = await uploadToS3Providers(target.s3, entry, '', `${newPath}/`, relationshipId);
            await managedUpload.promise();
        }
    },
    {
        source: 'azure',
        target: 's3',
        type: TYPE_SYMBOLIC,
        copyHandler: async (source, target, entry, relationshipId) => {
            const sourceSymlinkPath = calculateObjectStoragePath(source.azure.path, entry.path);
            const abortController = new AbortController();
            addInTransitItem(relationshipId, sourceSymlinkPath, [{
                type: 'aborter', resource: abortController
            }]);
            const linkFileContent = await getBlobToText(source.azure.container, sourceSymlinkPath, relationshipId, abortController.signal);
            const targetPath = calculateObjectStoragePath(target.s3.path, entry.path);
            const managedUpload = await uploadToS3Providers(target.s3, entry, linkFileContent, targetPath, relationshipId);
            await managedUpload.promise();
        }
    },

    // gcp -> s3
    {
        source: 'gcp',
        target: 's3',
        type: TYPE_FILE,
        copyHandler: async (source, target, entry, relationshipId) => {
            const targetPath = calculateObjectStoragePath(target.s3.path, entry.path);
            const newPath = calculateObjectStoragePath(source.gcp.path, entry.path);
            const readStream = getGcpReadStream(source.gcp.bucket, newPath, relationshipId);
            const managedUpload = await uploadToS3Providers(target.s3, entry, readStream, targetPath, relationshipId);

            addInTransitItem(relationshipId, targetPath, [{
                type: 'readStream', resource: readStream
            }, {
                type: 'managedUpload', resource: managedUpload
            }]);

            await managedUpload.promise();
            removeInTransitItem(relationshipId, targetPath);
        }
    },
    {
        source: 'gcp',
        target: 's3',
        type: TYPE_DIRECTORY,
        copyHandler: async (source, target, entry, relationshipId) => {
            const newPath = calculateObjectStoragePath(target.s3.path, entry.path);
            const managedUpload = await uploadToS3Providers(target.s3, entry, '', `${newPath}/`, relationshipId);
            await managedUpload.promise();
        }
    },
    {
        source: 'gcp',
        target: 's3',
        type: TYPE_SYMBOLIC,
        copyHandler: async (source, target, entry, relationshipId) => {
            const sourcePath = calculateObjectStoragePath(source.gcp.path, entry.path);
            const linkFileMetaData = await getGcpObjectMetadata(source.gcp.bucket, sourcePath);
            const linkFileContent = linkFileMetaData.metadata.symlink;
            const targetPath = calculateObjectStoragePath(target.s3.path, entry.path);
            const managedUpload = await uploadToS3Providers(target.s3, entry, linkFileContent, targetPath, relationshipId);
            await managedUpload.promise();
        }
    },

    // s3 -> gcp
    {
        source: 's3',
        target: 'gcp',
        type: TYPE_FILE,
        copyHandler: async (source, target, entry, relationshipId) => {
            const { s3: { glacier, deepArchive } } = source;
            if (glacier || deepArchive) {
                const isRestored = await restoreS3ObjectIfNeeded(relationshipId, source, entry);
                if (isRestored) {
                    return;
                }
            }
            const readStream = await createReadStreamFromS3Provider(source.s3, entry, relationshipId);
            const targetPath = calculateObjectStoragePath(target.gcp.path, entry.path);
            const writeStream = getGcpWriteStream(target.gcp.bucket, targetPath, createGcpMetadata(target.gcp.tags, entry));

            addInTransitItem(relationshipId, targetPath, [{
                type: 'readStream', resource: readStream
            }, {
                type: 'writeStream', resource: writeStream
            }]);

            await pipe(readStream, writeStream);
            removeInTransitItem(relationshipId, targetPath);
        }
    },
    {
        source: 's3',
        target: 'gcp',
        type: TYPE_DIRECTORY,
        copyHandler: async (source, target, entry) => {
            const targetPath = calculateObjectStoragePath(target.gcp.path, entry.path);
            await createGcpFolder(target.gcp.bucket, targetPath, createGcpMetadata(target.gcp.tags, entry));
        }
    },
    {
        source: 's3',
        target: 'gcp',
        type: TYPE_SYMBOLIC,
        copyHandler: async (source, target, entry, relationshipId) => {
            const { s3: { glacier, deepArchive } } = source;
            if (glacier || deepArchive) {
                const isRestored = await restoreS3ObjectIfNeeded(relationshipId, source, entry);
                if (isRestored) {
                    return;
                }
            }
            const symlinkSrcPath = calculateObjectStoragePath(source.s3.path, entry.path);
            const FilePointedFromSymlinkObject = await getObjectFromS3Provider(source.s3, symlinkSrcPath, relationshipId);
            const linkFileContent = FilePointedFromSymlinkObject.toString();
            const targetPath = calculateObjectStoragePath(target.gcp.path, entry.path);
            await createGcpSymlink(target.gcp.bucket, targetPath, linkFileContent, createGcpMetadata(target.gcp.tags, entry, linkFileContent), relationshipId);
        }
    },

    // azure -> gcp
    {
        source: 'azure',
        target: 'gcp',
        type: TYPE_FILE,
        copyHandler: async (source, target, entry, relationshipId) => {
            const newPath = calculateObjectStoragePath(source.azure.path, entry.path);
            const abortController = new AbortController();
            const readStream = await createBlobReadStream(source.azure.container, newPath, relationshipId, entry.size, abortController.signal);
            const targetPath = calculateObjectStoragePath(target.gcp.path, entry.path);
            const writeStream = getGcpWriteStream(target.gcp.bucket, targetPath, createGcpMetadata(target.gcp.tags, entry));

            addInTransitItem(relationshipId, targetPath, [{
                type: 'aborter', resource: abortController
            }, {
                type: 'writeStream', resource: writeStream
            }]);

            await pipe(readStream, writeStream);
            removeInTransitItem(relationshipId, targetPath);
        }
    },
    {
        source: 'azure',
        target: 'gcp',
        type: TYPE_DIRECTORY,
        copyHandler: async (source, target, entry) => {
            const targetPath = calculateObjectStoragePath(target.gcp.path, entry.path);
            await createGcpFolder(target.gcp.bucket, targetPath, createGcpMetadata(target.gcp.tags, entry));
        }
    },
    {
        source: 'azure',
        target: 'gcp',
        type: TYPE_SYMBOLIC,
        copyHandler: async (source, target, entry, relationshipId) => {
            const sourceSymlinkPath = calculateObjectStoragePath(source.azure.path, entry.path);
            const abortController = new AbortController();
            addInTransitItem(relationshipId, sourceSymlinkPath, [{
                type: 'aborter', resource: abortController
            }]);
            const linkFileContent = await getBlobToText(source.azure.container, sourceSymlinkPath, relationshipId, abortController.signal);
            const targetPath = calculateObjectStoragePath(target.gcp.path, entry.path);
            await createGcpSymlink(target.gcp.bucket, targetPath, linkFileContent, createGcpMetadata(target.gcp.tags, entry, linkFileContent), relationshipId);
        }
    },

    // gcp -> azure
    {
        source: 'gcp',
        target: 'azure',
        type: TYPE_FILE,
        copyHandler: async (source, target, entry, relationshipId) => {
            const targetPath = calculateObjectStoragePath(target.azure.path, entry.path);
            const sourcePath = calculateObjectStoragePath(source.gcp.path, entry.path);
            const readStream = getGcpReadStream(source.gcp.bucket, sourcePath, relationshipId);
            const metadata = createMetadata(target.azure.tags, entry);
            const abortController = new AbortController();

            addInTransitItem(relationshipId, targetPath, [{
                type: 'stream', resource: readStream
            }, {
                type: 'aborter', resource: abortController
            }]);

            await createWriteStreamToBlockBlobByReadStream(target.azure.container, targetPath, metadata, relationshipId, readStream, abortController.signal);
            await setBlobAccessTier(target.azure.container, targetPath, relationshipId, get(target, 'azure.blobTier'), abortController.signal);
            removeInTransitItem(relationshipId, targetPath);
        }
    },
    {
        source: 'gcp',
        target: 'azure',
        type: TYPE_DIRECTORY,
        copyHandler: async (source, target, entry, relationshipId) => {
            const targetPath = calculateObjectStoragePath(target.azure.path, entry.path);
            const metadata = createMetadata(target.azure.tags, entry);
            const abortController = new AbortController();

            addInTransitItem(relationshipId, targetPath, [{
                type: 'aborter', resource: abortController
            }]);

            await createBlockBlobFromText(target.azure.container, `${targetPath}/`, metadata, relationshipId, abortController.signal);
            await setBlobAccessTier(target.azure.container, `${targetPath}/`, relationshipId, get(target, 'azure.blobTier'), abortController.signal);
            removeInTransitItem(relationshipId, targetPath);
        }
    },
    {
        source: 'gcp',
        target: 'azure',
        type: TYPE_SYMBOLIC,
        copyHandler: async (source, target, entry, relationshipId) => {
            const sourcePath = calculateObjectStoragePath(source.gcp.path, entry.path);
            const linkFileMetaData = await getGcpObjectMetadata(source.gcp.bucket, sourcePath);
            const linkFileContent = linkFileMetaData.metadata.symlink;
            const targetPath = calculateObjectStoragePath(target.azure.path, entry.path);
            const metadata = createMetadata(target.azure.tags, entry);
            const abortController = new AbortController();

            addInTransitItem(relationshipId, targetPath, [{
                type: 'aborter', resource: abortController
            }]);

            await createBlockBlobFromText(target.azure.container, targetPath, metadata, relationshipId, abortController.signal, linkFileContent);
            await setBlobAccessTier(target.azure.container, targetPath, relationshipId, get(target, 'azure.blobTier'), abortController.signal);
            removeInTransitItem(relationshipId, targetPath);
        }
    },

    // gcp -> gcp
    {
        source: 'gcp',
        target: 'gcp',
        type: TYPE_FILE,
        copyHandler: async (source, target, entry) => {
            const targetPath = calculateObjectStoragePath(target.gcp.path, entry.path);
            const sourcePath = calculateObjectStoragePath(source.gcp.path, entry.path);
            const metadata = createGcpMetadata(target.gcp.tags, entry);
            await copyGcpObject(source.gcp.bucket, sourcePath, target.gcp.bucket, targetPath);
            await setMetadata(target.gcp.bucket, targetPath, metadata);
        }
    },
    {
        source: 'gcp',
        target: 'gcp',
        type: TYPE_DIRECTORY,
        copyHandler: async (source, target, entry) => {
            const targetPath = calculateObjectStoragePath(target.gcp.path, entry.path);
            await createGcpFolder(target.gcp.bucket, targetPath, createGcpMetadata(target.gcp.tags, entry));
        }
    },
    {
        source: 'gcp',
        target: 'gcp',
        type: TYPE_SYMBOLIC,
        copyHandler: async (source, target, entry) => {
            const targetPath = calculateObjectStoragePath(target.gcp.path, entry.path);
            const sourcePath = calculateObjectStoragePath(source.gcp.path, entry.path);
            const metadata = createGcpMetadata(target.gcp.tags, entry);
            await copyGcpObject(source.gcp.bucket, sourcePath, target.gcp.bucket, targetPath);
            await setMetadata(target.gcp.bucket, targetPath, metadata);
        }
    }
];

function sendStatisticsMessage(statisticsMessage) {
    // send status message back
    const isValid = validateMessage(statisticsMessage, 'SQS', copyStatisticsSchema);
    if (isValid) {
        process.send(statisticsMessage);
    }
}

async function handleDoneRemoveAbortUpdateMessage(message) {
    const {
        source, target, relationshipId, type
    } = message;

    logger.info(`Start ${type} message`);
    if (META_DATA[relationshipId]) {
        META_DATA[relationshipId].status = IPC_TYPE_ABORT;
        delete META_DATA[relationshipId].queues;
        INTERNAL_DISPATCHER_QUEUE.remove(task => task.data.body.relationshipId === relationshipId);

        const itemsInProgress = META_DATA[relationshipId].items;
        logger.info(`Command ${type}, aborting ${keys(itemsInProgress).length} items`);

        const items = Object.keys(itemsInProgress).flatMap(key => itemsInProgress[key]);

        await Promise.all(items.map(item => triggerAbort(item)));
    }

    relationshipsLogger.debug(`[${relationshipId}] ${type} copy from:`, source, 'to:', target);

    return new Promise(resolve => {
        setTimeout(() => {
            logger.info(`Done handling command: ${type} for rid: ${relationshipId}`);
            delete META_DATA[relationshipId];
            resolve();
        }, CLEAN_RELATIONSHIP_RESOURCES_TIMEOUT);
    });
}

async function handleCopy(source, target, entry, relationshipId, attributes, originalCopy, originalTarget) {
    const sourceProtocol = _(source).keys().head();
    const targetProtocol = _(target).keys().head();
    const match = find(COPY_TABLE, {
        source: sourceProtocol,
        target: targetProtocol,
        type: entry.type
    });
    const startTime = Date.now();
    try {
        await retry(async bail => {
            if (isRelationshipAborted(relationshipId)) {
                // in case of relationship abort - do not retry
                relationshipsLogger.info('Bailing - relationship aborted', relationshipId);
                bail();
            } else {
                await match.copyHandler(source, target, entry, relationshipId, attributes, originalCopy, originalTarget);
            }
        }, {
            retries: META_DATA[relationshipId].settings.retries,
            onRetry: async err => {
                relationshipsLogger.warn(`[${relationshipId}] RETRYING copy from:`, JSON.stringify(source), 'to:', JSON.stringify(target), 'entry:', JSON.stringify(entry), err);
                const targetPath = get(target, 'local.path');
                if (targetPath) {
                    const tempTargetPath = gracefulJoin(targetPath, entry.tempEntryName);
                    if (await pathExists(tempTargetPath)) {
                        try {
                            relationshipsLogger.info('Removing temporary file due to copy failure:', tempTargetPath);
                            await retry(async () => {
                                await remove(tempTargetPath);
                            }, { retries: META_DATA[relationshipId].settings.retries });
                        } catch (error) {
                            relationshipsLogger.error(`Failed removing temporary file ${tempTargetPath}`, error);
                        }
                    }
                }
            }
        });
        relationshipsLogger.info(`[${relationshipId}] COMPLETED copy from:`, JSON.stringify(source), 'to:', JSON.stringify(target), 'entry:', JSON.stringify(entry), `took: ${Date.now() - startTime}`);
        sendStatisticsMessage({
            step: TRANSFER_STEP,
            relationshipId,
            targetRequiredAction: ACTION_COPY,
            bytes: !(isNil(entry.partsDetails)) ? entry.partsDetails.end - entry.partsDetails.start : entry.size,
            items: (entry.partsDetails && entry.size === entry.partsDetails.end) || (!entry.partsDetails) ? 1 : 0,
            objectType: entry.type,
            path: entry.path,
            status: STATUS_SUCCEEDED,
            attributes
        });
    } catch (err) {
        // retry bail function throws 'Aborted' error
        if (err.message !== 'Aborted') {
            relationshipsLogger.error(`[${relationshipId}] FAILED copy from:`, JSON.stringify(source), 'to:', JSON.stringify(target), 'entry:', JSON.stringify(entry), err);
            throw err;
        }
    }
}

async function handleDelete(endpoint, entry, relationshipId, attributes, action, mountSuffix) {
    const match = find(DELETE_TABLE, { endpoint: _(endpoint).keys().head(), type: entry.type });

    relationshipsLogger.info(`[${relationshipId}] STARTING deletion from:`, endpoint, 'entry:', entry);

    try {
        await retry(() => match.deleteHandler(endpoint, entry, relationshipId, mountSuffix), {
            retries: META_DATA[relationshipId].settings.retries,
            onRetry: err => relationshipsLogger.info(`[${relationshipId}] RETRYING deletion from:`, endpoint, 'entry:', entry, err)
        });
        relationshipsLogger.info(`[${relationshipId}] COMPLETED deletion from:`, endpoint, 'entry:', entry);
        if (!isNetappEntry(entry.path)) {
            sendStatisticsMessage({
                step: TRANSFER_STEP,
                relationshipId,
                targetRequiredAction: action,
                bytes: entry.size,
                items: 1,
                objectType: entry.type,
                status: STATUS_SUCCEEDED,
                path: entry.path,
                attributes
            });
        }
    } catch (err) {
        relationshipsLogger.error(`[${relationshipId}] FAILED deletion from:`, endpoint, 'entry:', entry, err);
        throw err;
    }
}

async function handleTransferrerMessage(message) {
    // TODO think about moving type into attributes in all SQS messages
    const {
        body: {
            source, target, entry, relationshipId, type
        },
        receiptHandle, attributes
    } = message;
    let { deleteOnSource } = META_DATA[relationshipId].settings;

    try {
        const newSource = getNewEndPointIfNeeded(relationshipId, META_DATA[relationshipId].requestId, MOUNT_SOURCE_SUFFIX, source);
        const newTarget = getNewEndPointIfNeeded(relationshipId, META_DATA[relationshipId].requestId, MOUNT_TARGET_SUFFIX, target);

        switch (type) {
            case ACTION_COPY:
                await handleCopy(newSource, newTarget, entry, relationshipId, attributes, source, target);
                break;
            case ACTION_DELETE_SOURCE: // we use as precaution 2 flags for delete on source
                if (deleteOnSource && config.get('workers.transferrer.delete-on-source')) {
                    // TODO for now we want to ignore this action in statistics, later we will not
                    try {
                        await handleDelete(newSource, entry, relationshipId, attributes, ACTION_DELETE_SOURCE, MOUNT_SOURCE_SUFFIX);
                    } catch (e) {} // eslint-disable-line  no-empty
                }
                break;
            case ACTION_DELETE_TARGET:
                await handleDelete(newTarget, entry, relationshipId, attributes, ACTION_DELETE_TARGET, MOUNT_TARGET_SUFFIX);
                // TODO - think - we should delay directories delete messages in scanner
                break;
            default:
                throw new Error(`Got unexpected operation type: ${type}`);
        }
    } catch (err) {
        deleteOnSource = false;
        const errorObject = generateErrorStatisticsObject(err);
        const errorMessage = {
            step: TRANSFER_STEP,
            relationshipId,
            targetRequiredAction: type,
            bytes: entry.size,
            items: 1,
            objectType: entry.type,
            path: entry.path,
            status: STATUS_FAILED,
            reason: errorObject,
            attributes
        };

        sendStatisticsMessage(errorMessage);
        failedRelationshipLogger.addContext('relationshipId', relationshipId);
        failedRelationshipLogger.info(`[${relationshipId}] [${type}] FAILED: source`, source, 'target:', target, 'entry:', entry, err);
    } finally {
        message.attributes.MessagesCount = (message.attributes.MessagesCount - 1).toString();
        const transferrerQueue = get(META_DATA, `${relationshipId}.queues.transferrerQueue`);
        if (transferrerQueue && message.attributes.MessagesCount === '0') {
            deleteMessage(transferrerQueue, receiptHandle);
        }
        // we use as precaution 2 flags for delete on source, and make sure the relationship wasn't aborted
        // for now we support only deletion of files TODO delete on source folders
        if (config.get('workers.transferrer.delete-on-source') && deleteOnSource && entry.type === TYPE_FILE && type === ACTION_COPY && !isRelationshipAborted(relationshipId)) {
            const newSource = getNewEndPointIfNeeded(relationshipId, META_DATA[relationshipId].requestId, MOUNT_SOURCE_SUFFIX, source);
            await handleDelete(newSource, entry, relationshipId, attributes, ACTION_DELETE_SOURCE, MOUNT_SOURCE_SUFFIX);
        }
    }
}

async function handleMessageVisibility() {
    const runningTasks = INTERNAL_DISPATCHER_QUEUE.workersList().map(task => task.data);
    const waitingTasks = INTERNAL_DISPATCHER_QUEUE._tasks.toArray(); // eslint-disable-line no-underscore-dangle
    const queues = compact(Object.values(META_DATA).map(value => get(value, 'queues.transferrerQueue')));
    logger.info('Internal dispatcher queue:', {
        runningTasks: runningTasks.length,
        waitingTasks: waitingTasks.length,
        queues: queues.length
    });
    const groupedMessagesByRelationshipId = groupBy(runningTasks.concat(waitingTasks), message => message.body.relationshipId);
    Object.keys(groupedMessagesByRelationshipId).forEach(relationshipId => {
        if (META_DATA[relationshipId]) {
            const receiptHandles = groupedMessagesByRelationshipId[relationshipId]
                .filter(task => Date.now() - task.lastVisibilityTimeoutChange > ms(SQS_MESSAGE_UNCHANGED_VISIBILITY_TIMEOUT_LIMIT))
                .map(task => task.receiptHandle);

            INTERNAL_DISPATCHER_QUEUE._tasks.toArray().forEach(message => { // eslint-disable-line no-underscore-dangle
                if (receiptHandles.includes(message.receiptHandle)) {
                    message.lastVisibilityTimeoutChange = Date.now();
                }
            }); // eslint-disable-line no-param-reassign, no-underscore-dangle
            INTERNAL_DISPATCHER_QUEUE.workersList().forEach(task => {
                if (receiptHandles.includes(task.data.receiptHandle)) {
                    task.data.lastVisibilityTimeoutChange = Date.now();
                }
            }); // eslint-disable-line no-param-reassign
            changeMessageVisibilityBatch(META_DATA[relationshipId].queues.transferrerQueue, uniq(receiptHandles));
        }
    });
}

// Sums up the total size of all the waiting and running tasks (objects)
function totalTasksSize() {
    const reducer = (accumulator, currentValue) => accumulator + currentValue;
    return INTERNAL_DISPATCHER_QUEUE._tasks.toArray().map(task => task.body.entry.size).concat(INTERNAL_DISPATCHER_QUEUE.workersList().map(task => task.data.body.entry.size)).reduce(reducer, 0); // eslint-disable-line no-underscore-dangle
}

async function getWork(concurrency) {
    const queues = compact(Object.values(META_DATA).map(value => get(value, 'queues.transferrerQueue')));

    if (queues.length > 0) {
        const waitingWorkersThreshold = config.get('workers.transferrer.concurrency-eagerness-ratio') * concurrency;
        if (waitingWorkersThreshold > INTERNAL_DISPATCHER_QUEUE.length()) {
            const transferrerQueue = head(shuffle(queues));
            try {
                const messages = await receiveMessages(transferrerQueue);
                messages.sort((a, b) => a.body.entry.size - b.body.entry.size); // ascending sorted array. We want to push first the smaller files in order to fill the queue as much as possible (helpful when the target is S3)
                logger.debug(`Got ${messages.length} messages`, {
                    running: INTERNAL_DISPATCHER_QUEUE.running(),
                    length: INTERNAL_DISPATCHER_QUEUE.length()
                });
                messages.forEach(message => {
                    try {
                        if (validateMessage(message.body, 'SQS', sqsTransferrerIncomingMessagesSchemes[message.body.type])) {
                            if (!message.body.target.s3 || prettyBytes(config.get('protocols.s3.max-object-size')) > totalTasksSize() + message.body.entry.size) {
                                INTERNAL_DISPATCHER_QUEUE.push(message);
                            }
                        }
                    } catch (err) {
                        logger.warn('Failed to validate transferrer message', err);
                        deleteMessage(transferrerQueue, message.receiptHandle);
                    }
                });
            } catch (err) {
                logger.error('Failed to receive messages from queue', transferrerQueue, err);
            }
        }
        setImmediate(() => getWork(concurrency));
    } else {
        setTimeout(() => getWork(concurrency), SQS_DEFAULT_WAIT_TIMEOUT * 1000);
    }
}

function boot() {
    const concurrency = config.get('workers.transferrer.concurrency');
    logger.info(`Starting transferrer #${cluster.worker.id}, PID ${process.pid}, concurrency ${concurrency} ...`);

    // queue should have 'async' keyword otherwise it does not work
    INTERNAL_DISPATCHER_QUEUE = queue(async transferrerMessage => {
        const { body: { relationshipId } } = transferrerMessage;
        if (META_DATA[relationshipId]) {
            return handleTransferrerMessage(transferrerMessage);
        }
    }, concurrency);
    INTERNAL_DISPATCHER_QUEUE.drain = () => logger.info('Drain -> no more work');
    INTERNAL_DISPATCHER_QUEUE.error = (err, transferrerMessage) => logger.error(`Got error for transferrer message ${transferrerMessage}`, err);
    getWork(concurrency);
    setInterval(() => handleMessageVisibility(), ms(SQS_CHANGE_MESSAGE_VISIBILITY_TIMEOUT_INTERVAL));
}

process.on('message', message => {
    logger.info('Got IPC message', message);

    if (validateMessage(message, 'IPC', transferrerIncomingIPCMessagesSchemes[message.type])) {
        switch (message.type) {
            case IPC_TYPE_COPY:
                META_DATA[message.relationshipId] = {
                    queues: {
                        transferrerQueue: message.transferrerQueue
                    },
                    status: IPC_TYPE_COPY,
                    requestId: message.requestId,
                    items: {},
                    settings: message.settings
                };
                if (message.certificate && message.serverName) {
                    META_DATA[message.relationshipId].snfs = {
                        certificate: message.certificate,
                        serverName: message.serverName
                    };
                }
                ////TODO - Delete when CS-4986 is resolved
                if (config.get('protocols.nfs.acl.active')) {
                    META_DATA[message.relationshipId].settings.copyAcl = 'nfs';
                }
                break;
            case IPC_TYPE_DONE:
            case IPC_TYPE_UPDATE:
            case IPC_TYPE_REMOVE:
            case IPC_TYPE_ABORT:
                handleDoneRemoveAbortUpdateMessage(message);
                break;
            default:
                logger.error('Got unexpected message type', message.type);
        }
    }
});

if (process.env.NODE_ENV !== 'test') {
    boot();
}

module.exports = { DELETE_TABLE, COPY_TABLE, isNetappEntry };
