const AWS = require('aws-sdk-mock');
const {
    listBuckets, getObjectTagging, getBucketTagging, getBucketLocation, createTagsArray
} = require('../../../src/lib/aws/s3_base_operations');
const { getS3 } = require('../../../src/lib/s3Clients');
const { TYPE_FILE, TYPE_SYMBOLIC } = require('../../../src/lib/consts');

const date = new Date('2018-01-30T13:09:52.000Z');
const mtimeSec = parseInt(date.getTime() / 1000);

describe('listBuckets with S3 client', () => {
    afterEach(() => {

    });

    test('listBuckets - no buckets- S3 client', async () => {
        const S3 = getS3();
        S3.listBuckets = jest.fn(() => ({ promise: () => ({ Buckets: [] }) }));
        const buckets = await listBuckets(S3);
        expect(buckets).toHaveLength(0);
    });
});

describe('getObjectTagging- S3', () => {
    test('getObjectTagging - key doesnt exist', async () => {
        const S3 = getS3();
        S3.getObjectTagging = jest.fn(() => ({
            promise: () => new Promise((resolve, reject) => {
                reject({ // eslint-disable-line prefer-promise-reject-errors
                    message: 'The specified key does not exist.',
                    code: 'NoSuchKey'
                });
            })
        }));

        const tags = await getObjectTagging(S3, 'testBucket', 'test.js');
        expect(tags).toHaveLength(0);
    });

    test('getObjectTagging - bucket doesnt exist', async () => {
        const S3 = getS3();
        S3.getObjectTagging = jest.fn(() => ({
            promise: () => new Promise((resolve, reject) => {
                reject({ // eslint-disable-line prefer-promise-reject-errors
                    message: 'The specified bucket does not exist',
                    code: 'NoSuchBucket'
                });
            })
        }));

        await expect(getObjectTagging(S3, 'doesntexist', 'test.js')).rejects.toEqual({
            message: 'The specified bucket does not exist',
            code: 'NoSuchBucket'
        });
    });

    test('getObjectTagging -tags returns', async () => {
        const S3 = getS3();
        S3.getObjectTagging = jest.fn(() => ({ promise: () => ({ TagSet: [{ Key: 'a', Value: 'b' }] }) }));

        const tags = await getObjectTagging(S3, 'dfiochen', 'secondfile.txt');
        expect(tags).toHaveLength(1);
    });
});

describe('getBucketTagging -S3 client ', () => {
    afterEach(() => AWS.restore('S3', 'getBucketTagging'));

    test('getBucketTagging - no tags on bucket', async () => {
        const S3 = getS3();
        S3.getBucketTagging = jest.fn(() => ({
            promise: () => new Promise((resolve, reject) => {
                reject({ message: 'The TagSet does not exist', code: 'NoSuchTagSet' });// eslint-disable-line prefer-promise-reject-errors
            })
        }));

        const tags = await getBucketTagging(S3, 'test');
        expect(tags).toHaveLength(0);
    });
});

describe('getBucketLocation - S3', () => {
    afterEach(() => AWS.restore('S3', 'getBucketLocation'));

    test('getBucketLocation -access denied', async () => {
        const S3 = getS3();
        S3.getBucketLocation = jest.fn(() => ({
            promise: () => new Promise((resolve, reject) => {
                reject({ message: 'Access Denied', code: 'AccessDenied' });// eslint-disable-line prefer-promise-reject-errors
            })
        }));

        await expect(getBucketLocation(S3, 'testBucket')).rejects.toEqual({ code: 'AccessDenied', message: 'Access Denied' });
    });
});

describe('createTagsArray', () => {
    function createTag(key, value) {
        return { Key: key, Value: `${value}` };
    }

    test('createTags-HappyPath without symbolic link', () => {
        const entry = {
            uid: 0, gid: 1, mode: 16822, type: TYPE_FILE, mtimeSec
        };
        expect(createTagsArray(undefined, entry)).toEqual([createTag('netapp-cs-uid', 0), createTag('netapp-cs-gid', 1), createTag('netapp-cs-mode', 16822), createTag('netapp-cs-mtimesec', mtimeSec)]);
    });

    test('createTags-HappyPath with symbolic link', () => {
        const entry = {
            uid: 0, gid: 1, mode: 16822, type: TYPE_SYMBOLIC, mtimeSec
        };
        expect(createTagsArray(undefined, entry)).toEqual([createTag('netapp-cs-uid', 0), createTag('netapp-cs-gid', 1), createTag('netapp-cs-mode', 16822), createTag('netapp-cs-mtimesec', mtimeSec),
            createTag('netapp-cs-slink', 'True')]);
    });

    test('createTags-HappyPath with symbolic link and tags', () => {
        const entry = {
            uid: 0, gid: 1, mode: 16822, type: TYPE_SYMBOLIC, mtimeSec
        };
        const tags = [createTag('testKey', 'testValue')];
        expect(createTagsArray(tags, entry)).toEqual([createTag('testKey', 'testValue'), createTag('netapp-cs-uid', 0), createTag('netapp-cs-gid', 1), createTag('netapp-cs-mode', 16822), createTag('netapp-cs-mtimesec', mtimeSec),
            createTag('netapp-cs-slink', 'True')]);
    });

    test('createTags-HappyPath with FILE type and 2 tags', () => {
        const entry = {
            uid: 0, gid: 1, mode: 16822, type: TYPE_FILE, mtimeSec
        };
        const tags = [createTag('testKey', 'testValue'), createTag('testKey1', 'testValue1')];
        expect(createTagsArray(tags, entry)).toEqual(tags.concat([createTag('netapp-cs-uid', 0), createTag('netapp-cs-gid', 1), createTag('netapp-cs-mode', 16822), createTag('netapp-cs-mtimesec', mtimeSec)]));
    });
});
