const swaggerTools = require('swagger-tools');
const express = require('express');
const config = require('config');
const compression = require('compression');
const path = require('path');
const http = require('http');
const https = require('https');
const helmet = require('helmet');
const SwaggerParser = require('swagger-parser');
const cors = require('cors');
const log4js = require('log4js');
require('./globals');
const {
    API_ROOT,
    SWAGGER_SPEC_PATH, SWAGGER_UI
} = require('./utils/consts');
const { getRuntimeEnvironment } = require('./utils/environment');
const bootstrap = require('./utils/bootstrap');
const healthRoutes = require('./routes/health');
const accountsRoutes = require('./routes/accounts');
const messagesRoutes = require('./routes/messages');
const storagesRoutes = require('./routes/storages');
const dataBrokersRoutes = require('./routes/data-brokers');
const relationshipsRoutes = require('./routes/relationships');
const groupsRoutes = require('./routes/groups');
const awsMarketplaceRegistrationRoutes = require('./routes/marketplace-registration-aws');
const azureMarketplaceRegistrationRoutes = require('./routes/marketplace-registration-azure');
const timelinesRoutes = require('./routes/timelines');
const authenticateMiddleware = require('./middlewares/authenticate');
const findAccountMiddleware = require('./middlewares/find-account');
const reactRouterFallbackMiddleware = require('./middlewares/react-router-fallback');
const { continuationLocalStorageMiddleware } = require('./middlewares/continuation-local-storage');
const errorHandlerMiddleware = require('./middlewares/error-handler');
const timelineMiddleware = require('./middlewares/timeline');
const validateSwaggerSpec = require('./utils/swagger');
const getLogger = require('./utils/logger');
require('./utils/backup');
require('./lib/axios');
require('./operations/marketplace-azure-charging');
// patch max sockets per origin otherwise node starts to cough on requests
// see: http://docs.aws.amazon.com/sdk-for-javascript/v2/developer-guide/node-configuring-maxsockets.html
// see: https://gist.github.com/nscott/8735628
http.globalAgent.maxSockets = config.get('server.max-sockets-per-origin');
https.globalAgent.maxSockets = config.get('server.max-sockets-per-origin');

const logger = getLogger();

process.on('uncaughtException', err => logger.error('Uncaught exception', err));

process.on('unhandledRejection', (err, promise) => logger.error('Unhandled promise rejection', promise, err));

// TODO upgrade swagger
async function initSwagger(app) {
    try {
        const swaggerSpec = await new SwaggerParser().dereference(path.resolve(__dirname, SWAGGER_SPEC_PATH));
        // validate swagger spec
        const api = await validateSwaggerSpec(swaggerSpec);
        return new Promise(resolve => {
            swaggerTools.initializeMiddleware(api, middlewares => {
                // serve the Swagger documents and Swagger UI
                app.use(middlewares.swaggerUi({
                    swaggerUi: SWAGGER_UI
                }));
                // interpret Swagger resources and attach metadata to request
                app.use(middlewares.swaggerMetadata());
                // add timeline to all the relevant requests (not get and not messages)
                app.use(API_ROOT, timelineMiddleware);
                // validate Swagger requests
                app.use(middlewares.swaggerValidator({ validateResponse: true }));
                resolve(app);
            });
        });
    } catch (err) {
        logger.error('Failed initializing swagger:', err);
        throw err;
    }
}

function initMiddlewares(app) {
    app.use(API_ROOT, express.json());
    app.use(API_ROOT, express.urlencoded({ extended: true }));
    // configure security helmet
    app.use(helmet({
        frameguard: false, contentSecurityPolicy:false
    }));
    // configure continuation local storage
    app.use(API_ROOT, continuationLocalStorageMiddleware);

    // health check hidden (from swagger) route (and not logged in access log)
    app.use(API_ROOT, healthRoutes);
    // configure access logger
    app.use(API_ROOT, log4js.connectLogger(getLogger('access-log'), {
        level: log4js.levels.INFO,
        format: ':remote-addr :protocol :http-version :method :url :status :content-length :response-time'
    }));
    // always add response compression
    app.use(compression());
    // enable all CORS requests
    app.use(cors({
        exposedHeaders:'X-DFIO-Req-Id'
    }));

    // configure static assets middleware
    logger.info('Serving UI from:', config.get('paths.ui'));
    const staticAssetsMiddleware = express.static(config.get('paths.ui'));
    app.use(staticAssetsMiddleware);
    // always return index.html in case no resource found and it's not an API call
    app.use(reactRouterFallbackMiddleware(staticAssetsMiddleware));
    // authenticate request
    app.use(API_ROOT, authenticateMiddleware);
    app.use(API_ROOT, findAccountMiddleware);
}

function initRoutes(app) {
    app.use(API_ROOT, accountsRoutes);
    app.use(API_ROOT, messagesRoutes);
    app.use(API_ROOT, storagesRoutes);
    app.use(API_ROOT, awsMarketplaceRegistrationRoutes);
    app.use(API_ROOT, azureMarketplaceRegistrationRoutes);
    app.use(API_ROOT, dataBrokersRoutes);
    app.use(API_ROOT, relationshipsRoutes);
    app.use(API_ROOT, groupsRoutes);
    app.use(API_ROOT, timelinesRoutes);
}

async function startServer() {
    try {
        const app = express();
        await bootstrap();
        initMiddlewares(app);
        await initSwagger(app);
        initRoutes(app);
        // init app exception handler
        app.use(errorHandlerMiddleware);
        app.listen(config.get('server.http-port'));
        logger.info('======================================================================');
        logger.info(`Server running in '${getRuntimeEnvironment()}' mode started listening.`);
    } catch (err) {
        logger.error('Failed staring server with error:', err);
        process.exit(0);
    }
}

process.once('SIGINT', () => {
    logger.info('CTRL-C caught');
    process.exit(0);
});

startServer();
