#!/bin/bash

usage () { echo "Usage: $0 [-a <access_key>] [-s <secret_key>] [ -h <proxy_host> -p <proxy_port> [ -u <proxy_username> -w <proxy_password> ] [ -c <ca_file> ] ] [-g <google_cloud_file_path>]"; }

while getopts :a:s:h:p:u:w:c:g: opt ; do
   case $opt in
      a) ACCESS_KEY=$OPTARG ;;
      s) SECRET_KEY=$OPTARG ;;
      h) PROXY_HOST=$OPTARG ;;
      p) PROXY_PORT=$OPTARG ;;
      u) PROXY_USERNAME=$OPTARG ;;
      w) PROXY_PASSWORD=$OPTARG ;;
      c) CA_FILE=$OPTARG ;;
      g) GOOGLE_FILE=$OPTARG ;;
      *) echo 'no valid input available' ;;
   esac
done

if ! [ -z "$ACCESS_KEY" ];then
    if [ -z "$SECRET_KEY" ];then
        echo "secret key is required"
        usage
        exit
    fi
fi

if ! [ -z "$PROXY_HOST" ] && ! [ -z "$PROXY_PORT" ];then
    PROXY_STRING=$PROXY_HOST:$PROXY_PORT
    if ! [ -z "$PROXY_USERNAME" ] && ! [ -z "$PROXY_PASSWORD" ];then
        PROXY_STRING=$PROXY_USERNAME:$PROXY_PASSWORD@$PROXY_STRING
    fi
    export http_proxy=http://$PROXY_STRING
    export https_proxy=http://$PROXY_STRING
fi

if ! [ -z "$GOOGLE_FILE" ];then
    if ! [ -f "$GOOGLE_FILE" ];then
        echo "Google credentials file does not exist"
        exit
    fi
fi

mkdir -p /opt/netapp/databroker/mnt
mkdir -p /opt/netapp/databroker/tmp
cd /opt/netapp/databroker

# disable core dumps
ulimit -c 0

# increase open files limit
ulimit -n 4096

# install third party dependencies
installRequiredRedHatAndCentOsPackages() {
yum -y update --security
yum -y update curl
yum -y install unzip
yum -y install nfs-utils
yum -y install cifs-utils
yum -y install samba-client
yum -y install python3
yum -y install tar

# install 7z for ASUP
curl --location https://dl.fedoraproject.org/pub/epel/epel-release-latest-7.noarch.rpm --output epel-release-latest-7.noarch.rpm
yum -y install epel-release-latest-7.noarch.rpm
yum -y install p7zip p7zip-plugins --nogpgcheck

# install and start nscd (cache getaddrinfo requests)
yum -y install nscd
service nscd start
chkconfig nscd on

# remove SSM agent
yum -y erase amazon-ssm-agent
}

installRequiredUbuntuAndDebianPackages() {
apt-get -y update
apt-get -y install unattended-upgrades
apt-get -y install unzip
apt-get -y install nfs-kernel-server
apt-get -y install cifs-utils
apt-get -y install samba-client
apt-get -y install python3
apt-get -y install python3-distutils

# install 7z for ASUP
apt-get -y install p7zip-full

# install and start nscd (cache getaddrinfo requests)
apt-get -y install nscd
service nscd start
update-rc.d nscd enable
}

installRequiredSusePackages() {
zypper -n refresh
zypper -n install unzip
zypper -n install nfs-kernel-server
zypper -n install cifs-utils
zypper -n install samba samba-client
zypper -n install python3

# install 7z for ASUP
zypper -n install p7zip p7zip

# install and start nscd (cache getaddrinfo requests)
zypper -n install nscd
service nscd start
chkconfig nscd on
}

case "$( grep -Eoi 'red hat|debian|suse|ubuntu|centos' /etc/os-release |  head -n +1 | awk '{print tolower($0)}' )" in
    "red hat" | "centos")
      installRequiredRedHatAndCentOsPackages
      ;;
    "ubuntu" | "debian")
      installRequiredUbuntuAndDebianPackages
      ;;
    "suse")
      installRequiredSusePackages
      ;;
  esac

# install n and stable node+npm
curl --verbose --location https://repo.cloudsync.netapp.com/n --connect-timeout 10 --retry 5 --output /usr/bin/n
chmod +x /usr/bin/n
N_PREFIX=// n latest

# download data broker bundle
curl --verbose --location "{agentBundleUrl}" --connect-timeout 10 --retry 5 --output data-broker.zip
unzip -o data-broker.zip -d .
\rm -f data-broker.zip

# generate certificate & key
mkdir certification
openssl req -newkey rsa:2048 -subj '/CN={dataBrokerId}' -nodes -keyout ./certification/dataBrokerKey.key -x509 -days 365 -out ./certification/dataBrokerCert.crt

# add relevant key-values to the data-broker json
if ! [ -z "{port}" ];then

cat <<EOT > config/data-broker.json
{
    "environment": "{environment}",
    "data-broker-id": "{dataBrokerId}",
    "type": "{type}",
    "commandsQueue": "{commandsQueue}",
    "statusesQueue": "{statusesQueue}",
    "port": {port},
    "aws":{
        "sqs":{
            "accessKeyId": "{sqsAccessKeyId}",
            "secretAccessKey": "{sqsSecretAccessKey}"
        }
    }
}
EOT

else

cat <<EOT > config/data-broker.json
{
    "environment": "{environment}",
    "data-broker-id": "{dataBrokerId}",
    "type": "{type}",
    "commandsQueue": "{commandsQueue}",
    "statusesQueue": "{statusesQueue}",
    "aws":{
        "sqs":{
            "accessKeyId": "{sqsAccessKeyId}",
            "secretAccessKey": "{sqsSecretAccessKey}"
        }
    }
}
EOT

fi

# create empty local.json file
cat <<EOT > config/local.json
{}
EOT

if ! [ -z "$ACCESS_KEY" ] && ! [ -z "$SECRET_KEY" ];then
    curl --verbose --location "https://s3.amazonaws.com/aws-cli/awscli-bundle.zip" --connect-timeout 10 --retry 5 --output awscli-bundle.zip
    unzip -o awscli-bundle.zip -d .
    sudo python3 ./awscli-bundle/install -i /usr/local/aws -b /usr/local/bin/aws
    sudo rm -rf ./awscli-bundle awscli-bundle.zip

    /usr/local/bin/aws configure set aws_access_key_id $ACCESS_KEY --profile data-broker
    /usr/local/bin/aws configure set aws_secret_access_key $SECRET_KEY --profile data-broker
fi

if ! [ -z "$PROXY_STRING" ];then
    sudo npm config set proxy http://$PROXY_STRING
    sudo npm config set https-proxy http://$PROXY_STRING
fi

if ! [ -z "$CA_FILE" ];then
    sudo npm config set cafile $CA_FILE
fi

sudo -E npm i --production

# install vault
VAULT_VERSION="1.5.0"
curl --verbose --location https://repo.cloudsync.netapp.com/vault_"$VAULT_VERSION"_linux_amd64.zip --connect-timeout 10 --retry 5 --output /opt/netapp/databroker/vault/vault_"$VAULT_VERSION"_linux_amd64.zip --create-dirs
unzip -o /opt/netapp/databroker/vault/vault_"$VAULT_VERSION"_linux_amd64.zip -d /usr/bin
chmod +x /usr/bin/vault
rm /opt/netapp/databroker/vault/vault_"$VAULT_VERSION"_linux_amd64.zip


# install PM2 globally
sudo npm i pm2 -g

sudo pm2 start --name "vault" --output /dev/null --error /dev/null /usr/bin/vault -- server --config=/opt/netapp/databroker/vault/config.hcl

# start
if ! [ -z "$PROXY_STRING" ];then
    if ! [ -z "$CA_FILE" ];then
        sudo AWS_PROFILE=data-broker GOOGLE_APPLICATION_CREDENTIALS=$GOOGLE_FILE no_proxy=127.0.0.1,169.254.169.254 http_proxy=http://$PROXY_STRING https_proxy=http://$PROXY_STRING NODE_EXTRA_CA_CERTS=$CA_FILE pm2 start app.js --name data-broker --output /dev/null --error /dev/null
    else
        sudo AWS_PROFILE=data-broker GOOGLE_APPLICATION_CREDENTIALS=$GOOGLE_FILE no_proxy=127.0.0.1,169.254.169.254 http_proxy=http://$PROXY_STRING https_proxy=http://$PROXY_STRING pm2 start app.js --name data-broker --output /dev/null --error /dev/null
    fi
else
    sudo AWS_PROFILE=data-broker GOOGLE_APPLICATION_CREDENTIALS=$GOOGLE_FILE pm2 start app.js --name data-broker --output /dev/null --error /dev/null
fi

sudo pm2 startup
sudo pm2 save

