const Promise = require('bluebird');
const {
    isEmpty, find, chunk, keyBy, concat, flatten
} = require('lodash');
const AWS = require('aws-sdk/index');
const {agenda} = require('../../utils/agenda');
const SaaSBilling = require('../../models/saas-billing');
const {
    SAAS_SUBSCRIPTION_PROVIDER_AWS,
    AGENDA_TASK_NAME_AWS_SAAS_METERING, SAAS_REGISTRATION_STATUS_RESOLVE_FAILED, AGENDA_TASK_NAME_UPDATE_CUSTOMER_IDENTIFIER_ON_AWS_ACCOUNTS,
    SAAS_REGISTRATION_STATUS_SUBSCRIBE_SUCCESS, PAYGO_SUBSCRIPTION, MP_PRODUCT_CODE_METERING, SERVICE_REGION, MP_SELLER_SECRET_KEY, MP_SELLER_ACCESS_KEY
} = require('../../utils/consts');
const Relationship = require('../../models/relationship');
const DataBroker = require('../../models/data-broker');
const {Account, findBillableAccounts} = require('../../models/account');
const {Cost, lastCostRecords} = require('../../models/cost');
const {validateMarketplaceProductCode} = require('../../utils/common');
const updateSubscription = require('../../operations/subscription-operations');
const {findAccountsByAwsAccountId} = require('../../operations/data-broker-operations');
const getLogger = require('../../utils/logger');

const logger = getLogger();

function getMarketplaceMetering() {
    return new AWS.MarketplaceMetering({
        region: SERVICE_REGION,
        accessKeyId: MP_SELLER_ACCESS_KEY,
        secretAccessKey: MP_SELLER_SECRET_KEY
    });
}

async function batchMeterUsage(records) {
    const marketplaceMetering = getMarketplaceMetering();
    return Promise.all(chunk(records, 25).map(chunk => {
        const batch = {
            ProductCode: MP_PRODUCT_CODE_METERING,
            UsageRecords: chunk
        };
        logger.info('Metering request with batch:', batch);

        return marketplaceMetering.batchMeterUsage(batch).promise();
    }));
}

function getCustomerTiers(allRelationships, byolReservedRelationships) {
    let remainingCount;
    let tier1 = 0;
    let tier2 = 0;
    let tier3 = 0;
    if (byolReservedRelationships !== 0) {
        remainingCount = allRelationships - byolReservedRelationships;
        if (remainingCount > 0) {
            tier3 = remainingCount;
        }
    } else {
        remainingCount = allRelationships;
        if (remainingCount < 0) {
            remainingCount = 0;
        }

        if (remainingCount > 5) {
            tier1 = 5;
            remainingCount -= tier1;
        } else {
            tier1 = remainingCount;
            remainingCount = 0;
        }

        if (remainingCount > 15) {
            tier2 = 15;
            remainingCount -= tier2;
        } else {
            tier2 = remainingCount;
            remainingCount = 0;
        }

        tier3 = remainingCount;
    }
    return {tier1, tier2, tier3};
}

function createMeteringEntry(customerId, tier, quantity, timeStamp) {
    return {
        CustomerIdentifier: customerId,
        Dimension: tier,
        Quantity: quantity,
        Timestamp: timeStamp
    };
}

async function resolveCustomerIdentifierByMpCustomerToken(mpCustomerToken) {
    logger.info(`Resolving customer identifier from token: ${mpCustomerToken}`);

    try {
        const resolvedToken = await getMarketplaceMetering().resolveCustomer({RegistrationToken: mpCustomerToken}).promise();
        logger.info(`Resolved customer Id - ${resolvedToken.CustomerIdentifier} for token - ${mpCustomerToken}: ProductCode - ${resolvedToken.ProductCode}`);

        const type = validateMarketplaceProductCode(resolvedToken.ProductCode);
        logger.info(`ProductCode - ${resolvedToken.ProductCode} with type - ${type} was validated from token -  ${mpCustomerToken} with customer Id: ${resolvedToken.CustomerIdentifier}`);

        return {
            customerIdentifier: resolvedToken.CustomerIdentifier,
            status: SAAS_REGISTRATION_STATUS_SUBSCRIBE_SUCCESS,
            type
        };
    } catch (err) {
        logger.error(`Failed to resolve customer token ${mpCustomerToken}:`, err);
        return {
            customerIdentifier: undefined,
            status: SAAS_REGISTRATION_STATUS_RESOLVE_FAILED,
            type: PAYGO_SUBSCRIPTION
        };
    }
}

// given records to bill => batches them and send them to saas metering service.
// saves the honored records to the DB and returns a collection of the unprocessed/unhonored records
function billRecords(records) {
    return batchMeterUsage(records)
        .then(responses => {
            if (responses.length === 0) {
                return {processedRecords: [], unprocessedRecords: []};
            }
            logger.debug('Metering responses for batches:', JSON.stringify(responses));
            const processedRecords = flatten(responses.map(response => response.Results));
            // logging unhonored. unlike unprocessed these are processed but not successfully [DuplicateRecord, CustomerNotSubscribed]
            logger.warn('Records not honored:', processedRecords.filter(record => record.Status !== 'Success'));
            logger.info('processed billing entries ', processedRecords);
            return SaaSBilling.insertMany(processedRecords).then(() => { // need to delete in the end
                const unprocessedRecords = flatten(responses.map(response => response.UnprocessedRecords));
                logger.info('Unprocessed records', unprocessedRecords);
                return {processedRecords, unprocessedRecords};
            });
        });
}


async function filterAndInsert(billingObjects) {
    const lastRecords = await lastCostRecords();
    lastRecords.forEach(record => {
        const {
            tier1, tier2, tier3, byol, _id: accountId
        } = record;
        const newRecordForAccount = billingObjects[accountId];
        if (newRecordForAccount) {
            const {tier1: newTier1, tier2: newTier2, tier3: newTier3, byol: newByol} = newRecordForAccount;
            if (((tier1 === newTier1) || !(tier1 || newTier1)) &&
                ((tier2 === newTier2) || !(tier2 || newTier2)) &&
                ((tier3 === newTier3) || !(tier3 || newTier3)) &&
                ((newByol && newByol.usedRelationships === byol.usedRelationships) || !(newByol || byol.usedRelationships))) {
                logger.debug('******DELETE record from Cost table', newRecordForAccount);// forDEBUG
                delete billingObjects[accountId];
            }
        }
    });
    logger.info('Insert all the success records to the Cost collection in MongoDB', billingObjects);

    if (!isEmpty(billingObjects)) {
        return Cost.insertMany(Object.values(billingObjects));
    }
}

function createObjectsForMongoSaaSAccounts(processedRecords, chargingTime, billableAccounts, byolObjects) {
    logger.info('Create MongoDB objects for SaaS accounts');
    const objectsForMongo = {};
    const accounts = {};
    const successRecords = processedRecords.filter(record => record.Status === 'Success');

    successRecords.forEach(record => {
        const customerIdentifier = record.UsageRecord.CustomerIdentifier;
        const accountsWithCustomerIdentifier = billableAccounts.filter(account => account.subscription.customerIdentifier === customerIdentifier);
        const dimension = record.UsageRecord.Dimension.toLowerCase();
        const quantity = record.UsageRecord.Quantity;
        accountsWithCustomerIdentifier.forEach(account => {
            const byol = byolObjects[account.subscription.customerIdentifier];
            if (objectsForMongo[account._id]) {
                objectsForMongo[account._id][dimension] = quantity;
            } else {
                objectsForMongo[account._id] = {timestamp: chargingTime, account: account._id, byol};
                objectsForMongo[account._id][dimension] = quantity;
            }

            accounts[account._id] = account;
        });
    });

    return objectsForMongo;
}

async function billSaaSAccounts(chargingTime) {
    logger.info('Find all the relationships for subscribed account');
    const billableAccounts = await findBillableAccounts();
    const billableAccountIds = billableAccounts.map(account => account._id);
    logger.info('For each customerId find all its billing');
    const groups = await Relationship.aggregate([
        {$match: {account: {$in: billableAccountIds}, isQstack: false}},
        {
            $lookup: {
                from: 'accounts',
                localField: 'account',
                foreignField: '_id',
                as: 'account'
            }
        },
        {$group: {_id: '$account.subscription.customerIdentifier', count: {$sum: 1}}}]).exec();
    logger.info('For each customerId find all its billing');
    const byolEntitlements = await Account.customerIdentifierByolEntitlement();
    const byolByCustomerIdentifier = keyBy(byolEntitlements, '_id');
    const byolObjects = {};
    const records = groups.map(group => {
        const [customerId] = group._id;
        const customerEntitlement = byolByCustomerIdentifier[customerId] ? byolByCustomerIdentifier[customerId].count : 0;
        const byolPerCustomerIdentifier = byolByCustomerIdentifier[customerId] ? (group.count - customerEntitlement <= 0 ? group.count : customerEntitlement) : 0;
        if (customerEntitlement !== 0) {
            byolObjects[customerId] = {
                reservedRelationships: customerEntitlement,
                usedRelationships: byolPerCustomerIdentifier
            };
        }
        const {tier1, tier2, tier3} = getCustomerTiers(group.count, customerEntitlement);
        const tiers = [];

        if (tier1 > 0) {
            tiers.push(createMeteringEntry(customerId, 'Tier1', tier1, chargingTime));
        }
        if (tier2 > 0) {
            tiers.push(createMeteringEntry(customerId, 'Tier2', tier2, chargingTime));
        }
        if (tier3 > 0) {
            tiers.push(createMeteringEntry(customerId, 'Tier3', tier3, chargingTime));
        }

        return tiers;
    }).flat();
    const billRecordsResults = await billRecords(records);
    const {processedRecords, unprocessedRecords} = billRecordsResults;
    const billObjectAfterRetry = await billRecords(unprocessedRecords);
    const processedRecordsAfterRetry = concat(processedRecords, billObjectAfterRetry.processedRecords);
    const unprocessedRecordsAfterRetry = billRecordsResults.unprocessedRecords;
    if (unprocessedRecordsAfterRetry.length) {
        const unprocessedEntriesForDB = unprocessedRecordsAfterRetry.map(meteringRecord => ({
            Status: 'Failed',
            MeteringRecordId: '',
            UsageRecord: meteringRecord
        }));
        await SaaSBilling.insertMany(unprocessedEntriesForDB);
    }
    logger.info('Create objects for mongo');
    return createObjectsForMongoSaaSAccounts(processedRecordsAfterRetry, chargingTime, billableAccounts, byolObjects);
}

async function addAccountsWithoutRelationshipsToMongoObject(mongoObjects, chargingTime) {
    const accounts = await Account.find({}).exec();
    accounts.map(account => {
        if (!mongoObjects[account._id]) {
            const byol = {usedRelationships: 0, reservedRelationships: 0};
            mongoObjects[account._id] = {
                account: account._id, tier1: 0, timestamp: chargingTime, byol
            };
        }
    });
    return mongoObjects;
}

// will update BYOL for NetApp users with trial expired as well
async function byolDetailsForUnsubscribedAccounts(objectsForMongo, chargingTime) {
    logger.info('find BYOL details for unsubscribed accounts');
    const accounts = await Account.find({trialExpiry: {$lt: new Date()}}).exec();
    const accountsId = accounts.map(account => account._id);
    const relationshipsByAccountId = await Relationship.aggregate([
        {$match: {account: {$in: accountsId}}},
        {
            $lookup: {
                from: 'accounts',
                localField: 'account',
                foreignField: '_id',
                as: 'account'
            }
        },
        {
            $group: {
                _id: '$account._id',
                count: {$sum: 1}
            }
        }]).exec();
    const byolEntitlements = await Account.accountByolEntitlementForExpiredTrial();
    const byolById = keyBy(byolEntitlements, '_id');
    relationshipsByAccountId.map(group => {
        const [accountId] = group._id;
        if (!objectsForMongo[accountId]) {
            const byolEntiteled = byolById[accountId] ? byolById[accountId].count : 0;
            const byolPerAccount = byolById[accountId] ? (group.count - byolEntiteled <= 0 ? group.count : byolEntiteled) : 0;
            const byol = {reservedRelationships: byolEntiteled, usedRelationships: byolPerAccount};
            objectsForMongo[accountId] = {
                account: accountId, timestamp: chargingTime, tier1: 0, byol
            };
        }
    });

    return objectsForMongo;
}


async function updateSubscriptionByAccountId(account, awsAccountId) {
    const {subscription} = account;
    // enable subscription distribution mechanism for AWS only
    // do not enable it for Azure
    if (subscription.provider === SAAS_SUBSCRIPTION_PROVIDER_AWS) {
        logger.info('Update subscription by tenancyAccountId function', JSON.stringify(account));
        const accounts = await findAccountsByAwsAccountId(awsAccountId);
        // in case current account has subscribed subscription
        if (subscription.status === SAAS_REGISTRATION_STATUS_SUBSCRIBE_SUCCESS) {
            return updateSubscription(subscription, accounts);
        }
        // in case there is some other account with subscribe subscription
        const otherAccountSubscription = find(accounts, {
            'subscription.status': SAAS_REGISTRATION_STATUS_SUBSCRIBE_SUCCESS,
            'subscription.type': PAYGO_SUBSCRIPTION,
            'subscription.provider': SAAS_SUBSCRIPTION_PROVIDER_AWS
        });
        return updateSubscription(otherAccountSubscription, account);
    }
}

agenda.define(AGENDA_TASK_NAME_UPDATE_CUSTOMER_IDENTIFIER_ON_AWS_ACCOUNTS, {lockLifetime: 10000},  async(job, done) => {
    try {
        const brokersWithAwsAccountId = await DataBroker.aggregate([{$match: {'awsAccountId': {$exists: true}}},
            {
                $lookup: {
                    from: 'accounts',
                    localField: 'account',
                    foreignField: '_id',
                    as: 'accountData'
                }
            },
            {$unwind: '$accountData'},
            {$project: {_id:0, dataBroker: '$_id', account: '$accountData', awsAccountId:'$awsAccountId'}}]).exec();
        await Promise.all(brokersWithAwsAccountId.map(broker => updateSubscriptionByAccountId(broker.account, broker.awsAccountId)));
    } catch (err) {
        logger.error('Failed to run update subscription id agenda', err);
    } finally {
        done();
    }
});


agenda.define(AGENDA_TASK_NAME_AWS_SAAS_METERING, {lockLifetime: 10000}, async(job, done) => {
    try {
        const chargingTime = new Date().toISOString();
        logger.info('Scheduled MarketPlace Metering Billing');
        let mongoObjects = await billSaaSAccounts(chargingTime);
        mongoObjects = await byolDetailsForUnsubscribedAccounts(mongoObjects, chargingTime);
        mongoObjects = await addAccountsWithoutRelationshipsToMongoObject(mongoObjects, chargingTime);
        await filterAndInsert(mongoObjects);
    } catch (err) {
        logger.error('Failed to report metering:', err);
    } finally {
        done();
    }
});

module.exports = {
    byolDetailsForUnsubscribedAccounts,
    batchMeterUsage,
    addAccountsWithoutRelationshipsToMongoObject,
    resolveCustomerIdentifierByMpCustomerToken,
    getCustomerTiers,
    filterAndInsert,
    billSaaSAccounts
};
