const { BadRequest } = require('../../utils/errors');
const { S3_PROVIDER, ALIBABA_PROVIDER, DFIO_METADATA_BUCKET_NAME } = require('../../utils/consts');

const S3_DEFAULT_REGION = 'us-east-1';
const ALIBABA_DEFAULT_REGION = 'oss-me-east-1';

function getS3(region = S3_DEFAULT_REGION) {
    return new AWS.S3({ region });
}

function getAlibaba(accessKey, secretKey, region = ALIBABA_DEFAULT_REGION) {
    return new AWS.S3({ endpoint:'oss-me-east-1.aliyuncs.com',
        accessKey, secretKey, region });
}

async function putObject(region, bucketName, key, file, grantType = false, acl = false) {
    const s3 = getS3(region);
    const params = {
        Bucket: bucketName,
        Key: key,
        Body: file
    };

    if (acl) {
        //public read for cloud formation ( because cloudFormation must use S3 url and so the object must have public-read
        params.ACL = 'public-read';
    } else if (grantType) {
        //relevant for developers users only - when you run from the DEV/PROD/E2E its ok
        params.GrantFullControl = 'id=af5e96df7b003f697242aff6d868db9c03ba8b5dd84633cd1d2402cdd1d151fb'; //bucket owner
        params.GrantRead = 'id=3c8aca0fd3f3dd5087bc099ad1ccf909b641ae780760aae079696e021f485faa44249f50b0c7e220971714247c93b0fc';
    }
    return s3.putObject(params).promise();
}

async function listObjectsAsync(region, bucketName, prefix) {
    const s3 = getS3(region);

    const data = await s3.listObjectsV2({
        Bucket: bucketName,
        Prefix: prefix
    }).promise();
    return data.Contents.map(content => ({
        key: content.Key,
        lastModified: new Date(content.LastModified).getTime(),
        size: content.Size
    }));
}

async function getObjectTaggingAsync(region, bucket, key) {
    const s3 = getS3(region);

    const data = await s3.getObjectTagging({
        Bucket: bucket,
        Key: key
    }).promise();
    return data.TagSet.reduce((accumulator, currentValue) => {
        accumulator[currentValue.Key] = currentValue.Value;
        return accumulator;
    }, {});
}

function getS3ObjectUrl(region, bucket, key, shouldReplace, isBundle) {
    const prefix = region === 'us-east-1' ? 's3' : `s3-${region}`;
    const s3Url = `https://${prefix}.amazonaws.com/${bucket}/${encodeURIComponent(key)}`;
    if (shouldReplace) {
        let regex;
        let aliasName;
        if (isBundle) {
            regex = new RegExp(`https:\/\/.*.\/${DFIO_METADATA_BUCKET_NAME}\/repo\/`, 'g');
            aliasName = 'repo';
        } else {
            regex = new RegExp(`https:\/\/.*.\/${DFIO_METADATA_BUCKET_NAME}\/.*\/`, 'g');
            aliasName = 'cf';
        }
        return decodeURIComponent(s3Url).replace(regex, `https://${aliasName}.cloudsync.netapp.com/`);
    }
    return s3Url;
}

async function headBucket(s3, bucket) {
    return new Promise((resolve, reject) => {
        s3.headBucket({Bucket: bucket})
            .on('complete', res => {
                if (res.httpResponse.headers['x-amz-bucket-region']) {
                    resolve(res.httpResponse.headers['x-amz-bucket-region']);
                } else {
                    const statusCode = res.httpResponse.statusCode;
                    reject(new BadRequest(`Got unexpected status code ${statusCode} - for headBucket operation on bucket ${bucket}`));
                }
            })
            .send();
    });
}

function getRegion(provider, bucket, options) {
    switch (provider) {
        case S3_PROVIDER: {
            const s3 = getS3('us-east-1');
            return headBucket(s3, bucket);
        }
        case ALIBABA_PROVIDER: {
            const { accessKey, secretKey} = options;
            const alibaba = getAlibaba(accessKey, secretKey);
            return headBucket(alibaba, bucket);
        }
    }
}

function getPreSignedUrl(bucket, key/*, expiredInSeconds = 60 * 60*/) {
    const s3 = getS3();
    return s3.getSignedUrl('getObject', {
        Bucket: bucket,
        Key: key,
        // Expires: expiredInSeconds //time to expire in seconds
    });
}

async function headObject(bucket, key) {
    const s3 = getS3();
    return s3.headObject({Bucket: bucket, Key: key}).promise();
}

async function deleteObject(region, bucket, key) {
    const s3 = getS3(region);
    return s3.deleteObject({
        Bucket:bucket,
        Key: key
    }).promise();
}

module.exports = {
    putObject, listObjectsAsync, getObjectTaggingAsync, deleteObject,
    getPreSignedUrl, getS3ObjectUrl, headBucket, getRegion, headObject
};
