const https = require('https');
const {mapValues} = require('lodash');
const {SERVICE_REGION} = require('../../utils/consts');
const zlib = require('zlib');
const getLogger = require('../../utils/logger');

const logger = getLogger();

let sqs;

function getSQS() {
    if (!sqs) {
        sqs = new AWS.SQS({
            region: SERVICE_REGION,
            correctClockSkew: true,
            httpOptions: { agent: new https.Agent({ keepAlive: true }) }
        });
    }

    return sqs;
}


async function unzip(zippedString) {
    return new Promise((resolve, reject) => {
        const buffer = Buffer.from(zippedString, 'base64');
        zlib.unzip(buffer, (err, bufferUnzipped) => {
            if (err) {
                reject(err);
            } else {
                resolve(bufferUnzipped.toString());
            }
        });
    });
}

async function createQueue(name, attributes = {}) {
    logger.info(`Creating SQS queue ${name}`);
    const sqsClient = getSQS();

    const data = await sqsClient.createQueue({
        QueueName: name,
        Attributes: Object.assign({
            ReceiveMessageWaitTimeSeconds: '20',
            VisibilityTimeout: '60'
        }, attributes)
    }).promise();

    const queueUrl = data.QueueUrl;
    logger.info(`Queue ${queueUrl} successfully created`);

    return queueUrl;
}

async function sendMessage(queueUrl, message, messageAttributes = {}) {
    const sqsClient = getSQS();
    const MessageAttributes = mapValues(messageAttributes, value => ({StringValue: value, DataType: 'String'}));

    const data = await sqsClient.sendMessage({
        MessageBody: JSON.stringify(message),
        QueueUrl: queueUrl,
        MessageAttributes
    }).promise();

    return data.MessageId;
}

async function deleteMessage(queueUrl, receiptHandle) {
    logger.debug('Deleting message, receipt handle:', receiptHandle);
    const sqsClient = getSQS();

    return sqsClient.deleteMessage({
        QueueUrl: queueUrl,
        ReceiptHandle: receiptHandle
    }).promise();
}

async function receiveMessage(queueUrl) {
    const sqsClient = getSQS();

    const data = await sqsClient.receiveMessage({
        QueueUrl: queueUrl,
        AttributeNames:['All'],
        MessageAttributeNames: ['All'],
        MaxNumberOfMessages: 10,
        WaitTimeSeconds: 20
    }).promise();

    try {
        if (data && data.Messages) {
            //TODO delete unzip - after the broker will not send zipped messages
            return Promise.all(data.Messages.map(async message => {
                try {
                    return await {
                        receiptHandle: message.ReceiptHandle,
                        body: message?.MessageAttributes?.zipped ? await unzip(message.Body) : message.Body,
                        attributes: {...mapValues(message.MessageAttributes, value => value.StringValue), ...message.Attributes}
                    };
                } catch (err) {
                    logger.error('Failed to parse messages', err, 'deleting');
                    await deleteMessage(queueUrl, message.ReceiptHandle);
                }
            }));
        }

        return [];

    } catch (err) {
        logger.error('Failed to parse messages from ', queueUrl, err);
        data.Messages.forEach(message => deleteMessage(queueUrl, message.ReceiptHandle));
        return [];
    }
}


async function deleteQueue(queueUrl) {
    const sqsClient = getSQS();
    return sqsClient.deleteQueue({QueueUrl: queueUrl}).promise();
}

async function getQueueArn(queueUrl) {
    logger.debug(`Getting queue: ${queueUrl} arn`);
    const sqsClient = getSQS();

    const data = await sqsClient.getQueueAttributes({
        QueueUrl: queueUrl,
        AttributeNames: ['QueueArn']
    }).promise();

    return data.Attributes.QueueArn;
}

async function setQueueAttribute(queueUrl, attributesObject) {
    const sqsClient = getSQS();

    return sqsClient.setQueueAttributes({
        QueueUrl: queueUrl,
        Attributes: attributesObject
    }).promise();
}

async function setQueuePolicy(queueUrl, policy) {
    return setQueueAttribute(queueUrl, {
        Policy: JSON.stringify(policy)
    });
}

async function getQueueAttributes(queueUrl, attributeNames) {
    const sqsClient = getSQS();

    try {
        const response = await sqsClient.getQueueAttributes({
            QueueUrl: queueUrl,
            AttributeNames: attributeNames
        }).promise();
        return response.Attributes;
    } catch (err) {
        if (err.code === 'AWS.SimpleQueueService.NonExistentQueue') {
            logger.warn(`Queue ${queueUrl} does not exist deleting queue`);
        } else {
            logger.error(`getQueueAttributes, got error for queue ${queueUrl}`, err);
            throw err;
        }
    }
}

async function allowArnSendMessage(queueUrl, queueArn, arn) {
    const policy = {
        Version: '2012-10-17',
        Id: 'arn-send-message',
        Statement: [
            {
                Sid: `Sid${Date.now()}`,
                Effect: 'Allow',
                Principal: '*',
                Action: 'SQS:SendMessage',
                Resource: queueArn,
                Condition: {
                    ArnEquals: {
                        'aws:SourceArn': arn
                    }
                }
            }
        ]
    };

    return setQueuePolicy(queueUrl, policy);
}


async function getQueueUrl(queueName) {
    const sqs = getSQS();
    const data = await sqs.getQueueUrl({
        QueueName:queueName
    }).promise();
    return data.QueueUrl;
}

module.exports = {
    createQueue,
    getQueueUrl,
    getQueueAttributes,
    setQueuePolicy,
    getQueueArn,
    deleteQueue,
    allowArnSendMessage,
    receiveMessage,
    deleteMessage,
    sendMessage
};
