const querystring = require('querystring');
const axios = require('axios');
const { pick } = require('lodash');
const getLogger = require('../utils/logger');

const logger = getLogger();

const CLIENT_ID = '86e44c48-4964-4a95-97b8-be2c8a4d93f7';
const TENANT_ID = '8e21f23a-10b9-46fb-9d50-720ef604be98';
const CLIENT_SECRET = 'vMt1OOOVc@gk24FAGR_/@D6l_Ablbh@J';

const SAAS_FULFILLMENT_API_VERSION = '2018-08-31';

/**
 * Returns access token
 * @returns {Promise<String>} access token
 */
async function getAccessToken() {
    logger.info('Getting access token...');
    try {
        const { data: { access_token: accessToken } } = await axios.post(
            `https://login.microsoftonline.com/${TENANT_ID}/oauth2/token`,
            querystring.stringify({
                grant_type: 'client_credentials',
                client_id: CLIENT_ID,
                client_secret: CLIENT_SECRET,
                resource: 'a0e1e353-1a3e-42cf-a8ea-3a9746eec58c'
            }),
            { headers: { 'Content-Type': 'application/x-www-form-urlencoded' } }
        );
        logger.info('Access token:', accessToken);

        return `Bearer ${accessToken}`;
    } catch (err) {
        logger.error('Failed to get token', err);
        throw err;
    }
}

/**
 * Resolves short lived marketplace token (received during redirect from the MP) to subscription id and plan id
 * @param marketplaceToken short lived marketplace token
 * @returns {Promise<{String, String}>} subscription id and plan id
 */
async function resolveMarketplaceToken(marketplaceToken) {
    logger.info('Resolving marketplace token:', marketplaceToken);
    const accessToken = await getAccessToken();

    try {
        const {
            data: {
                id, subscriptionName, offerId, planId, quantity
            }
        } = await axios.post(`https://marketplaceapi.microsoft.com/api/saas/subscriptions/resolve?api-version=${SAAS_FULFILLMENT_API_VERSION}`, {}, {
            headers: {
                authorization: accessToken,
                'x-ms-marketplace-token': marketplaceToken
            }
        });
        logger.info('Marketplace token resolved:', {
            id, subscriptionName, offerId, planId, quantity
        });

        return { id, planId };
    } catch (err) {
        logger.error('Failed to resolve marketplace token', err);
        throw err;
    }
}

/**
 * Activates subscription
 * @param subscriptionId
 * @param planId
 * @returns {Promise<void>}
 */
async function activateSubscription(subscriptionId, planId) {
    logger.info('Activating subscription:', { subscriptionId, planId });
    const accessToken = await getAccessToken();

    try {
        await axios.post(`https://marketplaceapi.microsoft.com/api/saas/subscriptions/${subscriptionId}/activate?api-version=${SAAS_FULFILLMENT_API_VERSION}`, { planId }, {
            headers: {
                authorization: accessToken
            }
        });
        logger.info('Subscription successfully activated');
    } catch (err) {
        logger.error('Failed to activate subscription', err);
        throw err;
    }
}

async function usageEvent(event) {
    logger.info('Sending usage event:', event);
    const accessToken = await getAccessToken();

    try {
        const { data } = await axios.post(`https://marketplaceapi.microsoft.com/api/usageEvent?api-version=${SAAS_FULFILLMENT_API_VERSION}`, event, {
            headers: {
                authorization: accessToken
            }
        });

        logger.info('Usage event successfully sent', data);
    } catch (err) {
        logger.error('Failed to send usage event', { status: err.response.status, statusText: err.response.statusText });
        // TODO-SK insert into SaaSBilling
    }
}

async function batchUsageEvent(batchEvents) {
    logger.info('Sending batch usage event:', batchEvents);
    const accessToken = await getAccessToken();

    try {
        const { data: { result } } = await axios.post(`https://marketplaceapi.microsoft.com/api/batchUsageEvent?api-version=${SAAS_FULFILLMENT_API_VERSION}`, {
            request: batchEvents
        }, {
            headers: {
                authorization: accessToken
            }
        });

        logger.info('Batch usage event successfully sent', JSON.stringify(result));

        const promises = result.map(event => {
            if (event.status === 'Accepted') {
                // TODO-SK insert into SaaSBilling
                logger.info('Accepted event:', event);
            } else {
                logger.warn('Got unaccepted usage event, retrying:', JSON.stringify(event));
                return usageEvent(pick(event, 'resourceId', 'quantity', 'dimension', 'effectiveStartTime', 'planId'));
            }
        });
        return Promise.all(promises);
    } catch (err) {
        logger.error('Failed to send batch usage event', err);
        throw err;
    }
}

/**
 * Returns subscription details
 * @param subscriptionId
 * @returns {Promise<{saasSubscriptionStatus, name, planId, termUnit}>}
 */
async function getSubscriptionDetails(subscriptionId) {
    logger.info('Getting subscription details, subscription id:', subscriptionId);

    const accessToken = await getAccessToken();
    try {
        const {
            data: {
                planId, saasSubscriptionStatus, name, term: {
                    termUnit
                }
            }
        } = await axios.get(`https://marketplaceapi.microsoft.com/api/saas/subscriptions/${subscriptionId}?api-version=${SAAS_FULFILLMENT_API_VERSION}`, {
            headers: {
                authorization: accessToken
            }
        });
        logger.info('Subscription plan:', {
            planId, saasSubscriptionStatus, name, termUnit
        });
        return {
            planId, saasSubscriptionStatus, name, termUnit
        };
    } catch (err) {
        logger.error('Failed to get subscription details', err);
        throw err;
    }
}

module.exports = {
    resolveMarketplaceToken, activateSubscription, getSubscriptionDetails, batchUsageEvent
};
