const {
    PRODUCTION_QSTACK_CLIENT_IDS, STAGING_QSTACK_CLIENT_IDS,
    API_PATH_MARKETPLACE_REGISTRATION_AWS, JWT_TOKEN_API_AUTH0_AUDIENCE,
    JWT_TOKEN_CLOUD_MANAGER_AUTH0_AUDIENCE, API_PATH_HEALTH_CHECK, API_PATH_MARKETPLACE_AZURE_LANDING,
    API_PATH_MARKETPLACE_AZURE_EVENTS
} = require('../utils/consts');
const {Unauthorized, Forbidden} = require('../utils/errors');
const jwtVerifyAsync = require('../lib/jwt');
const {isEmptyString} = require('../utils/utils');
const {isProduction} = require('../utils/environment');
const getLogger = require('../utils/logger');

const logger = getLogger();

const JWT_TOKEN_AUTH0_PAYLOAD_FIELD_USER_ID = 'sub';
const JWT_TOKEN_AUTH0_PAYLOAD_FIELD_USER_NAME = 'http://cloud.netapp.com/full_name';

const SKIP_AUTHENTICATION_REQUESTS_PATH = [
    API_PATH_MARKETPLACE_REGISTRATION_AWS,
    API_PATH_HEALTH_CHECK,
    API_PATH_MARKETPLACE_AZURE_LANDING,
    API_PATH_MARKETPLACE_AZURE_EVENTS
];

function findClientType(info) {
    const {sub} = info;
    const [clientId] = sub.split('@');
    const approvedClientsIds = isProduction() ? PRODUCTION_QSTACK_CLIENT_IDS : STAGING_QSTACK_CLIENT_IDS;
    if (approvedClientsIds.includes(clientId)) {
        return 'qstack';
    }
    if (info.scope && info.scope.includes('occm-access')) {
        return 'occm';
    }
    throw new Forbidden('Unauthorized client');
}

async function loadClientToken(payload, req, authorization, onBehalf, client) {
    const portalIdFromToken = payload[JWT_TOKEN_AUTH0_PAYLOAD_FIELD_USER_ID];
    const userName = payload[JWT_TOKEN_AUTH0_PAYLOAD_FIELD_USER_NAME];
    if (isEmptyString(portalIdFromToken)) {
        logger.error('Missing user ID in client JWT token payload');
        throw new Unauthorized();
    }

    req.isQstack = (client === 'qstack');
    req.isCvo = (client === 'occm');
    req.auth0Id = portalIdFromToken;
    req.userName = userName;
}

async function loadRequestData(req) {
    // 86ba39f2-ed8d-4044-bd3c-7beb630cb8d5 - allows us to invoke REST API on behalf of NetApp
    if (!SKIP_AUTHENTICATION_REQUESTS_PATH.includes(req.path) && req.get('x-netapp-internal') !== '86ba39f2-ed8d-4044-bd3c-7beb630cb8d5') {
        // 2 cases:
        // #1 - Regular client - got single authorization header 'authorization', part of the payload is:
        //      "audience": "https://api.cloud.netapp.com"
        //      "sub": "auth0|user_id"
        // #2 - OCCM - got two authorization headers: 'authorization' and 'x-netapp-on-behalf'
        //      authorization:
        //      "audience": "https://api.cloud.netapp.com"
        //      "sub": "xxxxxxxxxxx|client"
        //
        //      on behalf:
        //      "audience": "https://ontap.cloud.netapp.com"
        //      "sub": "auth0:user_id"

        const authorization = req.get('Authorization') || req.get('authorization');
        logger.debug('Authorization header:', authorization);

        const onBehalf = req.get('x-netapp-on-behalf');
        logger.debug('On behalf header:', onBehalf);
        try {
            let payload = await jwtVerifyAsync(authorization, JWT_TOKEN_API_AUTH0_AUDIENCE);
            const client = onBehalf ? findClientType(payload) : undefined;
            if (client === 'occm') {
                payload = await jwtVerifyAsync(onBehalf, JWT_TOKEN_CLOUD_MANAGER_AUTH0_AUDIENCE);
            } else if (client === 'qstack') {
                payload = await jwtVerifyAsync(onBehalf, JWT_TOKEN_API_AUTH0_AUDIENCE);
            }
            logger.debug('Successfully verified JWT token. Payload:', payload);
            return loadClientToken(payload, req, authorization, onBehalf, client);
        } catch (error) {
            logger.debug(`Failed authenticating user with JWT token - authorization ${authorization}, x-netapp-on-behalf ${onBehalf}:`, error);
            throw new Unauthorized(error);
        }
    } else {
        logger.info('No authentication needed');
    }
}

async function authenticate(req, res, next) {
    logger.debug('Authenticating request to:', req.originalUrl);
    try {
        await loadRequestData(req);
        logger.debug('Authentication completed');
        next();
    } catch (err) {
        logger.error('Authentication failed with error:', err.message);
        next(err);
    }
}

module.exports = authenticate;
