const {Account} = require('../models/account');
const {
    API_PATH_MARKETPLACE_REGISTRATION_AWS,
    API_PATH_HEALTH_CHECK,
    API_PATH_MARKETPLACE_AZURE_LANDING,
    API_PATH_MARKETPLACE_AZURE_EVENTS, TRIAL_EXPIRATION_MS,
    API_PATH_ACCOUNTS, MONGO_DB_ERROR_DUPLICATE_KEY
} = require('../utils/consts');
const {BadRequest} = require('../utils/errors');
const {generateSerialNumber} = require('../operations/ipa-opertions');
const {findSubscription} = require('../operations/account-operations');
const getUserEmailByJwtToken = require('../operations/auth0-operations');
const {getAccounts} = require('../lib/tenancy');
const {setAccountId} = require('../utils/async-local-storage');
const getLogger = require('../utils/logger');
const logger = getLogger();


const SKIP_FIND_ACCOUNTS_REQUESTS_PATH = [
    API_PATH_MARKETPLACE_REGISTRATION_AWS,
    API_PATH_HEALTH_CHECK,
    API_PATH_MARKETPLACE_AZURE_LANDING,
    API_PATH_MARKETPLACE_AZURE_EVENTS,
    API_PATH_ACCOUNTS
];

async function shouldAutomaticExtend() {
    try {
        const email = await getUserEmailByJwtToken();
        logger.info(`New account in the DB for user ${email} `);
        return ['cpoccloud.org', 'netapp.com', 'unitek.com', 'fastlaneus.com', 'ihcnetappu.com', 'trainocate.in', 'netapp-test.com'].some(domain => (email).includes(domain));
    } catch (err) {
        logger.error('Received error while trying to get userInfo ', err);
        // this is an unknown email - it will not get automatic extension
        return false;
    }
}

async function handleExistingTenancyAccount(req, existingTenancyAccount, tenancyAccounts) {
    const {accountName: name, accountPublicId: tenancyAccountId} = existingTenancyAccount;
    logger.debug(`AccountId: ${tenancyAccountId} found in tenancy`);
    try {
        const account = await Account.findOne({tenancyAccountId}).exec();
        const lastLogin = new Date();
        const automaticExtension = account ? account.automaticExtension : await shouldAutomaticExtend();
        const subscribedSubscription = await findSubscription(tenancyAccountId, tenancyAccounts);
        const subscription = subscribedSubscription  ? subscribedSubscription : (account ? account.subscription : {});
        const trialExpiry = account ? (automaticExtension ? new Date(Date.now() + TRIAL_EXPIRATION_MS) : account.trialExpiry) : new Date(Date.now() + TRIAL_EXPIRATION_MS);
        try {
            req.account = await Account.findOneAndUpdate({tenancyAccountId}, {
                $setOnInsert: {
                    name,
                    'support.serialNumber': generateSerialNumber(),
                    automaticExtension,
                }, $set: {
                    lastLogin, subscription, trialExpiry
                }
            }, {upsert: true, new: true});
        } catch (err) {
            if (err.code === MONGO_DB_ERROR_DUPLICATE_KEY) {
                logger.info('Account is already exist on DB');
                req.account = await Account.findOne({tenancyAccountId}).exec();
            }
        }
    } catch (err) {
        logger.error('Error handling existing tenancy account', tenancyAccountId, err);
        throw err;
    }
}

async function findAccountAndUpdateDb(req) {
    const tenancyAccountId = req.headers['x-account-id'];

    if (!SKIP_FIND_ACCOUNTS_REQUESTS_PATH.includes(req.path)
        || req.method === 'DELETE' && req.path === API_PATH_ACCOUNTS) {

        if (req.path.includes('/accounts/info') && req.query.noCreate === 'true') {
            req.account = null;
        } else {
            const tenancyAccounts = await getAccounts();
            if (!tenancyAccountId) {
                logger.debug('x-account-id header is missing. Looking for unique tenancy account');
                if (tenancyAccounts.length === 0) {
                    throw new BadRequest('No tenancy accounts exist');
                } else if (tenancyAccounts.length === 1) {
                    const [existingTenancyAccount] = tenancyAccounts;
                    logger.debug(`Found specific tenancy accountId: ${existingTenancyAccount.accountPublicId}`);
                    await handleExistingTenancyAccount(req, existingTenancyAccount, tenancyAccounts);
                } else if (tenancyAccounts.length > 1) {
                    throw new BadRequest('Multiple tenancy accounts. Could not select specific tenancy account');
                }
            } else {
                const existingTenancyAccount = tenancyAccounts.find(tenancyAccount => tenancyAccount.accountPublicId === tenancyAccountId);
                if (existingTenancyAccount) {
                    await handleExistingTenancyAccount(req, existingTenancyAccount, tenancyAccounts);
                } else {
                    logger.error(`AccountId: ${tenancyAccountId} not found in tenancy`);
                    throw new BadRequest(`Account with id ${tenancyAccountId} is not found on tenancy for current user`);
                }
            }
        }
    }
}

async function findAccount(req, res, next) {
    logger.debug('Find account request to:', req.originalUrl);

    try {
        await findAccountAndUpdateDb(req);
        logger.debug('account found');
        setAccountId(req.account ? req.account.tenancyAccountId : 'anonymous');
        next();
    } catch (err) {
        logger.debug('Failed to find account with error:', err.message);
        next(err);
    }
}


module.exports = findAccount;
