const mongoose = require('mongoose');
const moment = require('moment');
const {
    SAAS_REGISTRATION_STATUS_UNSUBSCRIBED_INIT,
    PAYGO_SUBSCRIPTION,
    CONTRACT_SUBSCRIPTION, SAAS_SUBSCRIPTION_PROVIDER_AWS,
    SAAS_SUBSCRIPTION_PROVIDER_AZURE, IPA_PENDING_STATUS, IPA_REGISTERED_STATUS, IPA_UNREGISTERED_STATUS
} = require('../utils/consts');

const licensesSchema = new mongoose.Schema({
    serialNumber: {type: String, required: true},
    reservedRelationships: {type: Number, required: false},
    expirationDate: {type: Date, required: false}
}, {_id: false});

const subscriptionSchema = new mongoose.Schema({
    customerIdentifier: {type: String, required: false},
    status: {type: String, required: false, default: SAAS_REGISTRATION_STATUS_UNSUBSCRIBED_INIT},
    updatedAt: {type: Date, required: false, default: new Date()},
    type: {
        type: String, enum: [PAYGO_SUBSCRIPTION, CONTRACT_SUBSCRIPTION], required: false, default: PAYGO_SUBSCRIPTION
    },
    provider: {
        type: String,
        enum: [SAAS_SUBSCRIPTION_PROVIDER_AWS, SAAS_SUBSCRIPTION_PROVIDER_AZURE],
        required: false,
        default: SAAS_SUBSCRIPTION_PROVIDER_AWS
    }
}, {_id: false});

const nssCredentialsSchema = new mongoose.Schema({
    nssUsername: {type: String, required: false},
    nssPassword: {type: String, required: false}
}, {_id: false});

const supportSchema = new mongoose.Schema({
    serialNumber: {type: String, required: true},
    status: {
        type: String,
        enum: [IPA_PENDING_STATUS, IPA_REGISTERED_STATUS, IPA_UNREGISTERED_STATUS],
        required: true,
        default: IPA_UNREGISTERED_STATUS
    }
}, {_id: false});

const accountSchema = new mongoose.Schema({
    tenancyAccountId: {type: String, required: true, index: {unique: true}},
    name: {type: String, required: true},
    automaticExtension: {type: Boolean, required: true, default: false},
    trialExpiry: {type: Date, required: true},
    lastLogin: {type: Date, required: false},
    byol: {
        licenses: [licensesSchema]
    },
    nssCredentials: {type: nssCredentialsSchema, required: false},
    support: {type: supportSchema, required: true},
    subscription: {type: subscriptionSchema, required: true, default: {}}
}, {
    timestamps: true
});

accountSchema.statics.accountByolEntitlementForExpiredTrial = function() {
    return this.aggregate([
        {$match: {trialExpiry: {$lt: new Date()}}},
        {$unwind: '$byol.licenses'},
        {$match: {'byol.licenses.expirationDate': {$gt: new Date()}}},
        {$group: {_id: '$_id', count: {$sum: '$byol.licenses.reservedRelationships'}}}]);
};


accountSchema.statics.customerIdentifierByolEntitlement = function() {
    return this.aggregate([
        {$match: {$and: [{'subscription.customerIdentifier': {$exists: true}}, {'subscription.type': 'PAYGO'}]}},
        {$unwind: '$byol.licenses'},
        {$match: {'byol.licenses.expirationDate': {$gt: new Date()}}},
        {$group: {_id: '$subscription.customerIdentifier', count: {$sum: '$byol.licenses.reservedRelationships'}}}]);
};


accountSchema.methods.reservedRelationships = function() {
    let reservedRelationships = 0;

    if (this.byol.licenses.length > 0) {
        this.byol.licenses.map(license => {
            if (moment(license.expirationDate).startOf('day').isAfter()) {
                reservedRelationships += license.reservedRelationships;
            }
        });
    }

    return reservedRelationships;
};


const Account = mongoose.model('Account', accountSchema);


async function findBillableAccounts() {
    return Account.find({
        'subscription.customerIdentifier': {$exists: true},
        'subscription.type': PAYGO_SUBSCRIPTION,
        'subscription.provider': SAAS_SUBSCRIPTION_PROVIDER_AWS,
        trialExpiry: {$lt: new Date()}
    }).exec();
}


module.exports = {Account, findBillableAccounts};
