const config = require('config');
const mongoose = require('mongoose');
const ms = require('ms');
const {
    DATA_BROKER_CF_FOLDER,
    DFIO_METADATA_BUCKET_NAME,
    DFIO_METADATA_BUCKET_REGION,
    AZURE_LOCATIONS, AWS_REGIONS
} = require('../utils/consts');
const {getS3ObjectUrl} = require('../lib/aws/s3');

const MESSAGE_DATA_BROKER_OFFLINE = 'Data Broker is offline';
const LIVENESS_GRACE_FACTOR = 1.1;
const DATA_BROKER_PHOENIX_STATUS_INTERVAL_MS = ms(`${config.get('queues.data-broker.phoenix-status-interval-seconds')}s`);
const DATA_BROKER_PHOENIX_LIVENESS_TIMEOUT_MS = LIVENESS_GRACE_FACTOR * DATA_BROKER_PHOENIX_STATUS_INTERVAL_MS;


function createOnPremPlacement(doc) {
    return {
        hostname: doc.hostname,
        platform: doc.platform,
        privateIp: doc.privateIp,
        version: doc.version,
        os: doc.os,
        release: doc.release,
        osTotalMem: doc.osTotalMem,
        node: doc.node,
        cpus: doc.cpus,
        processMaxMem: doc.processMaxMem,
        awsAccountId: doc.awsAccountId, // ONPREM have an awsAccountId - the user entered awsAccess keys
        gcpProjectId: doc.gcpProjectId, // ONPREM have an gcpProjectId - the user entered awsAccess keys
        publicIp: doc.publicIp
    };
}

function createCloudPlacement(provider, doc) {
    return {
        instanceId: doc.instanceId,
        availabilityZone: doc.availabilityZone,
        vpcId: doc.vpcId,
        region: ((provider === 'AZURE') ? AZURE_LOCATIONS[doc.region] : AWS_REGIONS[doc.region]) ||  doc.region,
        resourceGroupName:(provider === 'AZURE') ? doc.resourceGroupName : undefined,
        vpcCidrBlock: doc.vpcCidrBlock,
        subnetId: doc.subnetId,
        zone :doc.zone,
        serviceAccount: doc.serviceAccount,
        network:doc.network,
        subnetCidrBlock: doc.subnetCidrBlock,
    };
}


function createPlacement(dataBrokerType, doc) {
    const placement = createOnPremPlacement(doc);
    if (['AWS', 'AZURE', 'GCP'].includes(dataBrokerType)) {
        const cloudPlacement = createCloudPlacement(dataBrokerType, doc);
        Object.assign(placement, cloudPlacement);
    }
    return placement;
}

function findResource(dataBrokerType, cftKey, installerKey) {
    if (['AWS', 'AZURE'].includes(dataBrokerType)) {
        return cftKey;
    } else if (dataBrokerType === 'GCP') {
        return `${cftKey}.jinja`;
    } else {
        //onprem
        return installerKey;
    }
}

const encryptionSchema = new mongoose.Schema({
    port: {
        type: Number, min: 0, max: 65535, required: false
    },
    certificate: {type: String, required: false}
}, {_id: false});

const dataBrokerSchema = new mongoose.Schema({
    account: {
        type: mongoose.Schema.Types.ObjectId, ref: 'Account', index: true
    },
    name: {type: String, required: true},
    instanceId: {type: String, required: false},
    privateIp: {type: String, required: false},
    publicIp: {type: String, required: false},
    transferRate: {type: Number, required: false},
    region: {type: String, required: false},
    availabilityZone: {type: String, required: false},
    vpcId: {type: String, required: false},
    vpcCidrBlock: {type: String, required: false},
    subnetId: {type: String, required: false},
    subnetCidrBlock: {type: String, required: false},
    version: {type: String, required: false},
    commandsQueueUrl: {type: String, required: false},
    statusesQueueUrl: {type: String, required: false},
    awsAccountId: {type: String, required: false},
    network:{type:String, required:false},
    zone:{type:String, required:false},
    serviceAccount:{type:String, required:false},
    proxy: {type: String, required: false},
    updateNewVersion: {type: Boolean, required: false},
    groupId: {type: mongoose.Schema.Types.ObjectId, ref: 'Group', required: false},
    lastPing: {type: Date, required: false},
    type: {
        type: String, enum: ['AWS', 'ONPREM', 'AZURE', 'GCP'], required: true, default: 'AWS'
    },
    userId: {type:String, required:false},
    encryption: {type: encryptionSchema, required: false},
    hostname: {type: String, required: false},
    platform: {type: String, required: false},
    os: {type: String, required: false},
    release: {type: String, required: false},
    osTotalMem: {type: String, required: false},
    node: {type: String, required: false},
    cpus: {type: String, required: false},
    processMaxMem: {type: String, required: false},
    versionUpdateSent: {type: Date},
    gcpProjectId: {type: String, required: false},
    resourceGroupName: {type:String, required:false},
    externalId : {type:String, required:false}
}, {
    strict:false,
    timestamps: true,
    toJSON: {
        transform: (doc, ret) => {
            ret.createdAt = ret.createdAt.getTime();
            delete ret.updatedAt;
            delete ret.__v;
            delete ret.account;
            delete ret._id;
            delete ret.versionUpdateSent;
            ret.id = doc.id;
            if (doc.lastPing) {
                ret.lastPing = doc.lastPing ? doc.lastPing.getTime() : 0;
            }
            if (['AWS', 'AZURE', 'GCP'].includes(ret.type) && ret.instanceId) {
                ret.placement = createPlacement(ret.type, ret);
            } else if ((['ONPREM'].includes(ret.type)) && ret.hostname) {
                ret.placement = createPlacement(ret.type, ret);
            }
            if (ret.encryption) {
                ret.port = ret.encryption.port;
            }
            if (doc.version) {
                ret.status = doc.isAlive() ? 'COMPLETE' : 'FAILED';
                if (ret.status !== 'COMPLETE') {
                    ret.message = MESSAGE_DATA_BROKER_OFFLINE;
                }
            }

            delete ret.instanceId;
            delete ret.privateIp;
            delete ret.publicIp;
            delete ret.region;
            delete ret.availabilityZone;
            delete ret.vpcId;
            delete ret.vpcCidrBlock;
            delete ret.subnetId;
            delete ret.subnetCidrBlock;
            delete ret.version;
            delete ret.commandsQueueUrl;
            delete ret.statusesQueueUrl;
            delete ret.awsAccountId;
            delete ret.os;
            delete ret.release;
            delete ret.osTotalMem;
            delete ret.node;
            delete ret.cpus;
            delete ret.processMaxMem;
            delete ret.hostname;
            delete ret.platform;
            delete ret.encryption;
            delete ret.resourceGroupName;

            const cftKey = `${DATA_BROKER_CF_FOLDER}/${ret.externalId}`;
            const installerKey = `${cftKey}_installer`;
            const resource = findResource(ret.type, cftKey, installerKey);

            ret.fileLink = getS3ObjectUrl(DFIO_METADATA_BUCKET_REGION, DFIO_METADATA_BUCKET_NAME, resource,
                ret.type !== 'AWS', false);
            return ret;
        }
    }
});

dataBrokerSchema.index({account: 1, name: 1}, {unique: true});

dataBrokerSchema.methods.isAlive = function() {
    const diff = (this.lastPing ? this.lastPing.getTime() : 0) - Date.now();
    return this.lastPing && (Math.abs(diff) < DATA_BROKER_PHOENIX_LIVENESS_TIMEOUT_MS);
};


module.exports = mongoose.model('DataBroker', dataBrokerSchema);
