const mongoose = require('mongoose');
const { defaultTransform } = require('../utils/utils');
const {createMessage} = require('./message-v2');

const endPointSchema = new mongoose.Schema({
    protocol: { type: String, enum: ['nfs', 'cifs', 's3', 'sgws', 'local', 'azure', 'gcp'], required: true },
    nfs: { type: mongoose.Schema.Types.Mixed, required: false },
    cifs: { type: mongoose.Schema.Types.Mixed, required: false },
    s3: { type: mongoose.Schema.Types.Mixed, required: false },
    sgws: { type: mongoose.Schema.Types.Mixed, required: false },
    local: { type: mongoose.Schema.Types.Mixed, required: false },
    azure: { type: mongoose.Schema.Types.Mixed, required: false },
    gcp: { type: mongoose.Schema.Types.Mixed, required: false }
}, {
    _id: false,
    toJSON: {
        transform: defaultTransform((doc, ret) => ret)
    }
});

const databrokerTypeSchema = mongoose.Schema({
    dataBrokerId: {type: mongoose.Schema.Types.ObjectId, ref: 'DataBroker', required: true},
    endpointType: {type: String, enum: ['src', 'trg'], required: true}
}, {
    _id: false,
    toJSON: {
        transform: defaultTransform((doc, ret) => ret)
    }
});

const encryptionSchema = mongoose.Schema({
    initiator: {type: databrokerTypeSchema, required: true},
    listener: {type: databrokerTypeSchema, required: true}
}, {
    _id: false,
    toJSON: {
        transform: defaultTransform((doc, ret) => ret)
    }
});

const relationshipSchema = new mongoose.Schema({
    source: { type: endPointSchema, required: true },
    target: { type: endPointSchema, required: true },
    relationshipId: { type: mongoose.Schema.Types.ObjectId, required: true },
    encryption: { type: encryptionSchema, required: false },
    activity: { type: mongoose.Schema.Types.Mixed, requires: false },
    group: { type: mongoose.Schema.Types.ObjectId, required: false },
    dataBrokerId: { type: mongoose.Schema.Types.ObjectId, required: false },
    isQstack: { type: Boolean, required: true },
    isCvo: { type: Boolean, required: true, default: false },
    isCm: {type: Boolean, required: true, default: false},
    activateRelationshipAt : {type: Date, required: false}
}, {
    _id: false,
    toJSON: {
        transform: defaultTransform((doc, ret) => ret)
    }
});

const dataBrokerSchema = new mongoose.Schema({
    dataBrokerId: { type: mongoose.Schema.Types.ObjectId, required: true },
    name: { type: String, required: false },
    groupId: { type: mongoose.Schema.Types.ObjectId, required: false },
    type: { type: String, require: false }
}, {
    _id: false,
    toJSON: {
        transform: defaultTransform((doc, ret) => ret)
    }
});

const groupSchema = new mongoose.Schema({
    groupId: { type: mongoose.Schema.Types.ObjectId, required: true },
    name: { type: String, required: true },
    dataBrokers : [dataBrokerSchema]
}, {
    _id: false,
    toJSON: {
        transform: defaultTransform((doc, ret) => ret)
    }
});


const timelinev2Schema = new mongoose.Schema({
    account: {
        type: mongoose.Schema.Types.ObjectId, ref: 'Account', index: true
    },
    requestId: { type: String, required: true },
    auth0Id: {type:String, default:'system'},
    userName:{type:String, default:'system'},
    status: { type: String, required: true, default: 'PROGRESS' },
    summary: { type: String, required: true },
    relationship: { type: relationshipSchema, required: false },
    dataBroker: { type: dataBrokerSchema, required: false },
    group: { type: groupSchema, required: false },
    failureMessage: { type: String, required: false }
}, {
    timestamps: true,
    toJSON: {
        transform: (doc, ret) => {
            ret.createdAt = ret.createdAt.getTime();
            delete ret.updatedAt;
            delete ret.__v;
            delete ret.account;
            delete ret._id;
            ret.id = doc.id;
            return ret;
        }
    }
});

timelinev2Schema.index({ 'relationship.relationshipId': 1, status: 1, summary: 1 }, { sparse: true });

timelinev2Schema.post('save', doc => {
    const accountId = doc.account._id ? doc.account._id : doc.account;
    return createMessage(accountId, doc.requestId, undefined, {'timeline': doc.toJSON()});
});

module.exports = mongoose.model('TimelineV2', timelinev2Schema);
