const moment = require('moment');
const {createMessage} = require('../models/message-v2');
const Relationship = require('../models/relationship');
const {Account} = require('../models/account');
const {
    CONTRACT_SUBSCRIPTION, PAYGO_SUBSCRIPTION, SAAS_REGISTRATION_STATUS_SUBSCRIBE_SUCCESS,
    SAAS_REGISTRATION_STATUS_UNSUBSCRIBE_SUCCESS, SAAS_SUBSCRIPTION_PROVIDER_AZURE, SAAS_SUBSCRIPTION_PROVIDER_AWS
} = require('../utils/consts');
const {resolveCustomerIdentifierByMpCustomerToken} = require('../lib/aws/marketplace-aws');
const {resolveMarketplaceToken, activateSubscription, getSubscriptionDetails} = require('../lib/marketplace-azure');
const {setHubSpotContactFirstSubscriptionSinceIfNeeded} = require('../lib/hubspot');
const getLogger = require('../utils/logger');
const logger = getLogger();


async function resolveAwsMarketplaceToken(mpCustomerToken, account) {
    logger.info('Resolving AWS Mp token for Tenancy account Id: ', account.tenancyAccountId);
    // Assumptions:
    // 1. We can get MP token only once per AWS account
    // 2. This function is invoked only when MP token is provided
    // 3. Resolve the token for customer identifier - in case of success (customer identifier is received) : find out which type the SAAS is by the Product code
    // -put customer identifier status and type on every user with the same AWS account id
    // 4. Do not update other users if the resolving failed - for example - user can "still" AWS account id, provide wrong MP token and reset subscription status for other users
    return resolveCustomerIdentifierByMpCustomerToken(mpCustomerToken);
}

async function updateAccountsSubscription(accounts, subscriptionStatus, subscriptionType, customerIdentifier, subscriptionProvider) {
    logger.info('Updating account subscription: ', {
        subscriptionStatus,
        subscriptionType,
        customerIdentifier,
        subscriptionProvider,
        accounts
    });

    const tenancyAccountIds = accounts.map(account => account.tenancyAccountId);
    const updateQuery = {
        $set: {
            'subscription.updatedAt': new Date(),
            'subscription.status': subscriptionStatus,
            'subscription.type': subscriptionType,
            'subscription.provider': subscriptionProvider,
        }
    };

    if (subscriptionStatus === SAAS_REGISTRATION_STATUS_UNSUBSCRIBE_SUCCESS) {
        updateQuery.$unset = {'subscription.customerIdentifier' : ''};
        updateQuery.$set['subscription.type'] = PAYGO_SUBSCRIPTION;
    } else {
        updateQuery.$set['subscription.customerIdentifier'] = customerIdentifier;
    }

    if (subscriptionType === CONTRACT_SUBSCRIPTION) {
        updateQuery.$set.trialExpiry = new Date();
    }

    await Account.updateMany({tenancyAccountId: {$in: tenancyAccountIds}}, updateQuery);

    await Promise.all(accounts.map(async account => {
        if (subscriptionStatus === SAAS_REGISTRATION_STATUS_SUBSCRIBE_SUCCESS) {
            setHubSpotContactFirstSubscriptionSinceIfNeeded(account);
        }
        return createMessage(account.id, undefined, undefined, {'update-subscription-status':{
            accountId: account.tenancyAccountId,
            subscriptionStatus: subscriptionStatus,
            subscriptionUpdatedAt: new Date(),
            subscriptionType: subscriptionType,
            subscriptionProvider: subscriptionProvider
        }});
    }));
}

async function handleAwsAccountSubscriptionAndTrialExpiry(account, mpCustomerToken) {
    logger.info('Handle AWS account subscription for account: ', account.tenancyAccountId);

    if (mpCustomerToken && account.subscription.type === PAYGO_SUBSCRIPTION) {
        const triple = await resolveAwsMarketplaceToken(mpCustomerToken, account);

        const {customerIdentifier, status, type} = triple;
        logger.info(`Handle customerIdentifier - ${customerIdentifier} login. status: ${status}`);

        await updateAccountsSubscription([account], status, type, customerIdentifier, SAAS_SUBSCRIPTION_PROVIDER_AWS);
    }
}

async function handleAzureAccountSubscriptionAndTrialExpiry(account, mpCustomerToken) {
    logger.info('Handle Azure account subscription for account: ', account.tenancyAccountId);

    if (account.subscription.status !== SAAS_REGISTRATION_STATUS_SUBSCRIBE_SUCCESS) {
        // Azure sends token in 2 cases:
        // 1. account creates the subscription (aka subscribes to the offer)
        // 2. already subscribed account clicks on 'Manage Account'
        const {id: customerIdentifier, planId} = await resolveMarketplaceToken(mpCustomerToken);
        await activateSubscription(customerIdentifier, planId);
        const {termUnit} = await getSubscriptionDetails(customerIdentifier);

        // subscription term unit might be:
        // P1Y for yearly subscriptions
        // P1M for monthly subscriptions
        const subscriptionType = termUnit === 'P1Y' ? CONTRACT_SUBSCRIPTION : PAYGO_SUBSCRIPTION;

        await updateAccountsSubscription([account], SAAS_REGISTRATION_STATUS_SUBSCRIBE_SUCCESS, subscriptionType, customerIdentifier, SAAS_SUBSCRIPTION_PROVIDER_AZURE);
    }
}

async function countRelationshipsForAccount(accountId, target) {
    const query = (target) ? {
        account: accountId, 'source.protocol': {$in: ['nfs', 'cifs']}, $or: [{'target.nfs.provider': target},
            {'target.cifs.provider': target}]
    } : {account: accountId};
    return Relationship.countDocuments(query).exec();
}

async function cvsAndAnfRelationshipsCount(account) {
    const availableProviders = ['anf', 'cvs'];

    return Relationship.countDocuments({
        account, $or: [
            {'source.nfs.provider': {$in: availableProviders}},
            {'source.cifs.provider': {$in: availableProviders}},
            {'target.nfs.provider': {$in: availableProviders}},
            {'target.cifs.provider': {$in: availableProviders}}]
    }).exec();
}

async function fanOutAccountSubscription(tenancyAccountId, tenancyAccounts) {
    logger.info('Update subscription for all of the user accounts: ', tenancyAccountId);

    const currentAccount = await Account.findOne({tenancyAccountId});
    const accountsIds = tenancyAccounts.map(currentAccount => currentAccount.accountPublicId);
    // Update subscription for all of the user accounts

    const accountsToUpdate = await Account.find({
        tenancyAccountId: {
            $in: accountsIds
        }
    }).exec();

    await updateAccountsSubscription(accountsToUpdate, currentAccount.subscription.status, currentAccount.subscription.type, currentAccount.subscription.customerIdentifier, currentAccount.subscription.provider);
}

//check the number of relationships that are not cvs or anf
async function isFreeOfCharge(account) {
    logger.info('Checking the number of relationships that are not cvs or anf');

    const allRelationshipsNumber = await countRelationshipsForAccount(account);
    const cvsOrAnfRelationshipsNumber = await cvsAndAnfRelationshipsCount(account);

    return allRelationshipsNumber > 0 && allRelationshipsNumber - cvsOrAnfRelationshipsNumber === 0;
}

async function findSubscription(tenancyAccountId, tenancyAccounts) {
    logger.debug('Find if a user has account with subscription. Current tenancyAccountId: ', tenancyAccountId);
    const accountsIds = tenancyAccounts.map(account => account.accountPublicId);

    const account = await Account.findOne({
        tenancyAccountId: {
            $in: accountsIds
        },
        'subscription.status' : 'subscribe-success'
    }).exec();

    return account ? account.subscription : undefined;
}

function calculateDaysLeftForTrial(trialExpiry) {
    const daysLeftForTrial = moment(trialExpiry).startOf('day').diff(moment().startOf('day'), 'days');
    return (daysLeftForTrial >= 0) ? daysLeftForTrial : -1;
}

async function addTrialDaysForInternalUsers(account) {
    //should happen for NetApp users or one of approved domain
    logger.info('Adding trial days for internal users');
    const {trialExpiry, subscription: {type: subscriptionType}, automaticExtension} = account;

    if (subscriptionType === PAYGO_SUBSCRIPTION && automaticExtension) {
        const today = moment();

        if (moment(trialExpiry).diff(today, 'days') <= 8) {
            logger.info('Trial was expired, adding 14 days to expiry date');
            account.trialExpiry = today.add(14, 'day').toDate();
            await account.save();

            return account.trialExpiry;
        }
    }

    return trialExpiry;
}

async function findAccountInfo(account) {

    const {
        tenancyAccountId: accountId, name, support: {serialNumber: supportSerialNumber, status: supportRegistrationStatus},
        subscription: {status: subscriptionStatus, type: subscriptionType, provider: subscriptionProvider}
    } = account;

    const trialExpiry = await addTrialDaysForInternalUsers(account);
    const noCharge = await isFreeOfCharge(account);
    const daysLeftForTrial = calculateDaysLeftForTrial(trialExpiry);

    return {
        accountId,
        name,
        noCharge,
        daysLeftForTrial,
        subscriptionStatus,
        subscriptionType,
        subscriptionProvider,
        supportSerialNumber,
        supportRegistrationStatus
    };
}


module.exports = {
    addTrialDaysForInternalUsers,
    isFreeOfCharge,
    findAccountInfo,
    calculateDaysLeftForTrial,
    cvsAndAnfRelationshipsCount,
    countRelationshipsForAccount,
    handleAwsAccountSubscriptionAndTrialExpiry,
    handleAzureAccountSubscriptionAndTrialExpiry,
    fanOutAccountSubscription,
    findSubscription,
    updateAccountsSubscription
};