const config = require('config');
const {cloneDeep} = require('lodash');
const {auditGroupRecordSchema} = require('../schemas/audit-schema');
const {validateSchema} = require('../utils/schema-validation');
const {hideSecretsHelper} = require('../utils/utils');
const {AUDIT_ACCESS_KEY, AUDIT_SECRET_KEY} = require('../utils/consts');
const putRecord = require('../lib/aws/kinesis');
const getLogger = require('../utils/logger');
const logger = getLogger();

const AUDIT_STATUS = {
    PROGRESS : 'pending',
    RUNNING:'pending',
    DONE:'success',
    FAILED:'failed',
    ABORTED: 'aborted',
    PENDING : 'pending',
    DELETED: 'failed',

};

function extractResourceId(timeline) {
    if (timeline.relationship) {
        return timeline.relationship.relationshipId.toString();
    } else if (timeline.dataBroker) {
        return timeline.dataBroker.dataBrokerId.toString();
    }
}


async function sendAuditGroup(timelineObj, accountId, request, response) {
    if (!process.env.IS_SIMULATOR) {
        try {
            // logger.info('send audit group', JSON.stringify({timelineObj, accountId, response}));
            const {
                requestId, createdAt: startTime, summary: actionName, status: timeLineStatus,
                auth0Id: principalId, updatedAt: endTime, userName: referrer
            } = timelineObj;
            const status = AUDIT_STATUS[timeLineStatus];
            const resourceId = extractResourceId(timelineObj);
            const requestData = cloneDeep(request);
            const errors = timelineObj.failureMessage ? [timelineObj.failureMessage] : undefined;
            hideSecretsHelper(requestData);
            const auditGroup = {
                requestId,
                serviceName: config.get('audit.service-name'),
                startTime: new Date(startTime).getTime(),
                actionName,
                status,
                principalId,
                accountId,
                resourceId,
                referrer,
                requestData: JSON.stringify(requestData),
                responseData: JSON.stringify(response),
                endTime: new Date(endTime).getTime(),
                errors
            };
            validateSchema(auditGroup, auditGroupRecordSchema);
            // logger.info('Sending to kineses', JSON.stringify(auditGroup, null, 4));
            await putRecord(AUDIT_ACCESS_KEY, AUDIT_SECRET_KEY, config.get('audit.kinesis-stream-name'), {auditGroup});
        } catch (err) {
            logger.error('Failed to send audit to audit service', err);
        }
    } else {
        logger.warn('Simulator - do not send audit to kinesis');
    }
}

module.exports = {sendAuditGroup, extractResourceId};