const os = require('os');
const hyperid = require('hyperid');
const {isArray, remove, assignIn, get} = require('lodash');
const {
    createQueue,
    deleteQueue, getQueueUrl,
    getQueueArn, sendMessage, getQueueAttributes, setQueuePolicy
} = require('../lib/aws/sqs');
const {deleteObject} = require('../lib/aws/s3');
const format = require('string-template');
const fs = require('fs');
const path = require('path');
const {timeout} = require('../utils/utils');
const {findRelationshipsForNewDataBrokerStartup} = require('../operations/relationship-operations');
const {SERVER_HOST, DATA_BROKER_CF_FOLDER, DFIO_METADATA_BUCKET_REGION, DFIO_METADATA_BUCKET_NAME} = require('../utils/consts');
const {BadRequest, NotFound, PreconditionFailed} = require('../utils/errors');
const {findRelationshipsWithSecretsInAGroup, findRelationshipsWithEncryptionInAGroup} = require('../operations/group-operations');
const {
    getUserAndPolicyName
} = require('../utils/common');
const {
    createAccessKey,
    createUserAsync,
    waitForUserExists,
    deleteKeys,
    putUserPolicyAsync,
    deleteUserPolicies,    deleteUser,
    createPolicyDocument
} = require('../lib/aws/iam');
const {putObject, getS3ObjectUrl} = require('../lib/aws/s3');
const Group = require('../models/group');
const DataBroker = require('../models/data-broker');
const Relationship = require('../models/relationship');
const getLogger = require('../utils/logger');
const {getDataBrokerLatestVersionAsync} = require('../utils/update-data-broker');
const {getUserArn} = require('../lib/aws/iam');
const config = require('config');

const logger = getLogger();

const [environment] = SERVER_HOST.split('.');

function getStatusesQueueName() {
    const [serverHost] = SERVER_HOST.split('.');
    return serverHost === 'local' ? `STATUSES_${os.hostname().replace(/[-.]/g, '_')}` : `STATUSES_${environment}`;
}

function sendListExportsCommand(requestId, host, dataBroker) {
    const {commandsQueueUrl, id: dataBrokerId, groupId} = dataBroker;
    return sendMessage(commandsQueueUrl, {host}, {type: 'exports', requestId, dataBrokerId, groupId:groupId.toString()});
}

function sendListExportsFoldersCommand(requestId, host, exportName, path, version, dataBroker) {
    const {commandsQueueUrl, id: dataBrokerId, groupId} = dataBroker;
    const payload = {host, export: exportName, version};
    if (path !== '') {
        payload.path = path;
    }
    return sendMessage(commandsQueueUrl, payload, {type: 'export-folders', requestId, dataBrokerId, groupId:groupId.toString()});
}

function sendListS3BucketsCommand(requestId, dataBroker, provider, options = {}) {
    const {commandsQueueUrl, id: dataBrokerId, groupId} = dataBroker;
    return sendMessage(commandsQueueUrl, Object.assign({provider}, {...options}), {type: 's3-buckets', requestId, dataBrokerId, groupId:groupId.toString()});
}

function sendS3ListBucketsFolders(requestId, bucketName, prefix, dataBroker, provider, options = {}) {
    const {commandsQueueUrl, id: dataBrokerId, groupId} = dataBroker;
    const {host, port, accessKey, secretKey} = options;
    const payload = {
        bucket: bucketName,
        prefix: prefix ? prefix.replace(/^\//, '') : prefix,
        provider,
        host,
        port,
        accessKey,
        secretKey
    };
    return sendMessage(commandsQueueUrl, payload, {type: 's3-bucket-folders', requestId, dataBrokerId, groupId:groupId.toString()});
}

function sendListKmsKeys(requestId, dataBroker, bucket) {
    const {commandsQueueUrl, id: dataBrokerId, groupId} = dataBroker;
    const payload = {
        bucket
    };
    return sendMessage(commandsQueueUrl, payload, {type: 'kms-keys', requestId, dataBrokerId, groupId:groupId.toString()});
}


async function addUserToRelationshipQueues(relationshipQueue, userArn) {
    const queuePolicy = await getQueueAttributes(relationshipQueue, ['Policy']);
    const jsonPolicy = JSON.parse(queuePolicy.Policy);
    const policyPrincipals = jsonPolicy.Statement[0].Principal.AWS;
    if (isArray(policyPrincipals)) {
        if (!policyPrincipals.includes(userArn)) {
            jsonPolicy.Statement[0].Principal.AWS.push(userArn);
            await setQueuePolicy(relationshipQueue, jsonPolicy);
        }
    } else if (policyPrincipals !== userArn) {
        jsonPolicy.Statement[0].Principal.AWS = [policyPrincipals, userArn];
        await setQueuePolicy(relationshipQueue, jsonPolicy);
    }
}

async function removeUserFromRelationshipQueues(relationshipQueue, userArn) {
    const queuePolicy = await getQueueAttributes(relationshipQueue, ['Policy']);
    const jsonPolicy = JSON.parse(queuePolicy.Policy);
    const policyPrincipals = jsonPolicy.Statement[0].Principal.AWS;
    if (isArray(policyPrincipals)) {
        if (policyPrincipals.includes(userArn)) {
            remove(jsonPolicy.Statement[0].Principal.AWS, user => user === userArn);
            await setQueuePolicy(relationshipQueue, jsonPolicy);
        }
    }
}

async function updateRelationshipQueuesWithDataBrokerUser(userName, groupId, removeUser) {
    logger.info('Update relationships queues', {userName, groupId});
    const relationships = await findRelationshipsForNewDataBrokerStartup(groupId);

    relationships.forEach(async relationship => {
        const userArn = await getUserArn(userName);

        const {scannerQueue, transferrerQueue} = relationship;

        if (removeUser) {
            await Promise.all([removeUserFromRelationshipQueues(scannerQueue, userArn),
                removeUserFromRelationshipQueues(transferrerQueue, userArn)]);
        } else {
            await Promise.all([addUserToRelationshipQueues(scannerQueue, userArn),
                addUserToRelationshipQueues(transferrerQueue, userArn)]);
        }
    });
}

async function sendStartupCommand(requestId, dataBroker, metadata = []) {
    const {commandsQueueUrl, id: dataBrokerId, groupId} = dataBroker;
    logger.info('sending startup command ', dataBrokerId);
    const relationships = await findRelationshipsForNewDataBrokerStartup(groupId);
    return sendMessage(commandsQueueUrl, {relationships, metadata}, {type: 'startup', requestId, dataBrokerId, groupId:groupId.toString()});
}

async function sendMetadataRequestCommand(requestId, dataBroker, targetDataBrokerId) {
    const {commandsQueueUrl, id: dataBrokerId, groupId} = dataBroker;
    logger.info(`Sending metadata command to databroker ${dataBrokerId}  to queue ${commandsQueueUrl}`);
    return sendMessage(commandsQueueUrl, {targetDataBrokerId}, {type: 'metadata', requestId, dataBrokerId, groupId:groupId.toString()});
}

async function sendShutdownCommand(requestId, dataBroker) {
    const {commandsQueueUrl, id: dataBrokerId, groupId} = dataBroker;
    logger.info(`Sending shutdown command to databroker ${dataBrokerId}`);
    return sendMessage(commandsQueueUrl, {}, {type: 'shutdown', requestId, dataBrokerId, groupId:groupId.toString()});
}

async function sendListAzureBlobStorageContainersFoldersCommand(requestId, dataBroker, container, prefix, storageAccountName,
    storageAccountConnectionString) {
    const {commandsQueueUrl, id: dataBrokerId, groupId} = dataBroker;
    let message = {
        container,
        prefix
    };
    if (storageAccountConnectionString) {
        message = Object.assign({}, message, {storageAccountName, storageAccountConnectionString});
    } else {
        throw new BadRequest('list-azure-blob-storage-containers request should contain storage-account-connection-string');
    }

    return sendMessage(commandsQueueUrl, message, {
        type: 'azure-blob-storage-container-folders',
        requestId,
        groupId:groupId.toString(),
        dataBrokerId
    });
}

async function sendListAzureBlobStorageContainers(requestId, dataBroker, storageAccountName, storageAccountConnectionString) {
    const {id: dataBrokerId, commandsQueueUrl, groupId} = dataBroker;
    let message;
    if (storageAccountConnectionString) {
        message = {storageAccountName, storageAccountConnectionString};
    } else {
        throw new BadRequest('list-azure-blob-storage-containers request should contain storage-account-connection-string');
    }
    return sendMessage(commandsQueueUrl, message, {
        type: 'azure-blob-storage-containers',
        requestId, groupId:groupId.toString(),
        dataBrokerId
    });
}

async function sendListGcpBucketsCommand(requestId, dataBroker) {
    const {id: dataBrokerId, commandsQueueUrl, groupId} = dataBroker;
    return sendMessage(commandsQueueUrl, {}, {
        type: 'gcp-buckets',
        requestId,
        groupId:groupId.toString(),
        dataBrokerId
    });
}

async function sendListGcpBucketFoldersCommand(requestId, dataBroker, bucket, prefix) {
    const {commandsQueueUrl, id: dataBrokerId, groupId} = dataBroker;
    return sendMessage(commandsQueueUrl, {
        bucket,
        prefix
    }, {type: 'gcp-bucket-folders', requestId, dataBrokerId, groupId:groupId.toString()});
}


function sendSharesCommand(requestId, dataBroker, host, version, username, password, domain) {
    const {id: dataBrokerId, commandsQueueUrl, groupId} = dataBroker;
    return sendMessage(commandsQueueUrl, {
        host, username, version, password, workgroup: domain
    }, {
        type: 'shares',
        requestId,
        dataBrokerId, groupId:groupId.toString()
    });
}

function sendDownloadLogCommand(requestId, dataBroker, name) {
    const {id: dataBrokerId, commandsQueueUrl, groupId} = dataBroker;
    return sendMessage(commandsQueueUrl, {name}, {
        type: 'download-logs',
        requestId, dataBrokerId, groupId:groupId.toString()
    });
}

function sendShareFoldersCommand(requestId, dataBroker, host, share, path, version, username, password, domain) {
    const {id: dataBrokerId, commandsQueueUrl, groupId} = dataBroker;
    return sendMessage(commandsQueueUrl, {
        host, username, password, share, path: path === '' ? undefined : path, version, workgroup: domain
    }, {
        type: 'share-folders',
        requestId,
        dataBrokerId, groupId:groupId.toString()
    });
}

async function sendDataBrokerAsupRequestCommand(requestId, serialNumber, dataBroker) {
    const {commandsQueueUrl, id: dataBrokerId, groupId} = dataBroker;
    logger.info('Sending send-asup command message: ', {commandsQueueUrl, dataBrokerId, requestId, serialNumber});

    return sendMessage(commandsQueueUrl, {serialNumber, asupUrl: config.get('asup.server-address')}, {
        type: 'send-asup',
        requestId,
        dataBrokerId, groupId:groupId.toString()
    });
}

async function findAccountsByAwsAccountId(awsAccountId) {
    return DataBroker.aggregate([
        {
            $match: {
                awsAccountId: `${awsAccountId}`
            }
        },
        {
            $lookup: {
                from: 'accounts',
                localField: 'account',
                foreignField: '_id',
                as: 'accountData'
            }
        },
        {
            $unwind: '$accountData'
        },
        {
            $group: {
                _id: '$account',
                subscription: {$first: '$accountData.subscription'}
            }
        },
        {
            $project: {
                _id: true,
                subscription: true
            }
        }
    ]);
}

function createCFN(installerLocation, type) {
    let template;
    if (type === 'AZURE') {
        template = fs.readFileSync(path.resolve(__dirname, config.get('paths.agent.newCfn.azure')), 'utf-8');
    } else if (type === 'GCP') {
        template = fs.readFileSync(path.resolve(__dirname, config.get('paths.agent.newCfn.gcp')), 'utf-8');
    } else {
        template = fs.readFileSync(path.resolve(__dirname, config.get('paths.agent.newCfn.aws')), 'utf-8');
    }
    return format(template, {
        installerLocation
    });
}

function createInstaller(dataBrokerId, agentBundleUrl, commandsQueue, statusesQueue, type, cloudSyncAwsAccessKey, cloudSyncAwsSecretKey, port) {
    logger.info('Create installer');
    const installer = fs.readFileSync(path.resolve(__dirname, config.get('paths.agent.newInstaller.aws')), 'utf-8');
    const installerParameters = {
        environment,
        agentBundleUrl,
        dataBrokerId,
        type,
        port,
        commandsQueue,
        statusesQueue,
        sqsAccessKeyId: cloudSyncAwsAccessKey,
        sqsSecretAccessKey: cloudSyncAwsSecretKey
    };

    return format(installer, installerParameters);
}

async function createDataBrokerQueue(dataBroker, userName, queueName, actions) {
    const {id: dataBrokerId, account: {tenancyAccountId}} = dataBroker;
    logger.info(`Create queue ${queueName} for account ${tenancyAccountId}`);

    try {
        const [queueUrl, userArn] = await Promise.all([createQueue(queueName), getUserArn(userName)]);
        const queueArn = await getQueueArn(queueUrl);
        await timeout(10000);

        const policy = {
            Version: '2012-10-17',
            Statement: [
                {
                    Sid: `Sid${Date.now()}`,
                    Effect: 'Allow',
                    Principal: {AWS: [`${userArn}`]},
                    Action: actions,
                    Resource: `${queueArn}`
                }]
        };

        await setQueuePolicy(queueUrl, policy);
        logger.info('Queue successfully created:', queueUrl);

        return {queueUrl, queueArn};
    } catch (err) {
        logger.error(`Failed to create a queue ${queueName} for data broker ${dataBrokerId} and account ${tenancyAccountId} with error`, err);
        throw err;
    }
}

async function deleteQueues(dataBroker) {
    logger.info(`Deleting SQS queues of data broker ${dataBroker.id}`);

    const {statusesQueueUrl, commandsQueueUrl} = dataBroker;
    const sharedQueueUrl = await getQueueUrl(getStatusesQueueName());
    const queuesToDelete = sharedQueueUrl === statusesQueueUrl ? [commandsQueueUrl] : [commandsQueueUrl, statusesQueueUrl];

    await Promise.all(queuesToDelete.map(async queueUrl => {
        try {
            logger.info(`Deleting queue ${queueUrl}`);
            await deleteQueue(queueUrl);
            logger.info('Done deleting queues');
        } catch (err) {
            logger.warn(`Failed to delete queue ${queueUrl}`, err);
        }
    }));
}

async function issueDataBrokerCommandAndRespond(req, res, next, fn) {
    const {params: {id: dataBrokerId}} = req;
    const dataBroker = await DataBroker.findById(dataBrokerId).exec();
    if (!dataBroker) {
        logger.warn(`Data Broker with id ${dataBrokerId} not found`);
        next(new NotFound(`Data Broker with id ${dataBrokerId} not found`));
    } else {
        await fn(dataBroker);
        res.status(202).json();
    }
}

async function attachPolicyToUserAndGenerateKeys(userAndPolicyName, commandsQueueArn, statusesQueueArn) {
    const policyDocument = await createPolicyDocument(commandsQueueArn, statusesQueueArn);
    await putUserPolicyAsync(policyDocument, userAndPolicyName, userAndPolicyName);
    return createAccessKey(userAndPolicyName);
}

async function addDataBrokerToAGroup(group, dataBroker) {
    const {groupId, name, account} = group;
    const filter = groupId ? {_id: groupId} : {name, account};
    return Group.findOneAndUpdate(filter, {$addToSet: {dataBrokers: dataBroker}}, {new: true, upsert: true});
}

async function deleteUserAndPolicyAndQueues(dataBroker) {
    const userAndPolicyName = getUserAndPolicyName(dataBroker);
    try {
        await Promise.all([deleteKeys(userAndPolicyName), deleteUserPolicies(userAndPolicyName)]);
        await deleteUser(userAndPolicyName);
        await deleteQueues(dataBroker);
    } catch (err) {
        logger.error(`An Error occurred while deleting data broker ${dataBroker.name} with id ${dataBroker._id} : ${err}`);
    }
}

async function removeDataBrokerFromGroupIfNeeded(dataBroker, group) {
    logger.info('remove data broker from group if needed');
    const {
        dataBrokers, _id: groupId
    } = group || {};
    if (group && dataBrokers.length <= 1) {
        logger.info(`Removing group with 1 dataBroker ${JSON.stringify(group)}`);
        await Group.findByIdAndRemove(groupId).exec();
    } else {
        logger.info(`Removing dataBroker ${dataBroker.id} from group ${groupId}`);
        await Group.findByIdAndUpdate(groupId, {$pull: {dataBrokers: dataBroker.id}});
        await updateRelationshipQueuesWithDataBrokerUser(getUserAndPolicyName(dataBroker), groupId, true);
    }
}

async function validateADataBrokerCanBeAddedToAGroup(group) {
    const {id: groupId, dataBrokers} = group;
    const brokersIds = dataBrokers.map(broker => broker._id);
    const relationshipsWithListener = await Relationship.find({'encryption.listener.dataBrokerId':{$in:brokersIds}}).exec();
    if (relationshipsWithListener.length) {
        throw new BadRequest('Acceleration is not supported with encrypted relationships');
    }
    logger.info('Validate that data brokers can be added to group: ', groupId);
    const numberOfRelationships = await findRelationshipsWithSecretsInAGroup(groupId);
    if (numberOfRelationships > 0) {
        logger.info('Validating there is at least one data-broker in the group');
        const {dataBrokers} = group;
        const aliveDataBroker = dataBrokers.find(dataBroker => dataBroker.isAlive());
        if (!aliveDataBroker) {
            throw new BadRequest('The existing data brokers in this relationship are not online. Start at least one of the data brokers and then try again');
        }
    }
    const numberOfEncryptedRelationships = await findRelationshipsWithEncryptionInAGroup(groupId);
    if (numberOfEncryptedRelationships > 0) {
        throw new BadRequest('Acceleration is not supported with encrypted relationships');
    }
}

async function validateGroupAndAddADataBroker(groupId, account, dataBroker) {
    let group;
    if (groupId) {
        group = await Group.findById(groupId).populate('dataBrokers').exec();
        if (!group) {
            throw new BadRequest(`There is no group with id ${groupId} `);
        }
        // need to check number of relationships that include the following as source or target : azure, cifs, sgws, gcp associated to the group
        // if it has more than one allow to add only if there is an alive data-broker in the group
        await validateADataBrokerCanBeAddedToAGroup(group);
        return addDataBrokerToAGroup({groupId: group.id}, dataBroker);
    }
    // no group was asked for this data-broker- we will take the data-broker name and create a new group- will be deleted once the UI have groups
    const {name} = dataBroker;
    return addDataBrokerToAGroup({name, account: account.id}, dataBroker);
}

async function validateDataBrokerDeletion(dataBrokers, group, dataBroker) {
    const filter = (group) ? {$or:[{group: group.id}, {'encryption.listener.dataBrokerId':dataBroker._id}]} : {$or:[{dataBroker: dataBroker.id}, {'encryption.listener.dataBrokerId':dataBroker._id}]};
    const numberOfRelationships = await Relationship.countDocuments(filter);
    if (numberOfRelationships > 0) {
        const anotherAliveDataBroker = dataBrokers.find(receivedDataBroker => receivedDataBroker.id !== dataBroker.id && receivedDataBroker.isAlive());
        if (!anotherAliveDataBroker) {
            // if we have no group we wouldn't find any other alive data-broker;
            throw new PreconditionFailed(`Data Broker '${dataBroker.name}' in use by ${numberOfRelationships} active relationships `);
        }
    }
}

async function createAndUploadInstaller(dataBroker) {
    // creating user and queues and keys
    const userAndPolicyName = getUserAndPolicyName(dataBroker);
    const {externalId, id:dataBrokerId, groupId} = dataBroker;
    const user = await createUserAsync(userAndPolicyName);
    await waitForUserExists(userAndPolicyName);
    let commandsQueueDetails, statusesQueueDetails;
    if (config.get('one-queue')) {
        commandsQueueDetails = await createDataBrokerQueue(dataBroker, userAndPolicyName, `COMMANDS_${environment}_${externalId ? externalId : dataBrokerId}`, ['SQS:DeleteMessage', 'SQS:ReceiveMessage']);
        const statusesQueueUrl = await getQueueUrl(getStatusesQueueName());
        const statusesQueueArn = await getQueueArn(statusesQueueUrl);
        statusesQueueDetails = {queueUrl: statusesQueueUrl, queueArn:statusesQueueArn};
    } else {
        ([commandsQueueDetails, statusesQueueDetails] = await Promise.all([
            createDataBrokerQueue(dataBroker, userAndPolicyName, `COMMANDS_${environment}_${externalId ? externalId : dataBrokerId}`, ['SQS:DeleteMessage', 'SQS:ReceiveMessage']),
            createDataBrokerQueue(dataBroker, userAndPolicyName, `STATUSES_${environment}_${externalId ? externalId : dataBrokerId}`, ['SQS:SendMessage'])
        ]));
    }
    dataBroker.commandsQueueUrl = commandsQueueDetails.queueUrl;
    dataBroker.statusesQueueUrl = statusesQueueDetails.queueUrl;
    dataBroker.userId = user.UserId;
    await dataBroker.save();
    const keys = await attachPolicyToUserAndGenerateKeys(userAndPolicyName, commandsQueueDetails.queueArn, statusesQueueDetails.queueArn);
    const {AccessKeyId, SecretAccessKey} = keys;
    // creating installer
    const installerKey = `${DATA_BROKER_CF_FOLDER}/${dataBroker.externalId}_installer`;
    const latestDataBrokerBundle = await getDataBrokerLatestVersionAsync();
    const installer = await createInstaller(dataBroker.id, latestDataBrokerBundle.url, commandsQueueDetails.queueUrl, statusesQueueDetails.queueUrl, dataBroker.type,
        AccessKeyId, SecretAccessKey, get(dataBroker, 'encryption.port'));
    await putObject(DFIO_METADATA_BUCKET_REGION, DFIO_METADATA_BUCKET_NAME, installerKey, installer, true);
    updateRelationshipQueuesWithDataBrokerUser(userAndPolicyName, groupId);
    return getS3ObjectUrl(DFIO_METADATA_BUCKET_REGION, DFIO_METADATA_BUCKET_NAME, installerKey,
        true, false);
}

async function validateBrokerCreationAndUpdateTimeline(req) {
    let group;
    const {
        account, body: {
            name, groupId, port, type
        }
    } = req;
    const dataBrokerType = type || 'AWS';// or on type is a temporary patch for OCCM integration
    const dataBroker = new DataBroker({account, name, type: dataBrokerType, externalId:hyperid().uuid});
    if (port) {
        dataBroker.encryption = {port: parseInt(port)};
    }
    if (!/^[a-zA-Z][a-zA-Z0-9-]{0,127}$/.test(name)) {
        throw new BadRequest('Data Broker name must contain only letters, numbers, dashes and start with alpha character');
    } else {
        logger.info('Check if another data broker exists with the same name');
        const anotherDataBroker = await DataBroker.findOne({name, account}).exec();
        if (anotherDataBroker) {
            throw new BadRequest(`Data Broker named '${name}' already exists`);
        } else {
            logger.info('Create a new data broker of type', dataBrokerType, 'to a group', groupId);
            req.timeline = assignIn(req.timeline, {dataBroker: name});
            req.timelineV2 = assignIn(req.timelineV2, {dataBroker: {name, dataBrokerId: dataBroker.id}});
            // first we create the group then assign the data-broker to the group.
            group = await validateGroupAndAddADataBroker(groupId, account, dataBroker);
            // Its like with aggregate and vol( first create the aggregate then creates the volume and adds the volume)
            dataBroker.groupId = group._id;
            //ui takes the first item in the array od dataBrokers and show in the timeline
            req.timelineV2 = assignIn(req.timelineV2, {
                dataBroker: {
                    name, dataBrokerId: dataBroker.id, groupId: group._id, type: dataBrokerType
                }
            });
        }
        return {dataBroker, group};
    }

}

async function deleteFileLinkIfNeeded(dataBroker, isNewBroker) {
    const {
        externalId, type
    } = dataBroker;
    if (isNewBroker) {
        //delete installer + delete template
        const installerKey = `${DATA_BROKER_CF_FOLDER}/${externalId}_installer`;
        logger.info(`deleting installer ${installerKey}`);
        await deleteObject(DFIO_METADATA_BUCKET_REGION, DFIO_METADATA_BUCKET_NAME, installerKey);
        if (type !== 'ONPREM') {
            const cftKeyPrefix = `${DATA_BROKER_CF_FOLDER}/${externalId}`;
            const cftKey = (type === 'GCP') ? `${cftKeyPrefix}.jinja` : cftKeyPrefix;
            logger.info(`deleting cfKey ${cftKey}`);
            await deleteObject(DFIO_METADATA_BUCKET_REGION, DFIO_METADATA_BUCKET_NAME, cftKey);
        }
    }
}

module.exports = {
    sendShareFoldersCommand,
    sendSharesCommand,
    sendS3ListBucketsFolders,
    sendListExportsFoldersCommand,
    sendListExportsCommand,
    sendStartupCommand,
    sendMetadataRequestCommand,
    sendShutdownCommand,
    sendListAzureBlobStorageContainers,
    sendListAzureBlobStorageContainersFoldersCommand,
    sendListGcpBucketFoldersCommand,
    sendListGcpBucketsCommand,
    sendDataBrokerAsupRequestCommand,
    sendListS3BucketsCommand,
    sendDownloadLogCommand,
    findAccountsByAwsAccountId,
    createCFN,
    deleteQueues,
    issueDataBrokerCommandAndRespond,
    deleteUserAndPolicyAndQueues,
    removeDataBrokerFromGroupIfNeeded,
    getStatusesQueueName,
    sendListKmsKeys,
    validateDataBrokerDeletion, validateBrokerCreationAndUpdateTimeline, createAndUploadInstaller, deleteFileLinkIfNeeded
};
