const ms = require('ms');
const config = require('config');
const Relationship = require('../models/relationship');
const {PreconditionFailed} = require('../utils/errors');
const {ONTAP_S3_PROVIDER, ALIBABA_PROVIDER, IBM_PROVIDER, SGWS_PROVIDER } = require('../utils/consts');
const {headObject} = require('../lib/aws/s3');
const getLogger = require('../utils/logger');

const logger = getLogger();


async function findRelationshipsWithSecretsInAGroup(groupId) {
    const secretsRequiredProtocols = ['cifs', 'azure'];
    const secretsRequiredProviders = [ONTAP_S3_PROVIDER, ALIBABA_PROVIDER, IBM_PROVIDER, SGWS_PROVIDER ];
    return Relationship.countDocuments({
        group: groupId,
        $or: [{'source.protocol': {$in: secretsRequiredProtocols}},
            {'target.protocol': {$in: secretsRequiredProtocols}},
            {'source.s3.provider': {$in: secretsRequiredProviders}},
            {'target.s3.provider': {$in: secretsRequiredProviders}}]
    }).exec();
}

async function findRelationshipsWithEncryptionInAGroup(groupId) {
    return Relationship.countDocuments({
        group: groupId,
        encryption:{$exists:true}
    }).exec();
}

function waitForObject(name) {
    let interval;
    let timeout;

    const setIntervalTime = '15s';
    const setTimeoutTime = '10m';

    const p1 = new Promise((resolve, reject) => {
        timeout = setTimeout(async() => {
            clearInterval(interval);
            reject(new Error(`Failed to find S3 object with name ${name}`));
        }, ms(setTimeoutTime));
    });

    const p2 = new Promise((resolve) => {
        interval = setInterval(async() => {
            try {
                logger.info('looking for object ', name, 'from bucket', config.get('dfio-metatdata-bucket.name'));
                await headObject(config.get('dfio-metatdata-bucket.name'), name);
                logger.info(`object exists ${name}`);
                clearInterval(interval);
                clearTimeout(timeout);
                resolve();
            } catch (err) {
                logger.info(`no object ${name} yet trying again`);
            }
        }, ms(setIntervalTime));
    });

    return Promise.race([p1, p2]);
}

function isThereAliveDataBroker(group) {
    const dataBrokers = group.dataBrokers;
    const aliveDataBroker = dataBrokers.find(dataBroker => dataBroker.isAlive());
    if (!aliveDataBroker) {
        throw new PreconditionFailed('Relationship must have at least one connected data-broker');
    }
}


module.exports = {
    findRelationshipsWithSecretsInAGroup,
    findRelationshipsWithEncryptionInAGroup,
    waitForObject,
    isThereAliveDataBroker
};
