const {chunk, compact} = require('lodash');
const {agenda} = require('../utils/agenda');
const {
    AGENDA_TASK_NAME_AZURE_SAAS_METERING, SAAS_REGISTRATION_STATUS_SUBSCRIBE_SUCCESS, SAAS_SUBSCRIPTION_PROVIDER_AZURE, CONTRACT_SUBSCRIPTION
} = require('../utils/consts');
const {Account} = require('../models/account');
const {getSubscriptionDetails, batchUsageEvent} = require('../lib/marketplace-azure');
const getLogger = require('../utils/logger');

const logger = getLogger();

const TIER_1_LIMIT = 5; // tier 1 -> 0 to 5
const TIER_2_LIMIT = 15; // tier 2 -> 6 to 20

const BATCH_USAGE_EVENT_SIZE = 25;
const YEARLY_SUBSCRIPTION_RELATIONSHIPS = 20;

async function getChargingData() {
    return Account.aggregate(
        [
            {
                $match: {
                    'subscription.status': SAAS_REGISTRATION_STATUS_SUBSCRIBE_SUCCESS,
                    trialExpiry: {$lt: new Date()},
                    'subscription.provider': SAAS_SUBSCRIPTION_PROVIDER_AZURE
                }
            },
            {
                $lookup: {
                    from: 'relationships',
                    localField: '_id',
                    foreignField: 'account',
                    as: 'relationshipData'
                }
            },
            {
                $project: {
                    customerIdentifier: '$subscription.customerIdentifier',
                    reservedRelationships: {$sum: '$byol.licenses.reservedRelationships'},
                    regularRelationships: {
                        $filter: {
                            input: '$relationshipData',
                            as: 'relationship',
                            cond: {$eq: ['$$relationship.isQstack', false]}
                        }
                    },
                    subscriptionType: '$subscription.type'
                }
            },
            {
                $project: {
                    customerIdentifier: 1,
                    reservedRelationships: 1,
                    numberOfRelationships: {$size: '$regularRelationships'},
                    subscriptionType: 1
                }
            },
            {
                $match: {numberOfRelationships: {$gt: 0}}
            },
            {
                $group: {
                    _id: {customerIdentifier: '$customerIdentifier', subscriptionType: '$subscriptionType'},
                    numberOfRelationships: {$sum: '$numberOfRelationships'},
                    reservedRelationships: {$sum: '$reservedRelationships'}
                }
            },
            {
                $project: {
                    customerIdentifier: '$_id.customerIdentifier',
                    subscriptionType: '$_id.subscriptionType',
                    numberOfRelationships: 1,
                    reservedRelationships: 1,
                    _id: 0
                }
            }
        ]
    )
        .exec();
}

async function charge() {
    logger.info('Charging in AZURE...');

    const chargingData = await getChargingData();
    logger.info('Charging data:', chargingData);

    const records = await Promise.all(chargingData.filter(entry => entry.numberOfRelationships - entry.reservedRelationships > 0).map(async entry => {
        const {planId} = await getSubscriptionDetails(entry.customerIdentifier);
        const effectiveStartTime = new Date().toUTCString();
        const nonReservedRelationships = entry.numberOfRelationships - entry.reservedRelationships;

        if (entry.subscriptionType === CONTRACT_SUBSCRIPTION && nonReservedRelationships > YEARLY_SUBSCRIPTION_RELATIONSHIPS) {
            return {
                dimension: 'tier3',
                quantity: nonReservedRelationships - YEARLY_SUBSCRIPTION_RELATIONSHIPS,
                resourceId: entry.customerIdentifier,
                planId,
                effectiveStartTime
            };
        }
        let tier1;
        let tier2 = 0;
        let remainingRelationships = nonReservedRelationships;

        if (remainingRelationships > TIER_1_LIMIT) {
            tier1 = TIER_1_LIMIT;
            remainingRelationships -= tier1;
        } else {
            tier1 = remainingRelationships;
            remainingRelationships = 0;
        }

        if (remainingRelationships > TIER_2_LIMIT) {
            tier2 = TIER_2_LIMIT;
            remainingRelationships -= tier2;
        } else {
            tier2 = remainingRelationships;
            remainingRelationships = 0;
        }
        //filter because azure doesnt accept 0 quantity
        return [{name: 'tier1', value: tier1}, {name: 'tier2', value: tier2}, {
            name: 'tier3',
            value: remainingRelationships
        }].filter(obj => obj.value).map(x => ({
            dimension: x.name,
            quantity: x.value,
            resourceId: entry.customerIdentifier,
            planId,
            effectiveStartTime
        }));
    }));

    const promises = chunk(compact(records.flat()), BATCH_USAGE_EVENT_SIZE).map(chunkedRecords => batchUsageEvent(chunkedRecords));
    return Promise.all(promises);
}

agenda.define(AGENDA_TASK_NAME_AZURE_SAAS_METERING, {lockLifetime: 10000}, async(job, done) => {
    try {
        await charge();
    } catch (err) {
        logger.error(`Failed to execute job ${AGENDA_TASK_NAME_AZURE_SAAS_METERING}`, err);
    } finally {
        done();
    }
});
