const config = require('config');
const ms = require('ms');
const retry = require('async-retry');
const moment = require('moment');
const TimelineV2 = require('../models/timeline-v2');
const {
    sumBy, isObject, uniq, omit, head, keys, get, first, isEmpty, sortBy, isNumber, inRange, isString, round, extend
} = require('lodash');
const {generateRequestId, timeout} = require('../utils/utils');
const {
    MESSAGE_STATUS_DONE, MESSAGE_STATUS_FAILED, IBM_PROVIDER, SGWS_PROVIDER, SERVER_HOST, MONGO_DB_ERROR_DUPLICATE_KEY,
    RUNNING_STATUS, ONTAP_S3_PROVIDER, S3_PROVIDER, ALIBABA_PROVIDER, MESSAGE_STATUS_PENDING, MESSAGE_STATUS_ABORTED, ABORTED_MESSAGE
} = require('../utils/consts');
const {getRegion} = require('../lib/aws/s3');
const {getQueueAttributes, sendMessage, deleteQueue} = require('../lib/aws/sqs');
const {assertIpOrHostname, getUserAndPolicyName} = require('../utils/common');
const {removeFirstSlash} = require('../utils/utils');
const {BadRequest} = require('../utils/errors');
const {createQueue, getQueueArn, setQueuePolicy} = require('../lib/aws/sqs');
const {getUserArn} = require('../lib/aws/iam');
const {
    ALLOWED_SOURCE_AND_TARGET, S3_TAG_KEY_REGEXP, S3_TAG_VALUE_REGEXP, S3_METADATA_NAME_REGEXP, S3_METADATA_VALUE_REGEXP, SCAN_MODE_SOURCE_AND_TARGET
} = require('../utils/consts');
const getLogger = require('../utils/logger');
const {settingsSchema, schedulerSchema} = require('../schemas/settings-schema');
const Relationship = require('../models/relationship');
const {sumStatsPerRelationship, RelationshipStats} = require('../models/relationship-statistics');
const DataBroker = require('../models/data-broker');
const {Message2} = require('../models/message-v2');
const {updateTimeLineNewDataBroker, findTimelineRecordV2} = require('./timeline-operations');
const {validateSchema} = require('../utils/schema-validation');
const {sendAuditGroup} = require('./audit-operations');

const [environment] = SERVER_HOST.split('.');

const logger = getLogger();


function addEndSlash(path) {
    return (path) ? (path.endsWith('/') ? path : `${path}/`) : '';
}

function createTimeLineObjectForRelationship(req, relationship) {
    const {
        source, target, dataBroker, group, id: relationshipId, isQstack, isCvo, isCm, encryption
    } = relationship || {};

    req.timelineV2 = extend(req.timelineV2, {
        status: (req.timelineV2.summary === 'Create a new Relationship' && !relationship.settings.schedule.syncWhenCreated) ? MESSAGE_STATUS_PENDING : 'PROGRESS',
        relationship: (relationship) ? {
            relationshipId,
            source,
            target,
            encryption,
            isQstack,
            isCvo,
            isCm,
            dataBrokerId: dataBroker,
            group: (group) ? group.id : undefined,
            activateRelationshipAt: (relationship.settings.schedule.syncWhenCreated) ? undefined : relationship.settings.schedule.nextTime,
        } : undefined,
        dataBroker: {dataBrokerId:dataBroker},
        group: (group) ? {
            name: group.name,
            groupId: group.id,
            dataBrokers: group.dataBrokers.map(broker => ({dataBrokerId: broker.id, name: broker.name}))
        } : undefined
    });
}

async function calculateRelationshipStatus(relationshipId) {
    const statusPerRelationship = await RelationshipStats.aggregate([
        { $match: { relationship: relationshipId } },
        {
            $group: {
                _id: '$relationship',
                statuses: { $addToSet: { status: '$status', reason: '$reason' } }
            }
        }]);
    if (isEmpty(statusPerRelationship)) { // there is an option there will not be a status - in case a data broker joined in the middle of a relationship
        return {status: RUNNING_STATUS};
    }
    const statusesScores = {
        [RUNNING_STATUS]: 4,
        [MESSAGE_STATUS_FAILED]: 5,
        [MESSAGE_STATUS_PENDING]: 6,
        [MESSAGE_STATUS_DONE]: 1,
        [MESSAGE_STATUS_ABORTED]: 2
    };
    const {statuses: statusArr} = first(statusPerRelationship);
    const relationshipStatusesWithScores = sortBy(statusArr.map(statusObj => ({
        statusObj,
        score: statusesScores[statusObj.status]
    })), 'score');

    return first(relationshipStatusesWithScores).statusObj;
}

function calculateProgress(status, bytesToTransfer, bytesTransferred, dirsToTransfer, dirsTransferred) {
    let progress;
    if (isNumber(bytesToTransfer) && isNumber(dirsToTransfer)) {
        if (bytesToTransfer === 0 && dirsToTransfer === 0) {
            progress = (status === MESSAGE_STATUS_DONE) ? 100 : 0;
        } else {
            progress = Math.round((bytesTransferred + dirsTransferred) / (bytesToTransfer + dirsToTransfer) * 100);
            if (progress >= 100) {
                if (status !== MESSAGE_STATUS_DONE && status !== MESSAGE_STATUS_FAILED) {
                    progress = 99;
                } else {
                    progress = (status === MESSAGE_STATUS_DONE) ? 100 : 0;
                }
            } else {
                progress = (status === MESSAGE_STATUS_DONE) ? 100 : ((status === MESSAGE_STATUS_FAILED) ? 0 : progress);
            }
        }
    }
    return progress;
}

function calculateTimeEstimation(status, bytesToTransfer, bytesTransferred, filesMarkedForTransfer, filesTransferred, filesMarkedForDelete, filesDeleted, dirsToTransfer, dirsTransferred, dataBrokersRate) {
    logger.debug('calculateTimeEstimation', {
        status,
        bytesToTransfer,
        bytesTransferred,
        filesMarkedForTransfer,
        filesTransferred,
        filesMarkedForDelete,
        filesDeleted,
        dirsToTransfer,
        dirsTransferred,
        dataBrokersRate
    });
    let estimation;
    //transferRate is relevant only when some of the bytes transferred
    if ((bytesToTransfer > 0 || filesMarkedForDelete > 0 || dirsToTransfer > 0 || filesMarkedForTransfer > 0) && (bytesTransferred > 0 || filesDeleted > 0 || dirsTransferred > 0 || filesTransferred > 0)) {
        const secondsToCopyOrDelete = 0.003;
        estimation = round((bytesToTransfer - bytesTransferred) /
            (dataBrokersRate)) + (secondsToCopyOrDelete * (dirsToTransfer - dirsTransferred)) +
            (secondsToCopyOrDelete * (filesMarkedForDelete - filesDeleted)) +
            (secondsToCopyOrDelete * (filesMarkedForTransfer - filesTransferred));
    }
    const secondsInADay = 60 * 60 * 24; // default time is 1 day instead of 1
    if (estimation > secondsInADay) {
        estimation = secondsInADay;
    }
    return estimation > 0 ? estimation : undefined; //CS-4449
}

function createErrorsObject(errors, step) {
    const errorObj = errors.reduce((acc, errorObj) => {
        const entries = Object.entries(errorObj);
        entries.forEach(entry => {
            const [key, value] = entry;
            if (acc[key]) {
                acc[key].counter = acc[key].counter + value.counter;
            } else {
                acc[key] = {step, errorCode: key, counter : value.counter, description:value.description};
            }
        });
        return acc;
    }, {});
    return Object.values(errorObj);
}

function getTopFiveMostCommonErrors(listErrors, transferErrors) {
    const relationshipListErrorsArray = createErrorsObject(listErrors, 'LIST');
    const relationshipTransferErrorsArray = createErrorsObject(transferErrors, 'TRANSFER');
    const sortedErrorsArray = sortBy(relationshipListErrorsArray.concat(relationshipTransferErrorsArray), 'counter');
    return sortedErrorsArray.slice(0, 4).reverse(); // return descending array
}

async function createActivity(relationship) {
    const dataBrokersRate = sumBy(relationship.group.dataBrokers.filter(dataBroker =>
        dataBroker.isAlive()), 'transferRate');
    const relationshipJson = relationship.toJSON();
    const relationshipsStats = await sumStatsPerRelationship(relationship.id);
    const relationshipStats = first(relationshipsStats);
    const bytesMarkedForCopy = get(relationshipStats, 'bytesMarkedForCopy', 0) + get(relationshipStats, 'symbolicBytesMarkedForCopy', 0);
    const bytesMarkedForRemove = get(relationshipStats, 'bytesMarkedForRemove', 0) + get(relationshipStats, 'symbolicBytesMarkedForRemove', 0);
    const bytesCopied = get(relationshipStats, 'bytesCopied', 0) + get(relationshipStats, 'symbolicBytesCopied', 0);
    const bytesRemoved = get(relationshipStats, 'bytesRemoved', 0) + get(relationshipStats, 'symbolicBytesRemoved', 0);
    const bytesFailed = get(relationshipStats, 'bytesFailed', 0) +  get(relationshipStats, 'symbolicBytesFailed', 0);
    const {status, reason} = await calculateRelationshipStatus(relationship._id);
    const filesCopied = get(relationshipStats, 'filesCopied', 0) + get(relationshipStats, 'symbolicCopied', 0);
    const filesFailed = get(relationshipStats, 'filesFailed', 0) + get(relationshipStats, 'symbolicFailed', 0);
    const filesMarkedForCopy = get(relationshipStats, 'filesMarkedForCopy', 0) + get(relationshipStats, 'symbolicMarkedForCopy', 0);
    const filesMarkedforRemove = get(relationshipStats, 'filesMarkedForRemove', 0) + get(relationshipStats, 'symbolicMarkedForRemove', 0);
    const filesRemoved = get(relationshipStats, 'filesRemoved', 0) + get(relationshipStats, 'symbolicRemoved', 0);
    const filesRemovedFailed = get(relationshipStats, 'filesRemovedFailed', 0) + get(relationshipStats, 'symbolicRemovedFailed', 0);
    const filesMarkedForIgnore = get(relationshipStats, 'filesMarkedForIgnore', 0) + get(relationshipStats, 'symbolicMarkedForIgnore', 0);
    const dirsCopied = get(relationshipStats, 'dirsCopied', 0);
    const dirsMarkedForCopy = get(relationshipStats, 'dirsMarkedForCopy', 0);
    const dirsScanned = get(relationshipStats, 'dirsScanned', 0);
    const filesScanned = get(relationshipStats, 'filesScanned', 0) + get(relationshipStats, 'symbolicScanned', 0);
    const bytesScanned = get(relationshipStats, 'bytesScanned', 0) + get(relationshipStats, 'symbolicBytesScanned', 0);
    const dirsFailedToScan = get(relationshipStats, 'dirsFailedToScan', 0);

    // get relationship errors
    const relationshipListErrors = get(relationshipStats, 'relationshipListErrors', []);
    const relationshipTransferErrors = get(relationshipStats, 'relationshipTransferErrors', []);
    const topFiveMostCommonRelationshipErrors = getTopFiveMostCommonErrors(relationshipListErrors, relationshipTransferErrors);

    const timeEstimation = calculateTimeEstimation(status,
        bytesMarkedForCopy,
        (bytesCopied + bytesFailed), filesMarkedForCopy, (filesCopied + filesFailed), filesMarkedforRemove, filesRemoved + filesRemovedFailed, dirsMarkedForCopy, dirsCopied, dataBrokersRate);
    relationshipJson.activity = {
        type: relationship.phase,
        status,
        failureMessage: (reason) || '',
        executionTime: (relationship.endTime) ? (relationship.endTime - relationship.startTime) : (Date.now() - relationship.startTime),
        startTime: relationship.startTime,
        endTime: (relationship.endTime) ? relationship.endTime : new Date().toISOString(),
        bytesMarkedForCopy,
        filesMarkedForCopy,
        dirsMarkedForCopy,
        filesCopied,
        bytesCopied,
        dirsCopied,

        filesFailed,
        bytesFailed,
        dirsFailed: get(relationshipStats, 'dirsFailed', 0),

        filesMarkedforRemove,
        bytesMarkedForRemove,
        dirsMarkedForRemove: get(relationshipStats, 'dirsMarkedForRemove', 0),
        filesRemoved,
        bytesRemoved,
        dirsRemoved: get(relationshipStats, 'dirsRemoved', 0),

        bytesRemovedFailed: get(relationshipStats, 'bytesRemovedFailed', 0),
        filesRemovedFailed,

        filesMarkedForIgnore,

        dirsScanned,
        filesScanned,
        dirsFailedToScan,
        bytesScanned,
        timeEstimation,
        progress: calculateProgress(status,
            (bytesMarkedForCopy + bytesMarkedForRemove),
            (bytesCopied + bytesRemoved + bytesFailed), dirsMarkedForCopy, dirsCopied),
        lastMessageTime: get(relationshipStats, 'lastMessageTime', 0),
        topFiveMostCommonRelationshipErrors
    };
    return relationshipJson;
}

// update all records related to relationship  and dataBrokers of relationship in relationshipStats
async function updateStatusForRelationship(relationship, status, reason) {
    logger.info(`updating status ${status} for relationship ${relationship.id}`);
    const {
        group, dataBroker, id: relationshipId, account
    } = relationship;
    try {
        // RUNNING and PENDING status in order to create a document for each broker -
        // if one is not created it is not so bad - it will be created when the statistics or failures will come
        if ([RUNNING_STATUS, MESSAGE_STATUS_PENDING].includes(status)) {
            const mastersToUpdate = (group) ? group.dataBrokers : [dataBroker];
            await Promise.all(mastersToUpdate.map(async dataBrokerToUpdate => {
                if (dataBrokerToUpdate.isAlive()) {
                    await RelationshipStats.create({
                        relationship: relationshipId, dataBroker: dataBrokerToUpdate, account, status
                    });
                }
            }));
        } else { // avoid  failing to update one of the documents
            await RelationshipStats.updateMany({
                relationship: relationshipId,
                account,
                status: {
                    $nin: [MESSAGE_STATUS_DONE, MESSAGE_STATUS_ABORTED,
                        MESSAGE_STATUS_FAILED]
                }
            }, {$set: {status, reason}});
        }
    } catch (err) {
        logger.error(`Failed to update status ${status} for relationship ${relationship.id}`);
        throw err;
    }
}

async function sendRelationshipMessage(relationshipResponse, requestId) {
    const rId = requestId || generateRequestId();
    return Message2.create({
        account: isObject(relationshipResponse.account) ? relationshipResponse.account._id : relationshipResponse.account,
        expiration: Date.now(),
        milliSeconds: Date.now(),
        payload: {
            header: {
                dataBrokerId: relationshipResponse.dataBroker,
                relationshipId: relationshipResponse.id,
                requestId: rId
            },
            payload: {
                'update-relationship-v2': relationshipResponse
            }
        }
    }
    );
}

async function createNfsEndPoint(details, encryption) {
    const {
        host, export: exportPath, path, version, provider
    } = details;
    const result = {host, export: exportPath, version};
    if (path !== '') {
        result.path = path;
    }
    if (encryption) {
        const listenerDatabroker = await DataBroker.findById(encryption.listener.dataBrokerId).exec();
        result.serverName = listenerDatabroker.id;
        result.serverIP = listenerDatabroker.privateIp;
        result.port = listenerDatabroker.encryption.port;
        result.certificate = listenerDatabroker.encryption.certificate;
    } else if (provider === 'efs') {
        result.version = '4.1';
    }
    return result;
}

function convertTags(tags) {
    return tags ? tags.map(tag => ({Key: tag.key, Value: tag.value})) : [];
}

async function createS3Endpoint(details, credentials, protocol = 's3', isEndOfRelationship) {
    const {
        bucket, prefix, encryption, tags: s3Tags, storageClass, provider, glacier, deepArchive
    } = details;

    const result = {
        bucket,
        tags: convertTags(s3Tags),
        provider,
        path: prefix
    };

    switch (provider) {
        case S3_PROVIDER: {
            result.storageClass = storageClass; // alibaba supports only Standard storageClass with S3 sdk
            if (encryption) {
                if (encryption.protocol === 'kms') {
                    result.encryption = {kms: {keyId: encryption.keyId.trim()}};
                } else { // protocol is aes256
                    result.encryption = { protocol: 'AES256' };
                }
            }
            result.region = (isEndOfRelationship) ? undefined : await getRegion(S3_PROVIDER, bucket);
            if (glacier) {
                result.glacier = glacier;
            }
            if (deepArchive) {
                result.deepArchive = deepArchive;
            }
            break;
        }
        case ALIBABA_PROVIDER: {
            if (credentials) {
                result.accessKey = credentials.s3.accessKey.trim();
                result.secretKey = credentials.s3.secretKey.trim();
            }
            result.cregion = isEndOfRelationship ? undefined : await getRegion(ALIBABA_PROVIDER, bucket, {
                accessKey: result.accessKey,
                secretKey: result.secretKey
            }); //works even without accessKey and secretKey
            if (encryption && encryption.protocol === 'kms') {
                result.encryption = {kms: {keyId: encryption.keyId.trim()}};
            }
            break;
        }
        case ONTAP_S3_PROVIDER:
        case SGWS_PROVIDER:
        case IBM_PROVIDER: {
            const {
                host, port
            } = details;
            Object.assign(result, {host, port});
            if (credentials) {
                result.accessKey = credentials[protocol].accessKey.trim();
                result.secretKey = credentials[protocol].secretKey.trim();
            }
            break;
        }
        default: {
            logger.error('Received endpoint without known provider', details);
            throw new Error(`Unknown provider ${provider}`);
        }
    }
    return result;
}

async function createEndPointForNewDataBroker(endPointObject, credentials, additionalArguments, isEndOfRelationship) {
    let result;
    let changeProtocolToSnfs = false;
    const {protocol} = endPointObject;
    const details = endPointObject[protocol];
    switch (protocol) {
        case 'nfs': {
            const {encryption, srcOrTrg} = additionalArguments;
            let isListener;
            if (encryption) {
                isListener = encryption.listener.endpointType === srcOrTrg;
            }
            result = await createNfsEndPoint(details,  isListener ? encryption : undefined);
            if (isListener) {
                changeProtocolToSnfs = true;
            }
            break;
        }
        case 's3': {
            result = await createS3Endpoint(details, credentials, 's3', isEndOfRelationship);
            break;
        }
        case 'local': {
            result = details;
            break;
        }
        case 'azure': {
            const {
                container, storageAccountName, prefix, tags: azureTags, blobTier
            } = details;
            const storageAccountConnectionString = credentials ? credentials.azure.storageAccountConnectionString : undefined;
            result = {
                container, storageAccountName, storageAccountConnectionString, tags: convertTags(azureTags), blobTier
            };

            if (prefix !== '') {
                result.path = prefix;
            }

            break;
        }
        case 'cifs': {
            const {
                host, share, path, version, aclServer
            } = details;

            result = {
                host, share, version, aclServer
            };
            if (path !== '') {
                result.path = path;
            }

            if (credentials) {
                result.username = credentials.cifs.username.trim();
                result.password = credentials.cifs.password;
                result.workgroup = credentials.cifs.domain === '' ? undefined : credentials.cifs.domain;
            }
            break;
        }
        case 'gcp': {
            const {bucket, prefix, tags: gcpTags} = details;
            result = {bucket, tags: convertTags(gcpTags)};
            if (prefix !== '') {
                result.path = prefix;
            }
            break;
        }
        default:
            throw Error(`Received unknown protocol ${protocol}`);
    }
    return changeProtocolToSnfs ? {snfs: result} : {[protocol === 'sgws' ? 's3' : protocol]: result};
}

async function findRelationshipsForNewDataBrokerStartup(group) {
    const filter = {
        scannerQueue: {$exists: true},
        transferrerQueue: {$exists: true},
        group
    };
    const relationships = await Relationship.find(filter).exec();
    const promises = relationships.map(async relationship => ({
        relationshipId: relationship.id,
        source: await createEndPointForNewDataBroker(relationship.source, undefined,
            {account: relationship.account, encryption: relationship.encryption, srcOrTrg: 'src'}),
        target: await createEndPointForNewDataBroker(relationship.target, undefined,
            {account: relationship.account, encryption: relationship.encryption, srcOrTrg: 'trg'}),
        scannerQueue: relationship.scannerQueue,
        transferrerQueue: relationship.transferrerQueue,
        settings: relationship.settings
    }));
    return Promise.all(promises);
}

async function sendRelationshipActionMessageToAllDataBrokersInAGroup(group, message, messageType, requestId, customHeaders) {
    const {dataBrokers, id :groupId} = group;
    const {relationshipId} = message;
    const commands = dataBrokers.map(async dataBroker => {
        // send to all the data brokers- if one data broker is down it wont receive done command
        // but he will also wont receive the relationship in the startup command
        const {commandsQueueUrl, id: dataBrokerId} = dataBroker;
        if (commandsQueueUrl) { // in case the data broker was not created yet
            try {
                await retry(() => sendMessage(commandsQueueUrl, message, Object.assign({
                    type: messageType,
                    requestId,
                    dataBrokerId :dataBrokerId.toString(),
                    groupId:groupId.toString()
                }, customHeaders)), {
                    retries: 3,
                    onRetry: err => logger.warn(`RETRYING sending ${messageType} command to data broker ${dataBrokerId} for relationship ${relationshipId} with error ${JSON.stringify(err)}`)
                });
            } catch (err) {
                logger.error(`Failed to send ${messageType} to dataBroker ${dataBrokerId}`);
            }
        }
    });
    await Promise.all(commands);
}

async function sendDeleteCommand(relationship, requestId, type) {
    const {
        source: sourceFormDb, target: targetFromDb, group,
        id: relationshipId, scannerQueue, transferrerQueue, account, encryption
    } = relationship;
    const source = await createEndPointForNewDataBroker(sourceFormDb, undefined, {account, encryption, srcOrTrg: 'src'}, true);
    const target = await createEndPointForNewDataBroker(targetFromDb, undefined, {account, encryption, srcOrTrg: 'trg'}, true);
    const message = {
        relationshipId,
        source,
        target,
        scannerQueue,
        transferrerQueue
    };
    logger.info(`sending ${type} command to relationship ${relationship.id} `);
    return sendRelationshipActionMessageToAllDataBrokersInAGroup(group, message, type, requestId);
}

async function deleteRelationshipQueuesIfExist(relationship) {
    const {transferrerQueue, scannerQueue} = relationship;
    if (transferrerQueue && scannerQueue) {
        return Promise.all([
            deleteQueue(relationship.scannerQueue),
            deleteQueue(relationship.transferrerQueue),
            relationship.updateOne({$unset: {transferrerQueue: '', scannerQueue: ''}})
        ]);
    }
}

async function handleEndOfRelationship(relationship, doneRequestId, commandType) {
    const {id: relationshipId} = relationship;
    // #1 send done command
    logger.info(`Sending ${commandType} command to relationship ${relationship.id}`);
    await sendDeleteCommand(relationship, doneRequestId, commandType);

    // #3 delete relationship queues
    try {
        logger.info(`${commandType} command was sent for relationship ${relationshipId},  deleting relationship queue`);
        await timeout(5000);
        await retry(() => deleteRelationshipQueuesIfExist(relationship), {
            retries: 3,
            onRetry: err => logger.warn(`RETRYING deleting relationship queues of relationship ${relationshipId} with error ${JSON.stringify(err)}`)
        });
    } catch (err) {
        logger.error(`Failed deleting relationship queues for relationship ${relationshipId} with error`, err);
    }
    if (['done', 'abort'].includes(commandType)) {
        // #2 wait 30 seconds - in order to avoid new sync before umount was done
        await timeout(ms(config.get('end-of-relationship-timeout')));
        logger.info(`Updating end time of relationship ${relationship.id}`);
        // update endTime of relationship
        relationship.endTime = Date.now();
        return  relationship.save();
    }
}

async function sendMessageToScanner(source, target, relationshipId, scannerQueue, attributes) {
    const {dataBrokerId, requestId, groupId} = attributes;
    logger.info(`Sending message to the scanner ${scannerQueue} of relationship ${relationshipId}`);
    await sendMessage(scannerQueue, {
        relationshipId,
        source,
        target,
        type: SCAN_MODE_SOURCE_AND_TARGET
    }, {type: 'copy', requestId, dataBrokerId, groupId});
}

function omitSecretKeys(object) {
    const SECRET_KEYS = ['password', 'privateKey', 'storageAccountKey', 'storageAccountConnectionString', 'secretKey', 'certificate', 'serverName'];
    return omit(object, SECRET_KEYS.map(key => `${head(keys(object))}.${key}`));
}

async function sendSaveMetadataCommand(relationship, requestId, sourceCredentials, targetCredentials) {
    const {
        source: sourceFromDb, target: targetFromDb, id: relationshipId, group, account, encryption
    } = relationship;
    const source = await createEndPointForNewDataBroker(sourceFromDb, sourceCredentials, {account, encryption, srcOrTrg: 'src'});
    const target = await createEndPointForNewDataBroker(targetFromDb, targetCredentials, {account, encryption, srcOrTrg: 'trg'});
    const message = {
        relationshipId,
        source,
        target
    };
    await sendRelationshipActionMessageToAllDataBrokersInAGroup(group, message, 'save-metadata', requestId);
}

async function sendCopyCommand(relationship, requestId, sourceCredentials, targetCredentials, customHeaders) {
    const {
        source: sourceFromDb, target: targetFromDb, id: relationshipId, scannerQueue, transferrerQueue,
        settings, dataBroker, group, encryption
    } = relationship;
    const source = await createEndPointForNewDataBroker(sourceFromDb, sourceCredentials, {encryption, srcOrTrg: 'src'});
    const target = await createEndPointForNewDataBroker(targetFromDb, targetCredentials, {encryption, srcOrTrg: 'trg'});
    const message = {
        relationshipId,
        source,
        target,
        scannerQueue,
        transferrerQueue,
        settings
    };

    await sendRelationshipActionMessageToAllDataBrokersInAGroup(group, message, 'copy', requestId, customHeaders);
    return sendMessageToScanner(omitSecretKeys(source), omitSecretKeys(target), relationshipId, scannerQueue, {dataBrokerId:dataBroker.toString(), requestId, groupId:group.id.toString()});
}

function validatePath(path, protocol) {
    if (path === '/.' || path.endsWith('/.') || path.includes('/./')) {
        throw new BadRequest(`${protocol} path cannot include '.' (current folder). Path inserted is ${path}`);
    }
    if (path === '/..' || path.endsWith('/..') || path.includes('/../')) {
        throw new BadRequest(`${protocol} path cannot include '..' (parent folder). Path inserted is ${path}`);
    }
    if (path.split('/').map(p => p.length).filter(length => length > 255).length) {
        throw new BadRequest(`${protocol} path's contains folders longer than 255 characters. Path inserted is ${path}`);
    }
}

function validateDifferentSourceAndTarget(protocol, relationship) {
    const adjustPath = (exportPath, path) => `${removeFirstSlash(exportPath)}/${addEndSlash(removeFirstSlash(path))}`;
    let sourceFullPath;
    let  targetFullPath;
    switch (protocol) {
        case 'nfs':
            if (relationship.source[protocol].host === relationship.target[protocol].host) {
                sourceFullPath = adjustPath(relationship.source[protocol].export, relationship.source[protocol].path);
                targetFullPath = adjustPath(relationship.target[protocol].export, relationship.target[protocol].path);
                if (sourceFullPath.startsWith(targetFullPath) || targetFullPath.startsWith(sourceFullPath)) {
                    throw new BadRequest('Circular relationships are not allowed');
                }
            }

            break;
        case 'cifs':
            if (relationship.source[protocol].host === relationship.target[protocol].host) {
                sourceFullPath = adjustPath(relationship.source[protocol].share, relationship.source[protocol].path);
                targetFullPath = adjustPath(relationship.target[protocol].share, relationship.target[protocol].path);
                if (sourceFullPath.startsWith(targetFullPath) || targetFullPath.startsWith(sourceFullPath)) {
                    throw new BadRequest('Circular relationships are not allowed');
                }
            }

            break;
        case 'local':
            if (relationship.source.local.path === relationship.target.local.path) {
                throw new BadRequest('Circular relationships are not allowed');
            }
            break;
        case 's3':
            if (relationship.source.s3.provider === relationship.target.s3.provider) {
                if (['sgws', 'ibm'].includes(relationship.source.s3.provider)) {
                    if (relationship.source.s3.host === relationship.target.s3.host) {
                        sourceFullPath = adjustPath(relationship.source.s3.bucket, relationship.source.s3.prefix);
                        targetFullPath = adjustPath(relationship.target.s3.bucket, relationship.target.s3.prefix);
                        if (sourceFullPath.startsWith(targetFullPath) || targetFullPath.startsWith(sourceFullPath)) {
                            throw new BadRequest('Circular relationships are not allowed');
                        }
                    }
                } else {
                    sourceFullPath = adjustPath(relationship.source.s3.bucket, relationship.source.s3.prefix);
                    targetFullPath = adjustPath(relationship.target.s3.bucket, relationship.target.s3.prefix);
                    if (sourceFullPath.startsWith(targetFullPath) || targetFullPath.startsWith(sourceFullPath)) {
                        throw new BadRequest('Circular relationships are not allowed');
                    }
                }
            }
            break;
        case 'azure':
            if (relationship.source.azure.storageAccountName === relationship.target.azure.storageAccountName) {
                if (relationship.source.azure.container === relationship.target.azure.container) {
                    sourceFullPath = adjustPath(relationship.source[protocol].container, relationship.source[protocol].prefix);
                    targetFullPath = adjustPath(relationship.target[protocol].container, relationship.target[protocol].prefix);
                    if (sourceFullPath.startsWith(targetFullPath) || targetFullPath.startsWith(sourceFullPath)) {
                        throw new BadRequest('Circular relationships are not allowed');
                    }
                }
            }

            break;
        case 'gcp':
            sourceFullPath = adjustPath(relationship.source.gcp.bucket, relationship.source.gcp.prefix);
            targetFullPath = adjustPath(relationship.target.gcp.bucket, relationship.target.gcp.prefix);
            if (sourceFullPath.startsWith(targetFullPath) || targetFullPath.startsWith(sourceFullPath)) {
                throw new BadRequest('Circular relationships are not allowed');
            }
            break;
        default:
            logger.debug(`There is no  ${protocol} to ${protocol} relationship`);
    }
}

//useTags indicates whether we use tags or metadata
function validateS3Tags(s3Tags, useTags = true) {
    const maxNumberOfTags = 5;
    if (s3Tags && s3Tags.length) {
        if (s3Tags.length > maxNumberOfTags) {
            throw new BadRequest(`Maximum of ${maxNumberOfTags} S3 tags may be provided`);
        }
        s3Tags.map(({key, value}) => {
            if (!key || !value) {
                throw new BadRequest('Each Amazon tag should contain both key and value');
            }
            if (useTags) {
                if (!S3_TAG_KEY_REGEXP.test(key)) {
                    throw new BadRequest('Amazon tag keys may not start with \'aws:\' and may contain letters, numbers, spaces, or the special characters + - . _ : / @');
                }
                if (!S3_TAG_VALUE_REGEXP.test(value)) {
                    throw new BadRequest('Amazon tag values may not start with \'aws:\' and may contain letters, numbers, spaces, or the special characters + - . _ : / @');
                }
            } else {
                if (!S3_METADATA_NAME_REGEXP.test(key)) {
                    throw new BadRequest('Amazon metadata names may contain letters, numbers or the special characters + - . _ : / @');
                }
                if (!S3_METADATA_VALUE_REGEXP.test(value)) {
                    throw new BadRequest('Amazon metadata values maximum length is 256');
                }
            }
        });
        if (s3Tags.length !== uniq(s3Tags.map(({key}) => key)).length) {
            throw new BadRequest('S3 tags keys should be distinct (no duplicated allowed)');
        }
    }
}

function validateEndPoint(endPoint, srcOrTrg, settings) {
    const {protocol} = endPoint;
    const details = endPoint[protocol];
    switch (protocol) {
        case 'nfs': {
            if (!details) {
                throw new BadRequest('source with protocol nfs should contain object : nfs: {host , export, path}');
            }
            const {
                host, export: exportPath, path, type, version
            } = details;
            assertIpOrHostname(host, 'nfs');
            if (!exportPath && type !== 'efs') {
                throw new BadRequest('source or target with protocol nfs must contain export');
            }
            if (path) {
                validatePath(path, 'nfs');
            }
            if (version && !['3', '4.0', '4.1', '4.2'].includes(version)) {
                throw new BadRequest('nfs version must be one of the following: 3, 4.0, 4.1, 4.2');
            }

            break;
        }
        case 'cifs': {
            if (!details) {
                throw new BadRequest('source with protocol cifs should contain object : cifs: {host , share, path}');
            }
            const {host, share, path, version} = details;
            assertIpOrHostname(host, 'cifs');
            if (!share) {
                throw new BadRequest('Source or target with protocol nfs must contain export');
            }
            if (path) {
                validatePath(path, 'cifs');
            }
            if (version && !['1.0', '2.0', '2.1', '3.0', '3.11'].includes(version)) {
                throw new BadRequest('cifs version must be one of the following: 1.0, 2.0, 2.1, 3.0, 3.11');
            }
            break;
        }
        case 's3': {
            if (!details) {
                throw new BadRequest('source with protocol s3 should contain object : s3: { bucket, prefix}');
            }
            const {
                prefix, bucket, encryption, tags, glacier, deepArchive, provider, host, port
            } = details;
            if (!bucket) {
                throw new BadRequest('source or target with protocol s3 must contain bucket');
            }
            if (prefix && prefix.length > 1024) {
                throw new BadRequest('S3 prefix maximum characters length can be 1024');
            }
            if (encryption) {
                if (!['kms', 'aes256'].includes(encryption.protocol)) {
                    throw new BadRequest('S3 encryption protocol should be kms or aes256');
                }
                if (encryption.protocol === 'kms' && !encryption.keyId) {
                    throw new BadRequest('If S3 encryption protocol is kms you should provide keyId');
                }
            }
            if (glacier) {
                if (srcOrTrg === 'trg') {
                    throw new BadRequest('S3 glacier object is only available for s3 as source endpoint');
                }
                if (!Number.isInteger(glacier.days)) {
                    throw new BadRequest('S3 glacier restore days - number must be a positive integer');
                }
                if (!inRange(glacier.days, 1, 2147483648)) {
                    throw new BadRequest('S3 glacier restore days - number must be a positive integer between 1 and 2147483647');
                }
                if (!isString(glacier.tier)) {
                    throw new BadRequest('S3 glacier restore retrieval option must be a string');
                }
                if (!(glacier.tier === 'Expedited' || glacier.tier === 'Standard' || glacier.tier === 'Bulk')) {
                    throw new BadRequest('S3 glacier restore retrieval option must be one of the following: "Expedited", "Standard", "Bulk"');
                }
            }
            if (deepArchive) {
                if (srcOrTrg === 'trg') {
                    throw new BadRequest('S3 deepArchive object is only available for s3 as source endpoint');
                }
                if (!Number.isInteger(deepArchive.days)) {
                    throw new BadRequest('S3 deep archive restore days - number must be a positive integer');
                }
                if (!inRange(deepArchive.days, 1, 2147483648)) {
                    throw new BadRequest('S3 deep archive restore days - number must be a positive integer between 1 and 2147483647');
                }
                if (!isString(deepArchive.tier)) {
                    throw new BadRequest('S3 deep archive restore retrieval option must be a string');
                }
                if (!(deepArchive.tier === 'Standard' || deepArchive.tier === 'Bulk')) {
                    throw new BadRequest('S3 deep archive restore retrieval option must be one of the following: "Standard", "Bulk"');
                }
            }
            if (tags) {
                validateS3Tags(tags, settings.objectTagging);
            }

            if ([SGWS_PROVIDER, IBM_PROVIDER].includes(provider)) {
                assertIpOrHostname(host, 'sgws');
                if (port) {
                    const portNum = Number.parseInt(port);
                    if (portNum < 0 || portNum > 165535) {
                        throw BadRequest('SGW port should be between 0 and 165535');
                    }
                }
            }
            break;
        }
        case 'azure': {
            const {container, storageAccountName, tags} = details;
            // storageAccountName validation
            if (storageAccountName.length < 3 || storageAccountName.length > 24) {
                // include alphanumeric
                throw new BadRequest('Azure storageAccount name should have between 3 to 24 characters');
            }
            const storageAccountNamePattern = /[^a-z0-9]/;
            if (storageAccountName.match(storageAccountNamePattern)) {
                throw new BadRequest('Azure storageAccount name can contain only lowercase alphanumeric characters');
            }
            // container validation
            // should include dash and alphanumeric
            if (container.length < 3 || container.length > 63) {
                throw new BadRequest('Azure container name should have between 3 to 63 characters');
            }
            const containerPattern = /[^a-z\-0-9]/;
            if (container.match(containerPattern)) {
                throw new BadRequest('Azure container name can contain only lowercase alphanumeric and dash characters');
            }
            if (tags) {
                const tagKeyPattern = /^[^a-zA-Z_]|[^\w]/;
                const tagValuePattern = /[^\x00-\x7F]+/; // eslint-disable-line no-control-regex
                tags.map(({key, value}) => {
                    if (key.match(tagKeyPattern)) {
                        throw new BadRequest('Azure metadata keys may not start with a number, and may contain letters, numbers or the special character _');
                    }
                    if (value.match(tagValuePattern)) {
                        throw new BadRequest('Azure metadata values may contain only ASCII characters');
                    }
                });
            }
            break;
        }
        case 'local': {
            if (!details) {
                throw new BadRequest('source with protocol local should contain object : local: { path }');
            }
            // TODO add validations
            break;
        }
        case 'gcp': {
            if (!details) {
                throw new BadRequest('source with protocol gcp should contain object : gcp: {storage, bucket, prefix}');
            }
            const {bucket, prefix} = details;
            if (!bucket) {
                throw new BadRequest('source or target with protocol gcp must contain bucket');
            }
            if (prefix && prefix.length > 1024) {
                throw new BadRequest('gcp prefix maximum characters length can be 1024');
            }
            break;
        }
        default:
            throw new BadRequest(`Protocol ${protocol} is not allowed`);
    }
}


function validateSyncSchedule(sync) {
    const {
        syncInDays, syncInHours, syncInMinutes, isEnabled, nextTime
    } = sync;
    validateSchema(sync, schedulerSchema);

    if (isEnabled && syncInDays === 0 && syncInHours === 0 && syncInMinutes === 0) {
        throw new BadRequest('Please provide one of \\\'syncInHours\\\' or \\\'syncInDays\\\' or \\\'syncInMinutes \\\' for sync interval\' when pause is false');
    }
    if (syncInHours > 0 && syncInDays > 0 && syncInMinutes > 0) {
        throw new BadRequest('Only one of \'syncInDays\' or \'syncInHours\' or \'syncInMinutes\' can be provided for sync interval');
    }
    if (syncInHours > 0 && syncInDays > 0) {
        throw new BadRequest('Only one of \'syncInDays\' or \'syncInHours\' can be provided for sync interval');
    }
    if (syncInHours > 0 && syncInMinutes > 0) {
        throw new BadRequest('Only one of \'syncInHours\' or \'syncInMinutes\' can be provided for sync interval');
    }
    if (syncInDays > 0 && syncInMinutes > 0) {
        throw new BadRequest('Only one of  \'syncInDays\' or \'syncInMinutes\' can be provided for sync interval');
    }
    if (moment().isAfter(moment(nextTime))) {
        throw new BadRequest('Next sync must be in the future');
    }
}

function validateSettings(settings) {
    validateSchema(settings, settingsSchema);
    if (settings.gracePeriod > 48 * 60 * 60) {
        throw new BadRequest('Grace period should be less than 48 hours');
    }
    if (!(settings.fileTypes.files || settings.fileTypes.directories || settings.fileTypes.symlinks)) {
        throw new BadRequest('Please select at least one of the following file types: files, directories, symlinks.');
    }
    if (settings.files.minSize > settings.files.maxSize) {
        throw new BadRequest('Minimum file size limit can\'t be larger then the maximum file size limit.');
    }
    if (settings.files.maxDate && moment(settings.files.minDate).isAfter(settings.files.maxDate)) {
        throw new BadRequest('Please make sure time range is valid.');
    }
    validateSyncSchedule(settings.schedule);
}

function validateNecessaryCredentials(sourceProtocol, targetProtocol, relationship) {
    let sourceProvider;
    let targetProvider;
    if (sourceProtocol === 's3') {
        sourceProvider = relationship.source.s3.provider;
    }
    if (targetProtocol === 's3') {
        targetProvider = relationship.target.s3.provider;
    }
    if (sourceProvider === 's3' || targetProvider === 's3') {
        if (relationship.group.dataBrokers.every(broker => !broker.awsAccountId)) {
            throw new BadRequest('The broker doesnt have aws credentials to copy to or from S3');
        }
    }
    if (sourceProtocol === 'gcp' || targetProtocol === 'gcp') {
        if (relationship.group.dataBrokers.every(broker => !broker.gcpProjectId)) {
            throw new BadRequest('The broker doesnt have aws credentials to copy to or from GCP');
        }
    }

}

function validateEncryption(initiator, listener) {
    if (initiator.endpointType === listener.endpointType) {
        throw new BadRequest('Initiator and listener data brokers must be located at different endpoints');
    }
}

function validateRelationship(relationship) {
    const {source, target, settings} = relationship;

    if (!isObject(source)) {
        throw new BadRequest('Source parameter should be an object');
    }

    if (!isObject(target)) {
        throw new BadRequest('Target parameter should be an object');
    }

    const {protocol: sourceProtocol} = source;
    const {protocol: targetProtocol} = target;

    if (!Object.keys(ALLOWED_SOURCE_AND_TARGET).includes(sourceProtocol)) {
        throw new BadRequest(`Source type can only be one of: ${Object.keys(ALLOWED_SOURCE_AND_TARGET).join(', ')}`);
    }
    if (!ALLOWED_SOURCE_AND_TARGET[sourceProtocol].includes(targetProtocol)) {
        throw new BadRequest(`Target type can only be one of: ${ALLOWED_SOURCE_AND_TARGET[sourceProtocol].join(', ')}`);
    }

    validateEndPoint(source, 'src', settings);
    validateEndPoint(target, 'trg', settings);
    const circularProtocols = ['nfs', 'cifs', 's3', 'azure', 'gcp'];
    if (sourceProtocol === targetProtocol && circularProtocols.includes(sourceProtocol)) {
        validateDifferentSourceAndTarget(sourceProtocol, relationship);
    }
    if (relationship.encryption) {
        validateEncryption(relationship.encryption.initiator, relationship.encryption.listener);
    }
    validateNecessaryCredentials(sourceProtocol, targetProtocol, relationship);
    // we convert the settings from the data base to a javascript regular object, using toObject().
    // see full explanation at put relationship function.
    const settingsObj = relationship.settings.toObject();
    validateSettings(settingsObj);
    relationship.settings = settingsObj;
}

async function getRelationshipWithActivity(relationshipId) {
    const relationship = await Relationship.findById(relationshipId).populate('account group').populate({
        path: 'group',
        populate: {path: 'dataBrokers'}
    }).exec();
    return createActivity(relationship);
}

async function updateRelationshipData(relationship) {
    const relationshipResponse = await createActivity(relationship);
    await sendRelationshipMessage(relationshipResponse, generateRequestId());
    await updateTimeLineNewDataBroker(relationshipResponse, relationshipResponse.activity.status, findTimelineRecordV2, 'relationship.activity');

}

async function areEmptyQueues(relationshipId, scannerQueue, transferrerQueue, count = 0) {
    const scannerAttributes = await getQueueAttributes(scannerQueue, ['ApproximateNumberOfMessages', 'ApproximateNumberOfMessagesNotVisible']);
    logger.info(`Scanner attribute of relationship ${relationshipId} are ${JSON.stringify(scannerAttributes)}`);

    if (scannerAttributes && scannerAttributes.ApproximateNumberOfMessages === '0' &&
        scannerAttributes.ApproximateNumberOfMessagesNotVisible === '0') {
        await timeout(5000);
        const transferrerAttributes = await getQueueAttributes(transferrerQueue, ['ApproximateNumberOfMessages', 'ApproximateNumberOfMessagesNotVisible']);
        logger.info(`Transferrer attribute of relationship ${relationshipId} are ${JSON.stringify(transferrerAttributes)}`);
        const isTransferrerEmpty = transferrerAttributes && transferrerAttributes.ApproximateNumberOfMessages === '0' && transferrerAttributes.ApproximateNumberOfMessagesNotVisible === '0';
        if (isTransferrerEmpty && !count) {
            //because sometimes we send DONE message too early - to avoide this we wait 10 seconds and check again
            await timeout(20000);
            return areEmptyQueues(relationshipId, scannerQueue, transferrerQueue, count + 1);
        }
        return isTransferrerEmpty;
    }

    return false;
}


async function relationshipsStatusPoller() {
    try {
        //populate account because of status calculation
        const relationships = await Relationship.find({
            scannerQueue: {$exists: true},
            transferrerQueue: {$exists: true},
            startTime: {$lte: (new Date()).getTime() - ms(config.get('relationships-poller-timeout'))}
        })
            .populate('account')
            .populate({path: 'group', populate: {path: 'dataBrokers'}}).exec();

        await Promise.all(relationships.map(async relationship => {
            const {
                account, id: relationshipId, scannerQueue, transferrerQueue
            } = relationship;
            try {
                await updateRelationshipData(relationship);
                const doneRequestId = generateRequestId();
                // must calculate status before checking the queue if master failed to initialize there will always be a scanner message
                const relationshipStatus = await calculateRelationshipStatus(relationshipId, account);
                if (relationshipStatus.status === MESSAGE_STATUS_FAILED) {
                    await handleEndOfRelationship(relationship, doneRequestId, 'done');
                    await updateRelationshipData(relationship);
                } else if (scannerQueue && transferrerQueue && relationshipStatus.status !== MESSAGE_STATUS_ABORTED) {
                    const emptyQueues = await areEmptyQueues(relationshipId, scannerQueue, transferrerQueue);
                    if (emptyQueues) {
                        //find relationship again in case a change was made to group
                        const foundRelationship = await Relationship.findOne({'_id':relationshipId}).populate('account')
                            .populate({path: 'group', populate: {path: 'dataBrokers'}}).exec();
                        // use updated relationship in order to get the updated endTime.
                        const updatedRelationship = await handleEndOfRelationship(foundRelationship, doneRequestId, 'done');
                        await updateStatusForRelationship(updatedRelationship, MESSAGE_STATUS_DONE, undefined, doneRequestId);
                        await updateRelationshipData(updatedRelationship);
                    }
                }
            } catch (err) {
                logger.error('failed checking relationship queues for relationship', relationshipId, err);
            }
        }));
    } catch (err) {
        logger.error('Failed to check relationships queues', err);
    } finally {
        setTimeout(() => relationshipsStatusPoller(), ms(config.get('relationship-poller-interval')));
    }
}


async function validateAbortRelationship(relationshipId) {
    // before aborting- calculating status
    const statusObj = await calculateRelationshipStatus(relationshipId);
    if (statusObj && statusObj.status !== RUNNING_STATUS) {
        logger.debug(`Cannot run Abort for relationship ${relationshipId} of : The relationship is not running`);
        throw new BadRequest('There is no running Sync right now, can not run abort action');
    }
}
async function abortNewDataBroker(relationship, requestId) {
    try {
        await handleEndOfRelationship(relationship, requestId, 'abort');
        await updateStatusForRelationship(relationship, MESSAGE_STATUS_ABORTED, ABORTED_MESSAGE, requestId);
        await updateRelationshipData(relationship);
    } catch (err) {
        logger.error('>>>>>>>ERROR', err);
    }
}

function updateRelationshipFields(endpoint) {
    if (endpoint.s3 && !endpoint.s3.provider) {
        endpoint.s3.provider = 's3';
    }
    if (endpoint.s3 && endpoint.provider === 's3' && !endpoint.s3.storageClass) {
        endpoint.s3.storageClass = 'STANDARD';
    }
    if (endpoint.azure && !endpoint.azure.blobTier) {
        endpoint.azure.blobTier = 'HOT';
    }
    if (endpoint.nfs && !endpoint.nfs.provider) {
        endpoint.nfs.provider = 'nfs';
    }
    if (endpoint.cifs && !endpoint.cifs.provider) {
        endpoint.cifs.provider = 'cifs';
    }
    if (endpoint.sgws && !endpoint.sgws.provider) {
        endpoint.sgws.provider = 'sgws';
    }
}

async function verifyOnlyOne(relationship) {
    const {
        source, target, account, dataBroker, group
    } = relationship;

    const query = {source, target, account};
    if (group) {
        query.group = group;
    } else {
        query.dataBroker = dataBroker;
    }

    const equalRelationship = await Relationship.findOne(query).exec();
    if (equalRelationship) {
        throw new BadRequest('Sync relationship already exists', MONGO_DB_ERROR_DUPLICATE_KEY);
    }
}

async function createRelationshipQueue(queueType, usersName, relationshipId) {
    const userArns = await Promise.all(usersName.map(userName => getUserArn(userName)));
    const queueUrl = await createQueue(`CS_${environment}_${queueType}_${relationshipId}_${Date.now()}`, {MessageRetentionPeriod: '1209600'}); //1209600 seconds equal to 14 days which is the max limit
    const queueArn = await getQueueArn(queueUrl);
    const usersArns = userArns.map(user => `${user}`);
    const policy = {
        Version: '2012-10-17',
        Statement: [
            {
                Sid: `Sid${Date.now()}`,
                Effect: 'Allow',
                Principal: {AWS: usersArns},
                Action: ['SQS:GetQueueAttributes', 'SQS:DeleteMessage', 'SQS:ReceiveMessage', 'SQS:SendMessage', 'SQS:ChangeMessageVisibility'],
                Resource: `${queueArn}`,
                Condition : { Bool : { 'aws:SecureTransport': true }}
            }]
    };


    await setQueuePolicy(queueUrl, policy);
    return queueUrl;
}

async function createPendingRelationship(relationship, requestId, sourceCredentials, targetCredentials) {
    const {
        id: relationshipId
    } = relationship;
    await relationship.save();
    await updateStatusForRelationship(relationship, MESSAGE_STATUS_PENDING, undefined, requestId); //update relationshipStats collection
    const relationshipResponse = await getRelationshipWithActivity(relationshipId);
    await sendRelationshipMessage(relationshipResponse, requestId);
    await updateTimeLineNewDataBroker(relationshipResponse, MESSAGE_STATUS_PENDING, findTimelineRecordV2, 'relationship.activity');
    await sendSaveMetadataCommand(relationship, requestId, sourceCredentials, targetCredentials);
    return relationship;
}

async function createRelationship(relationship, requestId, isSync, sourceCredentials, targetCredentials, customHeaders) {
    const {
        id: relationshipId, group
    } = relationship;
    const usersName = group.dataBrokers.map(dataBroker => getUserAndPolicyName(dataBroker));
    const [scannerQueue, transferrerQueue] = await Promise.all([createRelationshipQueue('scanner', usersName, relationshipId),
        createRelationshipQueue('transferrer', usersName, relationshipId)]);
    relationship.scannerQueue = scannerQueue;
    relationship.transferrerQueue = transferrerQueue;
    relationship.endTime = undefined;
    relationship.startTime = Date.now();
    if (isSync) {
        relationship.phase = 'Sync';
    }

    const updatedRelationship = await relationship.save(); // in order for the relationship to be with the right phase for the simulator;
    await updateStatusForRelationship(updatedRelationship, RUNNING_STATUS, undefined, requestId); // updates timeline if needed - so there will be a timeline to update
    await sendCopyCommand(updatedRelationship, requestId, sourceCredentials, targetCredentials, customHeaders);
    await updateRelationshipData(updatedRelationship);
    return updatedRelationship;
}

async function startNewDataBrokerSync(relationship, requestId, isTimeLine, throwError) {
    const {
        source, target, dataBroker, _id: relationshipId, account, isQstack, isCvo, group
    } = relationship;
    const statusObj = await calculateRelationshipStatus(relationshipId);
    if (!statusObj || statusObj.status === RUNNING_STATUS) {
        logger.info(`Cannot run Sync for relationship ${relationshipId}: There is a running operation right now`);
        if (throwError) {
            throw new BadRequest('A sync action is running right now, can not run another action');
        }
    } else {
        if (isTimeLine && statusObj.status !== MESSAGE_STATUS_PENDING) {
            try {
                const timeline = await new TimelineV2({
                    account,
                    requestId,
                    auth0Id: 'system',
                    userName: 'system',
                    relationship: {
                        source,
                        target,
                        isQstack,
                        isCvo,
                        relationshipId,
                        group: group.id,
                        dataBroker: {dataBrokerId: dataBroker},
                    },
                    dataBroker: {dataBrokerId: dataBroker},
                    group: (group) ? {
                        name: group.name,
                        groupId: group.id,
                        dataBrokers: group.dataBrokers.map(broker => ({dataBrokerId: broker.id, name: broker.name}))
                    } : undefined,
                    summary: 'Scheduled sync'
                }).save();
                await sendAuditGroup(timeline, relationship.account.tenancyAccountId);
            } catch (err) {
                logger.error('Failed to update timeline for scheduled sync of relationship', relationshipId, err);
            }
        }
        await RelationshipStats.deleteMany({relationship: relationshipId}).exec();
        try {
            await createRelationship(relationship, requestId, true);
        } catch (err) {
            // relevant only to sync and scheduled sync- in case create relationship is failed the relationship will be deleted
            // in the catch part of the POST relationship route.
            logger.error(`Failed to sync relationship with id ${relationshipId} error ${err}`);
            // #1 update relationship status with FAILED
            await RelationshipStats.updateMany({relationship: relationshipId}, {
                status: MESSAGE_STATUS_FAILED,
                reason: err.message || 'sync operation failed please try again'
            }, {
                multi: true,
                upsert: true,
                strict: false
            }).exec();
            // #2 send information to the UI
            await updateRelationshipData(relationship);
            throw err;
        }
    }
}

function getNextSyncTime(lastSyncTime, inDays, inHours, inMinutes) {
    let nextSync = moment(lastSyncTime);
    while (nextSync.isBefore()) {
        nextSync = nextSync.add(inMinutes, 'minutes').add(inDays, 'days').add(inHours, 'hours');
    }
    return nextSync.toISOString();
}


module.exports = {
    abortNewDataBroker,
    validateRelationship,
    validateSettings,
    validateSyncSchedule,
    sendRelationshipMessage,
    createActivity,
    getRelationshipWithActivity,
    validateS3Tags,
    relationshipsStatusPoller,
    createRelationship,
    handleEndOfRelationship,
    findRelationshipsForNewDataBrokerStartup,
    calculateProgress,
    updateRelationshipFields,
    startNewDataBrokerSync,
    createPendingRelationship,
    getNextSyncTime,
    verifyOnlyOne,
    createTimeLineObjectForRelationship, validateAbortRelationship
};
