const moment = require('moment');
const express = require('express');
const numeral = require('numeral');
const {extend, sortBy} = require('lodash');
const {Cost} = require('../models/cost');
const {Account} = require('../models/account');
const Relationship = require('../models/relationship');
const DataBroker = require('../models/data-broker');
const TimelineV2 = require('../models/timeline-v2');
const randomize = require('randomatic');
const {isEmptyString} = require('../utils/utils');
const {authenticateNssCredentials, getLicense} = require('../lib/ipa-server');
const {registerForSupport, deregisterSupport, encryptNssPassword, decryptNssPassword} = require('../operations/ipa-opertions');
const requestForAsup = require('../operations/asup-operations');
const {findBytesCopiedForAnAccountInTheLast30Days, findTotalBytesAndFilesCopiedForAnAccount} = require('../operations/timeline-operations');
const {findAccountInfo, countRelationshipsForAccount, cvsAndAnfRelationshipsCount, fanOutAccountSubscription} = require('../operations/account-operations');
const {calculateSumOfBillingInATier, validateMonthAndYear} = require('../operations/cost-operations');
const {API_PATH_ACCOUNTS, SAAS_SUBSCRIPTION_PROVIDER_AWS, SAAS_SUBSCRIPTION_PROVIDER_AZURE, IPA_REGISTERED_STATUS} = require('../utils/consts');
const {BadRequest, NotFound} = require('../utils/errors');
const {getAccounts, createAccount} = require('../lib/tenancy');
const {handleAwsAccountSubscriptionAndTrialExpiry, handleAzureAccountSubscriptionAndTrialExpiry} = require('../operations/account-operations');
const getLogger = require('../utils/logger');

const logger = getLogger();

const router = express.Router();


router.get(`${API_PATH_ACCOUNTS}`, async(req, res, next) => {
    try {
        logger.debug('Getting accounts');
        const tenancyAccounts = await getAccounts();

        if (!tenancyAccounts.length) {
            logger.info('No existing tenancy accounts, creating new account');

            const accountName = `CS_${randomize('0', 5)}`;
            const {accountPublicId} = await createAccount(accountName);
            tenancyAccounts.push({accountPublicId, accountName});
        }

        const accounts = sortBy(tenancyAccounts.map(tenancyAccount => {
            const {accountPublicId, accountName} = tenancyAccount;
            return {accountId: accountPublicId, name: accountName};
        }), 'name');

        res.json(accounts);
    } catch (err) {
        next(err);
    }
});

router.get(`${API_PATH_ACCOUNTS}/info`, async(req, res, next) => {
    try {
        const {headers: {'x-account-id': tenancyAccountId}, query: {noCreate}} = req;
        const account = await Account.findOne({tenancyAccountId}).exec();
        if (account) {
            const accountInfo = await findAccountInfo(account);
            res.json(accountInfo);
        } else {
            if (noCreate) {
                res.status(202).json();
            } else {
                next(new NotFound('Could not find matching account'));
            }
        }
    } catch (err) {
        next(err);
    }
});

router.put(`${API_PATH_ACCOUNTS}/register`, async(req, res, next) => {
    try {
        const {account, body: {mpProvider, mpCustomerToken}} = req;

        logger.info('Register to marketplace for account', {
            accountId: account.tenancyAccountId,
            mpProvider,
            mpCustomerToken
        });

        if (mpProvider === SAAS_SUBSCRIPTION_PROVIDER_AWS) {
            await handleAwsAccountSubscriptionAndTrialExpiry(account, mpCustomerToken);
        } else if (mpProvider === SAAS_SUBSCRIPTION_PROVIDER_AZURE) {
            await handleAzureAccountSubscriptionAndTrialExpiry(account, mpCustomerToken);
        }

        const tenancyAccounts = await getAccounts();
        await fanOutAccountSubscription(account.tenancyAccountId, tenancyAccounts);

        res.json();
    } catch (err) {
        next(err);
    }
});

router.delete(API_PATH_ACCOUNTS, async(req, res, next) => {
// #1 check if account has relationships and data brokers if do throw ERROR
// #2 If the account doesnt have relationships and databrokers delete it from and COST and timelinev2
    const deleteAccountFromCollection = account => Promise.all([Cost, TimelineV2].map(collection => collection.deleteMany({account})));
    const {account: {subscription, tenancyAccountId, id: accountDocumentId}} = req;
    try {
        if (subscription.customerIdentifier) {
            logger.info('Account is subscribed, cannot be deleted');
            next(new BadRequest('Subscribed accounts cannot be deleted'));
        } else {
            const [relationship, dataBroker] = await Promise.all([Relationship.findOne({account: accountDocumentId}).exec(), DataBroker.findOne({account: accountDocumentId}).exec()]);
            if (relationship || dataBroker) {
                logger.info('The user has relationships or data brokers, cannot be deleted');
                next(new BadRequest('User with relationships or data brokers cannot be deleted'));
            } else {
                await deleteAccountFromCollection(accountDocumentId);
                await Account.deleteOne({tenancyAccountId}).exec();
                res.json();
            }
        }
    } catch (err) {
        next(err);
    }
});

router.get(`${API_PATH_ACCOUNTS}/relationships-stats`, async(req, res, next) => {
    const {account:{_id:accountId}, query:{target}} = req;
    try {
        const {totalBytes, totalFiles} = await findTotalBytesAndFilesCopiedForAnAccount(accountId, target);
        const bytesCopiedInTheLast30Days = await findBytesCopiedForAnAccountInTheLast30Days(accountId, target);
        const numberOfRelationships = await countRelationshipsForAccount(accountId, target);
        const result = {
            totalBytes,
            totalFiles,
            numberOfRelationships,
            bytesCopiedInTheLast30Days
        };
        res.json(result);
    } catch (err) {
        next(err);
    }
});

router.put(`${API_PATH_ACCOUNTS}/nss-credentials`, async(req, res, next) => {
    try {
        const {body: {nssUsername, nssPassword}, account} = req;
        logger.debug('Authenticate NSS credentials for tenancyAccountId: ', account.tenancyAccountId);

        if (isEmptyString(nssUsername) || isEmptyString(nssPassword)) {
            next(new BadRequest('Invalid NSS credentials'));
        }

        const responseStatus = await authenticateNssCredentials(nssUsername, nssPassword);
        logger.info('ResponseStatus from authenticate NSS credentials is: ', responseStatus);

        const {statusResp: {statusCode}} = responseStatus;

        if (['E001', 'E999'].includes(statusCode)) {
            next(new BadRequest('Invalid NSS credentials'));
        }

        account.nssCredentials = {
            nssUsername,
            nssPassword: encryptNssPassword(nssPassword)
        };

        await account.save();

        res.json();
    } catch (err) {
        next(err);
    }
});

router.put(`${API_PATH_ACCOUNTS}/licenses`, async(req, res, next) => {
    try {
        const {account, body: {serialNumber}} = req;
        logger.debug('Get licenses for tenancyAccountId: ', account.tenancyAccountId);

        if (isEmptyString(serialNumber)) {
            next(new BadRequest('Invalid serialNumber'));
        }

        if (!account.nssCredentials) {
            next(new BadRequest('Please insert NSS credentials first'));
        }

        const accountWithSerialNumber = await Account.findOne({'byol.licenses.serialNumber': {$eq: serialNumber}}).exec();

        if (accountWithSerialNumber) {
            if (accountWithSerialNumber.id === account.id) {
                next(new BadRequest(`License with serial number: ${serialNumber} is already defined`));
            } else {
                next(new BadRequest(`License with serial number: ${serialNumber} in use by other account than account with id ${account.id}`));
            }
        }

        const licenseResponse = await getLicense(account.nssCredentials.nssUsername, decryptNssPassword(account.nssCredentials.nssPassword), serialNumber);

        if (licenseResponse) {
            const response = licenseResponse.statusResp;
            logger.info('License response: ', licenseResponse);

            if (!response.licenses || !response.licenses.capacity || !response.licenses.endDate) {
                logger.error(`License response for tenancyAccountId: ${account.id} does not contain valid licenses`);
                next(new BadRequest('Invalid serial number'));
            } else {
                if (response.statusCode === 'S007') {
                    const expirationDate = new Date(response.licenses.endDate);

                    account.byol.licenses.push({
                        serialNumber,
                        reservedRelationships: 20,
                        expirationDate
                    });

                    await account.save();
                    const jsonResponse = {
                        serialNumber,
                        reservedRelationships: 20,
                        expirationDate: expirationDate.getTime()
                    };
                    req.timelineV2 = extend(req.timelineV2, {jsonResponse});
                    return res.json(jsonResponse);
                } else {
                    logger.error(`Invalid serial number for the inserted NSS credentials got response with statusCode: ${response.statusCode} for tenancyAccountId: ${account.id}`);
                    next(new BadRequest('Invalid serial number for the inserted NSS credentials'));
                }
            }

        } else {
            logger.error(`failed to find license for nssUsername: ${account.nssCredentials.nssUsername} with serialNumber: ${serialNumber}`);
            next(new BadRequest('Invalid serial number'));
        }
    } catch (err) {
        logger.error('Failed to get license. ', err);
        next(err);
    }
});

router.get(`${API_PATH_ACCOUNTS}/licenses`, async(req, res, next) => {
    const {account} = req;
    logger.debug('Get licenses for tenancyAccountId: ', account.tenancyAccountId);

    if (account) {
        try {
            const accountFromDb = await Account.findById(account.id).exec();
            const {byol: {licenses}} = accountFromDb;
            const licensesWithTimeStamp = licenses.map(license => {
                const {serialNumber, reservedRelationships, expirationDate} = license;
                return {
                    serialNumber,
                    reservedRelationships,
                    expirationDate: expirationDate.getTime()
                };
            });
            return res.json({licenses: licensesWithTimeStamp});
        } catch (err) {
            next(err);
        }
    } else {
        res.json({licenses: []});
    }
});

router.get(`${API_PATH_ACCOUNTS}/license-settings`, async(req, res, next) => {
    try {
        const {account} = req;
        logger.debug('Get licenses for tenancyAccountId: ', account.tenancyAccountId);

        const totalReservedRelationships = account.reservedRelationships();

        const existingRelationships = await countRelationshipsForAccount(account.id);
        const cvsAndAnfRelationships = await cvsAndAnfRelationshipsCount(account.id);
        logger.info(`Number of relationships per account with id ${account.id} is ${existingRelationships}`);

        const freeTrial = Boolean(account.trialExpiry && account.trialExpiry >= new Date());

        let chargedRelationships = existingRelationships - cvsAndAnfRelationships - totalReservedRelationships;
        let byolRelationships = totalReservedRelationships;

        if (chargedRelationships <= 0) {
            byolRelationships = existingRelationships;
            chargedRelationships = 0;
        }

        res.json({
            freeTrial,
            existingRelationships,
            reservedRelationships: byolRelationships,
            totalReservedRelationships,
            chargedRelationships
        });
    } catch (err) {
        next(err);
    }
});

router.put(`${API_PATH_ACCOUNTS}/delete-license`, async(req, res, next) => {
    const {body: {serialNumber}} = req;
    try {
        logger.debug('Delete license with serial number: ', serialNumber);
        await Account.findOneAndUpdate({'byol.licenses.serialNumber': serialNumber},
            {$pull: {'byol.licenses': {serialNumber}}}).exec();

        res.json();
    } catch (err) {
        next(err);
    }
});

router.post(`${API_PATH_ACCOUNTS}/support/register`, async(req, res, next) => {
    try {
        const {account, body: {nssUsername, nssPassword}} = req;
        logger.debug('Received register for support request: ', nssUsername);

        if (isEmptyString(nssUsername) || isEmptyString(nssPassword)) {
            next(new BadRequest('Invalid NSS credentials'));
        }

        const responseStatus = await registerForSupport(account, nssUsername, nssPassword);
        req.timelineV2 = extend(req.timelineV2, {response:responseStatus});

        return res.json(responseStatus);
    } catch (err) {
        logger.error('Failed to register for support. ', err);
        next(err);
    }
});

router.put(`${API_PATH_ACCOUNTS}/support/deregister`, async(req, res, next) => {
    try {
        const {account: {id: accountId, tenancyAccountId}} = req;
        logger.debug('Received de-registration request for tenancyAccountId: ', tenancyAccountId);

        const accountFromDb = await Account.findOne({tenancyAccountId}).exec();

        if (accountFromDb.support.status === IPA_REGISTERED_STATUS) {
            const {support: {serialNumber}} = accountFromDb;
            await deregisterSupport(serialNumber, accountId, tenancyAccountId);

            return res.json();
        } else {
            next(new BadRequest(`Account: ${tenancyAccountId} was not registered for support`));
        }
    } catch (err) {
        logger.error('Failed to deregister from support. ', err);
        next(err);
    }
});

router.post(`${API_PATH_ACCOUNTS}/support/send-asup`, async(req, res, next) => {
    try {
        const {account, id: requestId} = req;
        logger.debug('Received request to sending Data Broker ASUP for tenancyAccountId: ', account.tenancyAccountId);

        if (account.support.status === IPA_REGISTERED_STATUS) {
            logger.debug('Account is registered for support');
            await requestForAsup(account, requestId);

            return res.json();
        } else {
            next(new BadRequest('Account is not registered to support'));
        }
    } catch (err) {
        logger.error('Failed to handle request for sending Data Broker ASUP. ', err);
        next(err);
    }
});

router.get(`${API_PATH_ACCOUNTS}/cost-times`, (req, res) => {
    const {account: {createdAt}} = req;
    const creationDate = moment(createdAt).startOf('month');
    const today = moment().utc();
    const monthsAndYears = [];

    while (today.isAfter(creationDate)) {
        monthsAndYears.unshift({year: creationDate.get('year'), month: creationDate.format('MMMM')});
        creationDate.add(1, 'M');
    }

    res.json(monthsAndYears);
});

router.get(`${API_PATH_ACCOUNTS}/cost`, async(req, res, next) => {
    // Missing most important thing- if the first timestamp is from the middle of the month I need to know the last record from last month and change its from to the start of the month
    const {account, query} = req;
    logger.debug('Received request to get cost for tenancyAccountId: ', account.tenancyAccountId);

    const {year, month} = query;
    validateMonthAndYear(month, year);
    const isCurrentMonthAndYear = (moment().format('MMMM') === month && moment().format('YYYY') === year);
    const currentTimeEpoch = Date.now();
    const endTimeStamp = moment().month(month).year(year).endOf('month')
        .toDate();
    const startTimeStamp = moment().month(month).year(year).startOf('month')
        .toDate();
    const startOfNextMonth = moment().month(month).year(year).add(1, 'M')
        .startOf('month')
        .toDate();
    const numeralObj = numeral();
    let totalCost = 0;
    let hasByol = false;
    let hasTier1 = false;
    let hasTier2 = false;
    let hasTier3 = false;

    function getCostRecord(record, from, to) {
        const numOfHours = moment(to).diff(from, 'hours');
        const tier1Sum = calculateSumOfBillingInATier(record.tier1, numOfHours, 1);
        const tier2Sum = calculateSumOfBillingInATier(record.tier2, numOfHours, 2);
        const tier3Sum = calculateSumOfBillingInATier(record.tier3, numOfHours, 3);
        const recordTotalCost = tier1Sum + tier2Sum + tier3Sum;
        totalCost += recordTotalCost;
        const totalRelationships = (record.tier1 + record.tier2 + record.tier3 + record.byol.usedRelationships);
        hasByol = hasByol || Boolean(record.byol && (record.byol.usedRelationships || record.byol.enabledRelationship));
        hasTier1 = hasTier1 || Boolean(record.tier1);
        hasTier2 = hasTier2 || Boolean(record.tier2);
        hasTier3 = hasTier3 || Boolean(record.tier3);
        return {
            from,
            to,
            hours: numOfHours,
            tier1: {
                numOfRelationships: record.tier1,
                sumOfBilling: numeralObj.set(tier1Sum).format('0.00')
            },
            tier2: {
                numOfRelationships: record.tier2,
                sumOfBilling: numeralObj.set(tier2Sum).format('0.00')
            },
            tier3: {
                numOfRelationships: record.tier3,
                sumOfBilling: numeralObj.set(tier3Sum).format('0.00')
            },
            byol: record.byol,
            totalCost: numeralObj.set(recordTotalCost).format('0.00'),
            totalRelationships
        };
    }

    try {
        const costRecords = await Cost.find({
            account,
            timestamp: {
                $gte: startTimeStamp,
                $lt: endTimeStamp
            }
        }).sort({timestamp: 1}).exec();
        let costArr = costRecords.map((record, indexOfTheArray, array) => {
            const to = (indexOfTheArray === array.length - 1) ? (isCurrentMonthAndYear ? currentTimeEpoch : startOfNextMonth.getTime()) : array[indexOfTheArray + 1].timestamp.getTime();
            const from = record.timestamp.getTime();
            return getCostRecord(record, from, to);
        });

        // need to go over all the costRecords and delete the records with 0 if there are no chargeable relationships
        if (!(hasByol || hasTier1 || hasTier2 || hasTier3)) {
            costArr = costArr.filter(cost => cost.tier1.numOfRelationships || cost.tier2.numOfRelationships || cost.tier3.numOfRelationships || (cost.byol && cost.byol.usedRelationships));
        }
        const response = {
            year,
            month,
            isTier1: hasTier1,
            isTier2: hasTier2,
            isTier3: hasTier3,
            isBYOL: hasByol,
            costArr,
            totalCost: numeralObj.set(totalCost).format('0.00')
        };
        res.json(response);
    } catch (err) {
        next(err);
    }
});


module.exports = router;
