const moment = require('moment');
const {partition, extend} = require('lodash');
const express = require('express');
const Group = require('../models/group');
const {createMessage} = require('../models/message-v2');
const TimelineV2 = require('../models/timeline-v2');
const {API_PATH_GROUPS, MESSAGE_STATUS_FAILED, MESSAGE_STATUS_DONE} = require('../utils/consts');
const {validateObjectId} = require('../utils/common');
const {sendDownloadLogCommand} = require('../operations/data-broker-operations');
const {waitForObject} = require('../operations/group-operations');
const {BadRequest} = require('../utils/errors');
const {sendAuditGroup} = require('../operations/audit-operations');
const getLogger = require('../utils/logger');

const logger = getLogger();

const router = express.Router();


router.get(API_PATH_GROUPS, async(req, res, next) => {
    const {account} = req;
    try {
        logger.debug('Getting groups for tenancyAccountId: ', account.tenancyAccountId);
        const groups = await Group.find({account}).sort({createdAt: 1}).populate('dataBrokers').exec();
        res.json(groups);
    } catch (err) {
        next(err);
    }
});

router.put(`${API_PATH_GROUPS}/:id/download-data-brokers-logs`, async(req, res, next) => {
    const {params: {id: groupId}, id: requestId, account:{tenancyAccountId, id:accountId}} = req;
    try {
        validateObjectId(groupId);
        const group = await Group.findById(groupId).populate('dataBrokers').exec();
        if (group) {
            extend(req.timelineV2, {
                group: {
                    name: group.name,
                    groupId: group.id,
                    dataBrokers: group.dataBrokers.filter(dataBroker => dataBroker.isAlive()).map(broker => ({
                        dataBrokerId: broker.id,
                        name: broker.name
                    }))
                }
            });

            const date = `${moment().format('DD-MMMM-YYYY_hh-mm-ss')}`;
            const dataBrokerLogsName = (dataBrokerId, dataBrokerName, date) => `${dataBrokerName}-${dataBrokerId}-${date}-logs`;
            //sending only to alive data brokers
            const aliveDataBrokersInAgroup = group.dataBrokers.filter(dataBroker => dataBroker.isAlive());

            if (aliveDataBrokersInAgroup.length) {
                await Promise.all(aliveDataBrokersInAgroup.map(async dataBroker => {
                    const {id: dataBrokerId, name: dataBrokerName} = dataBroker;
                    try {
                        const name = dataBrokerLogsName(dataBrokerId, dataBrokerName, date);
                        await sendDownloadLogCommand(requestId, dataBroker, name);
                    } catch (err) {
                        logger.error(`Failed to upload databroker logs for dataBroker with  id ${dataBrokerId}`, err);
                    }
                }));
                res.status(202).json();
                //send UI message
                const downloadObjects = await Promise.all(aliveDataBrokersInAgroup.map(async dataBroker => {
                    const {id: dataBrokerId, name: dataBrokerName} = dataBroker;

                    try {
                        const name = dataBrokerLogsName(dataBrokerId, dataBrokerName, date);
                        await waitForObject(`data-broker-logs/${name}`);
                        return {
                            status: MESSAGE_STATUS_DONE,
                            details: {dataBrokerId, dataBrokerName, url: `https://logs.cloudsync.netapp.com/${name}`}
                        };
                    } catch (err) {
                        logger.error(`Failed to download data-broker logs for dataBroker with  id ${dataBrokerId}`, err);
                        return {status: MESSAGE_STATUS_FAILED, details: {dataBrokerId, dataBrokerName}};
                    }
                }));
                const [success, failed] = partition(downloadObjects, obj => obj.status === MESSAGE_STATUS_DONE);
                //updateTimeline
                const timelineRecord = await TimelineV2.findOne({
                    summary: 'Download data brokers logs',
                    status: 'PROGRESS',
                    'group.groupId': groupId
                }).sort({updatedAt: -1})
                    .limit(1)
                    .exec();
                if (failed.length) {
                    timelineRecord.status = MESSAGE_STATUS_FAILED;
                    timelineRecord.failureMessage = `Failed to download logs of brokers ${failed.map(obj => obj.details.dataBrokerName).join(',')}`;
                } else {
                    timelineRecord.status = MESSAGE_STATUS_DONE;
                }

                await createMessage(accountId,  requestId, undefined, {'update-data-broker-logs': {logsDetails: success.map(obj => obj.details)}});
                await timelineRecord.save();
                await sendAuditGroup(timelineRecord, tenancyAccountId,
                    undefined);

            } else {
                next(new BadRequest(`All data brokers : ${group.dataBrokers.map(dataBroker => dataBroker.name).join(',')}  are offline`));
            }
        } else {
            next(new BadRequest(`Group with id ${groupId} was not found`));
        }
    } catch (err) {
        next(err);
    }
});


module.exports = router;
