const Joi = require('@hapi/joi');
const {
    SERVER_HOST, ALIBABA_REGIONS,
    ASTERIX_EXPIRY_DATE, AWS_REGIONS, ALIBABA_PROVIDER, S3_PROVIDER, MP_PRODUCT_CODE_CONTRACT, MP_PRODUCT_CODE_METERING, PAYGO_SUBSCRIPTION, CONTRACT_SUBSCRIPTION
} = require('./consts');
const { BadRequest } = require('../utils/errors');

const [environment] = SERVER_HOST.split('.');

function assertIpOrHostname(ipOrHostname, protocol) {
    if (!ipOrHostname) {
        throw new BadRequest(`source or target with protocol ${protocol} must contain ip or host`);
    }
    const hostNameSchema = Joi.string().hostname();
    const ipSchema = Joi.string().ip();
    const uriSchema = Joi.string().uri();
    const { error: ipHostError } = ipSchema.validate(ipOrHostname);
    const { error: hostNameError } = hostNameSchema.validate(ipOrHostname);
    const { error: uriError } = uriSchema.validate(ipOrHostname);
    if (ipHostError && hostNameError && uriError) {
        throw new BadRequest(`Invalid ${protocol} ip or hostname ${ipOrHostname}`);
    }
}

function isNewDataBroker(dataBroker) {
    const { name, createdAt } = dataBroker;
    return (name.startsWith('phoenix') && new Date(createdAt) < new Date(ASTERIX_EXPIRY_DATE)) || new Date(createdAt) > new Date(ASTERIX_EXPIRY_DATE);
}

function getUserAndPolicyName(dataBroker) {
    const {externalId} = dataBroker;
    if (externalId) {
        return `CS_${environment}_${dataBroker.externalId}`;
    }
    return `CS_${environment}_${dataBroker.id}`;
}

function s3RegionCodeToDisplayName(region, provider = S3_PROVIDER) {
    if (provider === ALIBABA_PROVIDER) {
        return ALIBABA_REGIONS[region] || region;
    } else {
        return AWS_REGIONS[region] || region;
    }
}

function validateObjectId(id) {
    const objectIdSchema = Joi.string().regex(/^[0-9a-fA-F]{24}$/);
    const { error: objectIdError } = objectIdSchema.validate(id);
    if (objectIdError) {
        throw new BadRequest('Please enter a valid objectId');
    }
}

// returns the AWS Subscription type by the product code
function validateMarketplaceProductCode(code) {
    if (code === MP_PRODUCT_CODE_METERING) {
        return PAYGO_SUBSCRIPTION;
    }

    if (code === MP_PRODUCT_CODE_CONTRACT) {
        return CONTRACT_SUBSCRIPTION;
    }

    throw new BadRequest(`Resolved product code is incorrect. Expected: ${MP_PRODUCT_CODE_METERING} or ${MP_PRODUCT_CODE_CONTRACT}, Got: ${code}`);
}


module.exports = {
    assertIpOrHostname, getUserAndPolicyName, isNewDataBroker, validateObjectId,
    s3RegionCodeToDisplayName, validateMarketplaceProductCode
};
