const {
    get, isObject, flattenDeep
} = require('lodash');
const DataBroker = require('../models/data-broker');
const Group = require('../models/group');
const {Account} = require('../models/account');
const {Message2} = require('../models/message-v2');
const Relationship = require('../models/relationship');
const {RelationshipStats} = require('../models/relationship-statistics');
const TimelineV2 = require('../models/timeline-v2');
const {getStatusesQueueName, sendStartupCommand, sendMetadataRequestCommand, deleteFileLinkIfNeeded} = require('../operations/data-broker-operations');
const {receiveMessage, deleteMessage, createQueue, getQueueUrl, getQueueArn, setQueuePolicy} = require('../lib/aws/sqs');
const {generateRequestId, timeout} = require('../utils/utils');
const {
    MESSAGE_STATUS_DONE, GCP_REGIONS, BUCKETS_MESSAGE_TYPE_BY_PROVIDER, BUCKET_FOLDERS_MESSAGE_TYPE_BY_PROVIDER,
} = require('../utils/consts');
const {findRelationshipsWithSecretsInAGroup} = require('../operations/group-operations');
const {s3RegionCodeToDisplayName} = require('../utils/common');
const {sendAuditGroup} = require('../operations/audit-operations');
const {updateDataBrokerWithVersion} = require('./update-data-broker');
const getAwsAccountId = require('../lib/aws/sts');

const getLogger = require('../utils/logger');

const logger = getLogger();


async function registerDataBrokerPlacement(dataBrokerId, dataBrokerType, placement, dataBrokerOperation) {
    const {certificate, ...rest} = placement;
    const update = {$set:{...rest,  'encryption.certificate': certificate}};
    dataBrokerOperation.push({
        updateOne: {
            filter: {'_id': dataBrokerId},
            update
        }
    });
}

async function generatePlacementTimelineIfNeeded(dataBroker, placement, isNewBroker, requestId, timelineOperations)  {
    let summary;
    const {
        version: dataBrokerVersion, account, name, id: dataBrokerId
    } = dataBroker;
    const {version: placementVersion} = placement;
    if (isNewBroker) {
        logger.info(`${dataBroker} created and registered`);
        summary = 'New Data Broker created and registered';
    } else if (dataBrokerVersion !== placement.version) {
        logger.info(`${dataBrokerId} version updated ${dataBrokerVersion} => ${placementVersion}`);
        summary = `Data-Broker version updated ${dataBrokerVersion} => ${placementVersion}`;
    }
    if (summary) {
        const dataBrokerObject = {dataBrokerId, name};
        const timelineObject = {
            status: MESSAGE_STATUS_DONE,
            requestId,
            createdAt: Date.now(),
            updatedAt:Date.now(),
            account,
            summary,
            auth0Id: 'system',
            userName: 'system',
            dataBroker: dataBrokerObject,
        };
        timelineOperations.push({
            updateOne: {
                filter: {
                    account,
                    requestId,
                },
                update: {
                    $set: timelineObject
                },
                upsert: true
            }
        });
        const accountFromDB = await Account.findById(account).exec();
        // send to kinesis audit
        await sendAuditGroup(timelineObject, accountFromDB.tenancyAccountId, dataBrokerObject);
    }
}

function createMessageOperation(account, dataBrokerId, requestId, type, value) {
    return {
        insertOne: {
            document: {
                expiration: Date.now() + 5000,
                milliSeconds: Date.now(),
                account,
                payload: {
                    header: {
                        requestId, agentId: dataBrokerId, clientId:account
                    },
                    payload: {
                        [type]: value

                    }
                }
            }
        }
    };
}

async function handleMessage(dataBroker, message, bulkOperations) {
    const {statusesQueueUrl, _id: dataBrokerId, account} = dataBroker;
    const {body, attributes: {requestId:requestIdFromBroker, type: commandToHandle}} = message;
    const requestId = requestIdFromBroker || generateRequestId();
    const {
        messages: {operations: messagesOperations}, relationshipsStats:
            {operations: relationshipsStatsBulkOperations}, timelines: {operations: timelineOperations},
        dataBrokers: {operations: dataBrokerOperations}
    } = bulkOperations;
    logger.debug(`Received message from queue '${statusesQueueUrl}' of ${dataBroker}:`, JSON.stringify(message));
    // parse message body
    const msg = JSON.parse(body);
    // handle message
    const isStatusSuccess = msg && msg.status === 'SUCCEEDED';

    switch (commandToHandle) {
        case 'startup': {
            // we will always get the same data broker and an account in a message do we can update the relaitionshipsIds at once
            // only relevant for failed messages
            const {relationships, status, reason} = msg;
            const relationshipsIds = relationships.map(relationship => relationship.relationshipId);
            relationshipsStatsBulkOperations.push({
                updateMany: {
                    filter: {
                        relationship: {$in: relationshipsIds},
                        dataBroker: dataBrokerId,
                        account
                    },
                    update: {$set: {status, reason}},
                    upsert: true
                }
            });
        }
            break;
        case 'health': {
            const {transferRate} = msg;
            dataBrokerOperations.push({
                updateOne: {
                    filter: {'_id': dataBrokerId},
                    update: {$set: {transferRate: transferRate}},
                }
            });
        }
            break;
        case 'placement': {
            const placementInfo = msg;
            const isNewBroker = !(dataBroker.platform);
            await deleteFileLinkIfNeeded(dataBroker, isNewBroker);
            await generatePlacementTimelineIfNeeded(dataBroker, placementInfo, isNewBroker, requestId, timelineOperations);
            await registerDataBrokerPlacement(dataBroker, dataBroker.type, placementInfo, dataBrokerOperations);
            const {groupId} = dataBroker;
            const numberOfRelationshipsWithPasswordInAGroup = await findRelationshipsWithSecretsInAGroup(groupId);
            if (numberOfRelationshipsWithPasswordInAGroup > 0) {
                const group = await Group.findById(groupId).populate('dataBrokers').exec();
                const {dataBrokers} = group;
                if (dataBrokers.length > 1) {
                    const aliveDataBrokerInAGroup = dataBrokers.find(dataBroker => dataBroker.isAlive() && dataBroker.id.toString() !== dataBrokerId.toString());
                    if (aliveDataBrokerInAGroup) {
                        logger.info(`There is at least one alive data broker in ${group._id.toString()}`);
                        await sendMetadataRequestCommand(generateRequestId(), aliveDataBrokerInAGroup, dataBrokerId);
                    } else {
                        await sendStartupCommand(generateRequestId(), dataBroker);
                        await updateDataBrokerWithVersion(dataBrokerId);

                    }
                }
            } else {
                await sendStartupCommand(generateRequestId(), dataBroker);
                await updateDataBrokerWithVersion(dataBrokerId);
            }
            msg.region = s3RegionCodeToDisplayName(msg.region);
            messagesOperations.push(createMessageOperation(account, dataBrokerId, requestId, 'data-broker-status', {
                id: dataBrokerId,
                status: 'COMPLETE',
                placement:placementInfo,
                isNewBroker
            }));
        }
            break;
        case 'metadata': {
            const {targetDataBrokerId, metadata} = msg;
            // receive metadata response -
            const targetDataBroker = await DataBroker.findById(targetDataBrokerId).exec();

            if (!targetDataBroker) {
                logger.error(`DataBroker with id ${targetDataBrokerId} was not found cant send metadata-response message `);
            } else if (isStatusSuccess) {
                // logger.info(`Received metadata message ${JSON.stringify(msg)} from ${dataBroker.id}, sending startup command`);
                // sending the message to the target metadata
                await sendStartupCommand(generateRequestId(), targetDataBroker, metadata);
            } else {
                logger.error('Failed to send metadata to databroker', targetDataBrokerId, targetDataBroker, 'status success', isStatusSuccess);
            }
        }
            break;
        case 'azure-blob-storage-containers':
            messagesOperations.push(createMessageOperation(account, dataBrokerId, requestId, 'list-azure-blob-storage-containers', msg));
            break;
        case 'azure-blob-storage-container-folders':
            messagesOperations.push(createMessageOperation(account, dataBrokerId, requestId,
                'list-azure-blob-storage-container-folders', msg));
            break;
        case 'gcp-buckets': {
            const {buckets, ...rest} = msg;
            messagesOperations.push(createMessageOperation(account, dataBrokerId, requestId, 'list-gcp-buckets', {
                ...rest, buckets: (buckets || []).map(bucket => {
                    const {region} = bucket;
                    bucket.region = {
                        name: region,
                        displayName: GCP_REGIONS[region] || region
                    };
                    return bucket;
                })
            }));
        }
            break;
        case 'gcp-bucket-folders': {
            const {bucket: {name: bucket}, folders: list, ...rest} = msg;
            messagesOperations.push(createMessageOperation(account, dataBrokerId, requestId, 'list-gcp-bucket-folders', {
                list, ...rest, bucket
            }));
        }
            break;
        case 'shares': {
            const {shares: paths, ...rest} = msg;
            messagesOperations.push(createMessageOperation(account, dataBrokerId, requestId, 'list-cifs-shares',
                {paths, ...rest}));
        }
            break;
        case 'share-folders': {
            messagesOperations.push(createMessageOperation(account, dataBrokerId, requestId, 'list-cifs-share-folders',
                msg));
        }
            break;
        case 'exports': {
            const {exports: paths, ...rest} = msg;
            messagesOperations.push(createMessageOperation(account, dataBrokerId, requestId, 'list-nfs-exports',
                {paths, ...rest}));
        }
            break;
        case 'export-folders': {
            const {export: exportName, path: rawPath, ...rest} = msg;
            const path = rawPath ? `${exportName}${rawPath}` : exportName;
            messagesOperations.push(createMessageOperation(account, dataBrokerId, requestId, 'list-nfs-export-folders',
                {path, ...rest}));
        }
            break;
        case 's3-buckets': {
            const {buckets, provider, ...rest} = msg;
            messagesOperations.push(createMessageOperation(account, dataBrokerId, requestId, BUCKETS_MESSAGE_TYPE_BY_PROVIDER[provider], {
                ...rest, buckets: (buckets || []).map(bucket => {
                    const {region} = bucket;
                    if (region) {
                        bucket.region = {
                            name: region,
                            displayName: s3RegionCodeToDisplayName(region, provider)
                        };
                    }
                    return bucket;
                })
            }));
        }
            break;
        case 's3-bucket-folders': {
            const {
                provider, folders: list,
                bucket: {name: bucket}, ...rest
            } = msg;
            messagesOperations.push(createMessageOperation(account, dataBrokerId, requestId, BUCKET_FOLDERS_MESSAGE_TYPE_BY_PROVIDER[provider], {
                ...rest,
                bucket,
                list
            }));
        }
            break;
        case 'kms-keys' : {
            messagesOperations.push(createMessageOperation(account, dataBrokerId, requestId, 'list-kms-keys', msg));
        }
            break;
        case 'abort':
        case 'copy': {
            const {relationshipId} = msg;
            // logger.info('Received massage', msg, 'for relationshipId', relationshipId);
            const isStatusMessage = Boolean(msg.status);
            const relationship = await Relationship.findById(relationshipId).exec();
            if (!relationship) {
                logger.warn(`Got  relationship message for relationship id '${relationshipId}' from ${dataBroker._id} but relationship ${relationship} not found (might have been deleted`);
            } else {
                const filter = {
                    relationship: relationshipId,
                    dataBroker: dataBrokerId,
                    account
                };

                let update;
                if (isStatusMessage) {
                    const {status, reason} = msg;
                    update = {$set: {status, reason: reason || undefined}};
                } else {
                    const incUpdateObj = {};
                    const setObj = {};

                    const findAllKeys = (accumString, obj) => {
                        if (isObject(obj)) {
                            return Object.keys(obj).map(key => {
                                const nextString = (accumString === '') ? `${key}` : `${accumString}.${key}`;
                                return findAllKeys(nextString, obj[key]);
                            });
                        }
                        return accumString;
                    };

                    const allKeys = flattenDeep(findAllKeys('', msg));
                    allKeys.forEach(key => {
                        if (key !== 'relationshipId') {
                            if (key.endsWith('.description')) {
                                setObj[key.toLowerCase()] = get(msg, key);
                            } else {
                                incUpdateObj[key.toLowerCase()] = get(msg, key);
                            }
                        }
                    });
                    update = {$inc: incUpdateObj, $set: setObj};
                }
                relationshipsStatsBulkOperations.push({
                    updateOne: {
                        filter,
                        update,
                        upsert: true,
                        timestamps: true
                    }
                });
            }
        }
            break;
        default:
            logger.debug(`Unknown command '${commandToHandle}' while processing message from SQS queue '${statusesQueueUrl}' of ${dataBroker}`, msg);
    }
}

async function pollQueue(dataBroker) {
    logger.debug('Polling dataBroker - old implementation', dataBroker.id);
    const {id, statusesQueueUrl} = dataBroker;

    try {
        logger.debug(`Polling messages from SQS queue '${statusesQueueUrl}' of ${dataBroker}`);
        const messages = await receiveMessage(statusesQueueUrl);
        if (messages.length > 0) {
            // retrieve data broker and update data broker last ping in one shot
            try {
                const updatedDataBroker = await DataBroker.findByIdAndUpdate(id, {$set: {'lastPing': new Date()}}, {new: true}).exec();
                const bulkOperations = {
                    messages: {operations: [], model: Message2},
                    relationshipsStats: {operations: [], model: RelationshipStats},
                    dataBrokers: {operations: [], model: DataBroker},
                    timelines: {operations: [], model: TimelineV2}
                };
                await Promise.all(messages.map(async message => {
                    try {
                        await handleMessage(updatedDataBroker, message, bulkOperations);
                    } catch (err) {
                        logger.error(`Failed handling SQS message from queue ${statusesQueueUrl} of ${dataBroker}`, {
                            message,
                            err
                        });
                    }
                }));

                await Promise.all(Object.values(bulkOperations).map(async object => {
                    const {operations, model} = object;
                    try {
                        await model.bulkWrite(operations);
                    } catch (err) {
                        logger.error('Failed to write operations', operations, 'to model', err);
                    }
                }));

            } finally {
                messages.forEach(message => deleteMessage(statusesQueueUrl, message.receiptHandle));
            }
        }

        pollQueue(dataBroker);
    } catch (err) {
        if (err.code === 'AWS.SimpleQueueService.NonExistentQueue') {
            logger.warn(`'${statusesQueueUrl}' SQS queue of ${dataBroker} does not exist anymore. Will not continue polling`);
        } else {
            logger.warn(`Failed to poll message - delaying polling for SQS queue '${statusesQueueUrl}' of ${dataBroker._id}`, err);
            setTimeout(() => pollQueue(dataBroker), 5000);
        }
    }
}

async function startListeningDataBrokersStatuses() {
    logger.info('Looking for status queues...');
    const statusesQueueUrl = await getQueueUrl(getStatusesQueueName());
    logger.info('StatusesQueueUrl', statusesQueueUrl);
    const databrokers = await DataBroker.find({statusesQueueUrl: {$ne:statusesQueueUrl}}).exec();
    logger.info('dataBrokers to listen to', databrokers);
    databrokers.forEach(dataBroker => pollQueue(dataBroker));
}

async function pollStatusesQueue(queueUrl) {
    try {
        const messages = await receiveMessage(queueUrl);
        const receiveTime = Date.now();
        if (messages.length > 0) {
            // retrieve data broker and update data broker last ping in one shot
            logger.info(`received ${messages.length} on ${receiveTime}`);
            try {
                const bulkOperations = {
                    messages: {operations: [], model: Message2},
                    relationshipsStats: {operations: [], model: RelationshipStats},
                    dataBrokers: {operations: [], model: DataBroker},
                    timelines: {operations: [], model: TimelineV2}
                };

                await Promise.all(messages.map(async message => {
                    try {
                        const {attributes:{dataBrokerId, SenderId:userId}} = message;
                        const dataBroker = await DataBroker.findOneAndUpdate({'_id':dataBrokerId, userId}, {$set: {'lastPing': new Date()}}, {new: true}).exec();
                        if (dataBroker) {
                            await handleMessage(dataBroker, message, bulkOperations);
                        } else {
                            logger.error(`Received message from databrokerId ${dataBrokerId} and userId ${userId} - databrokreId and userId dont match`, JSON.stringify(message));
                        }
                    } catch (err) {
                        logger.error('Failed handling SQS message from queue', {message, err});
                    }
                }));

                await Promise.all(Object.values(bulkOperations).map(async object => {
                    const {operations, model} = object;
                    try {
                        await model.bulkWrite(operations);
                    } catch (err) {
                        logger.error('Failed to write operations', operations, 'to model', err);
                    }
                }));

                logger.info('Done handling messages took', Date.now() - receiveTime);
            } finally {
                await Promise.all(messages.map(message => deleteMessage(queueUrl, message.receiptHandle)));
            }
        }

        pollStatusesQueue(queueUrl);
    } catch (err) {
        // TODO - what do we do in case the queue is deleted??
        logger.error(`Failed to poll message - delaying polling for SQS queue '${queueUrl}'`, err);
        setTimeout(() => pollStatusesQueue(queueUrl), 5000);

    }
}


async function startListeningStatusesQueue() {
    //#1 create statuses queue
    //#2 start listening to the queue
    try {
        const name = getStatusesQueueName();
        logger.info('Creating statuses queue', name);
        const queueUrl = await createQueue(name);
        logger.info(` statuses queue ${queueUrl} was created`);
        const queueArn = await getQueueArn(queueUrl);
        logger.info('Statuses queue arn', queueArn);
        const accountId = await getAwsAccountId();
        await timeout(10000);
        const policy = {
            Version: '2012-10-17',
            Statement: [
                {
                    Sid: `Sid${Date.now()}`,
                    Effect: 'Allow',
                    Principal: {AWS: [`${accountId}`]},
                    Action: ['SQS:SendMessage'],
                    Resource: `${queueArn}`,
                    Condition : { Bool : { 'aws:SecureTransport': true }}
                }]
        };
        await setQueuePolicy(queueUrl, policy);
        logger.info('Queue successfully created:', queueUrl);
        pollStatusesQueue(queueUrl);
    } catch (err) {
        logger.error('Failed to create statuses sqs', err);
    }
}

module.exports = {startListeningStatusesQueue, startListeningDataBrokersStatuses, pollQueue};
