const fs = require('fs');
const path = require('path');
const {
    getCustomerTiers, billSaaSAccounts, byolDetailsForUnsubscribedAccounts, filterAndInsert, addAccountsWithoutRelationshipsToMongoObject
} = require('../../src/lib/aws/marketplace-aws');
const DataBroker = require('../../src/models/data-broker');
const Relationship = require('../../src/models/relationship');
const {Account} = require('../../src/models/account');
const {Cost} = require('../../src/models/cost');
const { expect } = require('../globals');

const AWS = require('mock-aws');

describe('marketplace-aws library', () => {
    describe('getCustomerTiers', () => {
        it('Tiers should be partitioned according to the number of relationships (only tier1 relationships)', () => {
            const result = getCustomerTiers(4, 0);
            result.tier1.should.equal(4);
            result.tier2.should.equal(0);
            result.tier3.should.equal(0);
        });

        it('Tiers should be partitioned according to the number of relationships (only tier1 and tier2 relationships)', () => {
            const result = getCustomerTiers(18, 0);
            result.tier1.should.equal(5);
            result.tier2.should.equal(13);
            result.tier3.should.equal(0);
        });

        it('Tiers should be partitioned according to the number of relationships (all tiers relationships)', () => {
            const result = getCustomerTiers(23, 0);
            result.tier1.should.equal(5);
            result.tier2.should.equal(15);
            result.tier3.should.equal(3);
        });

        it('Tiers should be partitioned according to the number of relationships (tiers and BYOL relationships)', () => {
            const result = getCustomerTiers(23, 20);
            result.tier1.should.equal(0);
            result.tier2.should.equal(0);
            result.tier3.should.equal(3);
        });

        it('Tiers should be partitioned according to the number of relationships  BYOL relationships)', () => {
            const result = getCustomerTiers(23, 40);
            result.tier1.should.equal(0);
            result.tier2.should.equal(0);
            result.tier3.should.equal(0);
        });
    });

    describe('marketplace function with mongoDb actions', () => {
        const generateMeteringId = () => Math.floor((1 + Math.random()) * 0x10000).toString(8);

        function createProcessedAwsResponse(customerIdentifier, dimension, quantyity, awsTimestamp, status) {
            return {
                MeteringRecordId: generateMeteringId(),
                Status: status,
                UsageRecord: {
                    CustomerIdentifier: customerIdentifier,
                    Dimension: dimension,
                    Quantity: quantyity,
                    Timestamp: awsTimestamp
                }
            };
        }

        function createUnprocessedAwsResponse(customerIdentifier, dimension, quantyity, awsTimestamp) {
            return {
                CustomerIdentifier: customerIdentifier,
                Dimension: dimension,
                Quantity: quantyity,
                Timestamp: awsTimestamp
            };
        }

        AWS.mock('MarketplaceMetering', 'batchMeterUsage', params => ({
            promise: () => {
                const testTimestamp = new Date().toISOString();
                const customerIdObj = {};
                // building results from the requested records themselves and adding some mixed data
                const usageRecords = params.UsageRecords;
                let results;
                let unproccessedRecords;
                if (usageRecords.length > 1) {
                    results = usageRecords.slice(0, usageRecords.length - 1);
                    unproccessedRecords = [createUnprocessedAwsResponse(
                        usageRecords[usageRecords.length - 1].CustomerIdentifier,
                        usageRecords[usageRecords.length - 1].Dimension, usageRecords[usageRecords.length - 1].Quantity, testTimestamp
                    )];
                } else {
                    results = usageRecords;
                    unproccessedRecords = [];
                }
                results = params.UsageRecords.map(record => {
                    if (customerIdObj[record.CustomerIdentifier]) {
                        if (customerIdObj[record.CustomerIdentifier][record.Dimension]) {
                            return createProcessedAwsResponse(record.CustomerIdentifier, record.Dimension, record.Quantity, testTimestamp, 'DuplicateRecord');
                        }
                        customerIdObj[record.CustomerIdentifier][record.Dimension] = true;
                        return createProcessedAwsResponse(record.CustomerIdentifier, record.Dimension, record.Quantity, testTimestamp, 'Success');
                    }
                    customerIdObj[record.CustomerIdentifier] = { [record.Dimension]: true };
                    return createProcessedAwsResponse(record.CustomerIdentifier, record.Dimension, record.Quantity, testTimestamp, 'Success');
                });
                results.push({
                    MeteringRecordId: generateMeteringId(),
                    Status: 'DuplicateRecord',
                    UsageRecord: {
                        CustomerIdentifier: params.UsageRecords[0].CustomerIdentifier,
                        Dimension: params.UsageRecords[0].Dimension,
                        Quantity: params.UsageRecords[0].Quantity,
                        Timestamp: testTimestamp
                    }
                });

                results.push({
                    MeteringRecordId: generateMeteringId(),
                    Status: 'CustomerNotSubscribed',
                    UsageRecord: {
                        CustomerIdentifier: 'unsubscrpibedUserId'
                    }
                });
                unproccessedRecords.push(createUnprocessedAwsResponse('someBandoned', 'Tier1', 5, testTimestamp));
                return {
                    Results: results,
                    UnprocessedRecords: unproccessedRecords
                };
            }
        }));

        const accounts = JSON.parse(fs.readFileSync(path.resolve('./test/resources', 'accounts.json'), 'utf8'));
        const relationships = JSON.parse(fs.readFileSync(path.resolve('./test/resources', 'relationships.json'), 'utf8'));
        const dataBrokers = JSON.parse(fs.readFileSync(path.resolve('./test/resources', 'dataBrokers.json'), 'utf8'));


        before(async() => {
            await Promise.all([Account.deleteMany(), Relationship.deleteMany(), DataBroker.deleteMany()]);
            const accountWithUnExpiredTrial = accounts.find(account => account._id === '58a40e987200292bd0f942f8');
            const date = new Date();
            accountWithUnExpiredTrial.trialExpiry = date.setMonth(date.getMonth() + 12);
            return Account.insertMany(accounts)
                .then(() => DataBroker.insertMany(dataBrokers))
                .then(() => Relationship.insertMany(relationships));
        });

        it('billSaaSAccounts- should return one account with tiers and no BYOL', () => {
            const chargingTime = new Date().toISOString();
            return billSaaSAccounts(chargingTime)
                .then(obj => {
                    expect(obj).to.be.an('object');
                    expect(obj['58a40e987200292bd0f942f8']).to.be.an('undefined');
                    expect(obj['58a40eba7200292bd0f942fa']).to.be.an('undefined');
                    expect(obj['58a40edb7200292bd0f942fc']).to.be.an('undefined');
                    obj['58a40efb7200292bd0f942fd'].tier1.should.equal(5);
                    obj['58a40efb7200292bd0f942fd'].tier2.should.equal(15);
                    obj['58a40efb7200292bd0f942fd'].tier3.should.equal(4);
                });
        });

        it('byolDetailsForUnsubscribedAccounts- should return 3 accounts with BYOL object- 1 with 0 relationships and 2 with 1 relationship', () => {
            const chargingTime = new Date().toISOString();
            return byolDetailsForUnsubscribedAccounts({}, chargingTime)
                .then(obj => {
                    expect(obj).to.be.an('object');
                    expect(obj['58a40e987200292bd0f942f8']).to.be.an('undefined');
                    obj['58a40efb7200292bd0f942fd'].byol.reservedRelationships.should.equal(0);
                    obj['58a40efb7200292bd0f942fd'].byol.usedRelationships.should.equal(0);
                    obj['58a40edb7200292bd0f942fc'].byol.reservedRelationships.should.equal(20);
                    obj['58a40edb7200292bd0f942fc'].byol.usedRelationships.should.equal(1);
                    obj['58a40eba7200292bd0f942fa'].byol.reservedRelationships.should.equal(20);
                    obj['58a40eba7200292bd0f942fa'].byol.usedRelationships.should.equal(1);
                });
        });

        it('Agenda flow- Cost collection should have records for each account, should have 1 account with 0 relationships and 5 others with relationships', () => {
            const chargingTime = new Date().toISOString();
            return billSaaSAccounts(chargingTime)
                .then(mongoObjects => byolDetailsForUnsubscribedAccounts(mongoObjects, chargingTime))
                .then(mongoObjects => addAccountsWithoutRelationshipsToMongoObject(mongoObjects, chargingTime))
                .then(mongoObjects => filterAndInsert(mongoObjects))
                .then(() => Cost.find({}).exec())
                .then(costRecords => {
                    const costForExpiredaccountWithByol = costRecords.find(cost => cost.account.toString() === '58a40edb7200292bd0f942fc');
                    costForExpiredaccountWithByol.tier1.should.equal(0);
                    costForExpiredaccountWithByol.tier2.should.equal(0);
                    costForExpiredaccountWithByol.tier3.should.equal(0);
                    costForExpiredaccountWithByol.byol.usedRelationships.should.equal(1);
                    costForExpiredaccountWithByol.byol.reservedRelationships.should.equal(20);
                    const costForSubscribedNotNetappaccount = costRecords.find(cost => cost.account.toString() === '58a40efb7200292bd0f942fd');
                    costForSubscribedNotNetappaccount.tier1.should.equal(5);
                    costForSubscribedNotNetappaccount.tier2.should.equal(15);
                    costForSubscribedNotNetappaccount.tier3.should.equal(4);
                    costForSubscribedNotNetappaccount.byol.usedRelationships.should.equal(0);
                    costForSubscribedNotNetappaccount.byol.reservedRelationships.should.equal(0);
                    const costForSubscribedNetappaccount = costRecords.find(cost => cost.account.toString() === '58a40eba7200292bd0f942fa');
                    costForSubscribedNetappaccount.tier1.should.equal(0);
                    costForSubscribedNetappaccount.tier2.should.equal(0);
                    costForSubscribedNetappaccount.tier3.should.equal(0);
                    costForSubscribedNetappaccount.byol.usedRelationships.should.equal(1);
                    costForSubscribedNetappaccount.byol.reservedRelationships.should.equal(20);
                    const costForaccountWithFreeTrial = costRecords.find(cost => cost.account.toString() === '58a40e987200292bd0f942f8');
                    costForaccountWithFreeTrial.tier1.should.equal(0);
                    costForaccountWithFreeTrial.tier2.should.equal(0);
                    costForaccountWithFreeTrial.tier3.should.equal(0);
                    costForaccountWithFreeTrial.byol.usedRelationships.should.equal(0);
                    costForaccountWithFreeTrial.byol.reservedRelationships.should.equal(0);
                    const costForaccountWithSubscriptionAndBYOL = costRecords.find(cost => cost.account.toString() === '58a4610c740ffd7cc870e5fb');
                    costForaccountWithSubscriptionAndBYOL.tier1.should.equal(0);
                    costForaccountWithSubscriptionAndBYOL.tier2.should.equal(0);
                    costForaccountWithSubscriptionAndBYOL.tier3.should.equal(4);
                    costForaccountWithSubscriptionAndBYOL.byol.usedRelationships.should.equal(20);
                    costForaccountWithSubscriptionAndBYOL.byol.reservedRelationships.should.equal(20);
                    const costForSubscribeaccountWithExistingCustomerIdentifierAndNoRelationships = costRecords.find(cost => cost.account.toString() === '58a47410bee5ae664080cc26');
                    costForSubscribeaccountWithExistingCustomerIdentifierAndNoRelationships.tier1.should.equal(0);
                    costForSubscribeaccountWithExistingCustomerIdentifierAndNoRelationships.tier2.should.equal(0);
                    costForSubscribeaccountWithExistingCustomerIdentifierAndNoRelationships.tier3.should.equal(4);
                    costForSubscribeaccountWithExistingCustomerIdentifierAndNoRelationships.byol.usedRelationships.should.equal(20);
                    costForSubscribeaccountWithExistingCustomerIdentifierAndNoRelationships.byol.reservedRelationships.should.equal(20);
                });
        });
    });
});
