const {Account} = require('../../src/models/account');
const DataBroker = require('../../src/models/data-broker');
const {should} = require('../globals');

describe('DataBroker Model', () => {
    let account;
    const name = 'DM1';
    const instanceId = 'i-122345656';
    const privateIp = '10.0.0.1';
    const region = 'us-west-1';
    const availabilityZone = 'az';
    const vpcId = 'vpc-1234567';
    const vpcCidrBlock = 'vpcCidrBlock';
    const subnetId = 'sub-1234567';
    const subnetCidrBlock = 'subnetCidrBlock';
    const lastPing = new Date();

    beforeEach(async() => {
        await Promise.all([Account.deleteMany(), DataBroker.deleteMany()]);
        account = await Account.create({
            name: 'User Name',
            trialExpiry: new Date(),
            tenancyAccountId: '1',
            support: {
                serialNumber: '12345',
                status: 'unregistered'
            }
        });
    });

    describe('CRUD', () => {
        it('create', async() => {
            const dataBroker = await DataBroker.create({account, name});
            should.exist(dataBroker);
            dataBroker.account.id.should.equal(account.id);
            dataBroker.name.should.equal(name);
            should.not.exist(dataBroker.instanceId);
            should.not.exist(dataBroker.privateIp);
            should.not.exist(dataBroker.region);
            should.not.exist(dataBroker.availabilityZone);
            should.not.exist(dataBroker.vpcId);
            should.not.exist(dataBroker.vpcCidrBlock);
            should.not.exist(dataBroker.subnetId);
            should.not.exist(dataBroker.subnetCidrBlock);
            should.exist(dataBroker.createdAt);
            should.exist(dataBroker.updatedAt);
        });
    });

    it('create with optional fields', async() => {
        const dataBroker = await DataBroker.create({
            account,
            name,
            instanceId,
            privateIp,
            region,
            availabilityZone,
            vpcId,
            vpcCidrBlock,
            subnetId,
            subnetCidrBlock,
            lastPing
        });
        should.exist(dataBroker);
        dataBroker.account.id.should.equal(account.id);
        dataBroker.name.should.equal(name);
        dataBroker.instanceId.should.equal(instanceId);
        dataBroker.privateIp.should.equal(privateIp);
        dataBroker.region.should.equal(region);
        dataBroker.availabilityZone.should.equal(availabilityZone);
        dataBroker.vpcId.should.equal(vpcId);
        dataBroker.vpcCidrBlock.should.equal(vpcCidrBlock);
        dataBroker.subnetId.should.equal(subnetId);
        dataBroker.subnetCidrBlock.should.equal(subnetCidrBlock);
        dataBroker.lastPing.should.equalTime(lastPing);
        should.exist(dataBroker.createdAt);
        should.exist(dataBroker.updatedAt);
    });

    it('get', async() => {
        const dataBroker = await DataBroker.create({
            account,
            name,
            instanceId,
            privateIp,
            region,
            availabilityZone,
            vpcId,
            vpcCidrBlock,
            subnetId,
            subnetCidrBlock,
            lastPing
        });
        const foundDataBroker = await DataBroker.findById(dataBroker.id).populate('account');
        should.exist(foundDataBroker);
        foundDataBroker.account.id.should.equal(account.id);
        foundDataBroker.name.should.equal(name);
        foundDataBroker.instanceId.should.equal(instanceId);
        foundDataBroker.privateIp.should.equal(privateIp);
        foundDataBroker.region.should.equal(region);
        foundDataBroker.availabilityZone.should.equal(availabilityZone);
        foundDataBroker.vpcId.should.equal(vpcId);
        foundDataBroker.vpcCidrBlock.should.equal(vpcCidrBlock);
        foundDataBroker.subnetId.should.equal(subnetId);
        foundDataBroker.subnetCidrBlock.should.equal(subnetCidrBlock);
        foundDataBroker.lastPing.should.equalTime(lastPing);
        should.exist(foundDataBroker.createdAt);
        should.exist(foundDataBroker.updatedAt);
    });

    it('update', async() => {
        const updatedRegion = 'us-west-2';
        const dataBroker = await DataBroker.create({
            account,
            name,
            instanceId,
            privateIp,
            region,
            availabilityZone,
            vpcId,
            vpcCidrBlock,
            subnetId,
            subnetCidrBlock,
            lastPing
        });
        dataBroker.region = updatedRegion;
        await dataBroker.save();
        const foundDataBroker = await DataBroker.findById(dataBroker.id).populate('account');
        should.exist(foundDataBroker);
        foundDataBroker.account.id.should.equal(account.id);
        foundDataBroker.name.should.equal(name);
        foundDataBroker.instanceId.should.equal(instanceId);
        foundDataBroker.privateIp.should.equal(privateIp);
        foundDataBroker.region.should.equal(updatedRegion);
        foundDataBroker.availabilityZone.should.equal(availabilityZone);
        foundDataBroker.vpcId.should.equal(vpcId);
        foundDataBroker.vpcCidrBlock.should.equal(vpcCidrBlock);
        foundDataBroker.subnetId.should.equal(subnetId);
        foundDataBroker.subnetCidrBlock.should.equal(subnetCidrBlock);
        foundDataBroker.lastPing.should.equalTime(lastPing);
        should.exist(foundDataBroker.createdAt);
        should.exist(foundDataBroker.updatedAt);
    });

    it('add optional fields', async() => {
        const dataBroker = await DataBroker.create({
            account,
            name,
            instanceId,
            privateIp,
            region,
            availabilityZone,
            vpcId,
            vpcCidrBlock,
            subnetId,
            subnetCidrBlock
        });
        dataBroker.lastPing = lastPing;
        await dataBroker.save();
        const foundDataBroker = await DataBroker.findOne({_id: dataBroker.id}).populate('account');
        should.exist(foundDataBroker);
        foundDataBroker.account.id.should.equal(account.id);
        foundDataBroker.name.should.equal(name);
        foundDataBroker.instanceId.should.equal(instanceId);
        foundDataBroker.privateIp.should.equal(privateIp);
        foundDataBroker.region.should.equal(region);
        foundDataBroker.availabilityZone.should.equal(availabilityZone);
        foundDataBroker.vpcId.should.equal(vpcId);
        foundDataBroker.vpcCidrBlock.should.equal(vpcCidrBlock);
        foundDataBroker.subnetId.should.equal(subnetId);
        foundDataBroker.subnetCidrBlock.should.equal(subnetCidrBlock);
        foundDataBroker.lastPing.should.equalTime(lastPing);
        should.exist(foundDataBroker.createdAt);
        should.exist(foundDataBroker.updatedAt);
    });

    it('delete', async() => {
        const dataBroker = await DataBroker.create({
            account,
            name,
            instanceId,
            privateIp,
            region,
            availabilityZone,
            vpcId,
            vpcCidrBlock,
            subnetId,
            subnetCidrBlock,
            lastPing
        });
        await DataBroker.deleteOne({_id: dataBroker.id});
        DataBroker.findOne({account}).should.eventually.not.exist;
    });
});
