const {should} = require('../globals');
const {Account} = require('../../src/models/account');
const Relationship = require('../../src/models/relationship');
const DataBroker = require('../../src/models/data-broker');
const Group = require('../../src/models/group');

describe('Relationship Model', () => {
    let account;
    let dataBroker;
    const nfsServer = 'source';
    const source = {
        protocol: 'nfs',
        nfs: {
            host: nfsServer,
            export: '/'
        }
    };
    const target = {
        protocol: 's3',
        s3: {
            bucket: 'bucket',
            prefix: 'folder'
        }
    };

    const dataBrokerName = 'dataBrokerName';

    const schedule = {
        syncInDays: 0, syncInHours: 1, syncInMinutes: 0, nextTime: new Date(), isEnabled: true
    };
    const settings = {gracePeriod: 30, schedule};

    beforeEach(async() => {
        await Promise.all([Account.deleteMany(), Relationship.deleteMany(), DataBroker.deleteMany()]);
        account = await Account.create({
            name: 'User Name',
            trialExpiry: new Date(),
            tenancyAccountId: '1',
            support: {
                serialNumber: '12345',
                status: 'unregistered'
            }
        });
        dataBroker = await DataBroker.create({account, name: dataBrokerName, group: 'aa'});
        await Group.create({account, name: 'aa', dataBrokers: [dataBroker.id]});
    });

    describe('CRUD', () => {
        it('create', async() => {
            const relationship = await Relationship.create({
                account, dataBroker, source, target, settings
            });
            should.exist(relationship);
            relationship.account.id.should.equal(account.id);
            relationship.dataBroker.id.should.equal(dataBroker.id);
            relationship.source.protocol.should.equal('nfs');
            relationship.source.nfs.host.should.equal(nfsServer);
            relationship.source.nfs.export.should.equal('/');
            relationship.target.protocol.should.equal('s3');
            relationship.target.s3.bucket.should.equal('bucket');
            relationship.target.s3.prefix.should.equal('folder');
            relationship.settings.schedule.isEnabled.should.equal(true);
            should.exist(relationship.createdAt);
            should.exist(relationship.updatedAt);
        });

        it('get', async() => {
            const relationship = await Relationship.create({
                account, dataBroker, source, target, settings
            });
            const foundRelationship = await Relationship.findById(relationship.id).populate('account');
            should.exist(foundRelationship);
            foundRelationship.account.id.should.equal(account.id);
            foundRelationship.source.protocol.should.equal('nfs');
            foundRelationship.source.nfs.host.should.equal(nfsServer);
            foundRelationship.source.nfs.export.should.equal('/');
            foundRelationship.target.protocol.should.equal('s3');
            foundRelationship.target.s3.bucket.should.equal('bucket');
            foundRelationship.target.s3.prefix.should.equal('folder');
            foundRelationship.settings.schedule.isEnabled.should.equal(true);
            should.exist(foundRelationship.createdAt);
            should.exist(foundRelationship.updatedAt);
        });
    });

    it('update', async() => {
        const updatedNfsServer = 'fs1.netapp.com';
        const updatedSource = {
            protocol: 'nfs',
            nfs: {
                host: updatedNfsServer, export: 'export1', path: 'folder2'
            }
        };
        const relationship = await Relationship.create({
            account, dataBroker, source, target, settings
        });
        relationship.source = updatedSource;
        await relationship.save();
        const foundRelationship = await Relationship.findById(relationship.id).populate('account');
        should.exist(foundRelationship);
        foundRelationship.account.id.should.equal(account.id);
        foundRelationship.source.protocol.should.equal('nfs');
        foundRelationship.source.nfs.host.should.equal(updatedNfsServer);
        foundRelationship.source.nfs.export.should.equal('export1');
        foundRelationship.source.nfs.path.should.equal('folder2');
        foundRelationship.target.protocol.should.equal('s3');
        foundRelationship.target.s3.bucket.should.equal('bucket');
        foundRelationship.target.s3.prefix.should.equal('folder');
        foundRelationship.settings.schedule.isEnabled.should.equal(true);
        should.exist(foundRelationship.createdAt);
        should.exist(foundRelationship.updatedAt);
    });

    it('delete', async() => {
        const relationship = await Relationship.create({
            account, dataBroker, source, target, settings
        });
        await Relationship.deleteOne({_id: relationship.id});
        return Relationship.findOne({account}).should.eventually.not.exist;
    });

    describe('Constraints', () => {
        it('no duplicate relationship with same parameters', async() => {
            await Relationship.create({
                account, dataBroker, source, target,
                settings
            });
            await Relationship.create({
                account, nfsServer, dataBroker, source, target, settings
            }).should.eventually.be.rejected;
        });
    });

    describe('JSON representation', () => {
        const sourceString = {protocol: 'nfs', nfs: {host: 'source', export: '/'}};
        const targetString = {protocol: 's3', s3: {bucket: 'bucket', prefix: 'folder'}};

        it('successfully transform', async() => {

            const relationship = await Relationship.create({
                account, dataBroker, source, target, settings
            });
            const json = relationship.toJSON();
            should.exist(json);
            should.exist(json.dataBroker);
            json.source.should.deep.equal(sourceString);
            json.target.should.deep.equal(targetString);
            json.settings.schedule.isEnabled.should.equal(true);
            should.not.exist(json._id);
            should.not.exist(json.__v);
            should.exist(json.createdAt);
            should.not.exist(json.updatedAt);
        });
    });
});
