const {repeat} = require('lodash');
const {
    validateRelationship, validateS3Tags, validateSyncSchedule, calculateProgress
} = require('../../src/operations/relationship-operations');
const Relationship = require('../../src/models/relationship');
const DataBroker = require('../../src/models/data-broker');
const {Account} = require('../../src/models/account');
const Group = require('../../src/models/group');
const {RUNNING_STATUS, MESSAGE_STATUS_DONE, MESSAGE_STATUS_FAILED} = require('../../src/utils/consts');
const {expect} = require('../globals');
const {BadRequest} = require('../../src/utils/errors');

let account;
let dataBroker;
let group;
const autoSync = 'OFF';

const dataBrokerName = 'dataBroker'; // TODO need to change tests
const sync = {
    inDays: 0, inHours: 1, inMinutes: 0, nextTime: new Date(), autoSync
};
const settings = {gracePeriod: 30};

describe('relationships-operations', () => {
    beforeEach(async() => {
        await Promise.all([Account.deleteMany(), Relationship.deleteMany(), DataBroker.deleteMany()]);
        account = await Account.create({
            name: 'User Name',
            trialExpiry: new Date(),
            tenancyAccountId: '1',
            support: {
                serialNumber: '12345',
                status: 'unregistered'
            }
        });
        group = await Group.create({name: 'group', account});
        const dataBroker = await DataBroker.create({
            awsAccountId: '282316784512',
            group, account, name: dataBrokerName, createdAt: 1531402152
        });
        group.dataBrokers = [dataBroker];
        await group.save();
    });


    describe('Relationship validations', () => {
        function validateRelationshipDoesNotThrow(source, target) {

            const relationship = new Relationship({
                source, target, sync, settings, dataBroker,
                group
            });
            expect(validateRelationship.bind(this, relationship)).not.to.throw();
        }

        async function validateRelationshipThrows(expectedError, source, target) {
            const relationship = new Relationship({
                source, target, sync, settings, dataBroker, group
            });
            expect(validateRelationship.bind(this, relationship)).to.throw(BadRequest, expectedError);
        }

        describe('validate source & target URLs and type combinations asterix', () => {
            it('validates invalid source type asterix', async() => {
                await validateRelationshipThrows(/Source type can only be one of:/, {
                    protocol: 'xyz',
                    xyz: {host: 'dksjhf', export: 'kslf'}
                }, {protocol: 'xyz', xyz: {host: 'dksjhf', export: 'kslf'}}, true);
            });


            it('validates invalid CIFS source', async() => {
                await validateRelationshipThrows('source or target with protocol cifs must contain ip or host', {
                    protocol: 'cifs',
                    cifs: {share: 'share'}
                }, {protocol: 's3', s3: {bucket: 'kslf', provider: 'sgws'}}, true);
            });

            it('validates invalid NFS source', async() => {
                await validateRelationshipThrows('source or target with protocol nfs must contain ip or host', {
                    protocol: 'nfs',
                    nfs: {export: 'export'}
                }, {protocol: 's3', s3: {bucket: 'kslf', provider: 'sgws'}}, true);
            });

            it('validates invalid S3 source', async() => {
                await validateRelationshipThrows('source or target with protocol s3 must contain bucket', {
                    protocol: 's3',
                    s3: {}
                }, {protocol: 'nfs', nfs: {host: '1.1.1.1', export: 'export'}}, true);
            });

            it('validates invalid target format', async() => {
                await validateRelationshipThrows('source or target with protocol s3 must contain bucket', {
                    protocol: 'nfs',
                    nfs: {host: '1.1.1.1', export: 'export'}
                }, {protocol: 's3', s3: {}}, true);
            });

            it('validates invalid target type', async() => {
                await validateRelationshipThrows(/Target type can only be one of:/, {
                    protocol: 's3',
                    s3: {bucket: 'kslf'}
                }, {protocol: 'xyz', xyz: {host: 'dksjhf', export: 'kslf'}});
            });

            it('validates limited target types for CIFS source', async() => {
                await validateRelationshipThrows('Target parameter should be an object', {
                    protocol: 'cifs',
                    cifs: {share: 'share'}
                }, 'xyz://dksjhf/kslf');
            });

            it('validates S3 target for CIFS source', async() => {
                await validateRelationshipDoesNotThrow({
                    protocol: 'cifs',
                    cifs: {host: '1.1.1.1', share: '/'}
                }, {protocol: 's3', s3: {bucket: 'some-bucket'}});
            });

            it('validates SGWS target for CIFS source', async() => {
                await validateRelationshipDoesNotThrow({
                    protocol: 'cifs',
                    cifs: {host: '1.1.1.1', share: '/'}
                }, {protocol: 's3', s3: {host: 'hostname', bucket: 'bucket', provider: 'sgws'}});
            });

            it('validates limited target types for NFS source', async() => {
                await validateRelationshipThrows('Target type can only be one of: s3, nfs, local, azure, cifs, sgws, gcp', {
                    protocol: 'nfs',
                    nfs: {host: '1.1.1.1', export: 'export'}
                }, {protocol: 'xyz', xyz: {host: 'dksjhf', export: 'kslf'}}, true);
            });

            it('validates S3 target for NFS source', async() => {
                await validateRelationshipDoesNotThrow({
                    protocol: 'nfs',
                    nfs: {host: '1.1.1.1', export: '/'}
                }, {protocol: 's3', s3: {bucket: 'some-bucket'}}, true);
            });

            it('validates SGWS target for NFS source', async() => {
                await validateRelationshipDoesNotThrow({
                    protocol: 'nfs',
                    nfs: {host: '1.1.1.1', export: '/'}
                }, {protocol: 's3', s3: {host: '1.1.1.1', bucket: 'some-bucket', provider: 'sgws'}}, true);
            });

            it('validates limited target types for S3 source', async() => {
                await validateRelationshipThrows('Target type can only be one of: nfs, local, cifs, s3, sgws', {
                    protocol: 's3',
                    s3: {bucket: 'some-bucket'}
                }, {protocol: 'xyz', xyz: {host: 'dksjhf', export: 'kslf'}});
            });

            it('validates NFS target for S3 source', () => {
                validateRelationshipDoesNotThrow({protocol: 's3', s3: {bucket: 'some-bucket'}}, {
                    protocol: 'nfs',
                    nfs: {host: '1.1.1.1', export: '/'}
                }, true);
            });

            it('validates CIFS target for S3 source', () => {
                validateRelationshipDoesNotThrow({protocol: 's3', s3: {bucket: 'some-bucket'}}, {
                    protocol: 'cifs',
                    cifs: {host: '1.1.1.1', share: '/'}
                }, true);
            });
        });

        describe('validate source & target URLs and type combinations regular dataBroker', () => {
            it('validates invalid source type', async() => {
                await validateRelationshipThrows(/Source type can only be one of:/, {
                    protocol: 'xyz',
                    xyz: {host: 'dksjhf', export: 'kslf'}
                }, {protocol: 'xyz', xyz: {host: 'dksjhf', export: 'kslf'}});
            });

            it('validates invalid SGWS source', async() => {
                await validateRelationshipThrows('source or target with protocol sgws must contain ip or host', {
                    protocol: 's3',
                    s3: {host: '1.2.3.4', bucket: 'kslf', provide: 'sgws'}
                }, {protocol: 's3', s3: {bucket: 'kslf', provider: 'sgws'}});
            });

            it('validates invalid CIFS source', async() => {
                await validateRelationshipThrows('source or target with protocol cifs must contain ip or host', {
                    protocol: 'cifs',
                    cifs: {share: 'share'}
                }, {protocol: 's3', s3: {bucket: 'kslf'}});
            });

            it('validates invalid NFS source', async() => {
                await validateRelationshipThrows('source or target with protocol nfs must contain ip or host', {
                    protocol: 'nfs',
                    nfs: {export: 'export'}
                }, {protocol: 's3', s3: {bucket: 'kslf'}});
            });

            it('validates invalid S3 source', async() => {
                await validateRelationshipThrows('source or target with protocol s3 must contain bucket', {
                    protocol: 's3',
                    s3: {}
                }, {protocol: 'nfs', nfs: {host: '1.1.1.1', export: 'export'}});
            });

            it('validates invalid target format', async() => {
                await validateRelationshipThrows('source or target with protocol s3 must contain bucket', {
                    protocol: 'nfs',
                    nfs: {host: '1.1.1.1', export: 'export'}
                }, {protocol: 's3', s3: {}});
            });

            it('validates invalid target type', async() => {
                await validateRelationshipThrows(/Target type can only be one of:/, {
                    protocol: 's3',
                    s3: {bucket: 'kslf', provider: 'sgws'}
                }, {protocol: 'xyz', xyz: {host: 'dksjhf', export: 'kslf'}});
            });

            it('validates limited target types for CIFS source', async() => {
                await validateRelationshipThrows('Target parameter should be an object', {
                    protocol: 'cifs',
                    cifs: {share: 'share'}
                }, 'xyz://dksjhf/kslf');
            });

            it('validates S3 target for CIFS source', async() => {
                await validateRelationshipDoesNotThrow({
                    protocol: 'cifs',
                    cifs: {host: '1.1.1.1', share: '/'}
                }, {protocol: 's3', s3: {bucket: 'some-bucket'}});
            });


            it('validates limited target types for NFS source ', async() => {
                await validateRelationshipThrows('Target type can only be one of: s3, nfs, local, azure, cifs, sgws, gcp', {
                    protocol: 'nfs',
                    nfs: {host: '1.1.1.1', export: 'export'}
                }, {protocol: 'xyz', xyz: {host: 'dksjhf', export: 'kslf'}});
            });

            it('validates S3 target for NFS source', async() => {
                await validateRelationshipDoesNotThrow({
                    protocol: 'nfs',
                    nfs: {host: '1.1.1.1', export: '/'}
                }, {protocol: 's3', s3: {bucket: 'some-bucket'}});
            });


            it('validates limited target types for S3 source', async() => {
                await validateRelationshipThrows('Target type can only be one of: nfs, local, cifs, s3, sgws', {
                    protocol: 's3',
                    s3: {bucket: 'some-bucket'}
                }, {protocol: 'xyz', xyz: {host: 'dksjhf', export: 'kslf'}});
            });

            it('validates NFS target for S3 source', async() => {
                await validateRelationshipDoesNotThrow({protocol: 's3', s3: {bucket: 'some-bucket'}}, {
                    protocol: 'nfs',
                    nfs: {host: '1.1.1.1', export: '/'}
                });
            });

            it('validates CIFS target for S3 source', async() => {
                await validateRelationshipDoesNotThrow({protocol: 's3', s3: {bucket: 'some-bucket'}}, {
                    protocol: 'cifs',
                    cifs: {host: '1.1.1.1', share: '/'}
                });
            });
        });

        describe('validate URLs IP/Hostname', () => {
            it('validates invalid source ip for CIFS source', () => {
                validateRelationshipThrows(/ip or hostname/, {
                    protocol: 'cifs',
                    cifs: {host: '1.', share: '/'}
                }, {protocol: 's3', s3: {bucket: 'some-bucket'}});
            });

            it('validates valid source ip for CIFS source', () => {
                validateRelationshipDoesNotThrow({
                    protocol: 'cifs',
                    cifs: {host: '1.1.1.1', share: '/'}
                }, {protocol: 's3', s3: {bucket: 'some-bucket'}});
            });

            it('validates invalid source hostname for CIFS source', () => {
                validateRelationshipThrows(/ip or hostname/, {
                    protocol: 'cifs',
                    cifs: {host: 'some-host.', share: '/'}
                }, {protocol: 's3', s3: {bucket: 'some-bucket'}});
            });

            it('validates valid source hostname for CIFS source', () => {
                validateRelationshipDoesNotThrow({
                    protocol: 'cifs',
                    cifs: {host: 'host-1.local.example.com', share: '/'}
                }, {protocol: 's3', s3: {bucket: 'some-bucket'}});
            });

            it('validates invalid source ip for NFS source', () => {
                validateRelationshipThrows(/ip or hostname/, {
                    protocol: 'nfs',
                    nfs: {host: '1.1.', export: '/'}
                }, {protocol: 's3', s3: {bucket: 'some-bucket'}});
            });

            it('validates valid source ip for NFS source', () => {
                validateRelationshipDoesNotThrow({
                    protocol: 'nfs',
                    nfs: {host: '231.233.234.235', export: '/'}
                }, {protocol: 's3', s3: {bucket: 'some-bucket'}});
            });

            it('validates invalid source hostname for NFS source', () => {
                validateRelationshipThrows(/ip or hostname/, {
                    protocol: 'nfs',
                    nfs: {host: 'host,', export: '/'}
                }, {protocol: 's3', s3: {bucket: 'some-bucket'}});
            });

            it('validates valid source hostname for NFS source', () => {
                validateRelationshipDoesNotThrow({
                    protocol: 'nfs',
                    nfs: {host: 'host-name.local', export: '/'}
                }, {protocol: 's3', s3: {bucket: 'some-bucket'}});
            });

            it('validates invalid target ip for CIFS source', () => {
                validateRelationshipThrows(/ip or hostname/, {
                    protocol: 's3',
                    s3: {bucket: 'some-bucket'}
                }, {protocol: 'cifs', cifs: {host: '1.', share: '/'}});
            });

            it('validates valid target ip for CIFS source', () => {
                validateRelationshipDoesNotThrow({protocol: 's3', s3: {bucket: 'some-bucket'}}, {
                    protocol: 'cifs',
                    cifs: {host: '1.1.1.1', share: '/'}
                });
            });

            it('validates invalid target hostname for CIFS source', () => {
                validateRelationshipThrows(/ip or hostname/, {
                    protocol: 's3',
                    s3: {bucket: 'some-bucket'}
                }, {protocol: 'cifs', cifs: {host: 'some-host.', share: '/'}});
            });

            it('validates valid target hostname for CIFS source', () => {
                validateRelationshipDoesNotThrow({protocol: 's3', s3: {bucket: 'some-bucket'}}, {
                    protocol: 'cifs',
                    cifs: {host: 'host-1.local.example.com', share: '/'}
                });
            });

            it('validates invalid target ip for NFS source', () => {
                validateRelationshipThrows(/ip or hostname/, {
                    protocol: 's3',
                    s3: {bucket: 'some-bucket'}
                }, {protocol: 'nfs', nfs: {host: '1.1.', export: '/'}});
            });

            it('validates valid target ip for NFS source', () => {
                validateRelationshipDoesNotThrow({protocol: 's3', s3: {bucket: 'some-bucket'}}, {
                    protocol: 'nfs',
                    nfs: {host: '231.233.234.235', export: '/'}
                });
            });

            it('validates invalid target hostname for NFS source', () => {
                validateRelationshipThrows(/ip or hostname/, {
                    protocol: 's3',
                    s3: {bucket: 'some-bucket'}
                }, {protocol: 'nfs', nfs: {host: 'host,', export: '/'}});
            });

            it('validates valid target hostname for NFS source', () => {
                validateRelationshipDoesNotThrow({protocol: 's3', s3: {bucket: 'some-bucket'}}, {
                    protocol: 'nfs',
                    nfs: {host: 'host-name.local', export: '/'}
                });
            });

            it('validates invalid target ip for SGWS source', () => {
                validateRelationshipThrows(/ip or hostname/, {
                    protocol: 'cifs',
                    cifs: {host: '1.1.1.1', share: '/'}
                }, {protocol: 's3', s3: {provider: 'sgws', host: '1.1.', port: '8080', bucket: 'bucket'}}, true);
            });

            it('validates valid target ip for SGWS source', () => {
                validateRelationshipDoesNotThrow({
                    protocol: 'cifs',
                    cifs: {host: '1.1.1.1', share: '/'}
                }, {
                    protocol: 's3',
                    s3: {provideR: 'sgws', host: '231.233.234.235', port: '8910', bucket: 'bucket'}
                }, true);
            });

            it('validates invalid target hostname for SGWS source', () => {
                validateRelationshipThrows(/ip or hostname/, {
                    protocol: 'nfs',
                    nfs: {host: 'host', export: '/'}
                }, {protocol: 's3', s3: {provider: 'sgws', host: 'host,', port: '8080', bucket: 'bucket'}}, true);
            });

            it('validates valid target hostname for SGWS source', () => {
                validateRelationshipDoesNotThrow({protocol: 'nfs', nfs: {host: 'host', export: '/'}}, {
                    protocol: 's3',
                    s3: {provider: 'sgws', host: 'host-name.example.com', port: '8080', bucket: 'bucket'}
                }, true);
            });
        });

        describe('validate path components', () => {
            it('validates invalid source CIFS path is \'/.\'', () => {
                validateRelationshipThrows('path cannot include \'.\' (current folder)', {
                    protocol: 'cifs',
                    cifs: {host: 'host', share: 'share', path: '/.'}
                }, {protocol: 's3', s3: {bucket: 'some-bucket'}});
            });

            it('validates invalid target CIFS path is \'/.\'', () => {
                validateRelationshipThrows('path cannot include \'.\' (current folder)', {
                    protocol: 's3',
                    s3: {bucket: 'some-bucket'}
                }, {protocol: 'cifs', cifs: {host: 'host', share: 'share', path: '/.'}});
            });

            it('validates invalid source CIFS path ends with \'/.\'', () => {
                validateRelationshipThrows('path cannot include \'.\' (current folder)', {
                    protocol: 'cifs',
                    cifs: {host: 'host', share: 'share', path: 'folder1/folder2/.'}
                }, {protocol: 's3', s3: {bucket: 'some-bucket'}});
            });

            it('validates invalid target CIFS path ends with \'/.\'', () => {
                validateRelationshipThrows('path cannot include \'.\' (current folder)', {
                    protocol: 's3',
                    s3: {bucket: 'some-bucket'}
                }, {protocol: 'cifs', cifs: {host: 'host', share: 'share', path: 'folder1/folder2/.'}});
            });

            it('validates invalid source CIFS path contains \'/./\'', () => {
                validateRelationshipThrows('path cannot include \'.\' (current folder)', {
                    protocol: 'cifs',
                    cifs: {host: 'host', share: 'share', path: 'folder1/./folder2'}
                }, {protocol: 's3', s3: {bucket: 'some-bucket'}});
            });

            it('validates invalid target CIFS path contains \'/./\'', () => {
                validateRelationshipThrows('path cannot include \'.\' (current folder)', {
                    protocol: 's3',
                    s3: {bucket: 'some-bucket'}
                }, {protocol: 'cifs', cifs: {host: 'host', share: 'share', path: 'folder1/./folder2'}});
            });

            it('validates invalid source NFS path is \'/.\'', () => {
                validateRelationshipThrows('path cannot include \'.\' (current folder)', {
                    protocol: 'nfs',
                    nfs: {host: 'host', export: 'share', path: '/.'}
                }, {protocol: 's3', s3: {bucket: 'some-bucket'}});
            });

            it('validates invalid target NFS path is \'/.\'', () => {
                validateRelationshipThrows('path cannot include \'.\' (current folder)', {
                    protocol: 's3',
                    s3: {bucket: 'some-bucket'}
                }, {protocol: 'nfs', nfs: {host: 'host', export: 'export', path: '/.'}});
            });

            it('validates invalid source NFS path ends with \'/.\'', () => {
                validateRelationshipThrows('path cannot include \'.\' (current folder)', {
                    protocol: 'nfs',
                    nfs: {host: 'host', export: 'export', path: 'folder1/folder2/.'}
                }, {protocol: 's3', s3: {bucket: 'some-bucket'}});
            });

            it('validates invalid target NFS path ends with \'/.\'', () => {
                validateRelationshipThrows('path cannot include \'.\' (current folder)', {
                    protocol: 's3',
                    s3: {bucket: 'some-bucket'}
                }, {protocol: 'nfs', nfs: {host: 'host', export: 'export', path: 'folder1/folder2/.'}});
            });

            it('validates invalid source NFS path contains \'/./\'', () => {
                validateRelationshipThrows('path cannot include \'.\' (current folder)', {
                    protocol: 'nfs',
                    nfs: {host: 'host', export: 'export', path: 'folder1/./folder2'}
                }, {protocol: 's3', s3: {bucket: 'some-bucket'}});
            });

            it('validates invalid target NFS path contains \'/./\'', () => {
                validateRelationshipThrows('path cannot include \'.\' (current folder)', {
                    protocol: 's3',
                    s3: {bucket: 'some-bucket'}
                }, {protocol: 'nfs', nfs: {host: 'host', export: 'export', path: 'folder1/./folder2'}});
            });

            it('validates invalid source CIFS path is \'/..\'', () => {
                validateRelationshipThrows('path cannot include \'..\' (parent folder)', {
                    protocol: 'cifs',
                    cifs: {host: 'host', share: 'share', path: '/..'}
                }, {protocol: 's3', s3: {bucket: 'some-bucket'}});
            });

            it('validates invalid target CIFS path is \'/..\'', () => {
                validateRelationshipThrows('path cannot include \'..\' (parent folder)', {
                    protocol: 's3',
                    s3: {bucket: 'some-bucket'}
                }, {protocol: 'cifs', cifs: {host: 'host', share: 'share', path: '/..'}});
            });

            it('validates invalid source CIFS path ends with \'/..\'', () => {
                validateRelationshipThrows('path cannot include \'..\' (parent folder)', {
                    protocol: 'cifs',
                    cifs: {host: 'host', share: 'share', path: 'folder1/folder2/..'}
                }, {protocol: 's3', s3: {bucket: 'some-bucket'}});
            });

            it('validates invalid target CIFS path ends with \'/..\'', () => {
                validateRelationshipThrows('path cannot include \'..\' (parent folder)', {
                    protocol: 's3',
                    s3: {bucket: 'some-bucket'}
                }, {protocol: 'cifs', cifs: {host: 'host', share: 'share', path: 'folder1/folder2/..'}});
            });

            it('validates invalid source CIFS path contains \'/../\'', () => {
                validateRelationshipThrows('path cannot include \'..\' (parent folder)', {
                    protocol: 'cifs',
                    cifs: {host: 'host', share: 'share', path: 'folder1/../folder2'}
                }, {protocol: 's3', s3: {bucket: 'some-bucket'}});
            });

            it('validates invalid target CIFS path contains \'/../\'', () => {
                validateRelationshipThrows('path cannot include \'..\' (parent folder)', {
                    protocol: 's3',
                    s3: {bucket: 'some-bucket'}
                }, {protocol: 'cifs', cifs: {host: 'host', share: 'share', path: 'folder1/../folder2'}});
            });

            it('validates invalid source NFS path is \'/..\'', () => {
                validateRelationshipThrows('path cannot include \'..\' (parent folder)', {
                    protocol: 'nfs',
                    nfs: {host: 'host', export: 'export', path: '/..'}
                }, {protocol: 's3', s3: {bucket: 'some-bucket'}});
            });

            it('validates invalid target NFS path is \'/..\'', () => {
                validateRelationshipThrows('path cannot include \'..\' (parent folder)', {
                    protocol: 's3',
                    s3: {bucket: 'some-bucket'}
                }, {protocol: 'nfs', nfs: {host: 'host', export: 'export', path: '/..'}});
            });

            it('validates invalid source NFS path ends with \'/..\'', () => {
                validateRelationshipThrows('path cannot include \'..\' (parent folder)', {
                    protocol: 'nfs',
                    nfs: {host: 'host', export: 'export', path: 'folder1/folder2/..'}
                }, {protocol: 's3', s3: {bucket: 'some-bucket'}});
            });

            it('validates invalid target NFS path ends with \'/..\'', () => {
                validateRelationshipThrows('path cannot include \'..\' (parent folder)', {
                    protocol: 's3',
                    s3: {bucket: 'some-bucket'}
                }, {protocol: 'nfs', nfs: {host: 'host', export: 'export', path: 'folder1/folder2/..'}});
            });

            it('validates invalid source NFS path contains \'/../\'', () => {
                validateRelationshipThrows('path cannot include \'..\' (parent folder)', {
                    protocol: 'nfs',
                    nfs: {host: 'host', export: 'export', path: 'folder1/../folder2'}
                }, {protocol: 's3', s3: {bucket: 'some-bucket'}});
            });

            it('validates invalid target NFS path contains \'/../\'', () => {
                validateRelationshipThrows('path cannot include \'..\' (parent folder)', {
                    protocol: 's3',
                    s3: {bucket: 'some-bucket'}
                }, {protocol: 'nfs', nfs: {host: 'host', export: 'export', path: 'folder1/../folder2'}});
            });

            it('validates invalid CIFS source path with folder longer than 255 characters', () => {
                const veryLongFolder = repeat('f', 256);
                validateRelationshipThrows(
                    'contains folders longer than 255 characters',
                    {
                        protocol: 'cifs',
                        cifs: {host: 'host', share: 'share', path: `folder1/${veryLongFolder}/folder3`}
                    }, {protocol: 's3', s3: {bucket: 'some-bucket', prefix: 'path'}}
                );
            });

            it('validates invalid CIFS target path with folder longer than 255 characters', () => {
                const veryLongFolder = repeat('f', 256);
                validateRelationshipThrows(
                    'contains folders longer than 255 characters',
                    {protocol: 's3', s3: {bucket: 'some-bucket', prefix: 'path'}}, {
                        protocol: 'cifs',
                        cifs: {host: 'host', share: 'share', path: `folder1/${veryLongFolder}/folder3`}
                    }
                );
            });

            it('validates invalid NFS source path with folder longer than 255 characters', () => {
                const veryLongFolder = repeat('f', 256);
                validateRelationshipThrows(
                    'contains folders longer than 255 characters',
                    {
                        protocol: 'nfs',
                        nfs: {host: 'host', export: 'export', path: `folder1/${veryLongFolder}/folder3`}
                    }, {protocol: 's3', s3: {bucket: 'some-bucket', prefix: 'path'}}
                );
            });

            it('validates invalid NFS target path with folder longer than 255 characters', () => {
                const veryLongFolder = repeat('f', 256);
                validateRelationshipThrows(
                    'contains folders longer than 255 characters',
                    {protocol: 's3', s3: {bucket: 'some-bucket', prefix: 'path'}}, {
                        protocol: 'nfs',
                        nfs: {host: 'host', export: 'export', path: `folder1/${veryLongFolder}/folder3`}
                    }
                );
            });

            it('validates invalid S3 source path length', () => {
                const folder = '/folderX';
                const path = repeat(folder, (1024 / folder.length) + 2);
                validateRelationshipThrows('S3 prefix maximum characters length can be 1024', {
                    protocol: 's3',
                    s3: {bucket: 'some-bucket', prefix: `${path}`}
                }, {protocol: 'nfs', nfs: {host: 'host', export: '/'}});
            });

            it('validates invalid S3 target path length', () => {
                const folder = '/folderX';
                const path = repeat(folder, (1024 / folder.length) + 2);
                validateRelationshipThrows('S3 prefix maximum characters length can be 1024', {
                    protocol: 'nfs',
                    nfs: {host: 'host', export: '/'}
                }, {protocol: 's3', s3: {bucket: 'some-bucket', prefix: `${path}`}});
            });
        });

        describe('validate S3 tags', () => {
            function validateS3TagsDoesNotThrow(s3Tags) {
                expect(validateS3Tags.bind(this, s3Tags)).not.to.throw();
            }

            function validateS3TagsThrows(expectedError, s3Tags) {
                expect(validateS3Tags.bind(this, s3Tags)).to.throw(BadRequest, expectedError);
            }

            it('should not fail if no tags provided', () => {
                validateS3TagsDoesNotThrow(undefined);
            });

            it('should not fail if empty tags list provided', () => {
                validateS3TagsDoesNotThrow([]);
            });


            it('should fail if provided more than 5 tags', () => {
                const s3Tags = [];
                for (let i = 1; i <= 5 + 1; i++) {
                    s3Tags.push({key: `key${i}`, value: `value${i}`});
                }
                validateS3TagsThrows('Maximum of 5 S3 tags may be provided', s3Tags);
            });

            it('should fail if one of the tags missing key', () => {
                const key = 'key';
                const value = 'value';
                const s3Tags = [{key, value}, {key}, {key, value}];
                validateS3TagsThrows('Each Amazon tag should contain both key and value', s3Tags);
            });

            it('should fail if one of the tag\'s key contains not allowed characters', () => {
                const value = 'value';
                const s3Tags = [{key: 'key', value}, {key: 'key$', value}];
                validateS3TagsThrows('Amazon tag keys may not start with \'aws:\' and may contain letters, numbers, spaces, or the special characters + - . _ : / @', s3Tags);
            });

            it('should fail if one of the tag\'s value contains not allowed characters', () => {
                const s3Tags = [{key: 'key1', value: 'value1'}, {key: 'key2', value: 'va^lue'}];
                validateS3TagsThrows('Amazon tag values may not start with \'aws:\' and may contain letters, numbers, spaces, or the special characters + - . _ : / @', s3Tags);
            });

            it('should fail if has duplicate key', () => {
                const value = 'value';
                const s3Tags = [{key: 'key1', value}, {key: 'key2', value}, {key: 'key2', value}, {key: 'key4', value}];
                validateS3TagsThrows('S3 tags keys should be distinct (no duplicated allowed)', s3Tags);
            });
        });


        describe('validate sync schedule', () => {
            function validateSyncIntervalsDoesNotThrow(syncInDays, syncInHours, syncInMinutes, isEnabled = true) {
                const sync = {
                    syncInDays, syncInHours, syncInMinutes, isEnabled
                };

                expect(validateSyncSchedule.bind(this, sync)).not.to.throw();
            }

            function validateSyncIntervalsThrows(expectedError, syncInDays, syncInHours, syncInMinutes, isEnabled = true) {
                const sync = {
                    syncInDays, syncInHours, syncInMinutes, isEnabled
                };
                expect(validateSyncSchedule.bind(this, sync)).to.throw(BadRequest, expectedError);
            }

            it('should fail if both intervals ( inHours and inDays) provided', () => {
                validateSyncIntervalsThrows('Only one of \'syncInDays\' or \'syncInHours\' can be provided for sync interval', 10, 10, 0);
            });

            it('should fail if inHours = 0  and inDays = 0 and inMinutes = 0 and autoSync = ON provided', () => {
                validateSyncIntervalsThrows('Please provide one of \\\'syncInHours\\\' or \\\'syncInDays\\\' or \\\'syncInMinutes \\\' for sync interval\' when pause is false', 0, 0, 0);
            });

            it('should fail if inHours = 0  and inDays = 0 and inMinutes = 0 and isEnabled false provided', () => {
                validateSyncIntervalsDoesNotThrow(0, 0, 0, false);
            });

            it('should fail if both intervals ( inMinutes and inDays) provided', () => {
                validateSyncIntervalsThrows('Only one of  \'syncInDays\' or \'syncInMinutes\' can be provided for sync interval', 10, 0, 10);
            });

            it('should fail if both intervals ( inMinutes and inHours) provided', () => {
                validateSyncIntervalsThrows('Only one of \'syncInHours\' or \'syncInMinutes\' can be provided for sync interval', 0, 10, 10);
            });

            it('should fail if all intervals provided', () => {
                validateSyncIntervalsThrows('Only one of \'syncInDays\' or \'syncInHours\' or \'syncInMinutes\' can be provided for sync interval', 10, 10, 10);
            });

            it('should fail on negative syncInDays', () => {
                validateSyncIntervalsThrows('\"syncInDays\" must be larger than or equal to 0', -1, 0, 0);
            });

            it('should fail on too high syncInDays', () => {
                validateSyncIntervalsThrows('\"syncInDays\" must be less than or equal to 365', 366, 0, 0);
            });

            it('should fail on negative syncInHours', () => {
                validateSyncIntervalsThrows('\"syncInHours\" must be larger than or equal to 0', 0, -1, 0);
            });

            it('should fail on too high syncInHours', () => {
                validateSyncIntervalsThrows('\"syncInHours\" must be less than or equal to 24', 0, 25, 0);
            });

            it('should fail on negative syncInMinutes', () => {
                validateSyncIntervalsThrows('\"syncInMinutes\" must be larger than or equal to 0', 0, 0, -1);
            });

            it('should fail on too high syncInMinutes', () => {
                validateSyncIntervalsThrows('\"syncInMinutes\" must be less than or equal to 60', 0, 0, 61);
            });

            it('should succeed on lower end valid inDays', () => {
                validateSyncIntervalsDoesNotThrow(1, 0, 0);
            });

            it('should succeed on higher end valid inDays', () => {
                validateSyncIntervalsDoesNotThrow(365, 0, 0);
            });

            it('should succeed on lower end valid inHours', () => {
                validateSyncIntervalsDoesNotThrow(0, 1, 0);
            });

            it('should succeed on higher end valid inHours', () => {
                validateSyncIntervalsDoesNotThrow(0, 24, 0);
            });

            it('should succeed on higher end valid inMinutes', () => {
                validateSyncIntervalsDoesNotThrow(0, 0, 1);
            });

            it('should succeed on higher end valid inMinutes', () => {
                validateSyncIntervalsDoesNotThrow(0, 0, 60);
            });
        });
    });
    describe('calculateProgress', () => {
        function validateProgress(status, bytesToTransfer, bytesTransferred, dirsToTransfer, dirsTransferred, expectedProgress) {
            const progress = calculateProgress(status, bytesToTransfer, bytesTransferred, dirsToTransfer, dirsTransferred);
            progress.should.equal(expectedProgress);
        }

        function validateUndefinedProgress(status, bytesToTransfer, bytesTransferred, dirsToTransfer, dirsTransferred) {
            const progress = calculateProgress(status, bytesToTransfer, bytesTransferred, dirsToTransfer, dirsTransferred);
            expect(progress).to.be.undefined;
        }

        it('Calculate progress when there are no counters status is Running', () => {
            validateUndefinedProgress(RUNNING_STATUS);
        });


        it('Calculate progress when bytesToTransfer = 0 and  dirsToTransfer = 0  and the status is DONE', () => {
            validateProgress(MESSAGE_STATUS_DONE, 0, 0, 0, 0, 100);
        });

        it(`Calculate progress when bytesToTransfer = 0 and  dirsToTransfer = 0 and the status is ${MESSAGE_STATUS_FAILED}`, () => {
            validateProgress(MESSAGE_STATUS_FAILED, 0, 0, 0, 0, 0);
        });

        it(`Calculate progress when bytesToTransfer = 0 and  dirsToTransfer = 0  and the status is ${RUNNING_STATUS}`, () => {
            validateProgress(RUNNING_STATUS, 0, 0, 0, 0, 0);
        });

        it(`Calculate progress when bytesToTransfer + dirsToTransfer = 100, bytesTransferred + dirsToTransfer =  200 and the status is ${RUNNING_STATUS}`, () => {
            validateProgress(RUNNING_STATUS, 50, 50, 100, 100, 99);
        });

        it(`Calculate progress when bytesToTransfer + dirsToTransfer = 100, bytesTransferred + dirsToTransfer =  100, and the status is ${MESSAGE_STATUS_FAILED}`, () => {
            validateProgress(MESSAGE_STATUS_FAILED, 50, 50, 50, 50, 0);
        });

        it(`Calculate progress  when bytesToTransfer + dirsToTransfer = 100, bytesTransferred + dirsToTransfer =  100 and the status is ${MESSAGE_STATUS_DONE}`, () => {
            validateProgress(MESSAGE_STATUS_DONE, 100, 0, 100, 0, 100);
        });

        it(`Calculate progress when bytesToTransfer + dirsToTransfer = 100, bytesTransferred + dirsToTransfer =  200 and the status is ${MESSAGE_STATUS_DONE}`, () => {
            validateProgress(MESSAGE_STATUS_DONE, 0, 100, 0, 200, 100);
        });

        it(`Calculate progress when bytesToTransfer + dirsToTransfer = 100, bytesTransferred + dirsToTransfer =  50 and the status is ${RUNNING_STATUS}`, () => {
            validateProgress(RUNNING_STATUS, 100, 0, 50, 0, 0);
        });

        it(`Calculate progress when bytesToTransfer + dirsToTransfer = 100, bytesTransferred + dirsToTransfer =  50 and the status is ${MESSAGE_STATUS_FAILED}`, () => {
            validateProgress(MESSAGE_STATUS_FAILED, 100, 0, 100, 0, 0);
        });
    });
});
