import React from 'react';
import {
    FM_AGENT,
    FM_AGENT_TYPE_AWS,
    FM_AGENT_TYPE_AZURE,
    FM_AGENT_TYPE_GCP,
    FM_AGENT_TYPE_ONPREM
} from '../../consts';
import {connect} from "react-redux";
import ProviderButton from './ProviderButton';
import {InputField} from '../_common_/forms/forms';
import {Field, reduxForm, SubmissionError} from 'redux-form';
import Validator from 'validatorjs';
import useRunOnce from '../../hooks/useRunOnce';
import DataBrokerLoadingPage from './DataBrokerLoadingPage';
import {createDataBrokerWithTemplate, createDataBroker} from "../../store/wizard/action_creators";
import './dataBrokerDetails.scss';


const DataBrokerDetails = props => {
    const {isGcpAccessRequired, isAwsAccessRequired, shouldDisplayPortBool, goToNextStep, setType, type, defaultProvider, initialize, existingGroupId, setDataBroker} = props;
    const {createDataBrokerWithTemplate, createDataBroker} = props;
    const {handleSubmit, submitting} = props;

    useRunOnce(() => {
        if (defaultProvider) {
            setType(defaultProvider);
        }
        if (shouldDisplayPortBool)
            initialize({port: "443"});
    });

    const onSubmit = async ({name, port}) => {
        const dataBrokerDetails = {name, type, port};

        if (existingGroupId) {
            dataBrokerDetails.groupId = existingGroupId;
        }
        const createBrokerFunc = type === FM_AGENT_TYPE_GCP.name ? createDataBrokerWithTemplate : createDataBroker;
        try {
            const newDataBroker = await createBrokerFunc(dataBrokerDetails);
            setDataBroker(newDataBroker);
            goToNextStep();
        }
        catch (err) {
            handleError(err);
        }
    };

    const handleError = (err) => {
        if (err.status !== 401) {
            const errorMessage = err.data.message ? err.data.message : err.data;
            throw new SubmissionError({name: errorMessage});
        }
    };

    const credentialsRequiredText = isGcpAccessRequired && isAwsAccessRequired ? 'AWS and Google Cloud' : isGcpAccessRequired ? 'Google Cloud' : 'AWS';
    const disabledTitle = `This type of data broker isn’t supported because ${credentialsRequiredText} credentials are required for this relationship.`;

    return <form className="data-broker-wizard-step data-broker-details" onSubmit={handleSubmit(onSubmit)}>
        {!submitting && <>

            <h3>Select a Provider</h3>

            <div className="content-area">
                <div className="provider-types">
                    <ProviderButton
                        displayName={FM_AGENT_TYPE_AZURE.providerName}
                        handleClick={() => {
                            setType(FM_AGENT_TYPE_AZURE.name)
                        }}
                        icon={FM_AGENT_TYPE_AZURE.colorIcon}
                        disabledIcon={FM_AGENT_TYPE_AZURE.disabledIcon}
                        selected={type === FM_AGENT_TYPE_AZURE.name}
                        disabled={isAwsAccessRequired || isGcpAccessRequired}
                        disabledTitle={disabledTitle}
                    />
                    <ProviderButton
                        displayName={FM_AGENT_TYPE_AWS.providerName}
                        handleClick={() => {
                            setType(FM_AGENT_TYPE_AWS.name)
                        }}
                        icon={FM_AGENT_TYPE_AWS.colorIcon}
                        disabledIcon={FM_AGENT_TYPE_AWS.disabledIcon}
                        selected={type === FM_AGENT_TYPE_AWS.name}
                        disabled={isGcpAccessRequired}
                        disabledTitle={disabledTitle}
                    />
                    <ProviderButton
                        displayName={`${FM_AGENT_TYPE_GCP.providerName} Platform`}
                        handleClick={() => {
                            setType(FM_AGENT_TYPE_GCP.name)
                        }}
                        icon={FM_AGENT_TYPE_GCP.colorIcon}
                        disabledIcon={FM_AGENT_TYPE_GCP.disabledIcon}
                        selected={type === FM_AGENT_TYPE_GCP.name}
                        disabled={isAwsAccessRequired}
                        disabledTitle={disabledTitle}
                    />
                    <ProviderButton
                        displayName="On-Prem Data Broker"
                        handleClick={() => {
                            setType(FM_AGENT_TYPE_ONPREM.name)
                        }}
                        icon={FM_AGENT_TYPE_ONPREM.colorIcon}
                        disabledIcon={FM_AGENT_TYPE_ONPREM.disabledIcon}
                        selected={type === FM_AGENT_TYPE_ONPREM.name}/>
                </div>
                {type && <div className="form-inputs">
                    <div className="inputs">
                        <Field className="name-input" name="name" component={InputField} label={`${FM_AGENT} Name`} disabled={submitting}/>
                        {shouldDisplayPortBool &&
                        <Field className="port-input" name="port" component={InputField} label="Port" disabled={submitting}/>
                        }
                    </div>
                </div>}
            </div>
        </>}

        {submitting && <DataBrokerLoadingPage/>}

        <div className="button-footer">
           <button className="button primary" type="submit" disabled={!type || submitting}>Continue</button>
        </div>
    </form>
};

const rules = {
    name: ['required', 'regex:/^[a-zA-Z][a-zA-Z0-9-]{0,127}$/'],
    port: ['required', 'numeric', 'between:1,65535']
};

const validate = (values) => {
    const validator = new Validator(values, rules,
        {
            'regex.name': 'Name must contain only letters, numbers, dashes, must start with a letter and may be up to 128 characters.',
            'between.port': 'Port must be between 1 to 65535'
        });
    validator.passes();
    return validator.errors.all();
};

const DataBrokerDetailsForm = reduxForm({
    form: 'providerSelection',
    validate
})(DataBrokerDetails);

export default connect(null, {createDataBrokerWithTemplate, createDataBroker})(DataBrokerDetailsForm);

