import React, {useCallback} from 'react';
import Validator from 'validatorjs';
import {reduxForm, Field, getFormValues} from 'redux-form';
import {connect} from 'react-redux';
import {InputField, RadioInputs, SelectField} from '../_common_/forms/forms';
import _ from 'lodash';
import {awsRegions} from '../../services/aws.api.service';
import useAwsForm from './useAwsForm';
import ButtonWithError from '../widgets/ButtonWithError';
import {components} from 'react-select';
import {AWS_BROKER_POLICY_REQUIREMENTS} from '../../consts';
import './instanceDetailsAwsForm.scss';
import InstanceDetailsColumnHeader from './InstanceDetailsColumnHeader';

const optionLabel = ({label, additionalLabel}) => (<div>
        <div>{label}</div>
        <div>{additionalLabel}</div>
    </div>
);

const SelectRenderedValue = props => (<components.SingleValue {...props}>
        {props.data.label}
    </components.SingleValue>
);

const InstanceDetailsAwsForm = ({submitting, handleDeployClick, handleError, change, credentials, handleSubmit, formValues}) => {
    const {selectedRegion, selectedVpc} = formValues || {};

    const {vpcsOptions, subnetsOptions, keyPairsOptions} = useAwsForm(credentials, selectedRegion?.value, selectedVpc?.value, handleError);

    const clearFields = useCallback((fieldArr) => fieldArr.forEach(fieldName => change(fieldName, '')), [change]);

    return (<form className="data-broker-wizard-step instance-details-aws" onSubmit={handleSubmit(handleDeployClick)}>
        <div className="title">Basic Settings</div>
        <div className="content-area">
            <div className="form-content">
                <div className="col left">
                    <InstanceDetailsColumnHeader text={"Location"}/>
                    <Field component={SelectField}
                           onChange={() => clearFields(['selectedVpc', 'selectedSubnet', 'selectedKeyPair'])}
                           name="selectedRegion"
                           options={awsRegions}
                           label="Region"
                           placeholder="Select Region"
                           valueField="value"
                           isDisabled={submitting}
                    />
                    <Field component={SelectField}
                           onChange={() => clearFields(['selectedSubnet'])}
                           name="selectedVpc"
                           formatOptionLabel={optionLabel}
                           SingleValue={SelectRenderedValue}
                           options={vpcsOptions}
                           label="VPC"
                           isLoading={selectedRegion && !vpcsOptions}
                           placeholder="Select VPC"
                           valueField="value"
                           isDisabled={submitting || !selectedRegion}
                           noOptionsMessage={() => `No VPCs available`}
                    />
                    <Field component={SelectField}
                           name="selectedSubnet"
                           formatOptionLabel={optionLabel}
                           SingleValue={SelectRenderedValue}
                           options={subnetsOptions}
                           label="Subnet"
                           isLoading={selectedVpc && !subnetsOptions}
                           placeholder="Select Subnet"
                           valueField="value"
                           isDisabled={submitting || !selectedVpc}
                           noOptionsMessage={() => `No Subnets available`}
                    />
                </div>
                <div className="col right">
                    <InstanceDetailsColumnHeader text={"Connectivity"}/>
                    <Field component={SelectField}
                           name="selectedKeyPair"
                           options={keyPairsOptions}
                           label="Key Pair"
                           isLoading={selectedRegion && !keyPairsOptions}
                           placeholder="Select Key Pair"
                           valueField="value"
                           isDisabled={submitting || !selectedRegion}
                           noOptionsMessage={() => `No Key Pair available`}
                    />

                    <div className="section-title">Enable Public IP?</div>
                    <Field name="publicIpStatus" component={RadioInputs}
                           disabled={submitting}
                           changeSelection={_.noop}
                           checkedIndex={formValues?.publicIpStatus === 'Disable' ? 1 : 0}
                           inputs={['Enable', 'Disable']}/>

                    <Field name="userRole"
                           component={InputField}
                           label="IAM Role (optional)"
                           disabled={submitting}
                           info={{
                               text: <div>The name of an existing IAM role to attach to the NetApp Data Broker. <a
                                   href={AWS_BROKER_POLICY_REQUIREMENTS}
                                   target="_blank"
                                   rel="noopener noreferrer">The role must have these permissions</a>. If you don’t
                                   provide an existing role, then Cloud Sync will create one for you.</div>
                           }}/>
                </div>
            </div>
        </div>
        <div className="button-footer">
            <ButtonWithError className="button primary" type="submit" disable={submitting}>
                Continue
            </ButtonWithError>
        </div>
    </form>);
};

const validate = (values) => {
    const rules = {
        selectedRegion: ['required'],
        selectedVpc: ['required'],
        selectedSubnet: ['required'],
        selectedKeyPair: ['required']
    };

    const validator = new Validator(values, rules);
    validator.setAttributeNames({
        selectedRegion: 'Region',
        selectedVpc: 'VPC',
        selectedSubnet: 'Subnet',
        selectedKeyPair: 'Key Pair'
    });
    validator.passes();
    return validator.errors.all();
};

const connectedToForm = reduxForm({form: 'instanceDetailsAws', validate})(InstanceDetailsAwsForm);

export default connect(state => ({
    formValues: getFormValues('instanceDetailsAws')(state),
    credentials: state.deploy.credentials.aws
}))(connectedToForm);
