import React, {useEffect, useState} from 'react';
import {Field, getFormValues, reduxForm} from 'redux-form';
import {connect} from 'react-redux';
import Validator from 'validatorjs';
import {SelectField, InputField, RadioInputs} from '../_common_/forms/forms';
import _ from 'lodash';
import * as azureService from '../../services/azure.api.service';
import {AZURE_SSH_INSTRUCTIONS} from '../../consts';
import {RESOURCE_GROUP_OPTIONS} from './InstanceDetailsAzure';
import useRunOnce from '../../hooks/useRunOnce';
import InstanceDetailsColumnHeader from './InstanceDetailsColumnHeader';
import {ENABLE_DISABLE_OPTIONS} from "../../consts";
import './instanceDetailsAzureForm.scss';
import ButtonWithNotification from '../widgets/ButtonWithNotification';

const INSTANCE_NAME_REGEX = '/[a-zA-Z][-a-zA-Z0-9]*/';
const USER_NAME_REGEX = '/(?!\\badmin\\b)(?!\\bAdmin\\b)(^[a-zA-Z][0-9a-zA-Z]{0,30}$)/';

const InstanceDetailsAzureForm = ({ handleDeployClick, dataBrokerName, handleError, formValues, clearFields, initialize, submitting, handleSubmit}) => {
    const selectedSubscription = formValues?.selectedSubscription;

    const [subscriptions, setSubscriptions] = useState(undefined);
    const [regions, setRegions] = useState(undefined);
    const [vnets, setVnets] = useState(undefined);
    const [resourceGroups, setResourceGroups] = useState(undefined);

    useRunOnce(() => {
        initialize({
            virtualMachineName: dataBrokerName || undefined,
            adminUserName: "databroker",
            authMethod: "Password",
            resourceGroupOption: RESOURCE_GROUP_OPTIONS[0],
            publicIp: ENABLE_DISABLE_OPTIONS[0]
        });
    });

    useRunOnce(async () => {
        try {
            const res = await azureService.getSubscriptions();
            setSubscriptions(res.value);
        } catch (error) {
            handleError(error);
        }
    });

    useEffect(() => {
        const fetchData = async () => {
            try {
                const info = await getInfoPerSubscription(selectedSubscription.value);
                setRegions(info.regions);
                setVnets(info.vnets);
                setResourceGroups(info.resourceGroups);
            } catch (error) {
                handleError(error);
            }
        };

        if (selectedSubscription) {
            fetchData();
        }
    }, [selectedSubscription, handleError]);

    return (<>
        {formValues && <form className="data-broker-wizard-step instance-details-azure" onSubmit={handleSubmit(handleDeployClick)}>
            <div className="title">Basic Settings</div>
            <div className="content-area">
                <div className="form-content">
                        <div className="col left">
                            <InstanceDetailsColumnHeader text={"Location"}/>
                            <AzureSelectField fieldName="selectedSubscription" clearFields={clearFields}
                                              data={subscriptions}
                                              propertyForLabel="displayName"
                                              propertyForValue="subscriptionId"
                                              isDisabled={submitting}
                                              placeholder="Select a subscription" label="Subscription"
                                              displayName="Subscriptions"
                                              fieldsToClearUponChange={['selectedRegion', 'selectedVnet', 'selectedSubnet']}
                            />

                            <AzureSelectField fieldName="selectedRegion" clearFields={clearFields}
                                              data={regions}
                                              propertyForLabel="displayName"
                                              propertyForValue="name"
                                              isDisabled={!formValues.selectedSubscription || submitting}
                                              disabledTitle={submitting ? '' : 'Waiting for an Azure subscription.'}
                                              placeholder="Select a region" label="Azure Region"
                                              displayName="Azure Regions"
                                              fieldsToClearUponChange={['selectedVnet', 'selectedSubnet', 'selectedResourceGroup']}
                            />

                            <AzureSelectField fieldName="selectedVnet" clearFields={clearFields}
                                              data={vnets}
                                              propertyForLabel="name"
                                              propertyForValue="id"
                                              isDisabled={!formValues.selectedRegion || submitting}
                                              disabledTitle={submitting ? '' : 'Waiting for an Azure region.'}
                                              placeholder="Select a VNet" label="VNet"
                                              displayName="VNets"
                                              filterFunction={formValues.selectedRegion ? (vnet) => vnet.location === formValues.selectedRegion.value : undefined}
                                              fieldsToClearUponChange={['selectedSubnet']}
                            />

                            <AzureSelectField fieldName="selectedSubnet" clearFields={clearFields}
                                              data={formValues.selectedVnet ? _.get(_.find(vnets, {id: formValues.selectedVnet.value}), 'properties.subnets') : undefined}
                                              propertyForLabel="name"
                                              propertyForValue="name"
                                              isDisabled={!formValues.selectedVnet || submitting}
                                              disabledTitle={submitting ? '' : 'Waiting for an Azure Vnet.'}
                                              placeholder="Select a subnet" label="Subnet"
                                              displayName="Subnets"
                                              filterFunction={(subnet) => subnet.properties.delegations ? subnet.properties.delegations.length === 0 : true}
                            />

                            <Field component={SelectField}
                                   name="publicIp"
                                   label="Public IP"
                                   isDisabled={submitting}
                                   options={ENABLE_DISABLE_OPTIONS}
                                   valueField="value"/>

                        </div>
                        <div className="col right">
                            <InstanceDetailsColumnHeader text={"Connectivity"}/>
                            <Field name="virtualMachineName" component={InputField} label="VM Name" disabled={submitting}
                                   info={{text: 'The name can include up to 64 alphanumeric characters and hyphens. The name cannot include only numbers.'}}/>
                            <Field name="adminUserName" component={InputField} label="User Name" disabled={submitting}
                                   info={{text: 'The name should start with an alphabetic character and contain only numbers, letters, and should be 30 or fewer characters in length. Must not be Admin.'}}/>

                            <div className="section-title">Authentication Method:</div>
                            <Field name="authMethod" component={RadioInputs} disabled={submitting}
                                   changeSelection={_.noop}
                                   checkedIndex={formValues.authMethod === 'Password' ? 0 : 1}
                                   inputs={['Password', 'Public Key']}
                                   className="radio-options"/>
                            {formValues.authMethod === 'Password' &&
                            <Field name="adminPassword"
                                   component={InputField} label="Enter Password" disabled={submitting}
                                   type="password"
                                   autoComplete="off"
                                   info={{text: 'Password requirements: 7-71 characters in length and three of the following: Uppercase, Lowercase, Special characters, Digits (0-9).'}}/>}
                            {formValues.authMethod === 'Public Key' &&
                            <Field name="sshKey"
                                   component={InputField} label="Enter SSH Public Key" disabled={submitting}
                                   info={{
                                       text: <div>The Linux VM uses an SSH public-private key pair for authentication. Generate an RSA
                                           SSH key file using ssh-keygen or PuTTYGen
                                           (see <a href={AZURE_SSH_INSTRUCTIONS} target="_blank" rel="noopener noreferrer">Azure
                                               docs</a>).
                                           Copy the key from the file (it starts with &#34;ssh-rsa&#34;) and paste it into this field.
                                       </div>
                                   }}/>}

                            <div className="section-title">Resource Group:</div>
                            <Field name="resourceGroupOption" component={RadioInputs} disabled={submitting}
                                   changeSelection={_.noop}
                                   checkedIndex={formValues.resourceGroupOption === RESOURCE_GROUP_OPTIONS[0] ? 0 : 1}
                                   inputs={RESOURCE_GROUP_OPTIONS}
                                   className="radio-options"/>
                            {formValues.resourceGroupOption === RESOURCE_GROUP_OPTIONS[1] &&
                            <AzureSelectField fieldName="selectedResourceGroupExisting" clearFields={clearFields}
                                              data={resourceGroups}
                                              propertyForLabel="name"
                                              propertyForValue="name"
                                              isDisabled={!formValues.selectedRegion || submitting}
                                              disabledTitle={submitting ? '' : 'Waiting for an Azure region.'}
                                              label=''
                                              placeholder="Select a Resource Group"
                                              displayName="Resource Groups"
                                              filterFunction={formValues.selectedRegion ? (resourceGroup) => resourceGroup.location === formValues.selectedRegion.value : undefined}
                            />}
                        </div>
                    </div>
                </div>
            <div className="button-footer">
                <ButtonWithNotification
                    type="submit"
                    text="Continue"
                    submitting={submitting}/>
            </div>
        </form>}
    </>);
};

const AzureSelectField = ({data, propertyForLabel, propertyForValue, fieldsToClearUponChange, placeholder, label, clearFields,
                           displayName, fieldName, isDisabled, disabledTitle, filterFunction = () => true}) => {
    const options = data ? data.filter(filterFunction)
        .map((item) => ({label: item[propertyForLabel], value: item[propertyForValue]})) : undefined;

    const optionsSorted = _.sortBy(options, option => option.label);
    const title = isDisabled ? disabledTitle : '';

    return (
        <div title={title}>
            <Field component={SelectField}
                   onChange={() => fieldsToClearUponChange && clearFields('azureCreateDb', false, false, ...fieldsToClearUponChange)
                   }
                   name={fieldName}
                   options={optionsSorted}
                   label={label}
                   isLoading={!isDisabled && !data}
                   placeholder={placeholder}
                   valueField="value"
                   isDisabled={isDisabled}
                   noOptionsMessage={() => `No ${displayName} available`}/>
        </div>
    )
};

const getInfoPerSubscription = async subscriptionId => {
    const [resRegions, resVnets, resResourceGroups] = await Promise.all([
        azureService.getAzureRegions(subscriptionId),
        azureService.getVnetsWithSubnets(subscriptionId),
        azureService.getResourceGroups(subscriptionId)]);
    return {
        regions: resRegions.value,
        vnets: resVnets.value,
        resourceGroups: resResourceGroups.value
    }
};

const validate = (values) => {
    // eslint-disable-next-line no-unused-vars
    Validator.register('azurePassword', (value, requirement, attribute) => {
        let passwordSatisfaction = 0;
        if (value.length > 6 && value.length < 72) {
            if (value.match(/[A-Z]/)) ++passwordSatisfaction;
            if (value.match(/[a-z]/)) ++passwordSatisfaction;
            if (value.match(/[0-9]/)) ++passwordSatisfaction;
            // eslint-disable-next-line no-useless-escape
            if (value.match(/[~!@#$%^&\*_\-\+=`|\/(){}\[\]:;"'<>,.?]/)) ++passwordSatisfaction;
        }
        // need to have at least 3 of those rules (according to Azure requirements)
        return passwordSatisfaction >= 3;
    }, `The :attribute field doesn't match the requirements.`);

    const rules = {
        virtualMachineName: ['required', `regex:${INSTANCE_NAME_REGEX}`],
        adminUserName: ['required', `regex:${USER_NAME_REGEX}`],
        selectedSubscription: ['required'],
        selectedRegion: ['required'],
        selectedVnet: ['required'],
        selectedSubnet: ['required'],
        adminPassword: values.authMethod === 'Password' ? ['required', 'azurePassword'] : [],
        sshKey: values.authMethod === 'Public Key' ? ['required'] : [],
        selectedResourceGroupExisting: values.resourceGroupOption === RESOURCE_GROUP_OPTIONS[1] ? ['required'] : []
    };

    const validator = new Validator(values, rules);
    validator.setAttributeNames({
        selectedResourceGroupExisting: 'Resource Group',
        virtualMachineName : 'VM name',
        sshKey: 'SSH public key',
        adminUserName: 'User Name',
        adminPassword: 'Password',
        selectedSubscription: 'Subscription',
        selectedRegion: 'Region',
        selectedVnet: 'VNet',
        selectedSubnet: 'Subnet'
    });
    validator.passes();
    return validator.errors.all();
};

const AzureForm = reduxForm({form: "azureCreateDb", validate})(InstanceDetailsAzureForm);

export default connect(state => ({formValues: getFormValues('azureCreateDb')(state)}))(AzureForm);
