import React, {Fragment, useEffect, useState} from "react"
import {Field, reduxForm, reset} from "redux-form";
import {connect} from 'react-redux';
import * as actionCreators from "../../store/global/action_creators";
import Validator from "validatorjs";
import {InputField} from "../_common_/forms/forms";
import ButtonWithError from "../widgets/ButtonWithError";
import Isvg from "react-inlinesvg";
import {
    COMMUNITIES,
    DOCUMENTATION_URL,
    FAQ,
    SALES,
    SUPPORT_IN_PROGRESS,
    SUPPORT_NOT_REGISTERED,
    SUPPORT_PHONE,
    SUPPORT_REGISTERED,
    SUPPORT_TICKET_URL
} from "../../consts";
import {MenuBar} from "../widgets/MenuBar";
import {ClickableDiv} from '../widgets/ClickableDiv';
import _ from 'lodash';
import LoadingBar from "../_common_/loadingBar/LoadingBar";
import {noConnectionToAllBrokers} from '../../utils/dashboard-utils';
import {dataBrokers} from "../../store/wizard/action_creators";

require("./supportContent.scss");

const rules = {
    nssUsername: 'required',
    nssPassword: 'required'
};

const validate = (values) => {
    const validator = new Validator(values, rules);
    validator.setAttributeNames({nssUsername: 'NSS User Name', nssPassword: "NSS Password"});
    validator.passes();
    return validator.errors.all();
};

const SupportInfoLine = ({ subTitle, value }) => {
    return (<div className="support-info-line">
        <div className="sub-title no-wrap">{subTitle}</div>
        <div className="value no-wrap" title={value}>{value}</div>
    </div>);
};

const SupportInfoStatusIcon = ({ iconName, text , className}) => {
    return (<div className={className}>
        <div className="icon">
            <span><Isvg src={require(`../../assets/svgs/${iconName}`)}/></span>
        </div>
        <div className="icon-title">
            {text}
        </div>
    </div>)
};

const getSupportUpperPanelUX = ({accountId, supportSerialNumber}) => {
    return (
        <SupportPanel title="Account Details">
            <SupportInfoLine subTitle="Account ID" value={accountId}/>
            <SupportInfoLine subTitle="NetApp Serial Number" value={supportSerialNumber}/>
        </SupportPanel>
    );
};

const SupportPanel = ({title, children}) => {
    return (<div className="upper-panel">
        <h3>{title}</h3>
        {children}
    </div>);
};

const SupportContent = (props) => {
    const {auth, submitting, handleSubmit, dispatch, _dataBrokers, dataBrokers:dataBrokersFunc} = props;
    const [error, setError] = useState('');
    const [showASUPAnimation, setShowASUPAnimation] = useState(false);
    const [loading, setLoading] = useState(!_dataBrokers || (_dataBrokers && _dataBrokers.inProgress));

    useEffect(() => {
        dataBrokersFunc();
    },[dataBrokersFunc]);

    useEffect(() => {
        setLoading(!_dataBrokers || (_dataBrokers && _dataBrokers.inProgress));
    }, [_dataBrokers]);

    const dataBrokers = _.get(_dataBrokers, 'data', []);
    const noDataBrokers = dataBrokers.length === 0;
    const allDataBrokersFailed = !noDataBrokers && noConnectionToAllBrokers(dataBrokers);

    const submit = (values, dispatch) => {
        setError('');
        return actionCreators.registerToSupport(values)(dispatch)
            .catch((err) => {
                if (err.status === 400 || err.status === 403) {
                    setError(err.data.message);
                }
            })
            .then(clearForm);
    };

    const clearForm = () => {
        dispatch(reset('SupportContent'));
    };

    const getSupportContentUX = (auth) => {
        return (
            <Fragment>
                {getSupportUpperPanelUX(auth)}
                {getSupportInnerPanel(auth)}
            </Fragment>
        );
    };

    const getSupportInnerPanel = ({supportRegistrationStatus}) => {

        if (supportRegistrationStatus === SUPPORT_IN_PROGRESS) {
            return (
                <div className="support-pending">
                    <SupportInfoStatusIcon className="status-icon" iconName="pending-with-arrow.svg" text="Your request to activate support was received. This Process can take up to 1 minute."/>
                </div>);
        } else if (supportRegistrationStatus === SUPPORT_REGISTERED) {
            return (
                <div className="scrollable">

                    <SupportPanel title="NetApp Support Site">
                        <SupportInfoStatusIcon className="status-icon-row" iconName="support-checkmark.svg" text="Your support is active"/>
                    </SupportPanel>

                    <SupportPanel title="ASUP (AutoSupport)">
                        {noDataBrokers && <div>You can’t send AutoSupport messages because you don’t have any data brokers.</div>}
                        {allDataBrokersFailed && <div>You can’t send AutoSupport messages because all data brokers are currently unavailable.</div>}
                        {!noDataBrokers && !allDataBrokersFailed && <div className="asup">
                            <div className="text">Send AutoSupport to NetApp</div>
                            <ClickableDiv onClick={() => {
                                    setShowASUPAnimation(true);
                                    actionCreators.sendAsup()(dispatch);
                                    setTimeout(() => {
                                        setShowASUPAnimation(false);
                                    }, 3000);
                                }}>
                                <img src={showASUPAnimation ? require('../../assets/images/asup-animation.gif') : require('../../assets/images/send-asup.png')} alt="...loading"/>
                            </ClickableDiv>
                        </div>}
                    </SupportPanel>
                </div>)
        } else if (supportRegistrationStatus === SUPPORT_NOT_REGISTERED || !supportRegistrationStatus) {
            return (
                <form onSubmit={handleSubmit(submit)}>
                    <div className="content scrollable">
                        <h3>NetApp Support Site Credentials</h3>
                        <p>Enter your NetApp Support Site (NSS) credentials to activate support for this subscription serial number.</p>
                        <div>
                            <div className="field">
                                <Field name="nssUsername" component={InputField}
                                       label="NSS User Name"
                                       type="text"/>
                            </div>
                            <div className="field">
                                <Field name="nssPassword" component={InputField}
                                       label="NSS Password"
                                       autoComplete="off"
                                       type="password"/>
                            </div>
                        </div>
                        <ButtonWithError submitting={submitting}
                                         className="button card-blue"
                                         type="submit"
                                         error={error}>
                            Activate Support
                        </ButtonWithError>
                        <div className="no-account">
                            If you don’t have NSS credentials, or if this is your first NetApp product/service,
                            <a href={process.env.REACT_APP_SUPPORT_BASE_URL}
                               target="_blank"
                               rel="noopener noreferrer"> click here to register</a>.
                        </div>
                    </div>
                </form>)
        }
    };

    return (<div className="support-form">
        <header>
            <div className="header-icon"><span><Isvg src={require('./../../assets/svgs/support.svg')}/></span>
                Support
            </div>
            <div className="header-menu">
                <MenuBar openHeight="360px">
                    <ul className="no-bullet">
                        <li><span className="float-left documentation"><Isvg
                            src={require('./../../assets/svgs/documentation.svg')}/></span><a
                            href={DOCUMENTATION_URL} target="_blank" rel="noopener noreferrer">Documentation</a>
                        </li>
                        <li><span className="float-left faq"><Isvg
                            src={require('./../../assets/svgs/faq.svg')}/></span><a href={FAQ} target="_blank"
                                                                                    rel="noopener noreferrer">FAQ</a>
                        </li>
                        <li><span className="float-left communities"><Isvg
                            src={require('./../../assets/svgs/communities.svg')}/></span><a href={COMMUNITIES}
                                                                                            target="_blank"
                                                                                            rel="noopener noreferrer">Communities</a>
                        </li>
                        <li><span className="float-left feedback"><Isvg
                            src={require('./../../assets/svgs/feedback.svg')}/></span><a
                            href={'mailto:ng-cloud-manager-architects@netapp.com'}>Feedback</a></li>
                        <li><span className="float-left sales"><Isvg
                            src={require('./../../assets/svgs/sales.svg')}/></span><a href={SALES} target="_blank"
                                                                                      rel="noopener noreferrer">Contact
                            Sales</a></li>
                        <li><span className="float-left"><Isvg
                            src={require('./../../assets/svgs/ticket.svg')}/></span><a href={SUPPORT_TICKET_URL}
                                                                                       target="_blank"
                                                                                       rel="noopener noreferrer">Open
                            a Support Ticket</a></li>
                        <li><span className="float-left phone"><Isvg
                            src={require('./../../assets/svgs/phone.svg')}/></span><a href={SUPPORT_PHONE}
                                                                                      target="_blank"
                                                                                      rel="noopener noreferrer">Technical
                            Support Phone</a></li>
                    </ul>
                </MenuBar>
            </div>
        </header>

        {loading && <LoadingBar/>}

        {!loading && getSupportContentUX(auth)}

    </div>)
};

let SupportContentForm = reduxForm({
    form: 'SupportContent',
    validate
})(SupportContent);

const mapStateToProps = state => {
    return {
        auth: state.global.auth.data,
        _dataBrokers: state.syncReducer._dataBrokers,
    }
};

export default connect(mapStateToProps, {dataBrokers})(SupportContentForm);
