import React from 'react';
import {
    getServersByType,
    serverCredentialsStep,
    serverIPStep,
    clearUrlParamsIfChanged,
    clearShowAction,
    updateUrlParams,
    wizardStepStarted,
    wizardStepExited
} from '../../store/wizard/action_creators';
import {connect} from 'react-redux';
import stepinize from "./../wizard/stepinize";
import {Field, formValueSelector, reduxForm} from 'redux-form';
import {InputField} from "../_common_/forms/forms";
import {
    CIFS_VERSIONS,
    DEFAULT_CIFS_VERSION,
    END_POINT_NAMES_MAP,
    HOST_IPV4_IPV6_REG_EXP,
    HOST_NAME_REG_EXP,
    IP_REG_EXP,
    SELECTED_PARAMS_CIFS_SERVER,
    SELECTED_PARAMS_TARGET_CIFS_SERVER
} from '../../consts';
import Validator from 'validatorjs';
import {Route, Switch} from 'react-router-dom';
import ServerStepTemplate from "./ServerStepTemplate";
import {get} from 'lodash';
import {calculateStepId, getCredentialsByType} from '../../utils/sync-utils';
import {encodeAll} from "../../utils/encoders";
import Isvg from 'react-inlinesvg';
import {
    getIsIntegratedStep,
    fetchAnfVolumes
} from '../../utils/wizard-utils';
import {
    buildServersList,
    getWizardParamsFromUrlParams,
    getPropFromState
} from '../../utils/mapStateToPropsUtils';
const indefinite = require('indefinite');

const rules = (step) => {
    if (step === "credentials") {
        const schema = {
            serverName: ['required', 'max:256'],
            username: ['required'],
            password: 'required'
        };
        return schema;
    }
    if (step === 'ip') {
        return {
            serverName: ['required', 'max:256']
        }
    }
};

const shouldValidate = ({values, nextProps, props, initialRender, structure,}) => {
    const {_ui: {cifs}} = props;
    const step = cifs ? cifs.step : "ip";
    return (
        initialRender ||
        !structure.deepEqual(values, nextProps.values) ||
        step === "credentials"
    );
};

const validate = (values, props) => {
    let errors = {};
    const {serverName} = values;
    const {_ui: {cifs}} = props;
    const step = cifs ? cifs.step : "ip";
    const validator = new Validator(values, rules(step), {
        "max.serverName": "SMB hostname or IP address must be less than 256 characters"
    });
    validator.setAttributeNames({serverName: 'SMB server', username: "User Name"});
    validator.passes();
    errors = validator.errors.all();
    if (serverName && !IP_REG_EXP.test(serverName) && !HOST_NAME_REG_EXP.test(serverName) && !HOST_IPV4_IPV6_REG_EXP.test(serverName)) {
        errors.serverName = errors.serverName || [];
        errors.serverName.push("Invalid hostname or IP address");
    }
    return errors;
};

const CredentialsUi = ({protocols, isTargetStepFromSameSourceType}) => {
    const showACL = protocols === "cifs-cifs" && !isTargetStepFromSameSourceType;

    return (
        <div>
            <div className="row-medium credentials-container credentials">
                <div className="credentials-title">Define SMB Credentials:</div>
                <div className="medium-4 columns no-padding-left">
                    <Field name="username" component={InputField} label="User Name" type="text"/>
                </div>
                <div className="medium-4 columns">
                    <Field name="password" component={InputField} label="Password" type="password" autoComplete="off"/>
                </div>
                <div className="medium-4 columns no-padding-right">
                    <Field name="domain" component={InputField} label="Domain (Optional)"
                           type="text"/>
                </div>
            </div>
            {showACL && <div className="row-medium credentials-container acl">
                <div className="acl-title">ACL - Access Control List</div>
                <div className="no-padding-left acl">
                    <Field id="copyAcl" type="checkbox" component={InputField} name="copyAcl" className="acl-check-box">
                        <label htmlFor="copyAcl">Copy Access Control Lists to the target</label>
                    </Field>
                </div>
                <div className="acl-notice">
                    <ul>
                        <li><Isvg className="bullet" src={require("./../../assets/svgs/bullet-arrow.svg")}/>Copying ACLs can affect sync performance.</li>
                        <li><Isvg className="bullet" src={require("./../../assets/svgs/bullet-arrow.svg")}/>You can change this setting after you create the relationship.</li>
                    </ul>
                </div>
            </div>}
        </div>)
};

const CIFSServerStep = (props) => {
    if (!props._wizard) return <div/>;
    const {_servers, _anfVolumes, endpointObj, protocolAndProviderKey, isTargetStepFromSameSourceType, _wizard, _ui: {cifs, isCmIntegration}} = props;
    const {isSourceStep, protocols, selectedWizardParams} = _wizard;
    const {selectedCIFS, selectedCIFSTarget, cifsVersion, targetCifsVersion, selectedAnfWe, selectedAnfWeTarget, selectedAnfVolTarget, selectedAnfVol} = selectedWizardParams;

    const isFullAnfStep = getIsIntegratedStep(isCmIntegration, endpointObj);
    const relevantAnfWe = isTargetStepFromSameSourceType ? selectedAnfWeTarget : selectedAnfWe;
    const selectedVolume = isTargetStepFromSameSourceType ? selectedAnfVolTarget : selectedAnfVol;

    const serverType = endpointObj.provider || "smb";
    const selectedServer = isTargetStepFromSameSourceType ? selectedCIFSTarget : selectedCIFS;

    const stepOptions = {
        storage: _servers[protocolAndProviderKey],
        server: selectedServer,
        serverSVG: `${endpointObj.provider || "cifs"}-default.svg`,
        step: cifs ? cifs.step : "ip",
        showVolumesTable: isFullAnfStep,
        selectedVolumeId: selectedVolume, //relevant for case of "showVolumesTable"
        volumes:  _anfVolumes[relevantAnfWe], //relevant for case of "showVolumesTable"
        serverType: "CIFS",
        serverTitle: serverType.toUpperCase(),
        selectType: "Server",
        serverPlaceholder: "For example: 217.70.21.18 or netapp.com",
        title: `Select ${indefinite(serverType.toUpperCase())} ${!isSourceStep || isTargetStepFromSameSourceType ? "Target" : "Source"}`,
        selectedVersion: (isTargetStepFromSameSourceType ? targetCifsVersion : cifsVersion) || DEFAULT_CIFS_VERSION,
        versionsList: isFullAnfStep ? null : CIFS_VERSIONS,
        clearURLOptions: [{
            changedParam: isTargetStepFromSameSourceType ? SELECTED_PARAMS_TARGET_CIFS_SERVER : SELECTED_PARAMS_CIFS_SERVER,
            newValueFromState: "selectedIp",
            urlParams: isTargetStepFromSameSourceType ? ["sharePathTarget", "rootShareTarget"] : ["sharePath", "rootShare"]
        }],
        clearShowOptions: {keyToClear: ['_showShares', encodeAll(selectedServer)]},
        validate,
        getCredentialsUi: () => <CredentialsUi protocols={protocols} isTargetStepFromSameSourceType={isTargetStepFromSameSourceType}/>
    };

    return <ServerStepTemplate stepOptions={stepOptions} {...props}/>;
};

const willMount = (props, currentThis) => {
    const {isTargetStepFromSameSourceType, endpointObj, protocolAndProviderKey, _ui: {isCmIntegration}, _wizard: {selectedWizardParams}} = props;

    if (getIsIntegratedStep(isCmIntegration, endpointObj)) {
        const {selectedAnfWe, selectedAnfWeTarget} =  selectedWizardParams;
        const relevantAnfWe = isTargetStepFromSameSourceType ? selectedAnfWeTarget : selectedAnfWe;
        const isNeedForTimeout = !relevantAnfWe;
        if (isNeedForTimeout) setTimeout(fetchAnfVolumes.bind(this, currentThis, props), 300);
        else fetchAnfVolumes.bind(this, currentThis, props)();
    }
    else {
        if (!props._servers || !props._servers[protocolAndProviderKey] || !props._servers[protocolAndProviderKey].succeeded) {
            return props.getServersByType(endpointObj.protocol, endpointObj.provider, protocolAndProviderKey);
        }
    }
};

const stepId = ({_wizard: {wizardInfo}, isTargetStepFromSameSourceType}) => {
    return calculateStepId(wizardInfo, isTargetStepFromSameSourceType, "cifs");
};

const isFetching = (props) => {
    const {_ui: {isCmIntegration}, endpointObj} = props;

    if (getIsIntegratedStep(isCmIntegration, endpointObj)) {
        const {_anfVolumes, _wizard: {selectedWizardParams}, isTargetStepFromSameSourceType} = props;
        const {selectedAnfWe, selectedAnfWeTarget} =  selectedWizardParams;
        const relevantAnfWe = isTargetStepFromSameSourceType ? selectedAnfWeTarget : selectedAnfWe;
        return _anfVolumes[relevantAnfWe] && _anfVolumes[relevantAnfWe].inProgress;
    }
    else {
        const {_servers, protocolAndProviderKey} = props;
        return _servers[protocolAndProviderKey].inProgress;
    }
};

const canShowNextStep = ({_wizard, isTargetStepFromSameSourceType}) => {
    if (!get(_wizard, ["selectedWizardParams"])) return false;
    else if (isTargetStepFromSameSourceType) return _wizard.selectedWizardParams.selectedCIFSTarget && getCredentialsByType("cifs", _wizard.selectedWizardParams.selectedCIFSTarget, isTargetStepFromSameSourceType);
    else return _wizard.selectedWizardParams.selectedCIFS && getCredentialsByType("cifs", _wizard.selectedWizardParams.selectedCIFS, isTargetStepFromSameSourceType);
};

const nextRoutes = ({match, isTargetStepFromSameSourceType, nextStep, NextStepComp, _ui: {isCmIntegration}, endpointObj}) => {
    const isFullAnfStep = getIsIntegratedStep(isCmIntegration, endpointObj);

    return (
        <Switch>
            {isTargetStepFromSameSourceType &&
                <Route path={`${match.url}/:cifsServerTarget/:targetCifsVersion${isFullAnfStep ? "/:anfVolTarget/:rootExportTarget" : ""}/${nextStep.path}`} render={(props) =>
                    <NextStepComp {...props} isTargetStepFromSameSourceType={nextStep.isTargetStepFromSameSourceType}/>
                }/> }
            {!isTargetStepFromSameSourceType &&
                <Route path={`${match.url}/:cifsServer/:cifsVersion${isFullAnfStep ? "/:anfVol/:rootExport" : ""}/${nextStep.path}`} render={(props) =>
                    <NextStepComp {...props} isTargetStepFromSameSourceType={nextStep.isTargetStepFromSameSourceType}/>
                }/>}
        </Switch>
    )
};


let CIFSServerStepForm = reduxForm({
    form: 'cifsServer',
    validate,
    shouldValidate
})(CIFSServerStep);

const selector = formValueSelector('cifsServer');
CIFSServerStepForm = connect(
    (state) => ({
        copyAcl : selector(state, "copyAcl")
    })
)(CIFSServerStepForm);

const mapStateToProps = (state, ownProps) => {
    return {
        _servers: buildServersList(state),
        _wizard: getWizardParamsFromUrlParams(state, ownProps),
        _anfVolumes:  getPropFromState(state, "_anfVolumes", {}),
        _relationships: getPropFromState(state, "_relationships", {}),
        _ui: {...state.syncReducer._ui, ...state.global._ui},
        routing: state.router,
        ...ownProps
    }
};

const mapDispatchToProps = {
    getServersByType,
    serverCredentialsStep,
    serverIPStep,
    clearUrlParamsIfChanged,
    clearShowAction,
    updateUrlParams,
    wizardStepStarted,
    wizardStepExited
};

export default connect(mapStateToProps, mapDispatchToProps)(stepinize(CIFSServerStepForm, {
    stepId,
    willMount,
    isFetching,
    canShowNextStep,
    nextRoutes,
    loading: ({endpointObj}) => {
        const serverType = endpointObj.provider || endpointObj.protocol;
        return {
            text: `${END_POINT_NAMES_MAP[serverType.toUpperCase()]}`,
            icon: require(`../../assets/svgs/${serverType}-default.svg`)
        }
    }
}));


