import React, {useState} from 'react';
import {connect} from 'react-redux';
import {
    dataBrokers,
    clearUrlParamsIfChanged,
    updateUrlParams,
    wizardStepStarted,
    wizardStepExited
} from '../../store/wizard/action_creators'
import stepinize from './../wizard/stepinize'
import DataBrokers from './DataBrokers';
import {SystemFlow} from './../systemFlow/SystemFlow';
import UiStatePusher from './../UiStatePusher';
import HowDoesItWorkButton from './../systemFlow/HowDoesItWorkButton';
import Isvg from "react-inlinesvg";
import {FM_AGENT} from '../../consts';
import classNames from 'classnames';
import {Route, Switch} from 'react-router-dom';
import useRightPanel from '../../hooks/useRightPanel';
import DataBrokerWizard from '../dataBrokerWizard/DataBrokerWizard';
import ButtonWithNotification from "../widgets/ButtonWithNotification";
import _ from 'lodash';
import {getPropFromState, getWizardParamsFromUrlParams} from '../../utils/mapStateToPropsUtils';

const DataBrokerStep = props => {

    const [howItWorksClosed, setHowItWorksClosed] = useState(false);

    const {setPanelContent, closePanel} = useRightPanel();

    const {_dataBrokers: {data}, _wizard, up, isTargetStepFromSameSourceType, _deployments} = props;
    const {wizardInfo, wizardTypeForSystemFlow, isSecure, stepsOrder, currentStep, selectedWizardParams: {selectedIsTargetDataBrokerInitiator}} = _wizard;

    const isFirstStepOfWizard = stepsOrder[0] === currentStep;
    const showHowItWorks = data && data.length === 0 && !howItWorksClosed && isFirstStepOfWizard;

    const shouldDisplayPortBool = shouldDisplayPort({
        isSecure,
        selectedIsTargetDataBrokerInitiator,
        isTargetStepFromSameSourceType
    });

    const isAwsAccessRequired = (wizardInfo.source.protocol === 's3' && !wizardInfo.source.provider) || (wizardInfo.target.protocol === 's3' && !wizardInfo.target.provider);
    const isGcpAccessRequired = wizardInfo.source.protocol === 'gcp' || wizardInfo.target.protocol === 'gcp';

    //filteredDataBrokers contain all brokers that has port in case shouldDisplayPortBool is true, and all brokers without port otherwise.
    //This is because this component can be used in different places in the sync wizard.
    const filteredDataBrokers = data ? data.filter(dataBroker => (shouldDisplayPortBool ? !!dataBroker.port : !dataBroker.port)) : [];

    const filteredDataBrokersWithDeployInfo = _.map(filteredDataBrokers, dataBroker => {
        if (dataBroker.placement) return dataBroker;
        const deployment = _.find(_deployments, {dataBrokerId: dataBroker.id});
        return {...dataBroker, deployment}
    });

    const startDataBrokerWizard = () => {
        setPanelContent(<DataBrokerWizard
            closePanel={closePanel}
            isGcpAccessRequired={isGcpAccessRequired}
            isAwsAccessRequired={isAwsAccessRequired}
            shouldDisplayPortBool={shouldDisplayPortBool}
        />)
    };

    return (<>
        {filteredDataBrokers.length > 0 &&
        <DataBrokers startDataBrokerWizard={startDataBrokerWizard} shouldDisplayPortBool={shouldDisplayPortBool}
                     isAwsAccessRequired={isAwsAccessRequired} isGcpAccessRequired={isGcpAccessRequired}
                     data={filteredDataBrokersWithDeployInfo} {...props} />}
        {filteredDataBrokers.length === 0 &&
        <>
            {showHowItWorks &&
            <div className="wizard-step create-data-broker">
                <div className={classNames("scrollable-area", {up, down: !up})}>
                    <UiStatePusher ui="wizard no-steps">
                        <div className="how-it-works-step">
                            <h3>How It Works</h3>
                            <SystemFlow wizardType={wizardTypeForSystemFlow} isSecure={isSecure}/>
                        </div>
                    </UiStatePusher>
                </div>
                <ButtonWithNotification onClick={() => {setHowItWorksClosed(true)}} text="Continue"/>
            </div>}
            {!showHowItWorks &&
            <div className="wizard no-existing-brokers">
                <HowDoesItWorkButton wizardType={wizardTypeForSystemFlow} isSecure={isSecure}/>
                <Isvg src={require("../../assets/svgs/data-broker-default.svg")}/>
                <div className="title">{`You don't have any existing data brokers`}</div>
                <button className="button primary" onClick={startDataBrokerWizard}>
                    <span>+</span>
                    <div>Create Data Broker</div>
                </button>
            </div>
            }
        </>}
    </>)
};

const shouldDisplayPort = ({selectedIsTargetDataBrokerInitiator, isTargetStepFromSameSourceType, isSecure}) => {
    if (isSecure) {
        return ((selectedIsTargetDataBrokerInitiator === 'true' && !isTargetStepFromSameSourceType) ||
            (selectedIsTargetDataBrokerInitiator === 'false' && isTargetStepFromSameSourceType));
    }
    return false;
};

const stepId = ({isTargetStepFromSameSourceType}) => {
    return isTargetStepFromSameSourceType ? 'targetDataBroker' : 'dataBroker';
};

const willMount = (props) => {
    if (!props._dataBrokers || !props._dataBrokers.succeeded) {
        props.dataBrokers();
    }
};
const isFetching = ({_dataBrokers}) => {
    return (_dataBrokers.inProgress);
};
const canShowNextStep = (props) => {
    return props.isTargetStepFromSameSourceType ? _.get(props, '_wizard.selectedWizardParams.selectedDataBrokerTarget') : _.get(props, '_wizard.selectedWizardParams.selectedDataBroker');
};

const nextRoutes = ({match, isTargetStepFromSameSourceType, nextStep, NextStepComp}) => {
    return (
        <Switch>
            {isTargetStepFromSameSourceType &&
            <Route path={`${match.url}/:dataBrokerIdTarget/${nextStep.path}`} render={(props) =>
                <NextStepComp {...props} isTargetStepFromSameSourceType={nextStep.isTargetStepFromSameSourceType}/>
            }/>}
            {!isTargetStepFromSameSourceType &&
            <Route path={`${match.url}/:dataBrokerId/${nextStep.path}`} render={(props) =>
                <NextStepComp {...props} isTargetStepFromSameSourceType={nextStep.isTargetStepFromSameSourceType}/>
            }/>}
        </Switch>
    )
};

const mapStateToProps = (state, ownProps) => {
    return {
        _wizard: getWizardParamsFromUrlParams(state, ownProps),
        _dataBrokers: getPropFromState(state, "_dataBrokers", {}),
        _deployments: state.global._deployments,
        routing: state.router,
        ...ownProps
    }
};

const mapDispatchToProps = {
    dataBrokers,
    clearUrlParamsIfChanged,
    updateUrlParams,
    wizardStepStarted,
    wizardStepExited
};

export default connect(mapStateToProps, mapDispatchToProps)(stepinize(DataBrokerStep, {
    stepId,
    willMount,
    isFetching,
    canShowNextStep,
    nextRoutes,
    loading: () => {
        return {
            className: 'data-movers',
            text: FM_AGENT + 's',
            icon: require('./../../assets/svgs/data-broker-default.svg')
        }
    }
}));
