import React, {useState, useEffect, useRef} from 'react';
import Isvg from 'react-inlinesvg';
import HostsTable from './HostsTable.jsx';
import {encodeAll} from "../../utils/encoders";
import {SystemFlow} from "./../systemFlow/SystemFlow";
import UiStatePusher from "./../UiStatePusher";
import HowDoesItWorkButton from "./../systemFlow/HowDoesItWorkButton";
import classNames from "classnames";
import {Field} from 'redux-form';
import {InputField, SelectField} from '../_common_/forms/forms.jsx';
import _ from "lodash";
import {appendItemToLocalStorage, getCredentialsByType} from "../../utils/sync-utils";
import {END_POINT_NAMES_MAP} from "../../consts";
import AnfVolSelection from './AnfVolSelection';
import useRunOnceWhenTruthy from '../../hooks/useRunOnceWhenTruthy';
import ButtonWithNotification from '../widgets/ButtonWithNotification';
import {scrollToTop} from '../../utils/wizard-utils';

import "./serverStepTemplate.scss";
import "./nfsServerStep.scss";

const ServerStepTemplate = (props) => {
    const {stepOptions, isTargetStepFromSameSourceType, change, untouch, reset} = props;
    const {step, title, serverSVG, serverType, serverPlaceholder, versionsList, serverTitle, selectType,
        getCredentialsUi, storage, server, selectedVersion, clearURLOptions, clearShowOptions, selectedFieldIsNarrow = false, showVolumesTable, selectedVolumeId, volumes} = stepOptions;
    const {_relationships, _wizard: {wizardTypeForSystemFlow, isSecure, isSourceStep}, up, handleSubmit, _ui: {isCmIntegration}} = props;

    const [serverInfo, setServerInfo] = useState({});
    //TODO consider having the volume state only in the parent
    const [volumeInfo, setVolumeInfo] = useState(); //relevant for anf in cmIntegration
    const [data, setData] = useState(null);
    const scrollArea = useRef();

    useEffect(() => {
        const data = showVolumesTable ? volumes : _.uniqBy(storage?.data, "host");
        setData(data);
    }, [storage, setData, showVolumesTable, volumes]);

    useEffect(() => {
        scrollToTop(scrollArea?.current)
    }, [step]);

    useRunOnceWhenTruthy(() => {
        setServerInfo({...serverInfo, version: selectedVersion});

        if (server) {
            reset();
            if (!_.find(data, ["host", server]) || (step === "credentials")) {//if not part of the servers table, put in the input field, or if in second step
                change("serverName", server);
            }
            if (serverInfo.selectedIp !== server || serverInfo.error) {
                setServerInfo({...serverInfo, selectedIp: server, version: selectedVersion, error: null});
            }
            if (getCredentialsUi) {
                const credentials = getCredentialsByType(serverType, server, isTargetStepFromSameSourceType);
                //if there are already saved credentials for this ip, go directly to the second step, otherwise start with ip step
                if (credentials) {
                    _.forEach(_.keys(credentials), (key) => change(key, credentials[key]));
                }
                props.serverCredentialsStep(serverType);
            }
        }
        else props.serverIPStep(serverType);
    }, data);

    const changeVersion = (selectedVersion) => {
        setServerInfo({...serverInfo, version: selectedVersion.value});
    };

    const clearError = () => {
        setServerInfo({...serverInfo, error: null});
    };

    const submit = (values) => {
        if (!serverInfo.selectedIp) {
            setServerInfo({...serverInfo, error: `In order to continue to the next step please select ${serverTitle} ${selectType}.`});
            return Promise.reject();
        }

        const continueToNextStep = () => {
            _.forEach(clearURLOptions, (clearOption) => props.clearUrlParamsIfChanged(clearOption.changedParam, serverInfo[clearOption.newValueFromState], clearOption.urlParams, clearOption.extraKeyToClear));
            clearError();
            const nextStep = props.nextStep;
            const versionPath = serverInfo.version ? `/${serverInfo.version}` : "";
            const anfPath = volumeInfo?.volumeId ? `/${encodeAll(volumeInfo.volumeId)}/${encodeAll(volumeInfo.volumeRoot)}` : "";
            const path = `/${encodeAll(serverInfo.selectedIp)}${versionPath}${anfPath}/${nextStep.path}`;
            props._wizard.nextStep(path, props.history);
        };

        return Promise.resolve(true)
            .then(() => {
                if (getCredentialsUi) {
                    const cred = getCredentialsByType(serverType, serverInfo.selectedIp, isTargetStepFromSameSourceType);
                    const noCredentials = _.isEmpty(cred);
                    if (step === "ip" && (noCredentials)) {
                        props.serverCredentialsStep(serverType);
                    }
                    else {
                        const credentialsChanged = !_.isEqual(values, cred);
                        if (step === "credentials" && (noCredentials || credentialsChanged)) {
                            appendItemToLocalStorage(serverType, serverInfo.selectedIp, values, isTargetStepFromSameSourceType);
                            //force clear decoded params event if newValueFromState hasn't changed. since the credentials changed
                            if (credentialsChanged && !noCredentials) {
                                // in case credentials changed, make sure to force clear on parameters. unless there are no credentials yet, hence no need to clear
                                _.forEach(clearURLOptions, (clearOption) => props.clearUrlParamsIfChanged(clearOption.changedParam, serverInfo[clearOption.newValueFromState], clearOption.urlParams, clearOption.extraKeyToClear, true));
                                if (clearShowOptions) props.clearShowAction(clearShowOptions.keyToClear);
                            }
                        }
                        continueToNextStep();
                    }
                }
                else {
                    continueToNextStep();
                }
            });
    };

    const setSelected = (serverName) => {
        if (serverInfo.selectedIp === serverName) {
            setServerInfo({...serverInfo, selectedIp: null, error: null});
            //we can't use reset here, since it will reset it to the last initialized value
            change("serverName", "");
        }
        else {
            setServerInfo({...serverInfo, selectedIp: serverName, error: null});
            change("serverName", serverName);
        }
    };

    const setVolumeInStep = (volume) => {
        if (!volume) return;
        const {version, ipAddress, id, root} = volume;
        setVolumeInfo({volumeId: id, volumeRoot: root});
        if (version) {
            setServerInfo({...serverInfo, selectedIp: ipAddress, version});

        } else {
            setServerInfo({...serverInfo, selectedIp: ipAddress});
        }
        change("serverName", ipAddress);
    };

    const resetCredentialsInput = () => {
        change("domain", "");
        change("username", "");
        change("password", "");
        untouch("username");
        untouch("password");
        change("accessKey", "");
        change("secretKey", "");
        change("port", "");
        untouch("secretKey");
        untouch("accessKey");
        untouch("port");
    };

    const setNewServer = (serverNameEvent) => {
        resetCredentialsInput();
        const serverName = serverNameEvent.target.value;
        const {stepOptions: {validate}} = props;
        const errors = validate({serverName}, props);
        setServerInfo({...serverInfo, selectedIp: _.isEmpty(errors.serverName) ? serverName : null, error: null});
        return serverName;
    };

    const onWards = () => {
        setServerInfo({...serverInfo, error: null, howItWorksClosed: true});
    };

    const returnToIpStep = () => {
        const {stepOptions: {serverType}} = props;
        props.serverIPStep(serverType);
    };


    const {error, howItWorksClosed, selectedIp} = serverInfo;
    const showHowItWorks = !isCmIntegration && _relationships && _relationships.data && _relationships.data.length === 0 && !howItWorksClosed && isSourceStep && !server;
    const firstTime = data?.length === 0;
    const className = `wizard-step nfs-server-step ${step}`;
    const selectedVersionValue = serverInfo.version || selectedVersion;
    const selectedOption = {value: selectedVersionValue, label: selectedVersionValue};

    return (<>
        {!data ? <div/> :
            <div className={className}>
                <form className="wizard-step-form" onSubmit={handleSubmit(submit)}>
                    <div className={classNames("scrollable-area", {up, down: !up})} ref={scrollArea}>
                        {showHowItWorks &&
                        <UiStatePusher ui="wizard no-steps">
                            <div className="how-it-works-step">
                                <h3>How It Works</h3>
                                <SystemFlow wizardType={wizardTypeForSystemFlow} isSecure={isSecure}/>
                            </div>
                        </UiStatePusher>}
                        {!showHowItWorks && <>
                            <div className="title-and-actions">
                                <div className="actions-left">
                                    <HowDoesItWorkButton wizardType={wizardTypeForSystemFlow} isSecure={isSecure}/>
                                </div>
                                <div className="title-middle">
                                    <h3>{title}</h3>
                                </div>
                                <div className="actions-right"/>
                            </div>
                            {versionsList &&
                            <div className="version">
                                <div className="version-label">{END_POINT_NAMES_MAP[serverType]} Version :</div>
                                <Field component={SelectField} selectedValue={selectedOption}
                                       onChange={changeVersion} valueField="value"
                                       name="version"
                                       options={versionsList} className="nfs-version-cb bottom-border"
                                       noSearch={true}/>
                            </div>
                            }
                            {(!showVolumesTable || (showVolumesTable && step === "credentials") ) &&
                            <div className="row-medium align-center selected-server">
                                <div className={classNames("new-nfs", {selected: step === "credentials", narrow: selectedFieldIsNarrow})}>
                                    <Isvg src={require(`./../../assets/svgs/${serverSVG}`)}/>
                                    <div className="field-container nfs-address">
                                        <Field name="serverName" onChange={setNewServer}
                                               component={InputField}
                                               placeholder={serverPlaceholder}
                                               label={step === "ip" ? `Define a New ${serverTitle} ${selectType}:` : `Selected ${serverTitle} ${selectType}:`}
                                               readOnly={step === "credentials" ? "readonly" : false}
                                               type="text"/>
                                    </div>
                                    {step === "credentials" &&
                                    <div className="close-step" onClick={returnToIpStep}>
                                        <span><Isvg src={require('./../../assets/svgs/x.svg')}/></span>
                                    </div>
                                    }
                                </div>
                            </div>}

                            {!firstTime && !showVolumesTable && step !== "credentials" && <div className="row-medium align-center">
                                <div className="nfs-server-table">
                                    <div className="or-divider">
                                        <div className="left-border"/>
                                        <div className="circle">or</div>
                                        <div className="right-border"/>
                                    </div>
                                    <div className="select-header">Select an Existing {serverTitle} {selectType}</div>
                                    <div className="hostsSelect"><HostsTable selector={setSelected}
                                                                             tableData={data}
                                                                             selectedId={selectedIp}/></div>
                                </div>
                            </div>}

                            {step === "credentials" && getCredentialsUi()}

                            {showVolumesTable && step === "ip" &&
                            <AnfVolSelection
                                protocol={serverType.toLowerCase() === 'cifs' ? 'smb' : serverType.toLowerCase()}
                                volumes={data}
                                defaultVolumeId={selectedVolumeId}
                                setVolumeInStep={setVolumeInStep}/>
                            }
                        </>}
                    </div>

                    {showHowItWorks && <ButtonWithNotification
                        onClick={onWards}
                        text="Continue"/>}

                    {!showHowItWorks && <ButtonWithNotification
                        type="submit"
                        error={error}
                        clearError={clearError}
                        text="Continue"/>}

            </form>
        </div>}
        </>)
};

export default ServerStepTemplate;