import React, {useState, useRef} from 'react';
import Isvg from 'react-inlinesvg';
import {encodeAll, clearDoubleSlash} from "../../utils/encoders"
import HowDoesItWorkButton from "./../systemFlow/HowDoesItWorkButton";
import DrillDownBreadCrumbs from "./../table/DrillDownBreadCrumbs";
import {NEW_TARGET_FOLDER_ID} from "../../consts";
import _ from "lodash";
import classNames from "classnames";
import TargetSelectionNotice from "./TargetSelectionNotice"
import {openDialog} from "../dialog/Dialog";
import RefreshButton from "../refreshButton/RefreshButton";
import ButtonWithNotification from '../widgets/ButtonWithNotification';
import {scrollToTop} from "../../utils/wizard-utils";

require("./showDirectoriesStepTemplate.scss");

const FOLDER_REGEX = /^[./]/;

const ShowDirectoriesStep = (props) => {
    const {stepOptions, isTargetStepFromSameSourceType, nextStep, _wizard, history, up, handleSubmit} = props;
    const {selectedDirectory, selectedRootDirectory, selectedDirectoryTarget, selectedRootDirectoryTarget} = stepOptions;
    const initialPath = isTargetStepFromSameSourceType ? selectedDirectoryTarget : selectedDirectory;
    const initialRootDirectory = isTargetStepFromSameSourceType ? selectedRootDirectoryTarget : selectedRootDirectory;

    const [state, setState] = useState({
        selectedId: initialPath || null,
        error: null,
        initialPath: initialPath || null,
        rootDirectory: initialRootDirectory || null,
        rootValue: null
    });
    const scrollArea = useRef();

    const setSelected = (selected) => {
        setState({...state, selectedId: selected.value, rootValue: selected.rootValue, error: null});
    };

    const submit = (values) => {
        return Promise.resolve(true)
            .then(() => {
                onContinue(values);
            });
    };

    const goToNextStep = (fullPath, rootValue) => {
        setState({...state, error: null});
        _wizard.nextStep(`/${encodeAll(clearDoubleSlash(fullPath))}/${encodeAll(clearDoubleSlash(rootValue ? rootValue : fullPath))}/${nextStep.path}`, history);
    };

    const onContinue = (values) => {
        let error;
        if (!state.selectedId) {
            error = "In order to continue to the next step please select a directory.";
        }
        // user chose a bucket
        else {
            // chose new target folder - validate
            if (state.selectedId.endsWith(NEW_TARGET_FOLDER_ID)) {
                const {targetFolder} = values;
                // validate folder name exists
                if (!targetFolder || !_.isString(targetFolder)) {
                    error = "Please enter a folder name.";
                }
                // block chars that are illegal in linux
                else if (FOLDER_REGEX.test(targetFolder)) {
                    error = "Target folder name shouldn't start with . or /";
                } else if (targetFolder.length > 255) {
                    error = "Target folder name's maximum length should be 255 characters";
                }

                // folder name exists
                if (!error) {
                    //use replace with the function as parameter and not string otheriwse $ will be used as escaped character causing unexpected target folder names...
                    //see SER-1678 Create target folder on NFS - problem with dollar sign
                    const targetFolderFullPath = state.selectedId.replace(NEW_TARGET_FOLDER_ID, () => targetFolder);
                    goToNextStep(targetFolderFullPath, state.rootValue);
                    return;
                }
            }
            // chose existing folder - no validations
            else {
                goToNextStep(state.selectedId, state.rootValue);
            }
        }

        // update errors only in case there are errors, otherwise we get errors in the console for trying to set state of unmounted component
        if (error)
            setState({...state, error});
    };

    const drillDown = (path, firstLayerPath) => {
        const {storage, storageKey, listSubDirectoriesAction, selectedServer, selectedServerTarget, additionalParam} = stepOptions;
        const {isSecure, selectedWizardParams: {selectedDataBroker, selectedDataBrokerTarget}} = _wizard;
        const isUsingDataBrokerTarget = isSecure && isTargetStepFromSameSourceType;
        const relevantDataBroker = isUsingDataBrokerTarget ? selectedDataBrokerTarget : selectedDataBroker;
        const actualServer = isTargetStepFromSameSourceType ? selectedServerTarget : selectedServer;
        const folders = storage[storageKey].folders;
        if (folders[path] && (folders[path].succeeded || folders[path].inProgress)) return;
        if (additionalParam)
            listSubDirectoriesAction(relevantDataBroker.id, actualServer, path, firstLayerPath, additionalParam, isTargetStepFromSameSourceType);
        else listSubDirectoriesAction(relevantDataBroker.id, actualServer, path, firstLayerPath, isTargetStepFromSameSourceType);
    };

    const isRefreshDisabled = () => {
        const {selectedServer, selectedServerTarget} = stepOptions;
        const {isSourceStep} = _wizard;
        return (selectedServer === selectedServerTarget && !isSourceStep);
    };

    const {failedMessage, notFoundMessage, serverTypeName, storage, storageKey, containerOptions, title,
        selectedServer, selectedServerTarget, innerColumns, stepFormName, addManualOptions, refreshStep, isStaticRoot} = stepOptions;
    const {manualForm: ManualForm, directoryType, addManualAction, addManualText} = addManualOptions ? addManualOptions : {manualForm: {}};
    const {wizardTypeForSystemFlow, isSecure, isSourceStep} = _wizard;
    const actualStorage = storage[storageKey];
    const {data, folders, retrievedAll, failedRetrieve} = actualStorage;
    const error = actualStorage.failed ? actualStorage.data : state.error;

    return (
        <div className="wizard-step show-export-step">
            <form className="wizard-step-form" onSubmit={handleSubmit(submit)}>
                <div className={classNames("scrollable-area", {up, down: !up})} ref={scrollArea}>
                    <div className="title-and-actions">
                        <div className="actions-left">
                            <HowDoesItWorkButton wizardType={wizardTypeForSystemFlow} isSecure={isSecure}/>
                        </div>
                        {(data && data.length > 0) && <div className="title-and-refresh">
                            <h3 className="title" id="show-directories-step">{title}</h3>
                            <RefreshButton onClick={refreshStep} storageObjectName={containerOptions.pluralTitle} disabled={isRefreshDisabled()}/>
                        </div>}
                        <div className="actions-right">
                            {addManualAction && (failedRetrieve || data.length === 0) && !state.rootValue &&
                            <button type="button" className="outline new-export"
                                    onClick={(el) => openDialog(<ManualForm className="new-bucket export"
                                                                            directoryType={directoryType}
                                                                            onSubmitFunction={addManualAction}/>, el.target)}>
                                <Isvg src={require('./../../assets/svgs/plus-fill-circle-icon.svg')}/>{addManualText}
                            </button>}
                        </div>
                    </div>
                    {(data && data.length > 0) &&
                    <div>
                        {!isSourceStep &&
                        <div className="header-text">Drill down in order to create a new folder inside an export or
                            inside an existing folder</div>}

                        {!isSourceStep && <TargetSelectionNotice selected={state.selectedId} directory={true}/>}
                        <DrillDownBreadCrumbs containerOptions={containerOptions}
                                              isStaticRoot={isStaticRoot}
                                              drillDown={drillDown} data={folders || []}
                                              selectionOptions={{
                                                  selector: setSelected,
                                                  selectedId: state.selectedId,
                                                  pathPrefix: '/'
                                              }}
                                              scrollToTop={() => scrollToTop(scrollArea.current)}
                                              innerColumns={innerColumns}
                                              canCreateTargetFolder={!isSourceStep || isTargetStepFromSameSourceType}
                                              resetForm={props.reset}
                                              initialPath={state.initialPath}
                                              rootExport={state.rootDirectory} formName={stepFormName}
                                              beginWithSlash={true} {...props}/>
                    </div>}

                    {(retrievedAll && ((data && data.length === 0) || !data)) &&
                    <div className="nothing-found">
                        <span><Isvg src={require("./../../assets/svgs/folder-icon.svg")}/></span>
                        <p>{failedRetrieve ? `${failedMessage}` : `${notFoundMessage}`}</p>
                        <p>in {serverTypeName} server {isTargetStepFromSameSourceType ? selectedServerTarget : selectedServer}</p>
                    </div>}

                </div>
                <ButtonWithNotification type="submit" error={error} text="Continue"/>
            </form>
        </div>
    );
};

export default ShowDirectoriesStep;