import React, {useState, useRef, useEffect} from 'react';
import Isvg from 'react-inlinesvg';
import {encodeAll} from "../../utils/encoders";
import HowDoesItWorkButton from "./../systemFlow/HowDoesItWorkButton";
import DrillDownBreadCrumbs from "./../table/DrillDownBreadCrumbs";
import {NEW_TARGET_FOLDER_ID, STORAGE_OBJECT_KEY_MAX_LENGTH} from "../../consts";
import _ from "lodash";
import classNames from "classnames";
import {openDialog} from "./../dialog/Dialog"
import {NewBucketForm} from "./../systemFlow/NewBucketDialog";
import TargetSelectionNotice from "./TargetSelectionNotice"
import RefreshButton from "../refreshButton/RefreshButton";
import ButtonWithNotification from '../widgets/ButtonWithNotification';
import useRunOnce from '../../hooks/useRunOnce';
import {scrollToTop} from '../../utils/wizard-utils';

require("./showObjectStorageStep.scss");

const OBJECT_STORAGE_FOLDER_REGEX = /\/+/;

const ShowObjectStorageStepTemplate = (props) => {
    const {stepOptions, nextStep, clearUrlParamsIfChanged, _wizard, history, addNewBucket: addNewBucketAction, up, isSourceStep, handleSubmit} = props;
    const {storageObjectName, storageObjectType, currentStep, updateStepInfo, containerOptions,
        selectedBaseObject, resetToFirstStepAction, clearURLOptions, extraUrlParam, storage,
        listSubFoldersAction, getStorageByBrokerId, failedMessage, notFoundMessage, bottomMessage,
        innerColumns, innerPathBeginWithSlash, stepFormName, title, getSecondStepUi, supportAddStorageObject, refreshStep} = stepOptions;
    const {directoryName, nameForSetupStep} = containerOptions;
    const {wizardTypeForSystemFlow, isSecure, selectedWizardParams: {selectedDataBroker}} = _wizard;

    const [state, setState] = useState({
        selectedId: selectedBaseObject || null,
        error: null,
        initialPath: selectedBaseObject || null,
        step: ""
    });
    const scrollArea = useRef();

    useEffect(() => {
        scrollToTop(scrollArea?.current)
    }, [currentStep]);


    useRunOnce(() => {
        if (resetToFirstStepAction) {
            resetToFirstStepAction();
        }
    });

    const setSelected = (selected) => {
        setState({...state, selectedId: selected.value, error: null});
    };

    const submit = (values) => {
        return Promise.resolve(true)
            .then(() => {
                onContinue(values);
            });
    };

    const onContinue = (values) => {
        let error;

        const whatNext = (step, targetFolderFullPath) => {
            let shouldContinue = true;
            if (updateStepInfo) {
                shouldContinue = updateStepInfo(step, targetFolderFullPath);
            }
            if (shouldContinue)
                goToNextStep(targetFolderFullPath);
        };

        if (!state.selectedId) {
            error = `In order to continue to the next step please select a ${storageObjectName}.`;
        }
        // user chose a storageObject
        else {
            const step = currentStep || "select";
            // chose new target folder - validate
            if (state.selectedId.endsWith(NEW_TARGET_FOLDER_ID)) {
                const path = state.selectedId.replace(NEW_TARGET_FOLDER_ID, ''); // remove target const from path
                const {targetFolder} = values;
                // validate folder name exists
                if (!targetFolder || !_.isString(targetFolder)) {
                    error = `Please enter ${storageObjectType} ${directoryName} name.`;
                }
                // folder name exists
                else {
                    // validate object key length
                    const storageObjectSlashIndex = path.indexOf('/');
                    const objectStoragePathLength = (storageObjectSlashIndex === -1) ? 0 : path.length - storageObjectSlashIndex; // path without the storageObject name
                    if (objectStoragePathLength + targetFolder.length >= STORAGE_OBJECT_KEY_MAX_LENGTH) {
                        error = `${storageObjectType} object key must be less than 1024 characters.`;
                    }
                    // forward slash is not allowed in S3 folder name
                    else if (OBJECT_STORAGE_FOLDER_REGEX.test(targetFolder)) {
                        error = `${storageObjectType} ${directoryName} name shouldn't contain /`;
                    }
                }

                // valid folder name - go to next path with storageObject path that the folder will be created in
                if (!error) {
                    const targetFolderFullPath = state.selectedId.replace(NEW_TARGET_FOLDER_ID, () => targetFolder);
                    setState({...state, selectedId: targetFolderFullPath});

                    whatNext(step, targetFolderFullPath);
                }
            }
            // chose existing folder - no validations
            else {
                whatNext(step, state.selectedId);
            }
        }

        // update errors only in case there are errors, otherwise we get errors in the console for trying to set state of unmounted component
        if (error)
            setState({...state, error});
    };

    const goToNextStep = (storageObjectPath) => {
        _.forEach(clearURLOptions, (clearOption) => clearUrlParamsIfChanged(clearOption.changedParam, state[clearOption.newValueFromState], clearOption.urlParams, clearOption.extraKeyToClear));
        // clear errors
        setState({...state, error: null});
        //set it back to selection step
        if (resetToFirstStepAction)
            resetToFirstStepAction();
        // go to next step
        if (_wizard.isSourceStep) _wizard.nextStep(`/${encodeAll(storageObjectPath)}/${nextStep.path}`, history);
        else {
            const path = extraUrlParam ? `/${encodeAll(storageObjectPath)}/${extraUrlParam}/${nextStep.path}` : `/${encodeAll(storageObjectPath)}/${nextStep.path}`;
            _wizard.nextStep(path, history);
        }
    };


    const drillDown = (path, storageObject) => {
        const {selectedWizardParams: {selectedDataBroker}} = _wizard;
        const storageObjectsByBrokerId = getStorageByBrokerId ? storage[selectedDataBroker.id] : storage;
        if (storageObjectsByBrokerId && storageObjectsByBrokerId.folders && storageObjectsByBrokerId.folders[path] && (storageObjectsByBrokerId.folders[path].succeeded || storageObjectsByBrokerId.folders[path].inProgress)) return;
        listSubFoldersAction(selectedDataBroker.id, storageObject, path);
    };

    const addNewBucket = (values) => {
        addNewBucketAction(values.bucketName, _wizard.selectedWizardParams.selectedDataBroker.id);
        setSelected({value: values.bucketName});
    };

    const isRefreshDisabled = () => {
        const {isSourceStep, wizardTypeForSystemFlow, selectedWizardParams} = _wizard;
        const {selectedStorageAccountName, selectedStorageAccountNameTarget, selectedS3Based, selectedS3BasedTarget} = selectedWizardParams;
        const isSameAzureHost = selectedStorageAccountNameTarget && (selectedStorageAccountName === selectedStorageAccountNameTarget);
        const isSameS3Based = selectedS3BasedTarget && (selectedS3Based === selectedS3BasedTarget);

        return ((wizardTypeForSystemFlow === 's3-s3' || wizardTypeForSystemFlow === 'gcp-gcp'|| isSameAzureHost || isSameS3Based) && !isSourceStep);
    };

    const storageObjectsByBrokerId = getStorageByBrokerId ? storage[selectedDataBroker.id] : storage;
    const step = currentStep || "select";
    const storageObjectPath = state.selectedId ? state.selectedId.split('/')[0] : "";
    const error = storageObjectsByBrokerId && storageObjectsByBrokerId.failed ? storageObjectsByBrokerId.data : state.error;
    const {data, folders, retrievedAll, failedRetrieve} = storageObjectsByBrokerId;
    const headerAddon = () => <button type="button" className="new-bucket"
                                      onClick={(el) => openDialog(<NewBucketForm className="new-bucket"
                                                                                 onSubmitFunction={addNewBucket}/>, el.target)}>
        <span><Isvg src={require('./../../assets/svgs/plus-fill-circle-icon.svg')}/></span>Add to the list</button>;

    return (
        <div className="wizard-step show-buckets-step">
            <form className="wizard-step-form" onSubmit={handleSubmit(submit)}>
                <div className={classNames("scrollable-area", {up, down: !up})} ref={scrollArea}>
                    <div className="title-and-actions">
                        <div className="actions-left">
                            <HowDoesItWorkButton wizardType={wizardTypeForSystemFlow} isSecure={isSecure}/>
                        </div>
                        {data && data.length > 0 && <div className={classNames("title-and-refresh", {"no-refresh": step !== "select"})}>
                            <h3 className="title" id="show-storage-step">{step === "select" ? title : nameForSetupStep}</h3>
                            {step === "select" && <RefreshButton onClick={refreshStep} storageObjectName={`${storageObjectName}s`} disabled={isRefreshDisabled()}/>}
                        </div>}
                        <div className="actions-right">
                        </div>
                    </div>

                    {data && data.length > 0 && step === "select" &&
                    <div className={classNames({"disable-change": step !== "select"})}>
                        <div>
                            {!isSourceStep &&
                            <div className="header-text">Drill down to select an existing {directoryName} within
                                the {storageObjectName} or
                                to select a new {directoryName} that you create inside the {storageObjectName}
                            </div>}
                            {!isSourceStep && <TargetSelectionNotice selected={state.selectedId}
                                                                     directory={directoryName === "directory"}/>}
                        </div>
                        <DrillDownBreadCrumbs containerOptions={containerOptions}
                                              drillDown={drillDown} data={folders}
                                              selectionOptions={{
                                                  selector: setSelected,
                                                  selectedId: state.selectedId,
                                                  pathPrefix: ''
                                              }}
                                              scrollToTop={() => scrollToTop(scrollArea?.current)}
                                              innerColumns={innerColumns} canCreateTargetFolder={!isSourceStep}
                                              resetForm={props.reset}
                                              initialPath={state.initialPath || state.selectedId}
                                              formName={stepFormName} {...props}
                                              beginWithSlash={innerPathBeginWithSlash}
                                              showHeaderAddonOnRootOnly={true}
                                              headerAddon={supportAddStorageObject ? headerAddon : undefined}/>
                    </div>}

                    {((retrievedAll && step === "select" && data && data.length === 0) || !data) &&
                    (<div className="nothing-found">
                        <span><Isvg src={require("./../../assets/svgs/bucket-icon.svg")}/></span>
                        <p>{failedRetrieve ? `${failedMessage}` : `${notFoundMessage}`}</p>
                        <p>{bottomMessage}</p>
                    </div>)
                    }

                    {step === 'secondStep' && getSecondStepUi(step, storageObjectPath)}
                </div>

                <ButtonWithNotification type="submit" error={error} text="Continue"/>
            </form>
        </div>
    );
};

export default ShowObjectStorageStepTemplate;