import React, {useState, useEffect, useRef} from 'react';
import _ from "lodash";
import SelectableTable from "./../table/SelectableTable";
import {defaultFilter} from "./../../utils/helpers"
import ItemCountWidget from "./../widgets/ItemCountWidget"
import SearchWidget from "./../widgets/SearchWidget"
import Isvg from 'react-inlinesvg';
import ScrollArea from "react-scrollbar"
import classNames from "classnames";
import CreateFolderRow from "./CreateFolderRow";
import {NEW_TARGET_FOLDER_ID} from "../../consts";
import {removeStartingAndEndingSlash} from "./../../utils/encoders";
import {breakFullPathToRootAndPath} from "./../../utils/sync-utils";
import useRunOnceWhenTruthy from '../../hooks/useRunOnceWhenTruthy';

require("./drilldownBreadcrumbs.scss");

const DrillDownBreadCrumbs = (props) => {
    const {initialPath, rootExport, data: folders, canCreateTargetFolder, isStaticRoot} = props;

    const [state, setState] = useState({paths: [], currentPath: "-"});
    const [filter, setFilter] = useState(null);
    const _scrollbar = useRef();

    useRunOnceWhenTruthy(() => {
        //In cmIntegration anf (isStaticRoot), the root is decided in the anfVolume step. So we need to programmatically drill down to this vol.
        //After "continue", the rootExport is added a "/" at the beginning, so we know not to do this again (back navigation) - see CS-4719
        if (rootExport && !rootExport.startsWith("/")) {
            drillDown({path: `/${rootExport}`, actualPath: `/${rootExport}`});
        }
    }, isStaticRoot);

    useRunOnceWhenTruthy(()=> {
        const {selectionOptions} = props;

        const {pathWithoutRoot, rootOfPath} = breakFullPathToRootAndPath(initialPath, rootExport);
        const trimmedRoot = removeStartingAndEndingSlash(rootOfPath);//remove first / from root path
        const paths = _.compact(pathWithoutRoot.split("/"));

        //if the path is /folder1/folder2. then currentLocation to show is folder1, and folder2 will be selected in it.
        const currentLocationFullPaths = paths.length < 1 ? paths : _.take(paths, paths.length - 1);
        const currentLocationPaths = _.concat([trimmedRoot], currentLocationFullPaths);

        //if root is /, then we shouldn't  add another / at the beginning.
        const currentLocation = rootExport === "/" ? currentLocationPaths.join("/") : selectionOptions.pathPrefix + currentLocationPaths.join("/");
        const updatedCurrentLocation = currentLocation;

        //if root is /, then we shouldn't  add another / at the beginning.
        const updatedInitialPath = rootExport === "/" ? rootExport + initialPath : initialPath;

        // console.log(`%c DrillDown folders ${JSON.stringify(folders)} paths ${paths} currentLocation ${updatedCurrentLocation} initialPath ${updatedInitialPath} rootExport ${rootExport}`, "font-weight: bold;");
        //in case of refresh, if we are in inner folders we can't have the folders ready for us, skip selection
        if (!(folders && _.has(folders, updatedCurrentLocation))) {
            return;
        }
        const innerFolders = folders[updatedCurrentLocation].data;
        const targetFolder = !_.find(innerFolders, {actualPath: updatedInitialPath}) && initialPath !== currentLocation && canCreateTargetFolder ? paths[paths.length - 1] : null;
        setState({...state,
            paths: currentLocationPaths,
            currentPath: updatedCurrentLocation,
            targetFolder
        });
        // console.log("!!!!!!!!!!!!!!!!! drillDown updating selector with ", updatedInitialPath, " and root value ", updatedCurrentLocation);
        props.selectionOptions.selector({value: updatedInitialPath, rootValue: rootExport});

    }, initialPath);

    useEffect(() => {
        setTimeout(() => {
            if (_scrollbar.current) {
                _scrollbar.current.scrollArea.refresh();
                _scrollbar.current.scrollRight();
            }
        }, 200);
    }, [state.currentPath]);

    const getRootPath = (currentPath) => {
        const {selectionOptions: {pathPrefix}} = props;
        const rootPath = state.paths.length > 0 ? pathPrefix + state.paths[0] : currentPath;
        return rootPath;
    };

    const drillDown = (row) => {
        const {selectionOptions: {pathPrefix}} = props;
        const path = row.path;
        const pathNoSlash = removeStartingAndEndingSlash(path);
        const initialPathForDrillDown = state.paths.length > 0 ? pathPrefix + state.paths[0] : row.actualPath;
        // console.log(`%c DrillDownBreadCrumbs root ${initialPathForDrillDown} path (${row.actualPath})`, "font-weight: bold;");
        props.drillDown(row.actualPath, initialPathForDrillDown);
        setState({...state,
            paths: state.paths.concat(pathNoSlash),
            currentPath: row.actualPath
        });
        // if not selected - select row path
        if (props.selectionOptions.selectedId !== row.actualPath) {
            props.selectionOptions.selector({value: row.actualPath, rootValue: getRootPath(row.actualPath)});
        }
        props.closeSearchWidget();
        props.scrollToTop();
    };

    const setSelected = (id) => {
        // if already selected - un-select
        if (props.selectionOptions.selectedId === id) {
            props.selectionOptions.selector(state.currentPath === "-" ? {
                value: null,
                rootValue: null
            } : {value: state.currentPath, rootValue: getRootPath(state.currentPath)});
        }
        // else not selected - select
        else {
            props.selectionOptions.selector({value: id, rootValue: getRootPath(id)});
        }
    };

    const renderRow = (innerColumns, row) => {
        return (
            <div>
                <td className="check"><span><Isvg src={require("./../../assets/svgs/checkmark-fill-circle-icon.svg")}/></span>
                </td>
                {innerColumns && innerColumns(row).map((td, key) => React.cloneElement(td, {key}))}
                <td className="dive">
                    <button type="button" onClick={(e) => {
                        drillDown(row);
                        e.stopPropagation();
                    }}><span><Isvg src={require("./../../assets/svgs/right-chevron.svg")}/></span></button>
                </td>
            </div>
        )
    };

    const goBackTo = (index) => {
        const {selectionOptions} = props;
        let newPathsArray = _.take(state.paths, index);
        const newPath = newPathsArray.length > 0 ? selectionOptions.pathPrefix + newPathsArray.join("/") : "-";
        const newState = {...state, paths: newPathsArray, currentPath: newPath};
        const rootPath = getRootPath(newPath);
        // console.log(`%c DrillDownBreadCrumbs goBackTo root @ ${rootPath} newPath ${newPath} paths ${newPathsArray}`, "font-weight: bold;");
        setState(newState);
        props.selectionOptions.selector(newPath === "-" ? {value: null, rootValue: null} : {
            value: newPath,
            rootValue: rootPath
        });
    };


    const {paths, targetFolder} = state;
    //isStaticRoot: In cmIntegration anf, the root is decided in the anfVolume step.
    //We want to block the option to show all the volumes in that specific ip, and to show the ip instead of "all directories" text.
    const {
        headerAddon: HeaderAddon, showHeaderAddonOnRootOnly, containerOptions: {itemIdentifier, rootName, targetName, icon, directoriesName},
        innerColumns,  dispatch, formName, isTargetStepFromSameSourceType,
        _wizard: {selectedWizardParams: {selectedNFS, selectedCIFS}}, _ui: {isCmIntegration}
    } = props;
    const currentPath = state.currentPath === "//" ? "/" : state.currentPath;
    let foldersList = folders[currentPath];
    const {data, retrievedAll} = foldersList;
    const filteredItems = filter ? data.filter(item => defaultFilter(item, filter)) : data || [];
    const selectionOptions = ({...props.selectionOptions, selector: setSelected});
    const selectedNewTargetFolder = selectionOptions.selectedId?.endsWith(NEW_TARGET_FOLDER_ID);
    const hasSelection = data?.find(item => item[itemIdentifier] === selectionOptions.selectedId) || selectedNewTargetFolder;
    const shouldShowHeaderAddon = (!showHeaderAddonOnRootOnly || (showHeaderAddonOnRootOnly && paths.length === 0)) && HeaderAddon;

    return (<div className={classNames("drilldown-breadcrumbs", {"has-selection": hasSelection})}>
        <div className="search-row clearfix">
            <span className="float-left icon"><Isvg src={icon}/></span>
            <ItemCountWidget className="float-left" itemCount={data ? data.length : 0}
                             filteredItemCount={filteredItems.length}/>
            <SearchWidget className="float-right" setFilter={setFilter} {...props}/>
        </div>
        <div className="breadcrumbs-row clearfix">
            <ScrollArea
                style={{width: "100%", height: "100%"}}
                className="breadcrumb-scroll"
                contentClassName="breadbox"
                smoothScrolling={true}
                vertical={false}
                minScrollSize={40}
                ref={e => _scrollbar.current = e}
            >
                <span className="path">
                    <button type="button" className="name"
                                            onClick={() => goBackTo(0)}
                                            disabled={paths.length === 0 || isStaticRoot}>
                    {isStaticRoot ? (selectedNFS || selectedCIFS) : rootName }
                    </button>
                </span>

                {shouldShowHeaderAddon && <HeaderAddon/>}
                {paths.map((path, index) => (
                    (isCmIntegration && isTargetStepFromSameSourceType && index === 0) ? <span key={index}/> :
                        <span className="path" key={index}>
                    <span className="chevron"><Isvg src={require('./../../assets/svgs/right-chevron.svg')}/></span>
                    <span className="check"><Isvg
                        src={require("./../../assets/svgs/checkmark-fill-circle-icon.svg")}/></span>
                    <span className="name"><button type="button" onClick={() => {
                        goBackTo(index + 1)
                    }} disabled={index === paths.length - 1}>{path || "/"}</button></span>
                </span>
                ))}
            </ScrollArea>
        </div>
        <div>
            {state.currentPath !== "-" && canCreateTargetFolder &&
            <CreateFolderRow itemIdentifier={itemIdentifier} selectionOptions={selectionOptions}
                             currentPath={state.currentPath} resetForm={props.resetForm}
                             targetPlaceholder={targetName} initialTargetFolder={targetFolder} dispatch={dispatch}
                             formName={formName}/>
            }

            {(filteredItems.length > 0 || !retrievedAll) &&
            <SelectableTable
                noAnimate={true}
                Table={{renderRow: renderRow.bind(this, innerColumns)}}
                data={filteredItems}
                itemIdentifier={itemIdentifier}
                selectionOptions={selectionOptions}
            />}

            {!retrievedAll && <div className="table-loader">
                <img src={require("./../../assets/images/table-loader.gif")} alt=""/>
            </div>}

            {(filteredItems.length === 0 && retrievedAll) &&
            <div className="folder-empty">
                <div><span><Isvg src={icon}/></span></div>
                No {directoriesName} Found
            </div>}
        </div>
    </div>)
};

export default DrillDownBreadCrumbs;
