import React, {useEffect, useState} from 'react';
import {connect} from "react-redux";
import _ from 'lodash';
import Isvg from 'react-inlinesvg';
import {createStringFromEndPointObject} from '../../utils/helpers';
import {Doughnut} from 'react-chartjs-2';
import SelectableTable from "./../table/SelectableTable";
import {routesContainer, STATUS_PROGRESS} from '../../store/global/action_creators_consts';
import {openDialog} from '../dialog/Dialog';
import DownloadLogsDialog from './DownloadLogsDialog';
import ReactPopover from '../_common_/reactPopover/ReactPopover';
import classNames from "classnames";
import {downloadDataBrokerLogs} from "../../store/wizard/action_creators"
import {checkIfFailExist, noConnectionToAllBrokers} from '../../utils/dashboard-utils';
import {getPropFromState} from '../../utils/mapStateToPropsUtils';

require("./failedDetails.scss");

const FailedDetails = (props) => {
    const {match, _relationships, _routing, _downloadLogsRequests} = props;
    const relationshipId = match.params.relationshipId;
    const [relationship, setRelationship] = useState();

    useEffect(() => {
        if (_relationships && !_relationships.data) {
            return;
        }
        const rel = _.find(_relationships.data, {relationshipId});
        if (!rel) {
            routesContainer.routes.push("/dashboard");
        }
        setRelationship(rel);
    }, [_relationships, relationshipId]);

    if (!relationship) return <div/>;
    const {activity} = relationship;

    const renderRow = (row) => (
        <div className="table-row">
            <td className="blank"/>
            <td className="value">{row.count}</td>
            <td>{row.failurePoint}</td>
            <td className="host-name" title={row.errorMessage}>{row.errorMessage}</td>
        </div>
    );
    const renderHeader = () => (
        <thead>
        <tr>
            <td className="blank" style={{width: "3%"}}/>
            <td style={{width: "13%"}}>Failed Items</td>
            <td style={{width: "13%"}}>Failure Point</td>
            <td style={{width: "80%"}}>Failure Reason</td>
        </tr>
        </thead>
    );

    const goBackClicked = () => {
        const {location: {query}} = _routing;
        routesContainer.routes.push(`/${query.calledFrom}`);
    };

    const phaseMap = {
        LIST: "Scan",
        TRANSFER: "Copy"
    };

    const tableData = activity.topFiveMostCommonRelationshipErrors.map((error, index)=>({
        id: index,
        count: error.counter,
        failurePoint: phaseMap[error.step],
        errorMessage: `${error.errorCode}${error.description ? `: ${error.description}`: ''}`
    }));

    const isFailExist = checkIfFailExist(relationship.ui.activityValues);

    const downloadLogsInProgress = _.get(_downloadLogsRequests, [relationshipId, 'status']) === STATUS_PROGRESS;
    const noConnectionToBrokers = noConnectionToAllBrokers(relationship.group.dataBrokers);
    const disableDownloadLogs = noConnectionToBrokers || downloadLogsInProgress;
    const downloadLogsTitle = noConnectionToBrokers ? 'No available data brokers' : '';

    return (
        <div className="scrollable-area">
            <div className="failed-files">
                <div className="go-back" onClick={goBackClicked}> <Isvg src={require("./../../assets/svgs/arrow-left.svg")}/>Back</div>
                <div className="header">
                    <h2>Sync Failures</h2>
                    <div className="relationship-info-container">
                        <EndPointInfo type="Source" data={relationship.source}/>
                        <EndPointInfo type="Target" data={relationship.target}/>
                    </div>
                </div>
                <div className="summary-graphs">
                    <PieWithDetails header="Failed to Scan (Directories)"
                                    labelInsidePie="Failed Directories"
                                    failedLabel="Failed to scan"
                                    successLabel="Succeeded to scan"
                                    failedCount={activity.dirsFailedToScan}
                                    successCount={activity.dirsScanned}
                                    svg="directory" subtitle="Directories"/>

                    <PieWithDetails header="Failed to Copy (Files)"
                                    labelInsidePie="Failed Files"
                                    failedLabel="Failed to copy"
                                    successLabel="Succeeded to copy"
                                    failedCount={activity.filesFailed + activity.filesRemovedFailed}
                                    successCount={activity.filesCopied}
                                    svg="file" subtitle="Files"/>
                </div>
                <div className="failed-files-table">
                    <div className="table-title">
                        <h5>Top 5 Failures</h5>
                        <ReactPopover popoverClass={classNames("download-logs info-tooltip", {'hidden': !downloadLogsInProgress})}
                                      key="dbLogs"
                                      trigger="mouseover"
                                      container={
                                          <button disabled={disableDownloadLogs} title={downloadLogsTitle} className="download-logs"
                                                  onClick={(el) => {openDialog(<DownloadLogsDialog className="default-dialog" relationship={relationship} {...props}/>, el.target);}}>
                                              <span className="icon"><Isvg src={require("./../../assets/svgs/download.svg")}/></span>
                                              <span className="text">Download Data Broker Logs</span>
                                          </button>}>
                            {`Request was sent. This process usually takes a couple of minutes. The logs will be saved to your web browser's download location.`}
                        </ReactPopover>
                    </div>
                    <SelectableTable Table={{renderRow, renderHeader}} data={tableData}>
                    </SelectableTable>
                    {tableData.length === 0 && isFailExist &&
                    <div className="table-no-data">Sync Failed Details is a new feature.<br/>
                        The errors will be listed starting from the next sync.</div>}
                </div>
            </div>
        </div>)
};

const mapStateToProps = (state, ownProps) => ({
    _relationships: getPropFromState(state, "_relationships", {}),
    _routing: getPropFromState(state, "_routing", {}),
    _downloadLogsRequests: getPropFromState(state, "_downloadLogsRequests", {}),
    ...ownProps
});

export default connect(mapStateToProps, {downloadDataBrokerLogs})(FailedDetails);

const EndPointInfo = ({type, data}) => {
    const title = createStringFromEndPointObject(data);
    return (
        <div className="endpoint-info" title={title}>
            <div className="info">
                <div className="path-text">{title}</div>
                <div className="small-text">{type}</div>
            </div>
        </div>
    )
};

const PieWithDetails = ({header, svg, subtitle, labelInsidePie, failedLabel, successLabel, failedCount, successCount}) => {
    const total = failedCount + successCount;
    const isTotalZero = total === 0;
    const percentFailed = isTotalZero ? 0 : Number(failedCount / total * 100);
    const percentFailedRounded = Math.round( percentFailed * 10 ) / 10;

    const dataForPie = [
        {
            value: failedCount,
            color: '#ee4866'
        }, {
            value: isTotalZero ? 1 : successCount, //in case of isTotalZero, give the success a value just so the pie will be painted
            color: '#039ae8'
        }];

    return (<div className="pie-chart-container">
        <div className="chart-title">{header}</div>
        <div className="chart-details">
            <div className="pie-chart">
                <Doughnut
                    options={{
                        preserveAspectRatio: true,
                        cutoutPercentage: 80,
                        tooltips: { enabled: false }}
                    }
                    legend={{ display: false }}
                    data={{
                        datasets: [{
                            data: dataForPie.map(item => item.value),
                            backgroundColor: dataForPie.map(item => item.color)}]}}
                />
                <div className="inside-pie">
                    <div className="big-text">{percentFailedRounded}%</div>
                    <div className="small-text">{labelInsidePie}</div>
                </div>
            </div>
            <div className="legend-area">
                <div className="subtitle">
                    <Isvg className={svg} src={require(`./../../assets/svgs/${svg}.svg`)}/>
                    {subtitle}
                </div>
                <div className="legend">
                    <LegendUnit value={failedCount} text={failedLabel} status="failed"/>
                    <LegendUnit value={successCount} text={successLabel} status="success"/>
                </div>
            </div>
        </div>
    </div>)
};

const LegendUnit = ({value, status, text}) => (
    <div className="legend-unit">
        <div className="value">{value}</div>
        <div className={`status ${status}`}>{text}</div>
    </div>
);