import React, {useEffect, useState} from 'react';
import SyncSettingsForm from './SyncSettingsForm';
import {connect} from "react-redux";
import {getFormValuesFromSettings,} from '../../../utils/settings-utils';
import _ from 'lodash';
import {getFixedProtocol, getSVGByProtocol} from '../../../utils/dashboard-utils';
import {createStringFromEndPointObject} from '../../../utils/helpers';
import {updateRelationshipSettings} from '../../../store/wizard/action_creators';
import * as actionCreatorsConsts from '../../../store/global/action_creators_consts';
import ButtonWithNotification from '../../widgets/ButtonWithNotification';
import {getPropFromState} from '../../../utils/mapStateToPropsUtils';

const SyncSettingsEdit = (props) => {
    const {updateRelationshipSettings, _relationships, match} = props;

    const [error, setError] = useState(null);
    const [initialFormValues, setInitialFormValues] = useState(null);
    const [relationship, setRelationship] = useState(null);
    const [submitting, setSubmitting] = useState(false);

    useEffect(()=>{
        if (relationship) return;
        const relationshipId = match.params.relationshipId;
        const relationshipInStore = _.find(_.get(_relationships, 'data'), relationship => relationship.relationshipId === relationshipId);
        if (relationshipInStore) {
            const initialFormValues = getFormValuesFromSettings(relationshipInStore.settings, true);
            setInitialFormValues(initialFormValues);
            setRelationship(relationshipInStore);
        }
    }, [match, _relationships, relationship]);


    const gotoDashboard = () => {
        actionCreatorsConsts.routesContainer.routes.push("/dashboard");
    };

    const doSubmit = (settings) => {
        setError(undefined);
        return (updateRelationshipSettings(relationship.relationshipId, settings))
            .then(gotoDashboard)
            .catch((err)=>{
                if (err.status !== 401) {
                    setError(err.data.message);
                }
            })
    };

    if (!relationship) return <div/>;
    const showAclSection =
        relationship.ui.source.protocol === 'cifs' && relationship.ui.target.protocol === 'cifs' && //this option is only relevant for cifs-cifs relationships
        !relationship.ui.source.cifs.aclServer; //backwards compatibility - "aclServer" is an indication of acl done in the old way

    const showObjectTaggingSection = relationship.ui.target.protocol === "s3" && relationship.ui.target.s3.provider === "s3";

    return (
        <div className="settings-edit">
            <div className="settings-header">
                <h1>Sync Relationship Settings</h1>
                <p>You can modify settings that define how source files and folders are synced and maintained
                    in the target location</p>

                <div className="relationship-info-container">
                    <EndPointInfo type="source" data={relationship.ui.source} isSecureRelationship={!!relationship.encryption}/>
                    <EndPointInfo type="target" data={relationship.ui.target} isSecureRelationship={!!relationship.encryption}/>
                </div>
            </div>
            <div className="scrollable-area">
                <div className="settings-content">
                    <SyncSettingsForm doSubmit={doSubmit} initialValues={initialFormValues} setSubmitting={setSubmitting}
                                      showAclSection={showAclSection} showObjectTaggingSection={showObjectTaggingSection}
                                      isEditMode={true}/>
                </div>
            </div>
            <ButtonWithNotification
                type="submit"
                form="syncSettings"
                error={error}
                submitting={submitting}
                onCancel={gotoDashboard}
                clearError={()=>{setError(null)}}
                text="Save Settings"/>
        </div>
    )
};

const mapStateToProps = (state, ownProps) => {
    return {
        _ui: {...state.syncReducer._ui, ...state.global._ui},
        routing: state.router,
        _relationships: getPropFromState(state, "_relationships", {}),
        ...ownProps
    }
};
export default connect(mapStateToProps, {updateRelationshipSettings})(SyncSettingsEdit);

const EndPointInfo = (props) => {
    const {type, data, isSecureRelationship} =  props;
    const title = createStringFromEndPointObject(data);
    const svg = getSVGByProtocol(getFixedProtocol(data), isSecureRelationship);

    return (
        <div className="endpoint-info" title={title}>
            {svg} <div className="info"><b>Relationship {type}</b> | {title}</div>
        </div>
    )
};
