import React , {useState} from 'react';
import {InputField, SelectField, RadioInputs, CheckboxInputs, MultiSelectInput, DatePicker} from '../../_common_/forms/forms';
import useRunOnceWhenTruthy from "../../../hooks/useRunOnceWhenTruthy";
import { Field, reduxForm, getFormValues, getFormSyncErrors, change} from 'redux-form';
import * as utils from '../../../utils/settings-utils';
import Isvg from 'react-inlinesvg';
import classNames from 'classnames';
import { connect } from 'react-redux';
import _ from 'lodash';
import Validator from 'validatorjs';
import {openDialog} from '../../dialog/Dialog';
import ResetSettingsDialog from './ResetSettingsDialog';
import {convertSizeToBytes} from '../../../utils/helpers';
require("./syncSettings.scss");

//******************* steps to add a new setting to the form ****************
// 1. In render:
//  Add AccordionComponent and edit its properties using the new setting name
//  Insert the setting's relevant input/s as children.
//
// 2. If necessary, add validation rules to "validate" function
//
// 3. In settings-utils:
//  add the default value to SETTINGS_DEFAULTS
//  add relevant code to getSummaryLines
//  add relevant code to createSettingsObjectFromFormValues
//  add relevant code to getFormValuesFromSettings
//
// 4. Style: in syncSettings.scss, go to "accordion type-specific" area and add your relevant css
//**************************************************************************

//This ref is needed to allow manually triggering the handleBlur of the input, in case the select has changed
//The input needs to be blurred so the error will be visible
let fileMaxInput;

const validate = values => {
    const rules = {}; // for validatorjs validations
    const errors = {}; // for custom validations
    const customErrors = {};

    // eslint-disable-next-line no-unused-vars
    Validator.register('customInteger', (value, requirement, attribute) => {
        //In order to allow leading zeros, e.g. 010 is an integer that equals 10. See jira CS-4097.
        const valueNum = +value;
        return Number.isInteger(valueNum);
    }, `The :attribute field must be an integer`);

    //grace period validation:
    const graceResolution = _.get(values.graceResolution, 'value');

    if (!_.includes(values.gracePeriodOn, utils.NOT_EXCLUDE_FILES_OPTION)) {
        rules.graceResolution = 'required';
        if (graceResolution === 'Minutes' || graceResolution === 'Seconds') {
            rules.graceValue = 'required|numeric|between:1,59|customInteger';
        } else if (graceResolution === 'Hours') {
            rules.graceValue = 'required|numeric|between:1,48|customInteger';
        }
    }

    //file types validation:
    rules.fileTypes = 'required';
    customErrors['required.fileTypes'] = 'At least one type must be selected.';

    //file size validation:
    if (_.includes(values.fileSizeOption, utils.FILE_SIZE_OPTION)) {
        rules.fileMinSizeValue = 'required|numeric';
        rules.fileMaxSizeValue = 'required|numeric';
        const fileMinSizeResolution = _.get(values.fileMinSizeResolution, 'value');
        const fileMaxSizeResolution = _.get(values.fileMaxSizeResolution, 'value');
        const minBytes = convertSizeToBytes(values.fileMinSizeValue, fileMinSizeResolution);
        const maxBytes = convertSizeToBytes(values.fileMaxSizeValue, fileMaxSizeResolution);
        if (fileMinSizeResolution === "B") rules.fileMinSizeValue = 'required|customInteger';
        if (fileMaxSizeResolution === "B") rules.fileMaxSizeValue = 'required|customInteger';

        if (maxBytes > utils.UI_MAX_FILE_SIZE_LIMIT) {
            errors.fileMaxSizeValue = "The Max File Size can't be larger than 4 PB";
        }
        if (minBytes > maxBytes) {
            errors.fileMaxSizeValue = "The Min File Size can't be larger then the Max File Size";
        }
    }

    //date modified validation:
    const dateModifiedOption = _.get(values.fileDateModifiedOption, 'value');
    if (dateModifiedOption !== utils.DATE_MODIFIED_OPTIONS.ALL) {
        const minRequired = dateModifiedOption === utils.DATE_MODIFIED_OPTIONS.ONLY_MIN || dateModifiedOption === utils.DATE_MODIFIED_OPTIONS.RANGE;
        const maxRequired = dateModifiedOption === utils.DATE_MODIFIED_OPTIONS.ONLY_MAX || dateModifiedOption === utils.DATE_MODIFIED_OPTIONS.RANGE;
        if (minRequired) {
            rules.fileDateModifiedMin = "required|date";
        }
        if (maxRequired) {
            rules.fileDateModifiedMax = "required|date";
        }
    }

    //schedule validation:
    if (utils.isScheduleActive(values.scheduleStatus)) {
        if (_.get(values.scheduleResolution, 'value') === 'Days') rules.scheduleInterval = 'required|numeric|between:1,365|customInteger';
        else if (_.get(values.scheduleResolution, 'value') === 'Hours') rules.scheduleInterval = 'required|numeric|between:1,24|customInteger';
        else if (_.get(values.scheduleResolution, 'value') === 'Minutes') rules.scheduleInterval = 'required|numeric|between:1,60|customInteger';
    }

    const validator = new Validator(values, rules, customErrors);
    validator.setAttributeNames({
        graceValue: `${graceResolution} value`,
        fileMinSizeValue: 'Min File Size',
        fileMaxSizeValue: 'Max File Size',
        fileDateModifiedMin: 'From date',
        fileDateModifiedMax: 'To date',
        scheduleInterval: 'Interval value'
    });
    validator.passes();

    // Putting fileMinSizeValue errors on fileMaxSizeValue, to allow locating all errors in the same place.
    if (validator.errors.errors.fileMinSizeValue || errors.fileMaxSizeValue) {
        validator.errors.errors.fileMaxSizeValue = validator.errors.errors.fileMinSizeValue;
        delete validator.errors.errors.fileMinSizeValue;
        //The input needs to be blurred so the error will be visible
        fileMaxInput.ref.current.handleBlur();
    }

    return {...validator.errors.all(), ...errors};
};

const AccordionComponent = ({title, summary, children, containerClassName, isOpen, handleSectionClick, error}) => {
    return (
        <div className={classNames("accordion-container", containerClassName)}>
            <div className={classNames("title", {"error": error})} onClick={handleSectionClick}>{title}</div>
            <div className={classNames("middle", {"show-content": isOpen})} >
                <div className={classNames("summary-area", {"show": !isOpen})} onClick={handleSectionClick}>

                    {error && <div className="error placeholder"><span><Isvg src={require("../../../assets/svgs/exclamation-mark-triangle-icon.svg")}/></span>
                        <div className="error-text">{error}</div></div>}
                    {!error && <div className="summary-text placeholder" title={typeof summary === 'string' ? summary : ''}>{summary}</div>}
                </div>
                {/*css is used to show/hide the input area, because if the element is unmounted from the DOM the fields are unregistered from the form, and no validation is done*/}
                <div className={classNames("input-area", {"show": isOpen})}>{children}</div>
            </div>
            <div className={classNames("chevron end", {"up" : isOpen})} onClick={handleSectionClick}>
                <span><Isvg src={require("./../../../assets/svgs/down-chevron.svg")}/></span>
            </div>
        </div>)
};

const calculateSchedulePlaceholder = (value) => {
    switch (value) {
        case "Hours" : return "1-24";
        case "Days" : return "1-365";
        case "Minutes" : return "1-60";
        default: return "";
    }
};

const SyncSettings = ({ handleSubmit, formValues, formErrors, doSubmit, initialize, initialValues, isEditMode,
                          showAclSection, showObjectTaggingSection, setSubmitting, dispatch }) => {

    useRunOnceWhenTruthy(() => {
        initialize(initialValues)
    }, initialValues);

    const [openedSection, setOpenedSection] = useState(null);

    const handleSectionClick = (section) => {
        setOpenedSection(section === openedSection ? null : section);
    };

    const handleSubmittedData = (values, dispatch) => {

        //update the parent (editSettings) to show the submitting loader on the button
        if (setSubmitting) {
            setSubmitting(true);
        }

        const settingsObj = utils.createSettingsObjectFromFormValues(values);
        doSubmit(settingsObj, dispatch);
    };

    const handleDeleteOnSourceChange = (selectedIndex) => {
        if (selectedIndex === 1) dispatch(change("syncSettings", "deleteOnTarget", utils.NOT_DELETE_FILES_FROM_TARGET_OPTION_TEXT));
    };

    const resetSettings = () => {
        const formDefaults = utils.getFormValuesFromSettings(utils.SETTINGS_DEFAULTS, isEditMode);
        initialize(formDefaults);
        setOpenedSection(null);
    };

    const summaryLines = utils.getSummaryLines(formValues) ;


    // "Schedule" section:
    const hideScheduleDefinitions = !utils.isScheduleActive(formValues.scheduleStatus);
    const schedulePlaceholder = formValues.scheduleResolution ? calculateSchedulePlaceholder(formValues.scheduleResolution.value) : "";
    const scheduleRadioInputs = isEditMode ? [utils.SCHEDULE_OPTIONS.ON, utils.SCHEDULE_OPTIONS.OFF] : [utils.SCHEDULE_OPTIONS.NOW, utils.SCHEDULE_OPTIONS.ON, utils.SCHEDULE_OPTIONS.ONCE];
    const scheduleCheckedIndex = isEditMode ? (formValues.scheduleStatus === utils.SCHEDULE_OPTIONS.ON ? 0 : 1) :
        (formValues.scheduleStatus === utils.SCHEDULE_OPTIONS.NOW ? 0 : formValues.scheduleStatus === utils.SCHEDULE_OPTIONS.ON ? 1 : 2);
    const scheduleOptionNote = {
        [utils.SCHEDULE_OPTIONS.NOW]: "The first sync will start after the wizard is completed.",
        [utils.SCHEDULE_OPTIONS.ON]: isEditMode ? "The sync can be scheduled to start within the next 24 hours." : "You can define the first sync to start within the next 24 hours.",
        [utils.SCHEDULE_OPTIONS.ONCE]: "The data sync will occur once. It won't be scheduled to repeat in the future.",
        [utils.SCHEDULE_OPTIONS.OFF]: "No syncs will be scheduled for this relationship."
    };

    //"Retries" section:
    const retriesSelect =  <Field component={SelectField} valueField="value" name="retries" className="bottom-border"
                                  options={utils.getRetriesOptions()} noSearch={true} selectStyleType="bottom-border"/>;

    // "Recently modified" section:
    const graceRes = _.get(formValues.graceResolution, "value", "");
    const gracePlaceholder = graceRes === 'Hours' ? '1-48' : '1-59';
    const disableGrace = formValues.gracePeriodOn === utils.NOT_EXCLUDE_FILES_OPTION;

    const isDeleteOnTargetDisabled = formValues.deleteOnSource === utils.DELETE_FILES_FROM_SOURCE_OPTION_TEXT;

    // "File size" section:
    const disableFileSize = formValues.fileSizeOption === utils.FILE_SIZE_OPTION_DISABLED;

    // "Date modified" section:
    const showMinCalendar = _.includes([utils.DATE_MODIFIED_OPTIONS.ONLY_MIN, utils.DATE_MODIFIED_OPTIONS.RANGE], _.get(formValues.fileDateModifiedOption, 'value'));
    const showMaxCalendar = _.includes([utils.DATE_MODIFIED_OPTIONS.ONLY_MAX, utils.DATE_MODIFIED_OPTIONS.RANGE], _.get(formValues.fileDateModifiedOption, 'value'));
    const dateDisabledTitle = "This option is not available based on your selection above.";

    return (
        <div className="sync-settings">
             <form onSubmit={handleSubmit(handleSubmittedData)} id="syncSettings">

                 <h5 className="category-title">General</h5>

                 <AccordionComponent title="Schedule" isOpen={openedSection === 'schedule'} error={formErrors.scheduleInterval}
                                     summary={summaryLines.schedule} handleSectionClick={handleSectionClick.bind(this, 'schedule')}
                                     containerClassName="schedule">
                     <Field component={RadioInputs} name="scheduleStatus"
                            changeSelection={_.noop}
                            checkedIndex={scheduleCheckedIndex}
                            inputs={scheduleRadioInputs}/>
                     <div className="note">{scheduleOptionNote[formValues.scheduleStatus]}</div>
                     <div className={classNames({"hidden" : hideScheduleDefinitions})}>
                         <div className={"schedule-definition"}>
                             <div className="text">
                                 The {isEditMode ? "next" : formValues.scheduleStatus === utils.SCHEDULE_OPTIONS.NOW ? "next" : "first"} sync will automatically start at
                             </div>
                             <div className="inline-field">
                                 <Field component={SelectField} name="scheduleNextTime"
                                        valueField="value" options={utils.getScheduleNextTimeOptions()} noSearch={true} selectStyleType="bottom-border"/>
                             </div>
                         </div>
                         <div className={"schedule-definition"}>
                             <div className="text">and will repeat every</div>
                             <div className="user-input">
                                 <div className="inline-field schedule-interval">
                                     <Field component={InputField} name="scheduleInterval"
                                            placeholder={schedulePlaceholder} className="bottom-border"/>
                                 </div>
                                 <div className="inline-field">
                                     <Field component={SelectField} name="scheduleResolution"
                                            valueField="value" options={utils.getScheduleResolutionOptions()} noSearch={true} selectStyleType="bottom-border"/>
                                 </div>
                             </div>
                         </div>
                     </div>
                 </AccordionComponent>

                 <AccordionComponent title="Retries" isOpen={openedSection === 'retries'}
                                     summary={summaryLines.retries} handleSectionClick={handleSectionClick.bind(this, 'retries')}
                                     containerClassName="retries">
                     <div className="placeholder">Retry {retriesSelect} times before skipping file (0-5)</div>
                 </AccordionComponent>

                 <h5 className="category-title">Files and Directories</h5>

                 <AccordionComponent title="Recently Modified Files" isOpen={openedSection === 'gracePeriod'} error={formErrors.graceValue}
                                     summary={summaryLines.gracePeriod} handleSectionClick={handleSectionClick.bind(this, 'gracePeriod')}
                                     containerClassName="grace-period">
                     <div className="row">
                         <Field component={RadioInputs}
                                changeSelection={_.noop}
                                checkedIndex={formValues.gracePeriodOn === utils.EXCLUDE_FILES_OPTION ? 0 : 1} name="gracePeriodOn"
                                inputs={[utils.EXCLUDE_FILES_OPTION, utils.NOT_EXCLUDE_FILES_OPTION]}
                                className="auto-sync"/>
                         <div className={classNames("grace-value-column", {"disable-change": disableGrace})}>
                             <Field component={InputField} disabled={disableGrace} className="grace-value bottom-border value"
                                    name="graceValue" placeholder={gracePlaceholder}/>
                         </div>
                         <div className={classNames("grace-resolution-column ", {"disable-change": disableGrace})}>
                             <Field component={SelectField} disabled={disableGrace} valueField="value" name="graceResolution"
                                    options={utils.getGraceResolutionOptions()} className="select" noSearch={true} selectStyleType="bottom-border"/>
                         </div>
                         <span className="last-part"> before the scheduled sync</span>
                     </div>
                 </AccordionComponent>

                 <AccordionComponent title="Delete Files On Source" isOpen={openedSection === 'deleteOnSource'}
                                     summary={summaryLines.deleteOnSource}
                                     handleSectionClick={handleSectionClick.bind(this, 'deleteOnSource')}
                                     containerClassName="delete-on-source">
                     <Field component={RadioInputs}
                            disabled={isEditMode}
                            disabledTitle={"This option can be changed only when creating a relationship."}
                            changeSelection={handleDeleteOnSourceChange}
                            checkedIndex={formValues.deleteOnSource === utils.NOT_DELETE_FILES_FROM_SOURCE_OPTION_TEXT ? 0 : 1}
                            name="deleteOnSource"
                            inputs={[utils.NOT_DELETE_FILES_FROM_SOURCE_OPTION_TEXT, utils.DELETE_FILES_FROM_SOURCE_OPTION_TEXT]}/>
                     {formValues.deleteOnSource === utils.DELETE_FILES_FROM_SOURCE_OPTION_TEXT &&
                     <div className="warning"><span><Isvg
                         src={require("../../../assets/svgs/exclamation-mark-triangle-icon.svg")}/></span>
                         {`Important: This option includes the risk of data loss because the source files are deleted after they're copied. 
                         After you enable this option, you also need to change a parameter in the local.json file on the data broker. Refer to the documentation for details.`}
                     </div>}
                 </AccordionComponent>

                 <AccordionComponent title="Delete Files On Target" isOpen={openedSection === 'deleteOnTarget'}
                                     summary={summaryLines.deleteOnTarget} handleSectionClick={handleSectionClick.bind(this, 'deleteOnTarget')}
                                     containerClassName="delete-on-target">
                     <Field component={RadioInputs}
                            disabled={isDeleteOnTargetDisabled}
                            changeSelection={_.noop}
                            checkedIndex={formValues.deleteOnTarget === utils.NOT_DELETE_FILES_FROM_TARGET_OPTION_TEXT ? 0 : 1}
                            name="deleteOnTarget"
                            inputs={[utils.NOT_DELETE_FILES_FROM_TARGET_OPTION_TEXT, utils.DELETE_FILES_FROM_TARGET_OPTION_TEXT]}/>
                     {formValues.deleteOnTarget === utils.DELETE_FILES_FROM_TARGET_OPTION_TEXT &&
                     <div className="warning"><span><Isvg src={require("../../../assets/svgs/exclamation-mark-triangle-icon.svg")}/></span>
                         Important: This option is not typically recommended, especially if the target <br/> already contains files. It might result in data loss.
                     </div>}
                     {isDeleteOnTargetDisabled && <div className="warning">{`We automatically set this option to "Never delete files from the target location" because you enabled the option to delete files from the source location.`}</div>}
                 </AccordionComponent>

                 {showObjectTaggingSection &&
                 <AccordionComponent title="Object Tagging" isOpen={openedSection === 'objectTagging'}
                                     summary={summaryLines.objectTagging} handleSectionClick={handleSectionClick.bind(this, 'objectTagging')}
                                     containerClassName="disable-object-tagging">
                     <Field component={RadioInputs}
                            changeSelection={_.noop}
                            checkedIndex={formValues.objectTagging === utils.OBJECT_TAGGING_USE_OPTION_TEXT ? 0 : 1}
                            name="objectTagging"
                            inputs={[utils.OBJECT_TAGGING_USE_OPTION_TEXT, utils.OBJECT_TAGGING_DISABLE_OPTION_TEXT]}/>
                 </AccordionComponent>}

                 <AccordionComponent title="File Types" isOpen={openedSection === 'fileTypes'} error={formErrors.fileTypes}
                                     summary={summaryLines.fileTypes} handleSectionClick={handleSectionClick.bind(this, 'fileTypes')}
                                     containerClassName="file-types">
                     <div className="placeholder">
                            <div className="instructions">Select at least one file type that you want to include:</div>
                            <Field component={CheckboxInputs} name="fileTypes" options={utils.getFileTypesOptions()}/>
                     </div>
                 </AccordionComponent>

                 <AccordionComponent title="Exclude File Extensions" isOpen={openedSection === 'filesExtensions'} error={formErrors.filesExtensions}
                                     summary={summaryLines.filesExtensions} handleSectionClick={handleSectionClick.bind(this, 'filesExtensions')}
                                     containerClassName="files-extensions">
                     <div className="placeholder">
                         <div className="instructions">Enter the file extensions to exclude (the limit is 15 extensions):</div>
                         <Field component={MultiSelectInput} label="File extensions to exclude"  name="filesExtensions"
                                placeholder="Type the file extension and press Enter"
                                limit={15} inputTransformation={utils.extensionsInputTransformation}/>
                     </div>
                 </AccordionComponent>

                 <AccordionComponent title="File Size" isOpen={openedSection === 'fileSize'} error={formErrors.fileMaxSizeValue}
                                     summary={summaryLines.fileSize} handleSectionClick={handleSectionClick.bind(this, 'fileSize')}
                                     containerClassName="file-size">
                     <Field component={RadioInputs} className="file-size-radio"
                            changeSelection={_.noop}
                            checkedIndex={formValues.fileSizeOption === utils.FILE_SIZE_OPTION ? 1 : 0} name="fileSizeOption"
                            inputs={[utils.FILE_SIZE_OPTION_DISABLED, utils.FILE_SIZE_OPTION]}/>
                     <div className={classNames("file-size-row", {"disable-change": disableFileSize})}>
                         <div className="row-title">Min File Size</div>
                         <Field component={InputField} disabled={disableFileSize} className="bottom-border value" name="fileMinSizeValue"/>
                         <Field component={SelectField} disabled={disableFileSize} valueField="value" name="fileMinSizeResolution"
                                options={utils.getFileSizeResolutionOptions()} className="select" noSearch={true} selectStyleType="bottom-border"/>
                     </div>
                     <div className={classNames("file-size-row", {"disable-change": disableFileSize})}>
                         <div className="row-title">Max File Size</div>
                         <Field component={InputField} disabled={disableFileSize} className="bottom-border value" name="fileMaxSizeValue"
                                ref={c => fileMaxInput = c}/>
                         <Field component={SelectField} disabled={disableFileSize} valueField="value" name="fileMaxSizeResolution"
                                options={utils.getFileSizeResolutionOptions()} className="select" noSearch={true} selectStyleType="bottom-border"/>
                     </div>
                 </AccordionComponent>

                 <AccordionComponent title="Date Modified" isOpen={openedSection === 'dateModified'} error={formErrors.fileDateModifiedMin || formErrors.fileDateModifiedMax}
                                     summary={summaryLines.dateModified} handleSectionClick={handleSectionClick.bind(this, 'dateModified')}
                                     containerClassName="date-modified">
                     <div className="placeholder">
                         <Field component={SelectField} valueField="value" name="fileDateModifiedOption"
                                options={utils.getDateModifiedOptions()} noSearch={true}/>
                         <div className="calendars-row">
                             <div className="calendar-item">
                                 <div> From: </div>
                                 <Field component={DatePicker} name="fileDateModifiedMin" isDisabled={!showMinCalendar} disabledTitle={dateDisabledTitle}
                                        minDatePossible={new Date(utils.MIN_DATE_POSSIBLE)}
                                        maxDate={showMaxCalendar ? formValues.fileDateModifiedMax : null}/>
                             </div>
                            <div className="calendar-item">
                                 <div> To: </div>
                                 <Field component={DatePicker} name="fileDateModifiedMax" isDisabled={!showMaxCalendar} disabledTitle={dateDisabledTitle}
                                        minDatePossible={new Date(utils.MIN_DATE_POSSIBLE)}
                                        minDate={showMinCalendar ? formValues.fileDateModifiedMin : null}/>
                             </div>
                         </div>
                     </div>
                 </AccordionComponent>

                 {showAclSection &&
                 <AccordionComponent title="ACL - Access Control List" isOpen={openedSection === 'acl'} error={formErrors.copyAcl}
                                     summary={summaryLines.acl} handleSectionClick={handleSectionClick.bind(this, 'acl')}
                                     containerClassName="acl">
                     <div className="placeholder">
                         <Field id="copyAcl" type="checkbox" component={InputField} name="copyAcl">
                             <label htmlFor="copyAcl">Copy Access Control Lists to the target</label>
                         </Field>
                         <div className="acl-notice">
                             <div><span className="acl-notice-label">Notice:</span>Copying ACLs can affect sync performance</div>
                         </div>
                     </div>
                 </AccordionComponent>}


             </form>
            <div className="reset-click" onClick={(el) => {openDialog(<ResetSettingsDialog className="default-dialog" resetSettings={resetSettings}/>, el.target)}}>
                <span><Isvg src={require("../../../assets/svgs/reset.svg")}/></span>
                Reset to defaults
            </div>
        </div>
    )
};

let SyncSettingsForm = reduxForm({
    form: 'syncSettings',
    touchOnChange: true,
    destroyOnUnmount: false, //During the wizard the settings component is unmounted (due to the scroll-area disabling) and we want to prevent redux-DESTROY from being called (the form has already been initialized).
    validate
})(SyncSettings);

SyncSettingsForm = connect(
    state => {
    const formValues = getFormValues('syncSettings')(state) || {};
    const formErrors = getFormSyncErrors('syncSettings')(state);
    return {formValues, formErrors}
})(SyncSettingsForm);

export default SyncSettingsForm;
