import AuthenticationContext from 'adal-angular';
import {store} from '../store/store';
import {setAzureCredentials} from '../store/deploy/action_creators';
import axios from 'axios';
import _ from 'lodash';

const BASE_URL = process.env.REACT_APP_AZURE_DOMAIN;

const extractData = (response) => {
    return response.data;
};

const loginCallback = (error, id_token) => {
    if (id_token) {
        const accessToken = authContext.getCachedToken('');
        updateStoreWithToken(accessToken);
        setRenewToken();
    }
    if (error) {
        // eslint-disable-next-line no-console
        console.log('Error during Azure login:', error);
    }
};

const config = {
    clientId: process.env.REACT_APP_AZURE_CLIENT_ID,
    cacheLocation: 'localStorage',
    popUp: true,
    redirectUri: window.location.origin,
    callback : loginCallback
};

const authContext = new AuthenticationContext(config);

// Override the "_getNavigateUrl" function in adal.js
// so the responseType will include 2 parameters and the response will have the correct Audience needed for the API
const that = authContext;
authContext._getNavigateUrl_Bak = authContext._getNavigateUrl;
authContext._getNavigateUrl = function(responseType, resource){
    if (responseType === 'id_token' && !resource) {
        responseType = 'id_token%20token';
        resource = 'https://management.azure.com/';
    }
    return that._getNavigateUrl_Bak(responseType, resource);
};

export const acquireAccessToken = () => {
    if (process.env.REACT_APP_AZURE_LOGIN_REQUIRED === "false") {
        updateStoreWithToken("dummyToken");
        return;
    }
    const accessToken = authContext.getCachedToken('');
    if (accessToken) {
        updateStoreWithToken(accessToken);
        setRenewToken();
    }
    else {
        authContext.login();
    }
};

const setRenewToken = () => {
    const expirationKey = _.get(authContext, 'CONSTANTS.STORAGE.EXPIRATION_KEY');
    const expiry = authContext._getItem(expirationKey);
    const now = Date.now();
    const timeToExpire = expiry * 1000 - now;

    setTimeout(() => {
        authContext.login();
    }, (timeToExpire - (10 * 1000) ));//renew token 10 seconds before the token is expired
};

export const azureLogout = () => {
    authContext.logOut();
};

const updateStoreWithToken = (accessToken) => {
    store.dispatch(setAzureCredentials({accessToken}));
};

const getAzureTokenHeader = () => {
    const accessToken = store.getState().deploy.credentials.azure.accessToken;
    return { headers: { Authorization: `Bearer ${accessToken}` } };
};

export const getSubscriptions = () => {
    return axios.get(`${BASE_URL}/subscriptions?api-version=2016-06-01`,  getAzureTokenHeader()).then(extractData);
};

export const getVnetsWithSubnets = (subscriptionId) => {
    return axios.get(`${BASE_URL}/subscriptions/${subscriptionId}/providers/Microsoft.Network/virtualNetworks?api-version=2019-07-01`,  getAzureTokenHeader()).then(extractData);
};

export const getAzureRegions = (subscriptionId) => {
    return axios.get(`${BASE_URL}/subscriptions/${subscriptionId}/locations?api-version=2016-06-01`,  getAzureTokenHeader()).then(extractData);
};

export const createResourceGroup = (subscriptionId, resourceGroupName, location) => {
    return axios.put(`${BASE_URL}/subscriptions/${subscriptionId}/resourcegroups/${resourceGroupName}?api-version=2018-02-01`,
        { 'location': location }, getAzureTokenHeader()).then(extractData);
};

export const deleteResourceGroup = (subscriptionId, resourceGroupName) => {
    return axios.delete(`${BASE_URL}/subscriptions/${subscriptionId}/resourcegroups/${resourceGroupName}?api-version=2019-05-10`, getAzureTokenHeader()).then(extractData);
};

export const azureTemplateDeploymentValidation = (deployDetails) => {
    return axios.post(`${BASE_URL}/subscriptions/${deployDetails.subscriptionId}/resourcegroups/${deployDetails.resourceGroupName}/providers/Microsoft.Resources/deployments/${deployDetails.deploymentName}/validate?api-version=2018-02-01`,
        {properties: deployDetails.properties}, getAzureTokenHeader()).then(extractData);
};

export const azureTemplateDeployment = (deployDetails) => {
    return axios.put(`${BASE_URL}/subscriptions/${deployDetails.subscriptionId}/resourcegroups/${deployDetails.resourceGroupName}/providers/Microsoft.Resources/deployments/${deployDetails.deploymentName}?api-version=2018-02-01`,
        {properties: deployDetails.properties}, getAzureTokenHeader()).then(extractData);
};

export const getDeploymentStatus = (subscriptionId, resourceGroupName, deploymentName) => {
    return axios.get(`${BASE_URL}/subscriptions/${subscriptionId}/resourcegroups/${resourceGroupName}/providers/Microsoft.Resources/deployments/${deploymentName}?api-version=2018-02-01`, getAzureTokenHeader()).then(extractData);
};

export const cancelDeployment = (subscriptionId, resourceGroupName, deploymentName) => {
    return axios.post(`${BASE_URL}/subscriptions/${subscriptionId}/resourcegroups/${resourceGroupName}/providers/Microsoft.Resources/deployments/${deploymentName}/cancel?api-version=2019-08-01`, null, Object.assign({}, getAzureTokenHeader(), {maxContentLength: 0})).then(extractData);
};

export const getResourceGroups = (subscriptionId) => {
    return axios.get(`${BASE_URL}/subscriptions/${subscriptionId}/resourcegroups?api-version=2018-02-01`, getAzureTokenHeader()).then(extractData);
};

export const getVm = (subscriptionId, resourceGroupName, vmName) => {
    return axios.get(`${BASE_URL}/subscriptions/${subscriptionId}/resourceGroups/${resourceGroupName}/providers/Microsoft.Compute/virtualMachines/${vmName}?api-version=2017-12-01`, getAzureTokenHeader()).then(extractData);
};

export const deleteVM = (subscriptionId, resourceGroupName, vmName) => {
    return axios.delete(`${BASE_URL}/subscriptions/${subscriptionId}/resourcegroups/${resourceGroupName}/providers/Microsoft.Compute/virtualMachines/${vmName}?api-version=2017-12-01`, getAzureTokenHeader()).then(extractData);
};

export const deleteDisk = (subscriptionId, resourceGroupName, diskName) => {
    return axios.delete(`${BASE_URL}/subscriptions/${subscriptionId}/resourcegroups/${resourceGroupName}/providers/Microsoft.Compute/disks/${diskName}?api-version=2017-03-30`, getAzureTokenHeader()).then(extractData);
};

export const deleteNetworkInterface = (subscriptionId, resourceGroupName, networkInterfaceName) => {
    return axios.delete(`${BASE_URL}/subscriptions/${subscriptionId}/resourcegroups/${resourceGroupName}/providers/Microsoft.Network/networkInterfaces/${networkInterfaceName}?api-version=2018-04-01`, getAzureTokenHeader()).then(extractData);
};

export const deletePublicIP = (subscriptionId, resourceGroupName, publicIpAddressName) => {
    return axios.delete(`${BASE_URL}/subscriptions/${subscriptionId}/resourcegroups/${resourceGroupName}/providers/Microsoft.Network/publicIPAddresses/${publicIpAddressName}?api-version=2018-04-01`, getAzureTokenHeader()).then(extractData);
};

export const deleteSecurityGroup = (subscriptionId, resourceGroupName, networkSecurityGroupName) => {
    return axios.delete(`${BASE_URL}/subscriptions/${subscriptionId}/resourcegroups/${resourceGroupName}/providers/Microsoft.Network/networkSecurityGroups/${networkSecurityGroupName}?api-version=2018-04-01`, getAzureTokenHeader()).then(extractData);
};

//*************** ANF **************
// export const getAnfAccounts = (subscriptionId, resourceGroupName) => {
//     return axios.get(`${BASE_URL}/subscriptions/${subscriptionId}/resourceGroups/${resourceGroupName}/providers/Microsoft.NetApp/netAppAccounts?api-version=2017-08-15`, getAzureTokenHeader()).then(extractData);
// };
//
// export const getAnfPools = (subscriptionId, resourceGroupName, accountName) => {
//     return axios.get(`${BASE_URL}/subscriptions/${subscriptionId}/resourceGroups/${resourceGroupName}/providers/Microsoft.NetApp/netAppAccounts/${accountName}/capacityPools?api-version=2017-08-15`, getAzureTokenHeader()).then(extractData);
// };
//
// export const getAnfVolumes = (subscriptionId, resourceGroupName, accountName, poolName) => {
//     return axios.get(`${BASE_URL}/subscriptions/${subscriptionId}/resourceGroups/${resourceGroupName}/providers/Microsoft.NetApp/netAppAccounts/${accountName}/capacityPools/${poolName}/volumes?api-version=2017-08-15`, getAzureTokenHeader()).then(extractData);
// };
//
// export const getAnfVolume = (subscriptionId, resourceGroupName, accountName, poolName, volumeName) => {
//     return axios.get(`${BASE_URL}/subscriptions/${subscriptionId}/resourceGroups/${resourceGroupName}/providers/Microsoft.NetApp/netAppAccounts/${accountName}/capacityPools/${poolName}/volumes/${volumeName}?api-version=2017-08-15`, getAzureTokenHeader()).then(extractData);
// };