import {
    handleErrors,
    handleUnhandledMessages,
    processTimeline,
    updateKeyWithBuiltState
} from "../../utils/reducerUtils";
import cloneDeep from 'lodash/cloneDeep';
import merge from 'lodash/merge';
import get from 'lodash/get';
import find from 'lodash/find';
import filter from 'lodash/filter';
import Immutable from "seamless-immutable";
import * as actionCreatorsConsts from "./action_creators_consts";
import * as consts from "../../consts";
import * as licenseUtils from "../../utils/license-utils";
import {bootGoogleAnalytics} from "../../utils/analytics";
import {UPDATE_DATA_BROKER_STATUS} from "../wizard/action_creators_consts";
import {NOTIFICATION_TYPES} from '../../consts';
import {v4 as uuidv4} from 'uuid';
import {BROKER_STATUS_COMPLETE} from '../../consts';


export const globalDefaultState = () => Immutable({
    _ui: {dashboardView: consts.CARDS_VIEW},
    unhandledMessages: [],
    _timeline: {data: []},
    _generalErrors: {_errorsMap: {}, _errorsPanelOpen: false, _errorsPanelSize: {}},
    _licenses: {},
    _notifications: [],
    _cmIntegration: {}
});

export const globalReducer = (initialState = globalDefaultState(), action = {}) => {
    const state = handleErrors(action, initialState);
    const updatedState = handleAction(state, action);
    return handleUnhandledMessages(updatedState, handleAction);
};

const handleAction = (state, action) => {
    action = cloneDeep(action);

    switch (action.type) {
        case actionCreatorsConsts.AUTH_REFRESH_SSO_SUCCESS: {
            const {idTokenPayload, accessToken} = action.payload;
            const name = idTokenPayload["http://cloud.netapp.com/user_metadata"].name;
            const email = idTokenPayload["http://cloud.netapp.com/user_metadata"].email;
            const company = idTokenPayload["http://cloud.netapp.com/user_metadata"].company;
            const uid = idTokenPayload.sub;
            return state.setIn(["auth", "data", "email"], email)
                .setIn(["auth", "data", "userName"], name)
                .setIn(["auth", "data", "company"], company)
                .setIn(["auth", "data", "uid"], uid)
                .setIn(["auth", "inProgress"], false)
                .setIn(["auth", "failed"], false)
                .setIn(["auth", "succeeded"], true)
                .setIn(["auth", "accessToken"], accessToken);
        }
        case actionCreatorsConsts.AUTH_REFRESH_SSO_FAILED: {
            return state.setIn(["auth", "inProgress"], false)
                .setIn(["auth", "succeeded"], false)
                .setIn(["auth", "failed"], true)
                .setIn(["auth", "error"], action.payload.error_description)
                .setIn(["auth", "needLogin"], action.payload.needLogin);
        }
        case actionCreatorsConsts.GET_ACCOUNTS: {
            let updatedState = updateKeyWithBuiltState(state, action, ["_accounts"]);
            if (action.status === actionCreatorsConsts.STATUS_ERROR) {
                updatedState = updatedState.setIn(['auth', 'accountError'], action.payload.data.message);
            }

            return updatedState;
        }
        case actionCreatorsConsts.GET_ACCOUNT_INFO: {
            let updatedState = updateKeyWithBuiltState(state, action, ["_accountInfo"]);

            if (action.status === actionCreatorsConsts.STATUS_SUCCESS && action.payload) {
                const {payload} = action;
                updatedState = updatedState.setIn(['auth', 'data'], merge({}, get(updatedState, "auth.data"), payload));
                return licenseUtils.licenseEnforcement(updatedState);
            }
            if (action.status === actionCreatorsConsts.STATUS_ERROR) {
                return updatedState.setIn(['auth', 'accountError'], action.payload.data.message);
            }
            return updatedState;
        }
        case actionCreatorsConsts.CLEAR_ACCOUNT_DATA: {
            const {auth, _accounts, _accountInfo, _routing, _cmIntegration, _ui} = state;
            return Immutable({
                ...globalDefaultState(),
                auth,
                _accounts,
                _accountInfo,
                _routing,
                _cmIntegration,
                _ui
            });
        }
        case actionCreatorsConsts.BOOT_ANALYTICS: {
            const {payload: {isCmIntegration}} = action;
            const force = get(state._routing, "location.query.analytics", "").toLowerCase() === "true";
            if (force || process.env.NODE_ENV === "production") {
                //The Google Analytics "CS-OCCM Integarion" view is filtered by "Application Type" = "OCCM integration"
                bootGoogleAnalytics(isCmIntegration ? "OCCM integration" : "Stand alone app", get(state,'auth.data.email'));
            }
            return state;
        }
        case actionCreatorsConsts.UPDATE_CM_INTEGRATION: {
            return state.setIn(["_ui", "isCmIntegration"], true)
        }
        case actionCreatorsConsts.UPDATE_DASHBOARD_VIEW: {
            const {payload} = action;
            return state.setIn(["_ui", "dashboardView"], payload)
        }
        default: {
            //do nothing
        }
    }

    if (!state.auth) return state;

    switch (action.type) {
        case actionCreatorsConsts.LOGOUT: {
            window.Intercom("shutdown");
            window.auth.logout();
            return globalDefaultState();
        }
        case actionCreatorsConsts.GET_LICENSE_SETTINGS: {
            const calculateLicenseType = ({freeTrial, totalReservedRelationships, chargedRelationships}) => {
                if (freeTrial) return consts.LICENSE_TYPE_FREE_TRIAL;
                else {
                    if (Number.isInteger(totalReservedRelationships)) {
                        if (totalReservedRelationships === 0) {
                            return consts.LICENSE_TYPE_PAYGO;
                        } else if (chargedRelationships > 0) {
                            return consts.LICENSE_TYPE_PAYGO_WITH_BYOL;
                        } else {
                            return consts.LICENSE_TYPE_BYOL
                        }
                    }
                }
                return consts.LICENSE_TYPE_UNKNOWN;
            };

            let builtState = updateKeyWithBuiltState(state, action, ["_licenseSettings"]);

            if (action.status === actionCreatorsConsts.STATUS_SUCCESS && action.payload) {
                builtState = licenseUtils.licenseEnforcement(builtState);

                return builtState.setIn(["_licenseSettings", "data", "licenseType"], calculateLicenseType(action.payload));
            }

            return builtState;
        }
        case actionCreatorsConsts.GET_BYOL_LICENSES: {
            return licenseUtils.licenseEnforcement(updateKeyWithBuiltState(state, action, ["_licenses"]));
        }
        case actionCreatorsConsts.ADD_BYOL_LICENSE: {
            if (action.status === actionCreatorsConsts.STATUS_SUCCESS && action.payload) {
                return state.updateIn(["_licenses", "data", "licenses"], licenses => {
                    var licensesMutable = licenses ? licenses.asMutable() : [];
                    licensesMutable.push(action.payload);
                    return Immutable(licensesMutable);
                });
            }
            return state;
        }
        case actionCreatorsConsts.VISITED_TRIAL_PAGE: {
            return state.setIn(["_ui", "visitedTrialPage"], true)
                .setIn(["_ui", "showTrialPage"], false);
        }
        case actionCreatorsConsts.UPDATE_SUBSCRIPTION_STATUS: {
            const builtState = state.setIn(['auth', 'data', 'subscriptionStatus'], action.payload[consts.PAYLOAD_UPDATE_SUBSCRIPTION_STATUS].subscriptionStatus);

            return licenseUtils.licenseEnforcement(builtState);
        }
        case actionCreatorsConsts.UPDATE_ACCOUNT_INFO: {
            const payload = action.payload[consts.PAYLOAD_UPDATE_ACCOUNT_INFO];
            if (state._accountInfo.data.accountId === payload.accountId) {
                const updatedState = state.setIn(["_accountInfo", "data"], merge({}, get(state, "_accountInfo.data"), payload))
                    .setIn(["auth", "data"], merge({}, get(state, "auth.data"), payload));
                return licenseUtils.licenseEnforcement(updatedState);
            }
            return state;
        }
        case actionCreatorsConsts.UPDATE_SUPPORT_INFO: {
            const {supportInfo: {status, serialNumber}} = action.payload[consts.PAYLOAD_UPDATE_SUPPORT_INFO];
            return state.setIn(["auth", "data", "supportRegistrationStatus"], status)
                .setIn(["auth", "data", "supportSerialNumber"], serialNumber);
        }
        case actionCreatorsConsts.START_REGISTER_TO_SUPPORT: {
            if (action.status === actionCreatorsConsts.STATUS_SUCCESS) {
                return state.setIn(["auth", "data", "supportRegistrationStatus"], consts.SUPPORT_IN_PROGRESS);
            } else return state;
        }
        case actionCreatorsConsts.GET_TIMELINE: {
            return updateKeyWithBuiltState(state, action, ["_timeline"], (builtState) => {
                if (action.status === actionCreatorsConsts.STATUS_SUCCESS && action.payload) {
                    builtState.data.map((timeline) => {
                        processTimeline(timeline);
                        return timeline;
                    })
                }

                return builtState;
            });
        }
        case actionCreatorsConsts.UPDATE_TIMELINE: {
            let actionTimeline = action.payload[consts.PAYLOAD_TIMELINE];
            let updatedTimelineState = state;
            if (state._timeline.data) {
                if (!find(state._timeline.data, (o) => //if same timeline row is not in the table already, and this relatioship requestId didnt get done status
                {
                    return (o.id === actionTimeline.id || (o.requestId === actionTimeline.requestId && o.status === "Success"))
                })) {
                    processTimeline(actionTimeline);
                    updatedTimelineState = state.updateIn(["_timeline", "data"], timelineActions => {
                        if (timelineActions) {
                            var actionsMutable = timelineActions.asMutable();
                            actionsMutable.unshift(actionTimeline);
                            return Immutable(actionsMutable);
                        }
                        return timelineActions;
                    });
                } else {
                    updatedTimelineState = state.updateIn(["_timeline", "data"], timelines => {
                        return timelines.map(timeline => {
                            if (timeline.id === actionTimeline.id) {
                                timeline = actionTimeline;
                                processTimeline(timeline);
                            }
                            return timeline;
                        });
                    })
                }
            }
            return updatedTimelineState;
        }
        case actionCreatorsConsts.ADD_NOTIFICATION: {
            return state.set("_notifications", [...state._notifications, {...action.payload, id: uuidv4()}]);
        }
        case actionCreatorsConsts.REMOVE_NOTIFICATION: {
            const {notificationIdToRemove} = action.payload;
            const updatedNotifications = filter(state._notifications, notification => notification.id !== notificationIdToRemove);
            return state.set("_notifications", updatedNotifications);
        }
        //This is a special case that a message in handles by 2 reducers
        case UPDATE_DATA_BROKER_STATUS: {
            const {status, isNewBroker} = action.payload[consts.PAYLOAD_DATA_BROKER_STATUS];
            if (status === BROKER_STATUS_COMPLETE && isNewBroker) {
                return state.set("_notifications", [...state._notifications, {
                    id: uuidv4(),
                    type: NOTIFICATION_TYPES.INFO,
                    text: "Data Broker deployment is completed"
                }]);
            }
            return state;
        }
        case actionCreatorsConsts.SET_CM_INTEGRATION_DETAILS: {
            return state.setIn(["_cmIntegration", "inputDetails"], action.payload);
        }
        case actionCreatorsConsts.CLEAR_CM_INTEGRATION_DETAILS: {
            return state.setIn(["_cmIntegration", "inputDetails"], null);
        }
        default: {
            return state
        }
    }
};