import _ from "lodash";
import sdk from '../../sdk.js';
import cvsApi from '../../services/cvs.api.service';
import {getCredentialsByType} from '../../utils/sync-utils';
import {fixPath, removeStartingAndEndingSlash} from "../../utils/encoders";
import {removeKeyFromPath} from "../../utils/helpers";
import * as actionCreatorsConsts from "./action_creators_consts";
import {create, createMultiple} from "../../utils/actions-utils";
import {GET_LICENSE_SETTINGS} from "../global/action_creators_consts";
import {rollbackDeployment} from "../../services/rollback.deployment.service";

export const startup = (dispatch) => {
    setTimeout(() => {
        dispatch(getRelationships());
    })
};


export const updateUrlParams = (params) => create(actionCreatorsConsts.UPDATE_URL_PARAMS, params);


export const createDataBroker = (values) => create(actionCreatorsConsts.DATA_BROKER, sdk.createDataBroker(values), undefined, true, {port: values.port}); //self error handling

export const createDataBrokerWithTemplate = (values) => create(actionCreatorsConsts.DATA_BROKER, sdk.createDataBrokerWithTemplate(values), undefined, true); //self error handling

export const dataBrokers = () => create(actionCreatorsConsts.GET_DATA_BROKERS, sdk.dataBrokers());

export const getServersByType = (protocol, provider, protocolAndProviderKey) => create(actionCreatorsConsts.GET_SERVERS, sdk.getServers(protocol, provider), undefined, false, {protocolAndProviderKey});

export const azureStorageAccounts = () => create(actionCreatorsConsts.GET_AZURE_STORAGE_ACCOUNTS, sdk.azureStorageAccounts());


export function deleteDataBroker(id, existingDeployment) {
    return dispatch =>
        dispatch({
            type: actionCreatorsConsts.DELETE_DATA_BROKER,
            payload: sdk.deleteDataBroker(id).then(() => {
                rollbackDeployment(existingDeployment);
                dispatch(dataBrokers());
            })
        });
}


export const clearCloseSearchWidget = () => create(actionCreatorsConsts.CLEAR_CLOSE_SEARCH_WIDGET);

export const closeSearchWidget = () => create(actionCreatorsConsts.CLOSE_SEARCH_WIDGET);

export const pushUiState = (state) => create(actionCreatorsConsts.PUSH_UI_STATE, state);

export const popUiState = () => create(actionCreatorsConsts.POP_UI_STATE);

export const clearUrlParamsIfChanged = (changedParam, newVal, without, extraKeyToClear, forceClearDueToCredentialsChange) => create(actionCreatorsConsts.CLEAR_URL_PARAMS, {changedParam, newVal, without, extraKeyToClear, forceClearDueToCredentialsChange});

export const clearShowAction = (keyToClear) => create(actionCreatorsConsts.CLEAR_SHOW_ACTION, {keyToClear});

export const bucketsSelectStep = () => create(actionCreatorsConsts.BUCKETS_SELECT_STEP);
export const bucketsEncryptionStep = () => create(actionCreatorsConsts.BUCKETS_ENCRYPTION_STEP);

export const azureSelectStep = () => create(actionCreatorsConsts.AZURE_SELECT_STEP);
export const azureAdditionalStorageSettingsStep = () => create(actionCreatorsConsts.AZURE_ADDITIONAL_SETTINGS_STEP);

export const serverCredentialsStep = (serverType) => create(actionCreatorsConsts.SERVER_CREDENTIALS_STEP, {serverType: _.camelCase(serverType.toLowerCase())});
export const serverIPStep = (serverType) => create(actionCreatorsConsts.SERVER_IP_STEP, {serverType: _.camelCase(serverType.toLowerCase())});

export const showExports = (dataBrokerId, fileServer) => {
    return dispatch => {
        dispatch({
            type: actionCreatorsConsts.SHOW_EXPORTS,
            extra: {host: fileServer, dataBrokerId},
            payload: sdk.showExports(dataBrokerId, fileServer)
        })
    }
};

export const showEFSExports = (dataBrokerId, fileServer) => {
    return dispatch => {
        dispatch({
            type: actionCreatorsConsts.SHOW_EXPORTS,
            extra: {host: fileServer, dataBrokerId},
            payload: sdk.showEFSExports(dataBrokerId, fileServer)
        })
    }
};

export const listFolders = (dataBrokerId, fileServer, path, exportPath, nfsVersion) => {
    const finalPath = removeKeyFromPath(path, exportPath);

    return dispatch => {
        dispatch({
            type: actionCreatorsConsts.LIST_FOLDERS,
            extra: {fileServer, path, dataBrokerId},
            payload: sdk.listFolder(dataBrokerId, fileServer, finalPath, exportPath, nfsVersion)
        });
    }
};

export const showShares = (dataBrokerId, host, cifsVersion, isTargetOfSameSrc) => {
    const cifsCredentials = getCredentialsByType("cifs", host, isTargetOfSameSrc);
    if (!cifsCredentials) {
        return dispatch => {
            dispatch({
                type: actionCreatorsConsts.SHOW_SHARES,
                status: actionCreatorsConsts.STATUS_ERROR,
                extra: {host, dataBrokerId},
                payload: {
                    data: {message: "Unable to get shares, no credentials defined"}
                }
            })
        }
    }
    else {
        const {username, password, domain} = cifsCredentials;
        return dispatch => {
            dispatch({
                type: actionCreatorsConsts.SHOW_SHARES,
                extra: {host, dataBrokerId, username, password, domain, cifsVersion},
                payload: sdk.showShares(dataBrokerId, host, username, password, domain, cifsVersion)
            })
        }
    }
};

export const listShareFolders = (dataBrokerId, host, path, share, cifsVersion, isTargetOfSameSrc) => {
    const cifsCredentials = getCredentialsByType("cifs", host, isTargetOfSameSrc);
    //remove the share part from the path
    const fixedShare = fixPath(share);
    const fixedPath = removeKeyFromPath(fixPath(path), fixedShare);
    // console.log(`%c listShareFolders fixedShare @ ${fixedShare} fixedPath @ (${fixedPath})`, "font-weight: bold;");
    return dispatch => {
        dispatch({
            type: actionCreatorsConsts.LIST_SHARE_FOLDERS,
            extra: {host, path, dataBrokerId}, //put the original path since we need the starting slash in the drilldown
            payload: sdk.listShareFolders(dataBrokerId, host, fixedShare, fixedPath, cifsCredentials.username, cifsCredentials.password, cifsCredentials.domain, cifsVersion)
        });
    }
};

export const showS3BasedBuckets = (dataBrokerId, host, isTargetOfSameSrc, provider) => {
    const osCredentials = getCredentialsByType("objectStorage", host, isTargetOfSameSrc);
    if (!osCredentials) throw new Error(`Missing credentials in local storage`);
    const {port, accessKey, secretKey} = osCredentials;
    if (!osCredentials) {
        return dispatch => {
            dispatch({
                type: actionCreatorsConsts.SHOW_S3BASED_BUCKETS,
                status: actionCreatorsConsts.STATUS_ERROR,
                payload: {
                    data: {message: "Unable to get buckets, no credentials defined"}
                },
                extra: {host}
            })
        }
    }
    else {
        return dispatch => {
            dispatch({
                type: actionCreatorsConsts.SHOW_S3BASED_BUCKETS,
                extra: {host: host, dataBrokerId, port, accessKey, secretKey},
                payload: sdk.fetchS3BasedBuckets(dataBrokerId, host, osCredentials, provider)
            })
        }
    }
};

export const listS3BasedBucketFolders = (dataBrokerId, host, bucket, path, isTargetOfSameSrc, provider) => {
    const osCredentials = getCredentialsByType("objectStorage",host, isTargetOfSameSrc);
    //remove the share part from the path
    const fixedPath = removeKeyFromPath(path, bucket);
    // console.log(`%c listS3BasedBucketFolders bucket @ ${bucket} fixedPath @ (${fixedPath})`, "font-weight: bold;");
    return dispatch => {
        dispatch({
            type: actionCreatorsConsts.LIST_S3BASED_BUCKETS_FOLDERS,
            extra: {host, bucket, path}, //keep the original path since we need it in the drill down
            payload: sdk.listS3BasedBucketFolders(dataBrokerId, host, bucket, fixedPath, osCredentials.accessKey, osCredentials.secretKey, osCredentials.port, provider)
        });
    }
};

export const showGCPBuckets = (dataBrokerId) => {
    return dispatch => {
        dispatch({
            type: actionCreatorsConsts.SHOW_GCP_BUCKETS,
            extra: {dataBrokerId},
            payload: sdk.listGCPBuckets(dataBrokerId)
        })
    }
};

export const listGCPBucketFolders = (dataBrokerId, bucket, path) => {
    //remove the share part from the path
    const fixedPath = removeKeyFromPath(path, bucket);
    // console.log(`%c listGCPBucketFolders bucket @ ${bucket} fixedPath @ (${fixedPath})`, "font-weight: bold;");
    return dispatch => {
        dispatch({
            type: actionCreatorsConsts.LIST_GCP_BUCKET_FOLDERS,
            extra: { dataBrokerId, bucket, path}, //keep the original path since we need it in the drill down
            payload: sdk.listGCPBucketFolders(dataBrokerId, bucket, fixedPath)
        });
    }
};

export const listAwsKmsKeys = (dataBrokerId, bucket) => {
    return dispatch => {
        dispatch({
            type: actionCreatorsConsts.LIST_AWS_KMS_KEYS,
            extra: {dataBrokerId, bucket},
            payload: sdk.listAwsKmsKeys(dataBrokerId, bucket)
        })
    }
};

export const clearSync = () => create(actionCreatorsConsts.CLEAR_SYNC);

export const clearWizardCache = (cifs, objectStorage, s3, azure) => create(actionCreatorsConsts.CLEAR_WIZARD_CACHE, {cifs, objectStorage, s3, azure});

export const clearFoldersCache = (props) => create(actionCreatorsConsts.CLEAR_FOLDERS_CACHE, props);

export const addNewBucket = (bucketName, dataBrokerId) => create(actionCreatorsConsts.ADD_BUCKET_NAME, {bucketName, dataBrokerId});

export const addNewExport = (exportName, nfsServer, dataBrokerId) => create(actionCreatorsConsts.ADD_EXPORT_MANUALLY, {exportName, nfsServer, dataBrokerId});

export const addNewShare = (shareName, cifsServer, dataBrokerId) => create(actionCreatorsConsts.ADD_SHARE_MANUALLY, {shareName, cifsServer, dataBrokerId});

export const createSync = (opts, isCmIntegration) => create(actionCreatorsConsts.CREATE_SYNC, sdk.createSync(opts, isCmIntegration));

export const syncNow = (relationshipId) => {
    return dispatch =>
        dispatch({
            type: actionCreatorsConsts.SYNC_NOW,
            extra: {relationshipId},
            payload: sdk.syncNow(relationshipId)
        });
};


export function abortSync(relationshipId) {
    return dispatch =>
        dispatch({
            type: actionCreatorsConsts.ABORT_SYNC,
            extra: {relationshipId},
            payload: sdk.abortSync(relationshipId)
        });
}

export const updateRelationshipSettings = (relationshipId, settings) => create(actionCreatorsConsts.UPDATE_RELATIONSHIP_SETTINGS, sdk.updateRelationship(relationshipId, settings), undefined, true); //self error handling

export const showBuckets = (dataBrokerId) => {
    return dispatch => {
        dispatch({
            type: actionCreatorsConsts.SHOW_BUCKETS,
            extra: {dataBrokerId},
            payload: sdk.fetchBuckets(dataBrokerId)
        });
    }
};

export const listBucketFolders = (dataBrokerId, bucket, path) => {
    const fixedPath = removeKeyFromPath(path, bucket);
    return dispatch => {
        dispatch({
            type: actionCreatorsConsts.LIST_BUCKETS_FOLDERS,
            extra: {bucket, path, dataBrokerId},
            payload: sdk.listBucketFolders(dataBrokerId, bucket, fixedPath)
        });
    }
};

export const showAzureBlobContainers = (dataBrokerId, storageAccountName, isTargetOfSameSrc) => {
    const azureCredentialsObject = getCredentialsByType("azure", storageAccountName, isTargetOfSameSrc);
    if (!azureCredentialsObject) {
        return dispatch => {
            dispatch({
                type: actionCreatorsConsts.SHOW_AZURE_BLOB_CONTAINERS,
                status: actionCreatorsConsts.STATUS_ERROR,
                extra: {storageAccountName, dataBrokerId},
                payload: {
                    data: {message: "Unable to get azure containers, storage account connection string is not defined"}
                }
            })
        }
    }
    else {
        const {connectionString} = azureCredentialsObject;
        return dispatch => {
            dispatch({
                type: actionCreatorsConsts.SHOW_AZURE_BLOB_CONTAINERS,
                extra: {dataBrokerId, storageAccountName, connectionString},
                payload: sdk.fetchAzureBlobContainers(dataBrokerId, storageAccountName, connectionString)
            })
        }
    }
};

export const listAzureBlobContainerFolders = (dataBrokerId, storageAccountName, container, path, isTargetOfSameSrc) => {
    const azureCredentialsObject = getCredentialsByType("azure", storageAccountName, isTargetOfSameSrc);
    const fixedPath = removeKeyFromPath(path, container);
    if (!azureCredentialsObject) {
        return dispatch => {
            dispatch({
                type: actionCreatorsConsts.SHOW_AZURE_BLOB_CONTAINERS,
                status: actionCreatorsConsts.STATUS_ERROR,
                extra: {storageAccountName, dataBrokerId},
                payload: {
                    data: {message: "Unable to get azure containers, storage account conenction string is not defined"}
                }
            })
        }
    }
    else {
        const {connectionString} = azureCredentialsObject;
        return dispatch => {
            dispatch({
                type: actionCreatorsConsts.LIST_CONTAINERS_FOLDERS,
                extra: {container, path, storageAccountName, connectionString, dataBrokerId},
                payload: sdk.listContainerFolders(dataBrokerId, storageAccountName, connectionString, container, removeStartingAndEndingSlash(fixedPath))
            })
        }
    }
};


export const getCost = (opts) => create(actionCreatorsConsts.GET_COST, sdk.getCost(opts));
export const getCostTimes = () => create(actionCreatorsConsts.GET_COST_TIMES, sdk.getCostTimes());


export function deleteRelationship(id) {
    return dispatch =>
        dispatch({
            type: actionCreatorsConsts.DELETE_RELATIONSHIP,
            extra: {id},
            payload: sdk.deleteRelationship(id).then(() => dispatch(getRelationships()))
        });
}

export const downloadDataBrokerLogs = (groupId, relationshipId) => {
    return dispatch =>
        dispatch({
            type: actionCreatorsConsts.DOWNLOAD_DATA_BROKER_LOGS,
            extra: {relationshipId},
            payload: sdk.downloadDataBrokerLogs(groupId)
        });
};

export const startWizard = (protocols, wizardInfo, isSecure, stepsOrder, stepInfoById) => create(actionCreatorsConsts.START_WIZARD, {
    protocols,
    wizardInfo,
    isSecure,
    stepsOrder,
    stepInfoById
});
export const wizardStepStarted = (stepId) => create(actionCreatorsConsts.STARTED_WIZARD_STEP, stepId);
export const wizardStepExited = (stepId) => create(actionCreatorsConsts.EXITED_WIZARD_STEP, stepId);
const createUpdater = (type, payload) => {
    const actualPayload = _.values(payload)[0];
    payload.data = actualPayload;

    return create(type, payload, actualPayload.status === "FAILED" ? actionCreatorsConsts.STATUS_ERROR : undefined)
};
export const updateDataBrokerStatus = (payload) => create(actionCreatorsConsts.UPDATE_DATA_BROKER_STATUS, payload, undefined); //handles errors by itself
export const updateExports = (payload) => createUpdater(actionCreatorsConsts.UPDATE_EXPORTS, payload);
export const updateShares = (payload) => createUpdater(actionCreatorsConsts.UPDATE_SHARES, payload);
export const updateFolders = (payload) => createUpdater(actionCreatorsConsts.UPDATE_FOLDERS, payload);
export const updateShareFolders = (payload) => createUpdater(actionCreatorsConsts.UPDATE_SHARE_FOLDERS, payload);
export const updateBuckets = (payload) => createUpdater(actionCreatorsConsts.UPDATE_BUCKETS, payload);
export const updateBucketsFolders = (payload) => createUpdater(actionCreatorsConsts.UPDATE_BUCKETS_FOLDERS, payload);
export const updateContainers = (payload) => createUpdater(actionCreatorsConsts.UPDATE_CONTAINERS, payload);
export const updateContainersFolders = (payload) => createUpdater(actionCreatorsConsts.UPDATE_CONTAINERS_FOLDERS, payload);
export const updateS3BasedBuckets = (payload) => createUpdater(actionCreatorsConsts.UPDATE_S3BASED_BUCKETS, payload);
export const updateS3BasedBucketsFolders = (payload) => createUpdater(actionCreatorsConsts.UPDATE_S3BASED_BUCKETS_FOLDERS, payload);
export const updateGCPBuckets = (payload) => createUpdater(actionCreatorsConsts.UPDATE_GCP_BUCKETS, payload);
export const updateGCPBucketFolders = (payload) => createUpdater(actionCreatorsConsts.UPDATE_GCP_BUCKET_FOLDERS, payload);
export const updateRelationship = (payload) => create(actionCreatorsConsts.UPDATE_RELATIONSHIP, payload);//handles errors by itself
export const updateDatabrokerTransferRate = (payload) => create(actionCreatorsConsts.UPDATE_DATABROKER_TRANSFER_RATE, payload);
export const updateKmsKeys = (payload) => create(actionCreatorsConsts.UPDATE_KMS_KEYS, payload);

export const updateDataBrokerLogs = (payload) => create(actionCreatorsConsts.UPDATE_DATA_BROKER_LOGS, payload);

export const toggleErrorsState = () => create(actionCreatorsConsts.TOGGLE_ERRORS_STATE);
export const removeError = (id) => create(actionCreatorsConsts.REMOVE_ERROR, id);
export const broadcastErrorsPanelSize = (rect) => create(actionCreatorsConsts.BROADCAST_ERRORS_PANEL_SIZE, rect);

export const getRelationships = () => createMultiple([{type:actionCreatorsConsts.GET_RELATIONSHIPS, payload:sdk.getRelationships()},{type: GET_LICENSE_SETTINGS, payload: sdk.getLicenseSettings()}]);

export const getCvsWorkingEnvironments = (accountId) => create(actionCreatorsConsts.GET_CVS_WORKING_ENVIRONMENTS, cvsApi.getWorkingEnvironments(accountId), undefined);
export const getAnfVolumes = (accountId, credentialsId) => create(actionCreatorsConsts.GET_ANF_VOLUMES, cvsApi.getAnfVolumes(accountId, credentialsId), undefined, false, {credentialsId});
