import * as actionCreatorsConsts from './action_creators_consts.js';
import * as consts from "../../consts";
import _ from "lodash";
import Immutable from "seamless-immutable";
import {costTime} from "../../components/tabs/cost/costUtils";
import * as listActionsUtils from "../../utils/listActionsUtils";
import {decodeAll, encodeAll} from "../../utils/encoders";
import {endPointTypes, getWizardTypeForSystemFlow, removeCredentialsByType} from '../../utils/sync-utils'
import {getNextTimeHour, removeSettingsFromSessionStorage} from '../../utils/settings-utils';
import {
    addUnhandledMessages,
    buildStateFromAction,
    handleErrors,
    handleUnhandledMessages,
    processAgent, processRelationship,
    resetActivityValues,
    updateCurrentStep,
    updateErrorsPanelStatus,
    updateKeyWithBuiltState,
    updateLastRequestsForRelationshipsData
} from "../../utils/reducerUtils";
import * as actions from './action_creators_consts';
import {sendEvents} from "../../utils/helpers";

export const syncDefaultState = (_routing={}) => Immutable({
    _routing,
    _ui: {},
    unhandledMessages: [],
    _generalErrors: {_errorsMap: {}, _errorsPanelOpen: false, _errorsPanelSize: {}},
    _servers: null,
    _dataBrokers: {inProgress: true},
    _wizard: null,
    _costTimes: {inProgress: true},
    _cost: {inProgress: true},
    _downloadLogsRequests: {},
    _azureStorageAccounts: {inProgress: true},
    _showExports: {},
    _showShares: {},
    _showBuckets: {},
    _showAzureBlobContainers: {},
    _showS3BasedBuckets: {},
    _showGCPBuckets: {},
    _relationships: {inProgress: true},
    _cvsWorkingEnvironments: {},
    _anfVolumes: {}
});

export const syncReducer = (initialState = syncDefaultState(), action = {}) => {
    const state = handleErrors(action, initialState);
    const stateWithWizardInfo = fillWizardData(state);
    const updatedState = handleAction(stateWithWizardInfo, action);
    return handleUnhandledMessages(updatedState, handleAction);
};

function fillWizardData(state) {
    if (!state._wizard) return state;
    const {_routing: {decodedParams}} = state;
    // console.log(`########### decodedParams from router ${JSON.stringify(decodedParams)}`);
    return state.update("_wizard", _wizard => _wizard.set("urlParams", Object.assign({}, _.get(_wizard, "urlParams"), decodedParams)));
}

const handleAction = ((state, action) => {
    action = _.cloneDeep(action);

    switch (action.type) {
        case actionCreatorsConsts.UPDATE_URL_PARAMS: {
            return state.set("_routing", action.payload);
        }
        case actionCreatorsConsts.SERVER_CREDENTIALS_STEP: {
            const {payload: {serverType}} = action;
            return state.setIn(["_ui", serverType, "step"], "credentials")
        }
        case actionCreatorsConsts.SERVER_IP_STEP: {
            const {payload: {serverType}} = action;
            return state.setIn(["_ui", serverType, "step"], "ip")
        }
        case actionCreatorsConsts.BUCKETS_ENCRYPTION_STEP: {
            return state.setIn(["_ui", "s3", "step"], "encryption")
        }
        case actionCreatorsConsts.AZURE_ADDITIONAL_SETTINGS_STEP: {
            return state.setIn(["_ui", "azure", "step"], "additionalSettings")
        }
        case actionCreatorsConsts.AZURE_SELECT_STEP: {
            return state.setIn(["_ui", "azure", "step"], "select")
        }
        case actionCreatorsConsts.BUCKETS_SELECT_STEP: {
            return state.setIn(["_ui", "s3", "step"], "select")
        }

        case actionCreatorsConsts.PUSH_UI_STATE: {
            return state.updateIn(["_ui", "states"], states => states ? states.concat(action.payload) : [action.payload])
        }
        case actionCreatorsConsts.POP_UI_STATE: {
            return state.updateIn(["_ui", "states"], states => _.dropRight(states, 1))
        }
        case actionCreatorsConsts.CLOSE_SEARCH_WIDGET: {
            return state.setIn(["_ui", "closeSearchWidget"], true);
        }
        case actionCreatorsConsts.CLEAR_CLOSE_SEARCH_WIDGET: {
            return state.setIn(["_ui", "closeSearchWidget"], false);
        }
        case actionCreatorsConsts.CLEAR_URL_PARAMS: {
            const {payload: {changedParam, newVal, without, extraKeyToClear, forceClearDueToCredentialsChange}} = action;
            //if the new parameter value !== old parameter value (in urlParams) then clear what is requested:
            const oldVal = decodeAll(state._wizard.urlParams[changedParam]);
            if (forceClearDueToCredentialsChange || (oldVal && newVal !== oldVal)) {
                //console.log(`%c%c reducer clearUrlParams changedParam: ${changedParam} oldVal ${oldVal} newVal ${newVal} without (${without}) extraKeyToClear (${extraKeyToClear}) forceClearDueToCredentialsChange ${forceClearDueToCredentialsChange}`, "font-weight:bold", "background:yellow");
                //clear the relevant params from _wizard.urlParams and extraKey; clear _routing.decodedParams completely
                const newState = extraKeyToClear ? state.setIn(extraKeyToClear, undefined) : state;
                return newState.setIn(["_routing", "decodedParams"], undefined)
                    .updateIn(["_wizard", "urlParams"], obj => obj ? Immutable.without(obj, without) : obj);
            }
            return state;
        }
        case actionCreatorsConsts.CLEAR_SHOW_ACTION: {
            const {payload: {keyToClear}} = action;
            if (keyToClear) return state.setIn(keyToClear, undefined);
            return state;
        }
        case actionCreatorsConsts.GET_DATA_BROKERS: {
            return updateKeyWithBuiltState(state, action, ["_dataBrokers"], (builtState) => {
                if (action.status === actionCreatorsConsts.STATUS_SUCCESS && action.payload) {
                    builtState.data = builtState.data.map((entry) => processAgent(entry, _.get(state, "_wizard.wizardInfo")));
                }
                return builtState;
            });
        }
        case actionCreatorsConsts.GET_SERVERS: {
            const originalState = _.get(state, "_servers." + action.extra.protocolAndProviderKey);
            if (originalState && action.status !== actions.STATUS_PROGRESS && originalState.id !== action.id) {
                // eslint-disable-next-line no-console
                console.log(`%c handleAction discarding action @ ${action.type} since the id doesn't match the id of the same action in the store`, "font-weight: bold;");
                return state;
            }
            return state.setIn(["_servers", action.extra.protocolAndProviderKey], buildStateFromAction(action));
        }
        case actionCreatorsConsts.GET_AZURE_STORAGE_ACCOUNTS: {
            return updateKeyWithBuiltState(state, action, ["_azureStorageAccounts"]);
        }
        case actionCreatorsConsts.DATA_BROKER: {
            if (action.status === actionCreatorsConsts.STATUS_SUCCESS && action.payload) {
                const {id, fileLink, type, name} = action.payload;
                const {port} = action.extra;
                const existingDataBrokers = _.get(state._dataBrokers, "data", []);
                const newDataBroker = processAgent({id, fileLink, type, name, port}, _.get(state, "_wizard.wizardInfo"));
                const updatedDataBrokers = [...existingDataBrokers, newDataBroker];
                return state.setIn(["_dataBrokers", "data"], updatedDataBrokers);
            }
            return state;
        }
        case actionCreatorsConsts.DELETE_DATA_BROKER: {
            return updateKeyWithBuiltState(state, action, ["_deleteDataBroker"]);
        }
        case actionCreatorsConsts.SYNC_NOW: {
            if (!action.extra.relationshipId) return state;
            return state.updateIn(["_relationships", "data"], relationships => {
                return relationships.map(relationship => {
                    if (relationship.id === action.extra.relationshipId) {
                        relationship = relationship.asMutable({deep: true});
                        if (action.status === actionCreatorsConsts.STATUS_PROGRESS) {
                            relationship.ui = {...relationship.ui, ...resetActivityValues()};
                            //Set the status as "STARTING" until "update relationship" message comes and updates the actual value
                            //In order for the UI to show "starting sync" in the meantime (and not "completed")
                            relationship.ui.statusText = 'Starting Sync';
                            relationship.ui.status = consts.STATUS_STARTING;
                        }
                        if (action.status === actionCreatorsConsts.STATUS_ERROR) {
                            processRelationship(relationship); // in order to restore the activity values that were reset during "in progress"
                        }
                    }
                    return relationship;
                });
            });
        }
        case actionCreatorsConsts.UPDATE_DATA_BROKER_STATUS: {
            const {status, message, id, placement} = action.payload[consts.PAYLOAD_DATA_BROKER_STATUS];
            let newState = state;

            if (_.get(state, "_relationships.data")) {
                newState = newState.updateIn(["_relationships", "data"], relationships => {
                    return relationships.map((relationship) => {
                        const clonedRelationship = _.cloneDeep(relationship);
                        clonedRelationship.group.dataBrokers = clonedRelationship.group.dataBrokers.map((dataBroker) => {
                            if (dataBroker.id === id) {
                                if (status) dataBroker.status = status;
                                if (placement) dataBroker.placement = placement;
                                if (message) dataBroker.message = message;
                            }
                            return dataBroker;
                        });
                        processRelationship(clonedRelationship);
                        return clonedRelationship;
                    });
                });
            }

            if (!_.get(state, "_dataBrokers.data")) {
                return newState;
            }

            newState = newState.updateIn(["_dataBrokers", "data"], dataBrokers => {
                return dataBrokers.map(dataBroker => {
                    if (dataBroker.id === id) {
                        dataBroker = _.cloneDeep(dataBroker);
                        if (status) dataBroker.status = status;
                        if (placement) dataBroker.placement = placement;
                        if (message) dataBroker.message = message;

                        processAgent(dataBroker, _.get(state, "_wizard.wizardInfo"));
                    }
                    return dataBroker;
                });
            });
            return newState;
        }
        case actionCreatorsConsts.UPDATE_RELATIONSHIP: {
            const {id, activity, group} = action.payload[consts.PAYLOAD_UPDATE_RELATIONSHIP];

            let newUpdatedState;
            let isNewRelationship = true;

            if (_.get(state, "_relationships.data")) {
                newUpdatedState = state.updateIn(["_relationships", "data"], relationships => {
                    return relationships.map(relationship => {
                        if (relationship.id === id) {
                            isNewRelationship = false;
                            const thisRequest = _.find(relationship.lastRequests, function (req) {
                                return req.id === action.payload.requestId;
                            });
                            if (thisRequest && thisRequest.isDone) return relationship; //dont get more update for this request after done status
                            else if (relationship.activity.status === consts.RESPONSE_STATUS_ABORTING && activity.status === consts.RESPONSE_STATUS_RUNNING) {
                                //if in aborting status, dot update to syncing
                                return relationship;
                            }
                            const clonedRelationship = _.cloneDeep(relationship);
                            clonedRelationship.activity = activity;
                            clonedRelationship.group = group;
                            processRelationship(clonedRelationship);
                            return clonedRelationship;
                        }
                        return relationship;
                    })
                });
            }

            if (isNewRelationship) {
                newUpdatedState = state.update("_ui", _ui => _ui.set("hasNewRelationship", true));
            }
            newUpdatedState = updateLastRequestsForRelationshipsData(newUpdatedState, action);

            //update only if sync already exists, hence we are in the wizard
            if (_.get(newUpdatedState, '_wizard.sync') && _.get(newUpdatedState, '_wizard.sync.relationshipId') === id) {
                if (activity.status === consts.RESPONSE_STATUS_FAILED) {
                    return newUpdatedState.updateIn(["_wizard", "sync"], sync => sync.set("completed", false).set("failed", true));
                } else if (activity.status === consts.RESPONSE_STATUS_DONE) {
                    return newUpdatedState.updateIn(["_wizard", "sync"], sync => sync.set("completed", true).set("percentCopied", 100));
                } else {
                    return newUpdatedState.updateIn(["_wizard", "sync"], sync => sync.set("percentCopied", activity.progress).set("completed", false));
                }
            }
            return newUpdatedState;
        }
        case actionCreatorsConsts.UPDATE_DATA_BROKER_LOGS: {

            const downloadFile = async ({url}) => {
                const urlSplit = url.split('/');
                const fileName = urlSplit[urlSplit.length - 1];
                try {
                    const response = await fetch(url);
                    if (response.ok) { //in case of 403, "ok" is false
                        const blob = await response.blob();
                        let url = window.URL.createObjectURL(blob);
                        let a = document.createElement('a');
                        document.body.appendChild(a);// For Firefox https://stackoverflow.com/a/32226068
                        a.href = url;
                        a.download = `${fileName}.7z`;
                        a.click();
                        a.remove();
                    }
                } catch (e) {
                    // eslint-disable-next-line no-console
                    console.error(e)
                }
            };

            //download the files:
            const {logsDetails} = action.payload[consts.PAYLOAD_DATA_BROKER_LOGS];
            _.forEach(logsDetails, downloadFile);

            //update the state:
            const {requestId} = action.payload;
            const relationshipId = _.findKey(state._downloadLogsRequests, (key) => {
                return key.requestId === requestId
            });
            if (!relationshipId) return state;
            return state.setIn(["_downloadLogsRequests", relationshipId, 'status'], actionCreatorsConsts.STATUS_SUCCESS);
        }

        case actionCreatorsConsts.UPDATE_DATABROKER_TRANSFER_RATE: {
            const {dataBrokerId, transferRate} = action.payload[consts.PAYLOAD_DATABROKER_TRANSFER_RATE];

            if (!_.get(state, "_relationships.data")) {
                return state;
            }

            return state.updateIn(["_relationships", "data"], relationships => {
                return relationships.map(relationship => {
                    if (relationship.activity.status !== consts.RESPONSE_STATUS_RUNNING) {
                        return relationship;
                    }

                    const clonedRelationship = relationship.asMutable({deep: true});
                    if (clonedRelationship.group && clonedRelationship.group.dataBrokers) {
                        clonedRelationship.group.dataBrokers.forEach((dataBroker) => {
                            if (dataBroker && dataBroker.id === dataBrokerId) {
                                if (transferRate) {
                                    dataBroker.transferRate = transferRate;
                                    processRelationship(clonedRelationship);
                                }
                            }
                        })
                    }
                    return clonedRelationship;
                });
            });

        }

        case actionCreatorsConsts.DELETE_RELATIONSHIP: {
            if (!_.get(state, "_relationships.data")) {
                return addUnhandledMessages(state, action);
            }
            if (action.extra.id) {
                return state.updateIn(["_relationships", "data"], relationships => {
                    return relationships.map(relationship => {
                        if (relationship.id === action.extra.id) {
                            relationship = relationship.asMutable({deep: true});
                            relationship.activity.isBeingDeleted = true;
                            processRelationship(relationship);
                        }
                        return relationship;
                    });
                });
            }
            return state;
        }
        case actionCreatorsConsts.ABORT_SYNC: {
            if (!_.get(state, "_relationships.data")) {
                return addUnhandledMessages(state, action);
            }
            if (action.extra.relationshipId) {
                return state.updateIn(["_relationships", "data"], relationships => {
                    return relationships.map(relationship => {
                        if (relationship.id === action.extra.relationshipId) {
                            relationship = relationship.asMutable({deep: true});
                            relationship.activity.status = consts.RESPONSE_STATUS_ABORTING;
                            processRelationship(relationship);
                        }
                        return relationship;
                    });
                });
            }
            return state;
        }
        case actionCreatorsConsts.START_WIZARD: {
            action.payload.sync = {};
            const wizardInfo = action.payload.wizardInfo;
            action.payload.wizardTypeForSystemFlow = getWizardTypeForSystemFlow(wizardInfo);
            action.payload.sourceType = endPointTypes[wizardInfo.source.protocol];
            action.payload.targetType = endPointTypes[wizardInfo.target.protocol];

            return state.set("_wizard", action.payload);
        }
        case actionCreatorsConsts.STARTED_WIZARD_STEP: {
            return updateCurrentStep(state.updateIn(["_wizard", "steps"], steps => (steps || Immutable({})).set(action.payload, {started: true})), action.payload, true);
        }
        case actionCreatorsConsts.EXITED_WIZARD_STEP: {
            return updateErrorsPanelStatus(updateCurrentStep(state.updateIn(["_wizard", "steps"], steps => steps.without(action.payload)), action.payload, false).updateIn(["_generalErrors", "_errorsMap"], () => ({})));
        }
        case actionCreatorsConsts.CLEAR_SYNC: {
            return state.update("_wizard", _wizard => _wizard.set("sync", {}));
        }
        case actionCreatorsConsts.CLEAR_WIZARD_CACHE: {
            const {cifs, objectStorage, s3, azure} = action.payload;
            //clear credentials from local storage
            if (cifs) removeCredentialsByType("cifs");
            if (objectStorage) removeCredentialsByType("objectStorage");
            if (s3) removeCredentialsByType("s3");
            if (azure) removeCredentialsByType("azure");

            removeSettingsFromSessionStorage();

            //clear cache of wizard as well
            return state.set("_showBuckets", undefined)
                .set("_showExports", undefined)
                .set("_showShares", undefined)
                .set("_showS3BasedBuckets", undefined)
                .set("_showGCPBuckets", undefined)
                .setIn(["_routing", "decodedParams"], undefined)
                .set("_servers", undefined)
                .set("_azureStorageAccounts", undefined)
                .set("_showAzureBlobContainers", undefined)
                .set("_dataBrokers", undefined);
        }
        case actionCreatorsConsts.CLEAR_FOLDERS_CACHE: {
            //clear cache of specific endpoint
            const {cacheName, storageKey} = action.payload;
            return state.setIn([cacheName, storageKey], undefined)
        }
        case actionCreatorsConsts.ADD_BUCKET_NAME: {

            const {bucketName, dataBrokerId} = action.payload;
            if (bucketName) {
                const bucketObject = {
                    actualPath: bucketName,
                    name: bucketName,
                    path: bucketName,
                    wasAddedManually: true,
                    region: {displayName: "Unknown", name: "Unknown"},
                    tags: []
                };
                const addNewBucket = (data) => {
                    var dataMutable = data ? data.asMutable() : [];
                    dataMutable.unshift(bucketObject);
                    return Immutable(dataMutable);
                };
                if (!state._showBuckets[dataBrokerId].data.find(bucket => bucket.name === bucketName.split('/')[0])) { //if bucket not in the list
                    return state.updateIn(["_showBuckets", dataBrokerId, "folders", "-", "data"], data => addNewBucket(data))
                        .updateIn(["_showBuckets", dataBrokerId, "data"], data => addNewBucket(data));
                }
            }
            return state;
        }
        case actionCreatorsConsts.ADD_SHARE_MANUALLY: {
            const {shareName, cifsServer, dataBrokerId} = action.payload;

            if (shareName) {
                const trimmedShare = shareName[shareName.length - 1] === "/" ? shareName.substring(0, shareName.length - 1) : shareName;
                const shareObject = {
                    actualPath: trimmedShare,
                    path: trimmedShare,
                    host: cifsServer
                };
                const addNewShare = (data) => {
                    const dataMutable = data ? data.asMutable() : [];
                    dataMutable.unshift(shareObject);
                    return Immutable(dataMutable);
                };
                const showSharesKey = `${cifsServer}-${dataBrokerId}`;
                if (!(state._showShares[showSharesKey].data && state._showShares[showSharesKey].data.find(exp => exp.path === trimmedShare))) { //if share not in the list
                    return state.updateIn(["_showShares", showSharesKey, "folders", "-", "data"], data => addNewShare(data))
                        .updateIn(["_showShares", showSharesKey, "data"], data => addNewShare(data))
                        .setIn(["_showShares", showSharesKey, "failedRetrieve"], false)
                        .updateIn(["_generalErrors", "_errorsMap"], () => ({}))
                        .updateIn(["_generalErrors", "_errorsPanelSize"], () => ({height: 0}));
                }
            }
            return state;
        }
        case actionCreatorsConsts.ADD_EXPORT_MANUALLY: {
            const {exportName, nfsServer, dataBrokerId} = action.payload;

            if (exportName) {
                const trimmedExport = exportName[exportName.length - 1] === "/" ? exportName.substring(0, exportName.length - 1) : exportName;
                const exportObject = {
                    actualPath: trimmedExport,
                    path: trimmedExport,
                    host: nfsServer
                };
                const addNewExport = (data) => {
                    var dataMutable = data ? data.asMutable() : [];
                    dataMutable.unshift(exportObject);
                    return Immutable(dataMutable);
                };
                const showExportsKey = `${nfsServer}-${dataBrokerId}`;
                if (!(state._showExports[showExportsKey].data && state._showExports[showExportsKey].data.find(exp => exp.path === trimmedExport))) { //if export not in the list
                    return state.updateIn(["_showExports", showExportsKey, "folders", "-", "data"], data => addNewExport(data))
                        .updateIn(["_showExports", showExportsKey, "data"], data => addNewExport(data))
                        .setIn(["_showExports", showExportsKey, "failedRetrieve"], false)
                        .updateIn(["_generalErrors", "_errorsMap"], () => ({}))
                        .updateIn(["_generalErrors", "_errorsPanelSize"], () => ({height: 0}));
                }
            }
            return state;
        }
        case actionCreatorsConsts.CREATE_SYNC: {
            const updatedState = state.update("_wizard", _wizard => _wizard.set("sync", buildStateFromAction(action)));
            if (action.status === actionCreatorsConsts.STATUS_PROGRESS) {
                return updatedState.update("_wizard", _wizard => _wizard.set("hideHeader", true));
            }
            if (action.status === actionCreatorsConsts.STATUS_ERROR) {
                return updatedState.update("_wizard", _wizard => _wizard.set("hideHeader", false));
            }
            if (action.status === actionCreatorsConsts.STATUS_SUCCESS) {
                //send event to GA
                sendEvents('relationship', "created_relationship");
                const updatedState2 = updatedState.updateIn(["_wizard", "sync"], sync => sync.set("completed", false).set("relationshipId", action.payload.id))
                    .update("_wizard", _wizard => _wizard.set("hideHeader", true));
                if (action.payload.activity.status === "PENDING") {
                    const startTime = action.payload.settings.schedule.nextTime;
                    const startTimeDisplay = getNextTimeHour(startTime);
                    return updatedState2.updateIn(["_wizard", "sync"], sync => sync.set("isPending", true).set("startTime", startTimeDisplay));
                } else {
                    return updatedState2.updateIn(["_wizard", "sync"], sync => sync.set("percentCopied", 0).set("bytesCopied", 0).set("isPending", false));
                }
            } else return updatedState;
        }
        case actionCreatorsConsts.SHOW_EXPORTS: {
            //exports cache is saved by host and dataBrokerId to handle cases of same host but different data broker (can be in encrypted relationship)
            const fullKey = ["_showExports", `${encodeAll(action.extra.host)}-${action.extra.dataBrokerId}`];
            return listActionsUtils.updateStoreForShowAction(fullKey, ['host', 'dataBrokerId'], state, action);
        }
        case actionCreatorsConsts.UPDATE_EXPORTS: {
            //exports cache is saved by both host and dataBrokerId to handle cases of same host but different data broker (can be in encrypted relationship)
            const host = action.payload[consts.PAYLOAD_SHOW_EXPORTS].host;
            const dataBrokerId = action.payload.dataBrokerId;
            const fullKey = ["_showExports", `${encodeAll(host)}-${dataBrokerId}`];
            return listActionsUtils.updateStoreForUpdateShowActionForFileServer(fullKey, consts.PAYLOAD_SHOW_EXPORTS, 'export', state, action, listActionsUtils.isHostEqual.bind(this, fullKey, consts.PAYLOAD_SHOW_EXPORTS, 'export', state, action));

        }
        case actionCreatorsConsts.SHOW_SHARES: {
            //shares cache is saved by both host and dataBrokerId to handle cases of same host but different data broker (can be in encrypted relationship)
            const fullKey = ["_showShares", `${encodeAll(action.extra.host)}-${action.extra.dataBrokerId}`];
            return listActionsUtils.updateStoreForShowAction(fullKey, ['host', 'dataBrokerId', 'domain', 'username', 'password'], state, action);
        }
        case actionCreatorsConsts.UPDATE_SHARES: {
            //shares cache is saved by both host and dataBrokerId to handle cases of same host but different data broker (can be in encrypted relationship)
            const host = action.payload[consts.PAYLOAD_LIST_CIFS_SHARES].host;
            const dataBrokerId = action.payload.dataBrokerId;
            const fullKey = ["_showShares", `${encodeAll(host)}-${dataBrokerId}`];
            return listActionsUtils.updateStoreForUpdateShowActionForFileServer(fullKey, consts.PAYLOAD_LIST_CIFS_SHARES, "share", state, action, listActionsUtils.isHostEqual.bind(this, fullKey, consts.PAYLOAD_LIST_CIFS_SHARES, "share", state, action));

        }
        case actionCreatorsConsts.LIST_FOLDERS: {
            //exports cache is saved by both host and dataBrokerId to handle cases of same host but different data broker (can be in encrypted relationship)
            const fullKey = ["_showExports", `${encodeAll(action.extra.fileServer)}-${action.extra.dataBrokerId}`];
            return listActionsUtils.updateStoreForListFoldersAction(fullKey, action.extra.path, state, action);

        }
        case actionCreatorsConsts.UPDATE_FOLDERS: {
            //exports cache is saved by both host and dataBrokerId to handle cases of same host but different data broker (can be in encrypted relationship)
            const nfsServer = action.payload[consts.PAYLOAD_LIST_FOLDERS].host;
            const dataBrokerId = action.payload.dataBrokerId;
            const fullKey = ["_showExports", `${encodeAll(nfsServer)}-${dataBrokerId}`];
            return listActionsUtils.updateStoreForUpdateFoldersActionForFileServer(fullKey, consts.PAYLOAD_LIST_FOLDERS, state, action);

        }
        case actionCreatorsConsts.LIST_SHARE_FOLDERS: {
            //shares cache is saved by both host and dataBrokerId to handle cases of same host but different data broker (can be in encrypted relationship)
            const fullKey = ["_showShares", `${encodeAll(action.extra.host)}-${action.extra.dataBrokerId}`];
            return listActionsUtils.updateStoreForListFoldersAction(fullKey, action.extra.path, state, action);

        }
        case actionCreatorsConsts.UPDATE_SHARE_FOLDERS: {
            //shares cache is saved by both host and dataBrokerId to handle cases of same host but different data broker (can be in encrypted relationship)
            const host = action.payload[consts.PAYLOAD_LIST_CIFS_SHARE_FOLDERS].host;
            const dataBrokerId = action.payload.dataBrokerId;
            const fullKey = ["_showShares", `${encodeAll(host)}-${dataBrokerId}`];
            return listActionsUtils.updateStoreForUpdateFoldersActionForFileServer(fullKey, consts.PAYLOAD_LIST_CIFS_SHARE_FOLDERS, state, action);

        }
        case actionCreatorsConsts.SHOW_BUCKETS: {
            return listActionsUtils.updateStoreForShowAction(["_showBuckets", action.extra.dataBrokerId], [], state, action);
        }
        case actionCreatorsConsts.UPDATE_BUCKETS: {
            return listActionsUtils.updateStoreForUpdateBucketsAction(["_showBuckets", action.payload.dataBrokerId], consts.PAYLOAD_LIST_BUCKETS, "buckets", state, action);
        }
        case actionCreatorsConsts.LIST_BUCKETS_FOLDERS: {
            return listActionsUtils.updateStoreForListFoldersAction(["_showBuckets", action.extra.dataBrokerId], action.extra.path, state, action, listActionsUtils.isObjExistInField.bind(this, ["_showBuckets", action.extra.dataBrokerId], state, {name: action.extra.bucket}));
        }
        case actionCreatorsConsts.UPDATE_BUCKETS_FOLDERS: {
            return listActionsUtils.updateStoreForUpdateBucketsFoldersAction(["_showBuckets", action.payload.dataBrokerId], consts.PAYLOAD_LIST_SUB_DIRECTORIES, state, action);
        }
        case actionCreatorsConsts.SHOW_AZURE_BLOB_CONTAINERS: {
            return listActionsUtils.updateStoreForShowAction(["_showAzureBlobContainers", action.extra.storageAccountName], ["storageAccountName", "connectionString", "dataBrokerId"], state, action);
        }
        case actionCreatorsConsts.UPDATE_CONTAINERS: {
            return listActionsUtils.updateStoreForUpdateBucketsAction(["_showAzureBlobContainers", action.payload[consts.PAYLOAD_LIST_CONTAINERS].storageAccountName], consts.PAYLOAD_LIST_CONTAINERS, "containers", state, action);
        }
        case actionCreatorsConsts.LIST_CONTAINERS_FOLDERS: {
            return listActionsUtils.updateStoreForListFoldersAction(["_showAzureBlobContainers", action.extra.storageAccountName], action.extra.path, state, action, listActionsUtils.isObjExistInField.bind(this, ["_showAzureBlobContainers", action.extra.storageAccountName], state, {name: action.extra.container}));
        }
        case actionCreatorsConsts.UPDATE_CONTAINERS_FOLDERS: {
            return listActionsUtils.updateStoreForUpdateBucketsFoldersAction(["_showAzureBlobContainers", action.payload[consts.PAYLOAD_LIST_CONTAINER_SUB_DIRECTORIES].storageAccountName], consts.PAYLOAD_LIST_CONTAINER_SUB_DIRECTORIES, state, action, "container", "folders");
        }

        case actionCreatorsConsts.SHOW_S3BASED_BUCKETS: {
            return listActionsUtils.updateStoreForShowAction(["_showS3BasedBuckets", action.extra.host], ["host", "dataBrokerId", "port", "accessKey", "secretKey"], state, action);
        }
        case actionCreatorsConsts.UPDATE_S3BASED_BUCKETS: {
            const payloadName = Object.keys(action.payload).find(key => key.startsWith("list"));
            const theHost = action.payload.data ? (action.payload.data.host ? action.payload.data.host : action.payload.data[payloadName].host) : action.payload[payloadName].host;
            return listActionsUtils.updateStoreForUpdateBucketsAction(["_showS3BasedBuckets", theHost], payloadName, "buckets", state, action);
        }
        case actionCreatorsConsts.LIST_S3BASED_BUCKETS_FOLDERS: {
            return listActionsUtils.updateStoreForListFoldersAction(["_showS3BasedBuckets", action.extra.host], action.extra.path, state, action, listActionsUtils.isObjExistInField.bind(this, ["_showS3BasedBuckets", action.extra.host], state, {name: action.extra.bucket}));
        }
        case actionCreatorsConsts.UPDATE_S3BASED_BUCKETS_FOLDERS: {
            const payloadName = Object.keys(action.payload).find(key => key.startsWith("list"));
            const relHost = action.payload.data ? (action.payload.data.host ? action.payload.data.host : action.payload.data[payloadName].host) : action.payload[payloadName].host;
            return listActionsUtils.updateStoreForUpdateBucketsFoldersAction(["_showS3BasedBuckets", relHost], payloadName, state, action);
        }
        case actionCreatorsConsts.SHOW_GCP_BUCKETS: {
            return listActionsUtils.updateStoreForShowAction(["_showGCPBuckets", action.extra.dataBrokerId], [], state, action);
        }
        case actionCreatorsConsts.UPDATE_GCP_BUCKETS: {
            return listActionsUtils.updateStoreForUpdateBucketsAction(["_showGCPBuckets", action.payload.dataBrokerId], consts.PAYLOAD_LIST_GCP_BUCKETS, "buckets", state, action);
        }
        case actionCreatorsConsts.LIST_GCP_BUCKET_FOLDERS: {
            return listActionsUtils.updateStoreForListFoldersAction(["_showGCPBuckets", action.extra.dataBrokerId], action.extra.path, state, action, listActionsUtils.isObjExistInField.bind(this, ["_showGCPBuckets", action.extra.dataBrokerId], state, {name: action.extra.bucket}));
        }
        case actionCreatorsConsts.UPDATE_GCP_BUCKET_FOLDERS: {
            return listActionsUtils.updateStoreForUpdateBucketsFoldersAction(["_showGCPBuckets", action.payload.dataBrokerId], consts.PAYLOAD_LIST_GCP_BUCKET_FOLDERS, state, action);
        }

        case actionCreatorsConsts.TOGGLE_ERRORS_STATE: {
            return state.updateIn(["_generalErrors", "_errorsPanelOpen"], errorsPanelOpen => !errorsPanelOpen);
        }
        case actionCreatorsConsts.REMOVE_ERROR: {
            return updateErrorsPanelStatus(state.updateIn(["_generalErrors", "_errorsMap"], errors => errors.without(action.payload)));
        }
        case actionCreatorsConsts.BROADCAST_ERRORS_PANEL_SIZE: {
            return state.updateIn(["_generalErrors"], generalErrors => generalErrors.set("_errorsPanelSize", {height: action.payload.height}));
        }

        case actionCreatorsConsts.GET_COST_TIMES: {
            return updateKeyWithBuiltState(state, action, ["_costTimes"], (builtState) => {
                if (action.status === actionCreatorsConsts.STATUS_SUCCESS && action.payload) {
                    builtState.data.map((time) => {
                        const stringTime = costTime(time.month, time.year);
                        time.title = stringTime;
                        time.label = stringTime;
                        time.value = stringTime;
                        return time;
                    });
                }
                return builtState;
            });
        }
        case actionCreatorsConsts.GET_COST: {
            if (action.status !== actionCreatorsConsts.STATUS_SUCCESS || !action.payload) return state;
            const {year, month} = action.payload;
            const stringTime = costTime(month, year);
            const builtState = buildStateFromAction(action);
            return state.setIn(["_cost", stringTime], builtState);
        }
        case actionCreatorsConsts.DOWNLOAD_DATA_BROKER_LOGS: {
            if (action.status === actionCreatorsConsts.STATUS_PROGRESS) {
                const {extra: {relationshipId}} = action;
                return state.setIn(["_downloadLogsRequests", relationshipId], {status: actionCreatorsConsts.STATUS_PROGRESS});
            }
            if (action.status === actionCreatorsConsts.STATUS_SUCCESS) {
                const {requestId, extra: {relationshipId}} = action;
                return state.setIn(["_downloadLogsRequests", relationshipId, 'requestId'], requestId);
            }
            return state;
        }
        case actionCreatorsConsts.GET_RELATIONSHIPS: {
            let nextState = updateKeyWithBuiltState(state, action, ["_relationships"], (builtState) => {
                if (action.status === actionCreatorsConsts.STATUS_SUCCESS && action.payload) {

                    builtState.data.map((entry) => {
                        entry.lastCopyMessage = entry.lastCopyMessage || {};
                        entry.lastScanMessage = entry.lastScanMessage || {};
                        entry.lastSyncMessage = entry.lastSyncMessage || undefined;
                        entry.lastRequests = entry.lastRequests || [];

                        processRelationship(entry);

                        return entry;
                    });
                }
                return builtState;
            });
            return nextState.update("_ui", _ui => _ui.set("hasNewRelationship", false));
        }
        case actionCreatorsConsts.UPDATE_KMS_KEYS: {
            const {keys, bucket} = action.payload[consts.PAYLOAD_LIST_KMS_KEYS];
            const {dataBrokerId} = action.payload;
            const updatedBuckets = _.map(state._showBuckets[dataBrokerId].data, item => (item.name === bucket) ? {...item, kmsKeys: keys} : item);
            return state.setIn(["_showBuckets", dataBrokerId, 'data'], updatedBuckets);
        }
        case actionCreatorsConsts.GET_CVS_WORKING_ENVIRONMENTS: {
            return updateKeyWithBuiltState(state, action, ["_cvsWorkingEnvironments"]);
        }
        case actionCreatorsConsts.GET_ANF_VOLUMES: {
            return updateKeyWithBuiltState(state, action, ["_anfVolumes", action.extra.credentialsId]);
        }
        default:
            return state;
    }
});


export const filteringFunc = (action) => {
    //handle update messages only in certain routes, to optimize the reducer
    const ignoreUpdateRelationship = action.type === actionCreatorsConsts.UPDATE_RELATIONSHIP && !(window.location.href.includes("dashboard") || window.location.href.includes("review") || window.location.href.includes("timeline"));
    const ignoreUpdateTransferRateCondition = action.type === actionCreatorsConsts.UPDATE_DATABROKER_TRANSFER_RATE && !window.location.href.includes("dashboard");
    return ignoreUpdateRelationship || ignoreUpdateTransferRateCondition;
};


// export default (history) => combineReducers({
//     syncReducer: ignoreActions(syncReducer, filteringFunc),
//     router: connectRouter(history),
//     form: formReducer
// });
