import _ from 'lodash';
import {
    FM_AGENT,
    DEFAULT_NFS_VERSION,
    DEFAULT_CIFS_VERSION,
    END_POINT_NAMES_MAP,
    DATA_BROKER_FUNCTION,
    OBJECT_STORAGE_NAMES_MAP,
    ENCRYPTION_OPTIONS,
    ENCRYPTION_TYPES_NAMES
} from '../consts';
import {stringToTags} from './helpers';
import {getSettingsFromSessionStorage} from './settings-utils';
import {removeStartingAndEndingSlash, decodeAll} from './encoders';

export const END_POINT_FILE_SYSTEM = 'fileSystem';
export const END_POINT_STORAGE = 'storage';

export const getWizardInfo = (protocols, providers) => {
    const protocolsObj = splitCombinationToSourceTarget(protocols);
    const providersObj = splitCombinationToSourceTarget(providers);

    return {
        source: getEndpointObject('source', protocolsObj, providersObj),
        target: getEndpointObject('target', protocolsObj, providersObj)
    }
};

const getEndpointObject = (sourceOrTarget, protocolsObj, providersObj) => {
    return {
        protocol: protocolsObj[sourceOrTarget],
        provider: providersObj[sourceOrTarget] === 'NA' ? undefined : providersObj[sourceOrTarget]
    }
};

export const getStepsOrder = (wizardInfo, isSecure, weSelectionNeeded) => {
    const isTargetOfSameSource = wizardInfo.source.protocol === wizardInfo.target.protocol;

    return _.concat(
        //source steps:
        getStepIdsByProtocol({
            protocol: wizardInfo.source.protocol,
            provider: wizardInfo.source.provider,
            isSource: true,
            isSecure,
            isTargetOfSameSource: false,
            weSelectionNeeded: weSelectionNeeded?.source
        }),
        //target steps:
        getStepIdsByProtocol({
            protocol: wizardInfo.target.protocol,
            provider: wizardInfo.target.provider,
            isTargetOfSameSource,
            isSecure,
            weSelectionNeeded: weSelectionNeeded?.target
        }));
};

const getStepIdsByProtocol = ({protocol, provider, isSource, isSecure, isTargetOfSameSource, weSelectionNeeded}) => {

    let result = [];

    if (isSecure) {
        result = isSource
            ? [createStepId(protocol, provider, ), 'dataBrokerFunction', 'dataBroker', 'directories']
            : [createStepId(protocol, provider, true), 'targetDataBroker', 'targetDirectories', 'settings', 'review'];
    }
    else {
        switch (protocol) {
            case 'nfs': {
                result = isSource
                    ? [createStepId(protocol, provider, ), 'dataBroker', 'directories']
                    : isTargetOfSameSource
                        ? [createStepId(protocol, provider, true, ), 'targetDirectories', 'settings', 'review']
                        : [createStepId(protocol, provider, ), 'directories', 'settings', 'review'];
                break;
            }

            case 's3': {
                if (!provider) {
                    result = isSource
                        ? ['dataBroker', 's3Buckets']
                        : isTargetOfSameSource
                            ? ['targetS3Buckets', 'settings', 'relationshipTags', 'review']
                            : ['s3Buckets', 'settings', 'relationshipTags', 'review'];
                }
                else {
                    result = isSource
                        ? [createStepId(protocol, provider), 'dataBroker', createStepId(protocol, provider, false, 'Buckets')]
                        : isTargetOfSameSource
                            ? [createStepId(protocol, provider, true), createStepId(protocol, provider, true, 'Buckets'), 'settings', 'review']
                            : [createStepId(protocol, provider), createStepId(protocol, provider, false, 'Buckets'), 'settings', 'review'];
                }
                break;
            }

            case 'cifs': {
                result = isSource
                    ? [createStepId(protocol, provider), 'dataBroker', 'shares']
                    : isTargetOfSameSource
                        ? [createStepId(protocol, provider, true), 'targetShares', 'settings', 'review']
                        : [createStepId(protocol, provider), 'shares', 'settings', 'review'];
                break;
            }

            case 'gcp': {
                result = isSource
                    ? ['dataBroker', 'gcpBuckets']
                    : isTargetOfSameSource
                        ? ['targetGcpBuckets', 'settings', 'review']
                        : ['gcpBuckets', 'settings', 'review'];
                break;
            }
            case 'azure': {
                result = isSource
                    ? ['azureBlob', 'dataBroker', 'blobContainers']
                    : isTargetOfSameSource
                        ? ['targetAzureBlob', 'targetBlobContainers', 'settings', 'relationshipTags', 'review']
                        : ['azureBlob', 'blobContainers', 'settings', 'relationshipTags', 'review'];
                break;
            }
            default: {
                throw new Error(`Unknown protocol ${protocol}`);
            }
        }
    }

    if (weSelectionNeeded) {
        result.unshift(isTargetOfSameSource ? 'targetAnfWe' : 'anfWe');
    }
    return result;
};

const createStepId = (protocol, provider, isTargetOfSameSource = false, suffix = '') => {
    const protocolAndProvider = getProtocolAndProviderKey(protocol, provider);
    const res = isTargetOfSameSource ? _.camelCase(`target ${protocolAndProvider}`) : protocolAndProvider;
    return `${res}${suffix}`
};

export const getProtocolAndProviderKey = (protocol, provider) => {
    return provider ? _.camelCase(`${protocol} ${provider}`) : protocol;
};

export const stepInfoById = {
    dataBroker: {name: FM_AGENT, path: 'agents'},
    dataBrokerFunction: {name: DATA_BROKER_FUNCTION, path: 'brokerfunction'},
    targetDataBroker: {
        name: `Target ${FM_AGENT}`,
        path: 'targetagents',
        isTargetStepFromSameSourceType: true
    },
    relationshipTags: {name: 'Relationship Tags', path: 'tags'},
    settings: {name: 'Settings', path: 'settings'},
    review: {name: 'Review', path: 'review'},
    nfs: {name: END_POINT_NAMES_MAP.NFS, path: 'nfs', protocol: 'nfs'},
    targetNfs: {
        name: `Target ${END_POINT_NAMES_MAP.NFS}`,
        path: 'targetnfs',
        isTargetStepFromSameSourceType: true,
        protocol: 'nfs'
    },
    nfsAnf: {name: END_POINT_NAMES_MAP.ANF, path: 'nfs', protocol: 'nfs'},
    targetNfsAnf: {
        name: `Target ${END_POINT_NAMES_MAP.ANF}`,
        path: 'targetnfs',
        isTargetStepFromSameSourceType: true,
        protocol: 'nfs'
    },
    nfsEfs: {name: END_POINT_NAMES_MAP.EFS, path: 'nfs', protocol: 'nfs'},
    targetNfsEfs: {
        name: `Target ${END_POINT_NAMES_MAP.EFS}`,
        path: 'targetnfs',
        isTargetStepFromSameSourceType: true,
        protocol: 'nfs'
    },
    nfsCvs: {name: END_POINT_NAMES_MAP.CVS, path: 'nfs', protocol: 'nfs'},
    targetNfsCvs: {
        name: `Target ${END_POINT_NAMES_MAP.CVS}`,
        path: 'targetnfs',
        isTargetStepFromSameSourceType: true,
        protocol: 'nfs'
    },
    nfsCvo: {name: END_POINT_NAMES_MAP.CVO, path: 'nfs', protocol: 'nfs'},
    targetNfsCvo: {
        name: `Target ${END_POINT_NAMES_MAP.CVO}`,
        path: 'targetnfs',
        isTargetStepFromSameSourceType: true,
        protocol: 'nfs'
    },
    cifs: {name: [END_POINT_NAMES_MAP.CIFS], path: 'cifs', protocol: 'cifs'},
    targetCifs: {
        name: `Target ${END_POINT_NAMES_MAP.CIFS}`,
        path: 'targetcifs',
        isTargetStepFromSameSourceType: true,
        protocol: 'cifs'
    },
    cifsAnf: {name: [END_POINT_NAMES_MAP.ANF], path: 'cifs', protocol: 'cifs'},
    targetCifsAnf: {
        name: `Target ${END_POINT_NAMES_MAP.ANF}`,
        path: 'targetcifs',
        isTargetStepFromSameSourceType: true,
        protocol: 'cifs'
    },
    cifsCvs: {name: [END_POINT_NAMES_MAP.CVS], path: 'cifs', protocol: 'cifs'},
    targetCifsCvs: {
        name: `Target ${END_POINT_NAMES_MAP.CVS}`,
        path: 'targetcifs',
        isTargetStepFromSameSourceType: true,
        protocol: 'cifs'
    },
    cifsCvo: {name: [END_POINT_NAMES_MAP.CVO], path: 'cifs', protocol: 'cifs'},
    targetCifsCvo: {
        name: `Target ${END_POINT_NAMES_MAP.CVO}`,
        path: 'targetcifs',
        isTargetStepFromSameSourceType: true,
        protocol: 'cifs'
    },
    s3Sgws: {name: END_POINT_NAMES_MAP.SGWS, path: 's3based', protocol: 's3'},
    targetS3Sgws: {
        name: `Target ${END_POINT_NAMES_MAP.SGWS}`,
        path: 'targets3based',
        isTargetStepFromSameSourceType: true,
        protocol: 's3'
    },
    s3Ontap: {name: END_POINT_NAMES_MAP.ONTAP, path: 's3based', protocol: 's3'},
    targetS3Ontap: {
        name: `Target ${END_POINT_NAMES_MAP.ONTAP}`,
        path: 'targets3based',
        isTargetStepFromSameSourceType: true,
        protocol: 's3'
    },
    s3Ibm: {name: END_POINT_NAMES_MAP.IBM, path: 's3based', protocol: 's3'},
    targetS3Ibm: {
        name: `Target ${END_POINT_NAMES_MAP.IBM}`,
        path: 'targets3based',
        isTargetStepFromSameSourceType: true,
        protocol: 's3'
    },
    azureBlob: {name: END_POINT_NAMES_MAP.AZURE, path: 'azure', protocol: 'azure'},
    targetAzureBlob: {
        name: `Target ${END_POINT_NAMES_MAP.AZURE}`,
        path: 'targetazure',
        isTargetStepFromSameSourceType: true,
        protocol: 'azure'
    },
    directories: {name: 'Directories', path: 'exports', protocol: 'nfs'},
    targetDirectories: {
        name: 'Target Directories',
        path: 'targetexports',
        isTargetStepFromSameSourceType: true,
        protocol: 'nfs'
    },
    shares: {name: 'Shares', path: 'shares', protocol: 'cifs'},
    targetShares: {name: 'Target Shares', path: 'targetshares', isTargetStepFromSameSourceType: true, protocol: 'cifs'},
    s3Buckets: {name: 'AWS S3 Bucket', path: 'buckets', protocol: 's3'},
    targetS3Buckets: {
        name: 'Target AWS S3 Bucket',
        path: 'targetbuckets',
        isTargetStepFromSameSourceType: true,
        protocol: 's3'
    },
    gcpBuckets: {name: `${END_POINT_NAMES_MAP.GCP} Bucket`, path: 'gcpbuckets', protocol: 'gcp'},
    targetGcpBuckets: {
        name: `Target ${END_POINT_NAMES_MAP.GCP} Bucket`,
        path: 'targetgcpbuckets',
        isTargetStepFromSameSourceType: true,
        protocol: 'gcp'
    },
    s3SgwsBuckets: {name: 'SG Bucket', path: 's3basedbuckets', protocol: 's3'},
    targetS3SgwsBuckets: {
        name: 'Target SG Bucket',
        path: 'targets3basedbuckets',
        isTargetStepFromSameSourceType: true,
        protocol: 's3'
    },
    s3OntapBuckets: {name: 'ONTAP S3 Bucket', path: 's3basedbuckets', protocol: 's3'},
    targetS3OntapBuckets: {
        name: 'Target ONTAP S3 Bucket',
        path: 'targets3basedbuckets',
        isTargetStepFromSameSourceType: true,
        protocol: 's3'
    },
    s3IbmBuckets: {name: 'IBM Bucket', path: 's3basedbuckets', protocol: 's3'},
    targetS3IbmBuckets: {
        name: 'Target IBM Bucket',
        path: 'targets3basedbuckets',
        isTargetStepFromSameSourceType: true,
        protocol: 's3'
    },
    blobContainers: {name: 'Azure Blob Containers', path: 'blobcontainers', protocol: 'azure'},
    targetBlobContainers: {
        name: 'Target Azure Blob Containers',
        path: 'targetblobcontainers',
        isTargetStepFromSameSourceType: true,
        protocol: 'azure'
    },
    anfWe: {name: 'Working Environment', path: 'anfwe'},
    targetAnfWe: {name: 'Target Working Environment', path: 'targetanfwe', isTargetStepFromSameSourceType: true},
};

const calculateSourceOrTarget = (wizardInfo, isTargetOfSameSource, protocol) => {
    let sourceOrTarget;

    if (wizardInfo.source.protocol === wizardInfo.target.protocol) {
        sourceOrTarget = isTargetOfSameSource ? 'target' : 'source';
    } else if (wizardInfo.source.protocol === protocol) sourceOrTarget = 'source';
    else if (wizardInfo.target.protocol === protocol) sourceOrTarget = 'target';

    if (!sourceOrTarget) throw new Error(`wizardInfo does not include ${protocol}`);
    return sourceOrTarget;
};

//This function identifies and returns the source/target object that is used by a specific step (the step that called the function)
export const getEndPointObjectForStep = (wizardInfo, isTargetOfSameSource, stepId) => {
    const stepProtocol = stepInfoById[stepId].protocol;
    if (!stepProtocol) return; //for neutral steps like dataBroker, settings...
    const sourceOrTarget = calculateSourceOrTarget(wizardInfo, isTargetOfSameSource, stepProtocol);
    return wizardInfo[sourceOrTarget];
};

export const calculateStepId = (wizardInfo, isTargetOfSameSource, protocol, suffix) => {
    const sourceOrTarget = calculateSourceOrTarget(wizardInfo, isTargetOfSameSource, protocol);
    return createStepId(wizardInfo[sourceOrTarget].protocol, wizardInfo[sourceOrTarget].provider, isTargetOfSameSource, suffix);
};

export const isSourceTargetSupported = (source, target) => {
    if (source.provider === "ontap") return target.provider === "sgws";
    if (target.provider === "ontap") return source.provider === "sgws";

    if (source.protocol === "cifs" && target.protocol === "nfs") return false;
    if (source.protocol === "nfs" && target.protocol === "cifs") return false;

    return true;
};

export const mapProtocolToSourceSteps = {
    'nfs': ['nfs', 'nfsAnf', 'nfsEfs', 'nfsCvs', 'nfsCvo', 'directories', 'anfWe'],
    's3': ['s3Sgws', 's3Ibm', 's3Buckets', 's3SgwsBuckets', 's3IbmBuckets', 's3Ontap', 's3OntapBuckets'],
    'cifs': ['cifs', 'cifsAnf', 'cifsCvo', 'cifsCvs', 'shares'],
    'azure': ['azureBlob', 'blobContainers'],
    'gcp': ['gcpBuckets']
};

//protocols and providers
export const sourceAndTargetDisplayName = {
    's3': 'AWS S3 Bucket',
    'sgws': END_POINT_NAMES_MAP.SGWS,
    'gcp': `${END_POINT_NAMES_MAP.GCP} Bucket`,
    'ibm': END_POINT_NAMES_MAP.IBM,
    'nfs': END_POINT_NAMES_MAP.NFS,
    'cifs': END_POINT_NAMES_MAP.CIFS,
    'efs': END_POINT_NAMES_MAP.EFS,
    'azure': END_POINT_NAMES_MAP.AZURE,
    'cvs': END_POINT_NAMES_MAP.CVS,
    'anf': END_POINT_NAMES_MAP.ANF,
    'ontap': END_POINT_NAMES_MAP.ONTAP,
    'cvo': END_POINT_NAMES_MAP.CVO
};

//only protocols
export const endPointTypes = {
    'nfs': END_POINT_FILE_SYSTEM,
    'cifs': END_POINT_FILE_SYSTEM,
    's3': END_POINT_STORAGE,
    'gcp': END_POINT_STORAGE,
    'azure': END_POINT_STORAGE
};

export const getWizardTypeForSystemFlow = (wizardInfo) => {
    const source = wizardInfo.source.provider || wizardInfo.source.protocol;
    const target = wizardInfo.target.provider || wizardInfo.target.protocol;
    return `${source}-${target}`;
};

export const calculateNextStepObj = (wizard, currentStep) => {
    const currentStepIndex = wizard.stepsOrder.indexOf(currentStep);
    const nextStepId = wizard.stepsOrder[currentStepIndex + 1];
    return wizard.stepInfoById[nextStepId];
};

export const getObjectStorageTargetTypeFromWizardInfo = (wizardInfo) => {
    switch (wizardInfo.target.protocol) {
        case 'azure':
            return 'Azure Blob Container';
        case 's3':
            return 'S3 bucket';
        default:
            throw new Error(`Object storage of type ${wizardInfo.target.protocol} is not defined for getObjectStorageTargetTypeFromWizardInfo`);
    }
};

export const splitCombinationToSourceTarget = (type) => {
    const split = type.split('-');
    return {source: split[0], target: split[1]};
};

const getVersion = (version, defaultVersion) => {
    return version || defaultVersion;
};

export const appendItemToLocalStorage = (type, key, data, isTargetOfSameSrc) => {
    const formattedType = isTargetOfSameSrc ? `${type}-trg` : type;
    const keyCamelCase = _.camelCase(formattedType);
    let localCred = window.localStorage.getItem(keyCamelCase);
    localCred = (localCred === null) ? {} : JSON.parse(localCred);
    localCred[key] = data;
    window.localStorage.setItem(keyCamelCase, JSON.stringify(localCred));
};

export const getS3EncryptionKey = (path) => {
    return getCredentialsByType('s3', path.split('/')[0]);
};

export const getCredentialsByType = (type, key, isTargetOfSameSrc, remove) => {
    const formattedType = isTargetOfSameSrc ? `${type}Trg` : type;
    const credObject = JSON.parse(window.localStorage.getItem(_.camelCase(formattedType)));
    const credObjectByKey = credObject ? credObject[key] : null;
    if (remove) {
        _.each(remove, keyToRemove => delete credObjectByKey[keyToRemove]);
    }
    return credObjectByKey;
};

export const removeCredentialsByType = (type) => {
    window.localStorage.removeItem(type);
    window.localStorage.removeItem(_.camelCase(`${type}-trg`));
};

export const getSourceAndTargetInfo = ({
                                           _wizard: {
                                               selectedWizardParams: {
                                                   selectedNFS, selectedNFSTarget,
                                                   selectedCIFS, selectedCIFSTarget,
                                                   selectedS3BasedBucketPath, selectedS3BasedBucketPathTarget, selectedS3Based, selectedS3BasedTarget,
                                                   selectedBucketPath, selectedBucketPathTarget, selectedGCPBucketPath, selectedGCPBucketPathTarget,
                                                   selectedExport, selectedExportTarget,
                                                   selectedShare, selectedShareTarget,
                                                   selectedStorageAccountName, selectedStorageAccountNameTarget,
                                                   selectedAzureBlobContainerPath, selectedAzureBlobContainerPathTarget
                                               }, wizardInfo
                                           }
                                       }) => {
    const getBucketPaths = (fullBucketPath) => {
        const [bucketName, ...restOfBucketPath] = removeStartingAndEndingSlash(fullBucketPath || '').split('/');
        const bucketPath = restOfBucketPath ? restOfBucketPath.join('/') : '';
        return {
            bucketName,
            bucketPath
        }
    };
    const infoByProtocol = (endpointObj, asTarget) => {
        let objectInfo = {};

        switch (endpointObj.protocol) {
            case 'nfs' : {
                objectInfo.title = endpointObj.provider ? END_POINT_NAMES_MAP[endpointObj.provider.toUpperCase()] : END_POINT_NAMES_MAP.NFS;
                objectInfo.description = asTarget && selectedNFSTarget ? selectedNFSTarget : selectedNFS;
                objectInfo.subtitle = 'Directory';
                objectInfo.subtitleDescription = asTarget && selectedExportTarget ? selectedExportTarget : selectedExport;
                break;
            }
            case 'cifs' : {
                objectInfo.title = endpointObj.provider ? END_POINT_NAMES_MAP[endpointObj.provider.toUpperCase()] : END_POINT_NAMES_MAP.CIFS;
                objectInfo.description = asTarget && selectedCIFSTarget ? selectedCIFSTarget : selectedCIFS;
                objectInfo.subtitle = 'Share';
                objectInfo.subtitleDescription = asTarget && selectedShareTarget ? selectedShareTarget : selectedShare;
                break;
            }
            case 's3': {
                if (!endpointObj.provider) {
                    const bucketPaths = asTarget && selectedBucketPathTarget ? getBucketPaths(selectedBucketPathTarget) : getBucketPaths(selectedBucketPath);
                    objectInfo.title = OBJECT_STORAGE_NAMES_MAP.S3;
                    objectInfo.description = bucketPaths.bucketName;
                    objectInfo.subtitle = 'Folder';
                    objectInfo.subtitleDescription = bucketPaths.bucketPath;
                    break;
                } else {
                    const host = asTarget && selectedS3BasedTarget ? selectedS3BasedTarget : selectedS3Based;
                    const port = getCredentialsByType('objectStorage', host, asTarget && selectedS3BasedTarget).port;
                    const bucketPaths = asTarget && selectedS3BasedBucketPathTarget ? getBucketPaths(selectedS3BasedBucketPathTarget) : getBucketPaths(selectedS3BasedBucketPath);
                    objectInfo.title = OBJECT_STORAGE_NAMES_MAP[endpointObj.provider.toUpperCase()];
                    objectInfo.description = `${host}${port ? `:${port}` : ""}/${bucketPaths.bucketName}`;//ontap s3 doesn't use port
                    objectInfo.subtitle = 'Folder';
                    objectInfo.subtitleDescription = bucketPaths.bucketPath;
                    break;
                }
            }
            case 'gcp': {
                const bucketPaths = getBucketPaths(asTarget && selectedGCPBucketPathTarget ? selectedGCPBucketPathTarget : selectedGCPBucketPath);
                objectInfo.title = OBJECT_STORAGE_NAMES_MAP.GCP;
                objectInfo.description = bucketPaths.bucketName;
                objectInfo.subtitle = 'Folder';
                objectInfo.subtitleDescription = bucketPaths.bucketPath;
                break;
            }
            case 'azure': {
                objectInfo.title = OBJECT_STORAGE_NAMES_MAP.AZURE;
                objectInfo.description = asTarget && selectedStorageAccountNameTarget ? selectedStorageAccountNameTarget : selectedStorageAccountName;
                objectInfo.subtitle = 'Container';
                objectInfo.subtitleDescription = asTarget && selectedAzureBlobContainerPathTarget ? selectedAzureBlobContainerPathTarget : selectedAzureBlobContainerPath;
                break;
            }
            default: {
                objectInfo.title = 'Unknown Protocol';
                objectInfo.description = 'Unknown Protocol';
                objectInfo.subtitle = 'Unknown Protocol';
                objectInfo.subtitleDescription = 'Unknown Protocol';
                break;
            }
        }
        return objectInfo;
    };
    const sourceInfo = infoByProtocol(wizardInfo.source);
    const targetInfo = infoByProtocol(wizardInfo.target, true);
    return {sourceInfo, targetInfo};
};


export const breakFullPathToRootAndPath = (fullPath, root) => {
    const trimedRoot = removeStartingAndEndingSlash(root);
    const trimedFullPath = removeStartingAndEndingSlash(fullPath);
    let pathWithoutRoot, rootOfPath;
    if (!_.isNil(trimedRoot)) {//exports and shares
        const rootLen = trimedRoot.length;
        rootOfPath = root;
        if (_.startsWith(trimedFullPath, trimedRoot))
            pathWithoutRoot = removeStartingAndEndingSlash(trimedFullPath.slice(rootLen));
        else pathWithoutRoot = trimedFullPath;
    }
    else {//in case of bucket, there is no root;
        const indexOfFirstSlash = trimedFullPath.indexOf('/');
        pathWithoutRoot = trimedFullPath.slice(indexOfFirstSlash + 1);
        rootOfPath = trimedFullPath.slice(0, indexOfFirstSlash)
    }

    return {pathWithoutRoot, rootOfPath};
};

export const buildSyncObject = ({
                                    _wizard: {
                                        selectedWizardParams: {
                                            selectedDataBroker, selectedDataBrokerTarget, selectedIsTargetDataBrokerInitiator, selectedNFS, selectedNFSTarget,
                                            selectedCIFS, selectedCIFSTarget,
                                            selectedS3Based, selectedS3BasedTarget, selectedS3BasedBucketPath, selectedS3BasedBucketPathTarget,
                                            selectedBucketPath, selectedBucketPathTarget, selectedStorageClass, selectedEncryptionType,
                                            selectedExport, selectedExportTarget,
                                            selectedShare, selectedShareTarget,
                                            selectedRootExport, selectedRootExportTarget,
                                            cifsVersion, targetCifsVersion,
                                            selectedRootShare, selectedRootShareTarget,
                                            relationshipTags, nfsVersion, targetNfsVersion,
                                            selectedStorageAccountName, selectedStorageAccountNameTarget,
                                            selectedAzureBlobContainerPath, selectedAzureBlobContainerPathTarget,
                                            selectedGCPBucketPath, selectedGCPBucketPathTarget, selectedBlobTier
                                        }, wizardInfo, isSecure
                                    }
                                }) => {

    const relationshipTagsArray = relationshipTags === 'noTags' || !relationshipTags ? [] : stringToTags(relationshipTags);
    const selectedIsTargetDataBrokerInitiatorBool = (selectedIsTargetDataBrokerInitiator === 'true');

    const encryption = isSecure ? {
        initiator: {
            dataBrokerId: selectedIsTargetDataBrokerInitiatorBool ? selectedDataBrokerTarget.id : selectedDataBroker.id,
            endpointType: selectedIsTargetDataBrokerInitiatorBool ? "trg" : "src"
        },
        listener: {
            dataBrokerId: selectedIsTargetDataBrokerInitiatorBool ? selectedDataBroker.id : selectedDataBrokerTarget.id,
            endpointType: selectedIsTargetDataBrokerInitiatorBool ? "src" : "trg"
        }
    } : undefined;

    const settings = getSettingsFromSessionStorage();
    let syncRequestObject = {
        dataBrokerId: selectedIsTargetDataBrokerInitiatorBool ? selectedDataBrokerTarget.id : selectedDataBroker.id,
        sourceCredentials: undefined,
        targetCredentials: undefined,
        settings,
        encryption
    };

    const breakBucketPathToBucketAndPrefix = (bucketPath) => {
        const splittedBucketPath = _.split(bucketPath, '/');
        const bucket = splittedBucketPath[0];
        const prefix = _.size(splittedBucketPath) <= 1 ? '' : _.join(_.drop(splittedBucketPath), '/');
        return {bucket, prefix};

    };

    const getEncryption = (selectedEncryptionType, path) => {
        const encryptionTypeObj = _.find(ENCRYPTION_OPTIONS, {name: selectedEncryptionType});
        const encryptionDetails = {protocol: selectedEncryptionType};
        if (encryptionTypeObj?.isKeyRequired) {
            encryptionDetails.keyId = getS3EncryptionKey(path);
        }
        return encryptionDetails;
    };


    function nfsKindObjBuilder(isTargetOfSameSource, isSourceOfSameTarget, provider) {
        const brokenPath = isTargetOfSameSource ? breakFullPathToRootAndPath(selectedExportTarget, selectedRootExportTarget) : breakFullPathToRootAndPath(selectedExport, selectedRootExport);
        const host = isTargetOfSameSource ? selectedNFSTarget : selectedNFS;
        const nfsType = provider || "nfs";
        const version = isTargetOfSameSource ? getVersion(targetNfsVersion, DEFAULT_NFS_VERSION[nfsType.toUpperCase()]) : getVersion(nfsVersion, DEFAULT_NFS_VERSION[nfsType.toUpperCase()]);
        return {
            host,
            export: brokenPath.rootOfPath,
            path: brokenPath.pathWithoutRoot,
            version,
            provider
        };
    }


    const getParamsByProtocol = (protocol, provider, isTargetOfSameSource, isSourceOfSameTarget, isSecure) => {
        let protocolObject;
        switch (protocol) {
            case 'nfs': {
                protocolObject = nfsKindObjBuilder(isTargetOfSameSource, isSourceOfSameTarget, provider, isSecure);
                break;
            }
            case 's3': {
                if (!provider) {
                    const brokenPath = breakBucketPathToBucketAndPrefix(isTargetOfSameSource ? selectedBucketPathTarget : selectedBucketPath);
                    protocolObject = {
                        bucket: brokenPath.bucket,
                        prefix: brokenPath.prefix,
                        encryption: (isSourceOfSameTarget || !selectedEncryptionType || selectedEncryptionType === ENCRYPTION_TYPES_NAMES.NONE)
                            ? undefined
                            : getEncryption(selectedEncryptionType, isTargetOfSameSource ? selectedBucketPathTarget : selectedBucketPath),
                        tags: (isSourceOfSameTarget || !relationshipTags) ? undefined : relationshipTagsArray,
                        storageClass: (isSourceOfSameTarget || !selectedStorageClass) ? undefined : selectedStorageClass
                    };
                    break;
                } else {
                    const brokenPath = breakBucketPathToBucketAndPrefix(isTargetOfSameSource ? selectedS3BasedBucketPathTarget : selectedS3BasedBucketPath);
                    const theHost = isTargetOfSameSource ? selectedS3BasedTarget : selectedS3Based;
                    protocolObject = {
                        provider,
                        host: theHost,
                        port: getCredentialsByType('objectStorage', theHost, isTargetOfSameSource).port,
                        bucket: brokenPath.bucket,
                        prefix: brokenPath.prefix
                    };
                    break;
                }
            }
            case 'cifs': {
                const brokenPath = isTargetOfSameSource ? breakFullPathToRootAndPath(selectedShareTarget, selectedRootShareTarget) : breakFullPathToRootAndPath(selectedShare, selectedRootShare);
                const version = isTargetOfSameSource ? getVersion(targetCifsVersion, DEFAULT_CIFS_VERSION) : getVersion(cifsVersion, DEFAULT_CIFS_VERSION);
                const host = isTargetOfSameSource ? selectedCIFSTarget : selectedCIFS;
                protocolObject = {host, share: brokenPath.rootOfPath, path: brokenPath.pathWithoutRoot, version, provider};
                break;
            }
            case 'azure': {
                const brokenPath = breakBucketPathToBucketAndPrefix(isTargetOfSameSource ? selectedAzureBlobContainerPathTarget : selectedAzureBlobContainerPath);
                protocolObject = {
                    storageAccountName: isTargetOfSameSource ? selectedStorageAccountNameTarget : selectedStorageAccountName,
                    container: brokenPath.bucket,
                    prefix: brokenPath.prefix,
                    tags: relationshipTagsArray,
                    blobTier: isSourceOfSameTarget ? undefined : selectedBlobTier
                };
                break;
            }
            case 'gcp': {
                const brokenPath = breakBucketPathToBucketAndPrefix(isTargetOfSameSource ? selectedGCPBucketPathTarget : selectedGCPBucketPath);
                protocolObject = {
                    bucket: brokenPath.bucket,
                    prefix: brokenPath.prefix
                };
                break;
            }
            default: {
                throw new Error(`getParamsByProtocol unknown protocol ${protocol}`);
            }
        }
        return {protocol, [protocol]: protocolObject}

    };

    const isTargetOfSameSource = wizardInfo.source.protocol === wizardInfo.target.protocol;

    //put the basic params
    syncRequestObject.source = getParamsByProtocol(wizardInfo.source.protocol, wizardInfo.source.provider, false, isTargetOfSameSource, isSecure);
    syncRequestObject.target = getParamsByProtocol(wizardInfo.target.protocol, wizardInfo.target.provider, isTargetOfSameSource, false, isSecure);

    //add param to source:
    switch (wizardInfo.source.protocol) {
        case 's3': {
            if (wizardInfo.source.provider) {
                syncRequestObject.sourceCredentials = {s3: getCredentialsByType('objectStorage', selectedS3Based, false, ['serverName', 'port'])};
                break;
            }
            break;
        }
        case 'cifs': {
            syncRequestObject.sourceCredentials = {cifs: getCredentialsByType('cifs', selectedCIFS, false, ['serverName', 'version', 'copyAcl'])};
            break;
        }
        case 'azure': {
            const {connectionString} = getCredentialsByType('azure', selectedStorageAccountName, false);
            syncRequestObject.sourceCredentials = {azure: {storageAccountConnectionString: connectionString}};
            break;
        }
        default: {
            //do nothing;
        }
    }

    //add param to target:
    switch (wizardInfo.target.protocol) {
        case 's3': {
            if (wizardInfo.target.provider) {
                if (isTargetOfSameSource) {
                    syncRequestObject.targetCredentials = {s3: getCredentialsByType('objectStorage', selectedS3BasedTarget, true, ['serverName', 'port'])};
                } else {
                    syncRequestObject.targetCredentials = {s3: getCredentialsByType('objectStorage', selectedS3Based, false, ['serverName', 'port'])};
                }
                break;
            }
            break;
        }
        case 'cifs': {
            if (isTargetOfSameSource) {
                syncRequestObject.targetCredentials = {cifs: getCredentialsByType('cifs', selectedCIFSTarget, true, ['serverName', 'version', 'copyAcl'])};
            }
            else {
                syncRequestObject.targetCredentials = {cifs: getCredentialsByType('cifs', selectedCIFS, false, ['serverName', 'version', 'copyAcl'])};
            }
            break;
        }
        case 'azure': {
            if (isTargetOfSameSource) {
                const {connectionString: connectionStringTarget} = getCredentialsByType('azure', selectedStorageAccountNameTarget, true);
                syncRequestObject.targetCredentials = {azure: {storageAccountConnectionString: connectionStringTarget}};
            } else {
                const {connectionString} = getCredentialsByType('azure', selectedStorageAccountName, false);
                syncRequestObject.targetCredentials = {azure: {storageAccountConnectionString: connectionString}};
            }
            break;
        }
        default: {
            //do nothing;
        }
    }
    return syncRequestObject;
};

export const getWizardSelectedParams = (props, wizardParams) => {
    const {_dataBrokers, _showExports, _showShares, _showBuckets, _showS3BasedBuckets, _showAzureBlobContainers, _showGCPBuckets, _wizard: {wizardInfo, isSecure}} = props;
    const {
        nfsServer, nfsServerTarget, nfsVersion, targetNfsVersion, cifsServer, cifsServerTarget, s3Based, s3BasedTarget, dataBrokerId, dataBrokerIdTarget, isTargetDataBrokerInitiator,
        s3BasedBucketPath, s3BasedBucketPathTarget, bucketPath, storageClass, encryptionType, bucketPathTarget, exportPath, rootExport, exportPathTarget, rootExportTarget, gcpBucketPath, gcpBucketPathTarget,
        sharePath, sharePathTarget, rootShareTarget, rootShare, cifsVersion, targetCifsVersion, relationshipTags,
        storageAccountName, storageAccountNameTarget, azureBlobContainerPath, azureBlobContainerPathTarget, blobTier, blobTierTarget,
        anfWe, anfWeTarget, anfVol, anfVolTarget
    } = wizardParams;

    const getSelectedParamByName = (selectedParamName) => {
        return selectedParamName ? decodeAll(_.get(props, `_wizard.selectedWizardParams.${selectedParamName}`)) : undefined;
    };

    const decodeServer = (server, selectedParamName) => {
        return server ? decodeAll(server) : getSelectedParamByName(selectedParamName);
    };

    const decodePathAndRootPath = (storagePart, server, path, rootPath, selectedPathName = null, selectedRootPathName = null) => {
        let result = {path: undefined, rootPath: undefined};
        const storagePartForServer = storagePart ? storagePart[server] : null;
        if (storagePartForServer && storagePartForServer.succeeded && storagePartForServer.data) {
            result.path = path ? decodeAll(path) : getSelectedParamByName(selectedPathName);
            result.rootPath = rootPath ? decodeAll(rootPath) : getSelectedParamByName(selectedRootPathName);
        }
        return result;
    };

    const decodeBucketAndBucketPath = (storagePart, bucketPath) => {
        const decodedBucketPath = decodeAll(bucketPath);
        const bucket = decodedBucketPath && storagePart && storagePart.succeeded && storagePart.data ? storagePart.data.find(bucket => bucket.name === decodedBucketPath.split('/')[0]) : null;
        return {bucket, bucketPath: bucket ? decodedBucketPath : null};
    };

    const getParamsByProtocol = (protocol, provider, isTargetOfSameSource) => {
        switch (protocol) {
            case 'nfs': {
                const isUsingDataBrokerTarget = isSecure && isTargetOfSameSource;
                const actualDataBrokerId = isUsingDataBrokerTarget ? dataBrokerIdTarget : dataBrokerId;
                const decodedPathAndRootPath = isTargetOfSameSource ? decodePathAndRootPath(_showExports, `${nfsServerTarget}-${actualDataBrokerId}`, exportPathTarget, rootExportTarget, 'selectedExportTarget', 'selectedRootExportTarget') :
                    decodePathAndRootPath(_showExports, `${nfsServer}-${actualDataBrokerId}`, exportPath, rootExport, 'selectedExport', 'selectedRootExport');
                return isTargetOfSameSource ? {
                    selectedNFSTarget: decodeServer(nfsServerTarget, null),
                    selectedExportTarget: decodedPathAndRootPath.path || decodedPathAndRootPath.rootPath,
                    selectedRootExportTarget: decodedPathAndRootPath.rootPath,
                    selectedAnfWeTarget: anfWeTarget,
                    selectedAnfVolTarget: decodeAll(anfVolTarget),
                    targetNfsVersion
                } : {
                    selectedNFS: decodeServer(nfsServer, 'selectedNFS'),
                    selectedExport: decodedPathAndRootPath.path,
                    selectedRootExport: decodedPathAndRootPath.rootPath,
                    selectedAnfWe: anfWe,
                    selectedAnfVol: decodeAll(anfVol),
                    nfsVersion
                };
            }
            case 's3': {
                if (provider) {
                    const relevantDecodedServer = isTargetOfSameSource ? decodeServer(s3BasedTarget) : decodeServer(s3Based);
                    const decodedBucketAndBucketPath = isTargetOfSameSource
                        ? decodeBucketAndBucketPath(_showS3BasedBuckets && _showS3BasedBuckets[relevantDecodedServer], s3BasedBucketPathTarget)
                        : decodeBucketAndBucketPath(_showS3BasedBuckets && _showS3BasedBuckets[relevantDecodedServer], s3BasedBucketPath);

                    return isTargetOfSameSource ? {
                        selectedS3BasedBucketTarget: decodedBucketAndBucketPath.bucket,
                        selectedS3BasedBucketPathTarget: decodedBucketAndBucketPath.bucketPath,
                        selectedS3BasedTarget: relevantDecodedServer
                    } : {
                        selectedS3BasedBucket: decodedBucketAndBucketPath.bucket,
                        selectedS3BasedBucketPath: decodedBucketAndBucketPath.bucketPath,
                        selectedS3Based: relevantDecodedServer
                    }
                } else {
                    const actualBucketPath = isTargetOfSameSource ? bucketPathTarget : bucketPath;
                    const decodedBucketAndBucketPath = decodeBucketAndBucketPath(_showBuckets && _showBuckets[dataBrokerId], actualBucketPath);
                    return {
                        [isTargetOfSameSource ? "selectedBucketPathTarget": "selectedBucketPath"]: decodedBucketAndBucketPath.bucketPath,
                        relationshipTags: relationshipTags ? decodeAll(relationshipTags) : undefined,
                        selectedStorageClass: storageClass,
                        selectedEncryptionType: encryptionType
                    }
                }
            }
            case 'cifs': {
                const decodedPathAndRootPath = isTargetOfSameSource ? decodePathAndRootPath(_showShares, `${cifsServerTarget}-${dataBrokerId}`, sharePathTarget, rootShareTarget) :
                    decodePathAndRootPath(_showShares, `${cifsServer}-${dataBrokerId}`, sharePath, rootShare);
                return isTargetOfSameSource ? {
                    selectedCIFSTarget: decodeServer(cifsServerTarget),
                    selectedShareTarget: decodedPathAndRootPath.path || decodedPathAndRootPath.rootPath,
                    selectedRootShareTarget: decodedPathAndRootPath.rootPath,
                    selectedAnfWeTarget: anfWeTarget,
                    selectedAnfVolTarget: decodeAll(anfVolTarget),
                    targetCifsVersion
                } : {
                    selectedCIFS: decodeServer(cifsServer),
                    selectedShare: decodedPathAndRootPath.path,
                    selectedRootShare: decodedPathAndRootPath.rootPath,
                    selectedAnfWe: anfWe,
                    selectedAnfVol: decodeAll(anfVol),
                    cifsVersion
                }
            }
            case 'gcp' : {
                const actualBucketPath = isTargetOfSameSource ? gcpBucketPathTarget : gcpBucketPath;
                const decodedBucketAndBucketPath = decodeBucketAndBucketPath(_showGCPBuckets && _showGCPBuckets[dataBrokerId], actualBucketPath);
                return isTargetOfSameSource ? {
                    selectedGCPBucketTarget: decodedBucketAndBucketPath.bucket,
                    selectedGCPBucketPathTarget: decodedBucketAndBucketPath.bucketPath
                } : {
                    selectedGCPBucket: decodedBucketAndBucketPath.bucket,
                    selectedGCPBucketPath: decodedBucketAndBucketPath.bucketPath
                }
            }
            case 'azure' : {
                const actualStoragePart = isTargetOfSameSource
                    ? _showAzureBlobContainers && _showAzureBlobContainers[storageAccountNameTarget]
                    : _showAzureBlobContainers && _showAzureBlobContainers[storageAccountName];
                const decodedBucketAndBucketPath = isTargetOfSameSource
                    ? decodeBucketAndBucketPath(actualStoragePart, azureBlobContainerPathTarget)
                    : decodeBucketAndBucketPath(actualStoragePart, azureBlobContainerPath);
                return isTargetOfSameSource ? {
                    relationshipTags: relationshipTags ? decodeAll(relationshipTags) : undefined,
                    selectedStorageAccountNameTarget: decodeAll(storageAccountNameTarget),
                    selectedAzureBlobContainerPathTarget: decodedBucketAndBucketPath.bucketPath,
                    selectedBlobTier: blobTierTarget

                } : {
                    relationshipTags: relationshipTags ? decodeAll(relationshipTags) : undefined,
                    selectedStorageAccountName: decodeAll(storageAccountName),
                    selectedAzureBlobContainerPath: decodedBucketAndBucketPath.bucketPath,
                    selectedBlobTier: blobTier
                }
            }
            default: {
                return {};
            }
        }
    };

    const getSpecificSelectedWizardParams = (wizardInfo) => {
        const isTargetOfSameSource = wizardInfo.target.protocol === wizardInfo.source.protocol;
        return Object.assign({}, getParamsByProtocol(wizardInfo.source.protocol, wizardInfo.source.provider), getParamsByProtocol(wizardInfo.target.protocol, wizardInfo.target.provider, isTargetOfSameSource));
    };

    const selectedDataBroker = _dataBrokers && _dataBrokers.succeeded && _dataBrokers.data.length > 0 ? _dataBrokers.data.find(dataBroker => dataBroker.id === dataBrokerId) : _.get(props, '_wizard.selectedWizardParams.selectedDataBroker');
    const selectedDataBrokerTarget = _dataBrokers && _dataBrokers.succeeded && _dataBrokers.data.length > 0 ? _dataBrokers.data.find(dataBroker => dataBroker.id === dataBrokerIdTarget) : _.get(props, '_wizard.selectedWizardParams.selectedDataBrokerTarget');
    const selectedIsTargetDataBrokerInitiator = isTargetDataBrokerInitiator ? isTargetDataBrokerInitiator : 'false';

    const selectedWizardParams = Object.assign({}, {
        selectedDataBroker,
        selectedDataBrokerTarget,
        selectedIsTargetDataBrokerInitiator
    }, getSpecificSelectedWizardParams(wizardInfo));

    return selectedWizardParams;
};

export const isSecureAllowed = (source, target) => {
    if (!source || !target) return false;
    const isNfsNfs = source.protocol === 'nfs' &&  !source.provider &&
        target.protocol === 'nfs' &&  !target.provider;
    const isAnfAnf = source.protocol === 'nfs' &&  source.provider === 'anf' &&
        target.protocol === 'nfs' &&  target.provider === 'anf';
    return isNfsNfs || isAnfAnf;
};
